package Acme.Serve;

import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
// Copyright (C) 1996 by Jef Poskanzer <jef@acme.com>.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//
// Visit the ACME Labs Java page for up-to-date versions of this and other
// fine Java utilities: http://www.acme.com/java/

/**
* Serves a file as a servlet under the javax servlet definition.
*
* @author Jef Poskanzer, Copyright (C) 1996 by Jef Poskanzer <jef@acme.com>.  All rights reserved.
* @author Nazario Irizarry, <naz@mitre.org> modifications to original code, 1 Jan 1998
*/

public class FileServlet extends HttpServlet {


	File fileToGet;
	int fileGetError;
	public File relativeLocalRoot;		/* null unless subclass fills in */
    public String alias;                    /* must be filled in by subclass if relativeLocalRoot is set */
	String filePathname;

    public FileServlet() {
	}

    public String getServletInfo() {
		return "servlet to serve iPAM files or directories";
	}

    public void log(String msg) {
		String name = getClass().getName();
		int pos = name.lastIndexOf(".");
		if (pos > 0)
			name = name.substring(pos+1);
		getServletContext().log(name + "\t"+ msg);
    }
/**
* Intercept the service call to reset our private attributes.
* Invokes javax.servlet.http.HttpServlet.service
*/
	protected void service (HttpServletRequest req, HttpServletResponse res)
	throws ServletException, IOException {
		fileToGet = null;
		fileGetError = 0;
		filePathname = null;
		super.service(req, res);
	}

/**
* Examines the requested path to try to resolve it to an existing file
* or directory.
*/
	private boolean resolvePath( HttpServletRequest req) {
	
		if (fileToGet != null)
			return true;
		
		String path = req.getServletPath();

		//System.out.println("request for "+path);

		// need to remove alias
		if (relativeLocalRoot != null) {
		    if ((alias != null) && path.startsWith(alias))
			path = path.substring(alias.length());
		}


		//System.out.println("new path is " + path);
	/*
	* Check for obvious errors
	*/
		if ( path == null || path.charAt( 0 ) != '/' ) {
		    fileGetError = HttpServletResponse.SC_BAD_REQUEST;
		    return false;
		}
	/*
	* Make sure user isn't trying to snoop our system
	*/
		if ( path.indexOf( "/../" ) != -1 || path.endsWith( "/.." ) ) {
			fileGetError = HttpServletResponse.SC_FORBIDDEN;
		    return false;
		}

	/*
	* Strip off leading /
	*/
		String pathname = path.substring( 1 );
//		if ( pathname.length() == 0 )
//		    pathname = "./";

	/*
	* Strip off trailing /
	*/
		String filename = pathname;
		filePathname = pathname;
		
		if ( filename.endsWith("/") )
		    filename = filename.substring( 0, filename.length() - 1 );

		filename = filename.replace( '/', File.separatorChar );

		filename = getServletContext().getRealPath( filename );
	/*
	* Create a file reference using the relative local root if provided
	*/
		if (relativeLocalRoot != null)
			fileToGet = new File(relativeLocalRoot, filename );
		else
			fileToGet = new File(filename );
		
		if (fileToGet.exists()) {
//		    System.out.println(" pathname is "+filePathname);

			return true;
		}

//		System.out.println(" fileServlet 2 - "+filename);
		
	    if ( filename.endsWith( "index.html" ) ) {
			filename = pathname.substring( 0, pathname.length() - 10 );
			filePathname = filePathname.substring( 0, filePathname.length() - 10 );
		}
			
		if (relativeLocalRoot != null)
			fileToGet = new File(relativeLocalRoot, filename );
		else
			fileToGet = new File(filename );
			
		if (fileToGet.exists()) {
//		    System.out.println(" pathname is "+filePathname);
			return true;
		}

//		System.out.println(" FileServlet can't find - "+filePathname);
		log(" FileServlet can't find - "+fileToGet.getAbsolutePath()+" from request "+fileToGet);
		log("   Relative root is - "+relativeLocalRoot.getAbsolutePath());
		

		fileGetError = HttpServletResponse.SC_NOT_FOUND;
		return false;
	}
/**
* Implement the GET method
*/
	protected void doGet (HttpServletRequest req, HttpServletResponse res)
	throws ServletException, IOException {

		if (!resolvePath( req)) {
			res.sendError(fileGetError);
			return;
		}

	    if ( fileToGet.isDirectory() )
	    	doGetDirectory( req, res, false, fileToGet );
	    else 
			doGetFile( req, res, false, fileToGet );
	}



    protected long getLastModified (HttpServletRequest req)
    {
		if (!resolvePath( req))
			return -1;
		return fileToGet.lastModified();
//		return System.currentTimeMillis();
    }

    private void doGetFile( 
    		HttpServletRequest req, 
    		HttpServletResponse res, 
    		boolean headOnly, 
    		File file ) throws IOException
	{
	log( "getting " + req.getRequestURI() );
	if ( ! file.canRead() )
	    {
	    res.sendError( HttpServletResponse.SC_FORBIDDEN );
	    return;
	    }

	res.setStatus( HttpServletResponse.SC_OK );

	res.setContentType( getServletContext().getMimeType( file.getName() ) );
	res.setContentLength( (int) file.length() );
	res.setDateHeader( "Last-Modified", file.lastModified() );
	OutputStream out = res.getOutputStream();
	if ( ! headOnly ) {
	    InputStream in = new FileInputStream( file );
	    copyStream( in, out );
	    in.close();
	    }
//	out.close();
	}
    
    private void copyStream( InputStream in, OutputStream out ) throws IOException
	{
		byte[] buf = new byte[4096];
		int len;
		while ( ( len = in.read( buf ) ) != -1 ) {
		    if (len > 0)
		    	out.write( buf, 0, len );
		}
	}


    private void doGetDirectory( 
    			HttpServletRequest req, 
    			HttpServletResponse res, 
    			boolean headOnly, 
    			File file ) throws IOException
	{
	String pathStr = req.getRequestURI();
	int t1 = pathStr.indexOf("//");
	int t2 = pathStr.indexOf('/',t1);
	if (t2 >= 0)
		pathStr = pathStr.substring(t2);
		
	log( "get dir " + pathStr );
	if ( ! file.canRead() )
	    {
	    res.sendError( HttpServletResponse.SC_FORBIDDEN );
	    return;
	    }
	res.setStatus( HttpServletResponse.SC_OK );
	res.setContentType( "text/html" );
	OutputStream out = res.getOutputStream();
	if ( ! headOnly )
	    {
	    PrintStream p = new PrintStream( out  );
	    p.println( "<HTML><HEAD>" );
	    p.println( "<TITLE>Index of " + pathStr + "</TITLE>" );
	    p.println( "</HEAD><BODY BGCOLOR=\"#a0c0a0\">" );
	    p.println( "<H2>Index of " + pathStr + "</H2>" );
	    p.println( "<PRE>" );
	    p.println( "mode     bytes  last-changed  name" );
	    p.println( "<HR>" );
	    String[] names = file.list();
	    Acme.Utils.sortStrings( names );
		for ( int i = 0; i < names.length; ++i ) {
			File aFile = new File( file.getAbsolutePath()+File.separatorChar+ names[i] );
			String aFileType;
			if ( aFile.isDirectory() )
			    aFileType = "d";
			else if ( aFile.isFile() )
			    aFileType = "-";
			else
			    aFileType = "?";
			String aFileRead = ( aFile.canRead() ? "r" : "-" );
			String aFileWrite = ( aFile.canWrite() ? "w" : "-" );
			String aFileExe = "-";
			String aFileSize = Acme.Fmt.fmt( aFile.length(), 8 );
			String aFileDate =
			    Acme.Utils.lsDateStr( new Date( aFile.lastModified() ) );
			String aFileDirsuf = ( aFile.isDirectory() ? "/" : "" );
			String aFileSuf = ( aFile.isDirectory() ? "/" : "" );
			String qualifiedFileName;
				qualifiedFileName = names[i];
			p.println(
			    aFileType + aFileRead + aFileWrite + aFileExe +
			    "  " + aFileSize + "  " + aFileDate + "  " +
			    "<A HREF=\"" + qualifiedFileName + aFileDirsuf + "\">" +
			    names[i] + aFileSuf + "</A>" );
			}
	    p.println( "</PRE>" );
	    p.println( "<HR>" );
	    ServeUtils.writeAddress( p );
	    p.println( "</BODY></HTML>" );
	    p.flush();
	    }
	}
    
}
