/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.daemon;

import java.util.*;
import java.io.*;

import org.mitre.pam.interfaces.*;
import org.mitre.pam.executive.PropertySet;
import org.mitre.pam.product.view.DictionaryProduct;
import org.mitre.pam.daemon.Memo;
import org.mitre.pam.interfaces.IDictionary;
import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.interfaces.ExecToEditor;
import org.mitre.pam.executive.Exec;
import org.mitre.pam.executive.Module;
import org.mitre.pam.util.SortedHash;
import org.mitre.pam.store.IntegrityException;
import org.mitre.pam.util.AttributeKey;

/**
* Each instance of this daemon stores a "dictionary" which maps field names to 
* array indexes, and datatypes which are used by DictionaryProduct. <p>
*
* @author Jones Leung
* @author John R. Fantry - added support for Java serialization
* @author Nazario Irizarry - completely redesigned
*
*/
final public class Dict extends Memo implements IDictionary, Serializable {
    private final static String BOOLEAN = "B";
    private final static String INTEGER = "I";
    public final static String LONG = "L";
    public final static String FLOAT = "F";
    public final static String DOUBLE = "D";
    public final static String STRING = "S";

	/** The common name used by pam for this class */
	public final static String PAM_ALIAS = "dictionary";
    /** Count of how many indexes have been used so far */
    public final static String USED_INDEX_COUNT = "usedIndexCount";

    /** Prefix attached to datatypes to avoid collision with regular agent field names */
    public final static String TYPE_OF_NAME_PREFIX = "TYPE$_";
    
    /** Prefix attached to indexes to avoid collision with regular agent field names */
    public final static String INDEX_OF_NAME_PREFIX = "IND$_";
                    
    private int uIndexCount = -1;

/**
* used by getDictionary() in Exec00 to instantiate a
* new Dict daemon when needed
*/
    public Dict (String dictionaryName) {
    	if (dictionaryName != null)
    		throw new IllegalArgumentException("Need non-null dictionary name in constructor");
//        setProperty (DictionaryProduct.DICTIONARY_NAME, dictionaryName); //why?
		clearUsedIndexCount(); // indicates empty
    }

	synchronized private void clearUsedIndexCount() {
		uIndexCount = -1;
		getMemento().setProperty(USED_INDEX_COUNT, uIndexCount);
		getMemento().setProperty(META_PERSIST, false);
	}        
	synchronized private void incrementUsedIndexCount() {
		uIndexCount++;
		getMemento().setProperty(USED_INDEX_COUNT, uIndexCount);
	}        
/**
* used by persistence subsystem 
*/
	public Dict() {	    
	    clearUsedIndexCount();     
	}

/**
* Returns the name of the dictionary stored by this daemon. Same as its id.
* @return the name of the dictionary
*/
    public String getDictionaryName () {
    	return getId();
    }
/**
* Returns the index for a given field name <p>
* Or, if the field name does not exist, an unused index
* is assigned to that field name, and that newly assigned
* index is returned. <p> 
* @param    propertyName   the property name      
* @return   the array index associated with this property name
*/
    public int getIndexByName (String propertyName) { 
    	return getIndexInternal (propertyName, null);
    }
/**
* Returns the index for a given field name.  Creates
* the property if it does not exist.  Tags the field
* with the specified type if the type is not null.
*
* @param    propertyName   the property name 
* @param    type the internal typing (may be null)     
* @return   the array index associated with this property name
*/    
    private int getIndexInternal(String propertyName, String type) {
    	if (propertyName == null)
    		throw new IllegalArgumentException("propertyName cannot be null");

		String indexString = INDEX_OF_NAME_PREFIX+propertyName;
        int index = getIntegerProperty(indexString, -1);
        
        if (index == -1) {
            incrementUsedIndexCount();
            index = getUsedIndexCount();
            getMemento().setProperty(indexString, index);
            if (type != null)
            	getMemento().setProperty(TYPE_OF_NAME_PREFIX+propertyName, type);
        }
        else if (type != null) {
        	String typeString = TYPE_OF_NAME_PREFIX+propertyName;
            String existingType = getProperty(typeString);
            if (existingType == null) {
            	getMemento().setProperty(typeString, type);
            } else if (!existingType.equals(type)) {
            	assert(false,"The requested property type '"+type+"' for property '"+
            		propertyName,"' conflicts with the existing type of '",existingType,
            		"', the type was not changed");
            }
        }
        return index;
    }

/**
* Define and set the type of the specified field
* @param key the key that represents the attribute
* @param type   PropertyMetaData.INTEGER, PropertyMetaData.LONG, PropertyMetaData.FLOAT, etc.
*/
    public void defineProperty(AttributeKey key, Class type) {
    	defineProperty(key.toString(), type);
    }

/**
* Define and set the type of the specified field
* @param name of the attribute
* @param type   PropertyMetaData.INTEGER, PropertyMetaData.LONG, PropertyMetaData.FLOAT, etc.
*/
    public void defineProperty(String name, Class type) {
    	String internalType = STRING;
    	
    	if (type == PropertyMetaData.STRING)
    		internalType = STRING;
    	else if (type == PropertyMetaData.BOOLEAN)
    		internalType = BOOLEAN;
    	else if (type == PropertyMetaData.LONG)
    		internalType = LONG;
    	else if (type == PropertyMetaData.INTEGER)
    		internalType = INTEGER;
    	else if (type == PropertyMetaData.FLOAT)
    		internalType = FLOAT;
    	else if (type == PropertyMetaData.DOUBLE)
    		internalType = DOUBLE;
    	else {
    		reporter.assert(false,"unknown type ",type,"treating as String",null);
    	}
    	getIndexInternal(name, internalType);
    }
/**
* Returns an Enumeration of field names
* @return an Enumeration of the field names stored by this Dictionary object
*/
    public Enumeration getNames() {
        int i = 0;
        Vector v = new Vector(getMemento().size()); 
        Enumeration e = getProperties();

        while (e.hasMoreElements()) {
        	String item = (String) e.nextElement();
        	if (item.startsWith(INDEX_OF_NAME_PREFIX)) {
        		v.addElement (item.substring(INDEX_OF_NAME_PREFIX.length()));
        	}
            
        }        
        return v.elements();
    }
/**
* Returns true, if this Dictionary object is empty.
* @return true, if this Dictionary object does not 
*  contain any associations of field names to array indexes.
*/
    public boolean isEmpty () {
        return (getUsedIndexCount() == -1);
    }
/**
* Returns the used index count (i.e. the highest index value used so far)
* @return the highest index value used so far
*/
    public int getUsedIndexCount () {
        return uIndexCount;
    }
/**
* Returns type of the specified field
* @param name   the field name
* @param dataType   PropertyMetaData.INTEGER, PropertyMetaData.LONG, PropertyMetaData.FLOAT, etc.
*/
    public Class getDataType(String name){
    	String type = getProperty(TYPE_OF_NAME_PREFIX + name);
    	if (type == null) {
    		return null;
    	}
    	char t = type.charAt(0);
    	if (t == STRING.charAt(0))
    		return PropertyMetaData.STRING;
    	if (t == BOOLEAN.charAt(0))
    		return PropertyMetaData.BOOLEAN;
    	if (t == LONG.charAt(0))
    		return PropertyMetaData.LONG;
    	if (t == INTEGER.charAt(0))
    		return PropertyMetaData.INTEGER;
    	if (t == FLOAT.charAt(0))
    		return PropertyMetaData.FLOAT;
    	if (t == DOUBLE.charAt(0))
    		return PropertyMetaData.DOUBLE;
    	reporter.assert(false,"unknown type ",type,"treating as String",null);
    	return PropertyMetaData.STRING;
    }
/**
* Returns type of the specified field
* @param name   the field name
* @param dataType   PropertyMetaData.INTEGER, PropertyMetaData.LONG, PropertyMetaData.FLOAT, etc.
*/
    public Class getDataType(AttributeKey key){
    	return getDataType(key.toString());
    }
/**
* Returns true if a field with a given name already exists
* @param name   the field name
* @return   true, if there exists a field with this name
*/
    public boolean propertyExists (String name) {
    	return (null != getProperty(INDEX_OF_NAME_PREFIX + name));
    }
/**    
* This method is synchronized to prevent the accidental creation of
* multiple Dictionary objects having the same name.  There should only
* be one Dictionary object for each dictionary name. <p> 
* This method should only be called by the ExecCallback getDictionary() 
* for the purpose of retrieving an existing dictionary or creation of a new one. <p>
*/
    public static synchronized Dict makeDict(ExecToEditor exec, String requestedDictionaryName) {

    	Dict agent=null;
    	String dictionaryName = requestedDictionaryName;
    	int iter = 1;
    	while (true) {
			try {
				agent = (Dict) exec.getActiveAgent(dictionaryName);
				if (agent == null)
					break;
				return agent;
			}
			catch (ClassCastException e) {
				dictionaryName = requestedDictionaryName+ (iter++); // build a nonconflicting name
			}
		}
	    
		// create a new dictionary, if not found
		try {
		    exec.declareNode(PAM_ALIAS, dictionaryName, null) ;
		}
		catch (IntegrityException e) {
			exec.getExecAsAgent().getReporter().reportException(e);
			return null;
		}

        agent = (Dict) exec.getActiveAgent(dictionaryName);

    	if (dictionaryName != requestedDictionaryName) {
    		exec.getExecAsAgent().getReporter().warning("A request for dictionary \'"+requestedDictionaryName+
    			"\' was changed to \'"+dictionaryName+"\' to avoid conflict with an existing agent's name");
    	}
    	return agent;
    }    

	/**
	* Determines if a dictionary with the provided name exists.
	*
	* @param the dictionary name to check for
	*/

	public static boolean exists(String dictionaryName) {
		Agent agent = Exec.getExec().getActiveAgent(dictionaryName);
		if (agent == null)
			return false;
		if (agent instanceof Dict)
			return true;
		
		throw new RuntimeException("Dictionary name: "+dictionaryName+" is already in use by a non dictionary agent");
	}

	/**
	* Keep track of whether a set of an indexed property is made.
	* If so we ignore the restored index and assign a new one.
	* This maintains the integrity of the dictionary every time it is loaded or built.
	* However, we don't anything special with restored types.  We take them as they are.
	*/
	public void setProperty(String name, String value) {
		if (name.startsWith(INDEX_OF_NAME_PREFIX)) {
			String property = name.substring(INDEX_OF_NAME_PREFIX.length());
			getIndexInternal(property,null);
		}
		else if (name.equals(USED_INDEX_COUNT)) {
			// do nothing cause the real used index count is set as new items are created
		}
		else {
			super.setProperty(name, value);
		}
	}
/**
* A factory method created to facilitate migration from iPAM 2.2
* to iPAM 2.3.  This method takes arguments similar to the defunct
* getIndexOf method. To use this
* change calls of the type dictionaryProduct.getIndexOf("attr", Dict.INTEGER)
* into dictionary.makeAttributeKey("attr", PropertyMetaData.INTEGER).
*/ 
	public AttributeKey makeAttributeKey(String name, Class type) {
		defineProperty(name, type);
		return new AttributeKey(name); // it will get resolved later
	}

}