/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.executive;

import java.util.Observable;
import java.util.Enumeration;
import org.mitre.pam.interfaces.Doer;
import org.mitre.pam.interfaces.Tasker;
import org.mitre.pam.interfaces.Task;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Sink;
import org.mitre.pam.interfaces.Agent;

/**
* A basic unscheduled command for the exec. 
*
* @author Nazario Irizarry, Jr.
*/
abstract public class Command implements Task {
/**
* A flag that indicates whether the command is still outstanding
*/
	private boolean isDone = true;  // initially waitTillDone returns immediately
/**
* A one-up id number, for debugging purposes
*/
	static private int nextIdNumber;
/**
* A one-up id number, for debugging purposes
*/
	protected int idNumber;
/**
* The exec reporter
*/
	static protected Reporter reporter;
/**
* The exec 
*/
	static protected ExecCallbacks exec;
/**
* The id of the initiating tasker
*/
	protected Tasker tasker;
/**
* Who will execute the command
*/
	protected Doer doer;
/**
* What is the command
*/
	protected String command;
/**
* The product that is the argument for a SEND_PRODUCT
*/
	protected Product productArg;
/**
* The sink that is the argument for a GET_PRODUCT
*/
	protected Sink sinkArg;
/**
* The put token for SEND_PRODUCT and GET_PRODUCT
*/
	protected Task putToken;
/**
* Create a generic command
*/
	public Command (Tasker tasker, Doer doer, String command) {
		idNumber = ++nextIdNumber;
		this.doer = doer;
		this.command = command;
		this.tasker = tasker;
/*		
		if (reporter.isDebugKey(ExecGlobal.TRACE_MSG_FLOW)) { //avoid unneeded complex string building
			reporter.debugMsg(this,"new (), ",toShortString(),null,ExecGlobal.TRACE_MSG_FLOW);
		}
*/
	}
/**
* Allows the exec to identify itself.  This must be done
* before any instances are created and used.
*/
	public static void setExec(ExecCallbacks exec) {
		Command.exec = exec;
		Command.reporter = exec.getExecAsAgent().getReporter();
	}
/**
* @return the tasking agent - the one that gets the done signal
*/
	public Tasker getTasker() {
		return tasker;
	}
/**
* @return the agent that will execute this command
*/
	public Doer getDoer(){
		return doer;
	}

/**
* @return the Product argument on SEND_PRODUCT
*/
	public Product getProduct(){
		return productArg;
	}
/**
* @param the Product argument on SEND_PRODUCT
*/
	synchronized public void setProduct(Product p){
		productArg = p;
	}
/**
* @return the put token argument for GET_PRODUCT and SEND_PRODUCT
*/
	public Task getPutToken(){
		return putToken;
	}
/**
* @return the Product argument on SEND_PRODUCT
*/
	public Sink getSink(){
		return sinkArg;
	}

/**
* the command as a string
*/
	public String getCommand(){
		return command;
	}

/**
* the object as a string
*/
	public String toShortString(){
		if (command == SEND_PRODUCT) {
			return "("+command+", product= "+productArg+", to=" + doer +" )";
		}
		else if (command == GET_PRODUCT) {
			return "("+command+", from= "+doer+", deliverTo= "+sinkArg+")";
		}
		else
			return "("+command+", doer= "+doer+")";
	}
/**
* the object as a string
*/
	public String toLongString(){
		String myDoerString=null;
		if (doer == null)
			myDoerString = "<>";
		else if (doer instanceof Command) 
			myDoerString = ((Command)doer).toShortString();
		else
			myDoerString =  doer.toString();
		String myTaskerString=null;
		if (doer == null)
			myTaskerString = "<>";
		else if (doer instanceof Command) 
			myTaskerString = ((Command)tasker).toShortString();
		else
			myTaskerString =  tasker.toString();

		if (command == SEND_PRODUCT) {
			return "(Command "+command+" (doer: "+myDoerString+")(product: "+productArg+")(tasker: "+myTaskerString+"))";
		}
		else if (command == GET_PRODUCT) {
			return "(Command "+command+" (doer: "+myDoerString+")(sink: "+sinkArg+")(tasker: "+myTaskerString+"))";
		}
		else
			return "(Command "+command+" (doer: "+myDoerString+")(tasker: "+myTaskerString+"))";
	}
/**
* the object as a string
*/
	public String toString(){
		return "Command"+Integer.toString(idNumber)+" "+command;
	}
/**
* a string form suitable for displaying on admin lists
*/
	public String getAsDisplayString(){
		return doer.getId()+", "+command;
	}
/**
* Perform the command that we represent.
* First checks to see if this command has been accepted for execution
* by the Doer.  If a putToken was specified in the SEND_PRODUCT
* constructor variant then that token is used as the accepted command.
*
*/
	public void perform() {
		while( !doer.getTaskingPermission(this)) {
			reporter.debugMsg(this,"waiting for tasking permission",null,null,ExecGlobal.TRACE_MSG_FLOW);
			doer.waitTillIdle();
		}
		reporter.debugMsg(this,"doCommittedTask()",null,null,ExecGlobal.TRACE_MSG_FLOW);
		doer.doCommittedTask();

	}
/**
* perform the command that we represent
*/
	public void signalDone() {
		if (tasker == null) {
			reporter.debugMsg(this,"signalDone(), <no tasker to signal to>",null,null,ExecGlobal.TRACE_MSG_FLOW);
		}
		else {
			reporter.debugMsg(this,"signalDone() ",null,null,ExecGlobal.TRACE_MSG_FLOW);
			tasker.taskDone(this);
		}
		synchronized(this) {
			isDone = true;
			notifyAll();
		}
	}
/**
* Get the sequence of tasks
*/
	public Enumeration getTaskEnumeration(){
		throw new RuntimeException("not implemented");
	}

/**
* Determine the real sender given a task.
* This is not really general purpose but it works
* for the current implementation where we are tasked
* by a CommandParallelOutputs.
*
* It would also work if we were tasked directly by an Agent
*/
	public Agent identifyAgentSource() {
		int iterations = 10;
		Tasker tasker = getTasker();
		while (tasker != null && iterations-- > 0) {
			if (tasker instanceof Agent) {
				return (Agent) tasker;
			}
			if (tasker instanceof CompletionSinkN) {
				return ((CompletionSinkN) tasker).getParent();
			}
			
			if (tasker instanceof Task) {
				tasker = ((Task)tasker).getTasker();
			}
			else {
				break;
			}
		}
		
		throw new RuntimeException("Unable to resolve sending agent, can not execute");
		
	}
/**
* Wait till done.  If we wait too long (greater than 3 minutes) the
* parent module is automatically disabled.
*/
	public synchronized void waitTillDone(Module parentModule) {
		int iter =36; // wait about 3 minutes
		while(!isDone) {
			try {
				wait(5000);
			}
			catch (InterruptedException e) {
				if (isDone)
					return;
			}
			iter--;
			if ((iter % 12) == 1) {
				parentModule.getReporter().error("Waiting for asynchronous output to complete.");
			}
			if (iter <= 0) {
				parentModule.getReporter().error("Waited too long for asynchronous output to complete; Will now auto-disable");
				parentModule.setEnabled(false);
				return;
			}
		}
	}
/**
* Clear the wait flag
*/
	public synchronized void setBusy() {
		isDone = false;
	}

}
