/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.executive;


import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Task;

/**
* A Juggler base class for a Juggler that has schedules.
*
* @author Nazario Irizarry, Jr.
*/
final public class DoTaskHelper {
	Module parentAgent;

/**
* A reference to the tasker that has reserved the right to task us
*/
	protected Task committedTask;
/**
* A local reference to our executing state
*/
	private String moduleState = ExecGlobal.IDLE;


/**
* The helper app acts on behalf of a parent
*/
	public DoTaskHelper(Module parentAgent) {
		this.parentAgent = parentAgent;
	}
/**
* Get the token to issue a task.
* 
* This class will return false to anybody if its state is not idle 
* (getState() != ExecGlobal.IDLE).  If a requestor that already has permission requests
* again (while the agent is idle, then it will get true (permission).
*
* @param task is what is to be done
* @return false if not idle or if already committed to another task
*/
	public boolean getTaskingPermission(Task task){
		if (committedTask == task)
			return true;
		if (moduleState != ExecGlobal.IDLE) {
			return false;
		}
		if (committedTask != null) {
			return false;
		}
		synchronized(this) {
			if (committedTask != null) {
				return false;
			}
			committedTask = task;
			return true;
		}
	}
/**
* Release tasking permission.  Only the tasker with current permission
* can release tasking permission.
*/
	synchronized public void releaseTaskingPermission(Task task){
		if (committedTask == task) {
			committedTask = null;
		}
		else {
			parentAgent.getReporter().assert(false, "Attempt to release tasking permission, new task: ",task,"\n\tcommitted task: ",committedTask);
		}
	}
/**
* Set module state.  If the state == ExecGlobal.IDLE then clear the
* tasker reservation lock.
*/
	synchronized public void setState(String newState) {
		if (newState == null || newState.length() == 0) {
			parentAgent.getReporter().assert(false, "can't set state to null",null,null,null);
			newState = ExecGlobal.IDLE;
		}
		if (newState != moduleState) {
			moduleState = newState;
			parentAgent.stateMemento.setProperty(ExecGlobal.STATE, newState);
		}
	/*
	* set the state after clearing the reservation lock
	* in case there are side effects from observers of the memento change
	*/
		if (newState == ExecGlobal.IDLE) {
			if (committedTask != null) {
				try {
					Task theTask = committedTask;
					committedTask = null;
					theTask.signalDone();
				}
				catch (Exception e) {
					parentAgent.getReporter().reportException(e);
				}
			}
			notifyAll();
		}
	}

/**
* Wait a specified amount of time to go idle
*/
	synchronized void waitTillIdle(long delay) throws InterruptedException {
		if (!isIdle()) {
			wait(delay);
		}
	}
/**
* Get state
*/
	public String getState() {
		if (moduleState == ExecGlobal.IDLE && committedTask != null)
			return ExecGlobal.COMMITTED;
		return moduleState;
	}
/**
* Get the task that was committed to
*/
	public Task getCommittedTask() {
		return committedTask;
	}
/**
* Are you idle?  Can only be idle if runstate is idle and if
* no taskers have reserved the right to task.
*
* @return true if idle
*/
	public boolean isIdle(){
		if (moduleState != ExecGlobal.IDLE) {
			return false;
		}
		if (committedTask != null) {
			return false;
		}
		return true;
	}
/**
* Checks the specified putToken
* to see if it is valid, before putting the product.
* This prevents asyncrhonouse sources from sneaking product
* in while the sink is in the middle of another put or task.
*
* @param product to put
* @param the token that grants authorization to put
*/
	public void pleasePutProduct(Product product, Task token){
		if (committedTask != token) {
			parentAgent.getReporter().assert(false,"a request to put product without obtaining task permission, authorized task=",committedTask," request=",token);	
			return;
		}
		parentAgent.putProduct(product);
	}
}