/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.executive;

import java.util.Enumeration;
import java.util.Hashtable;

import org.mitre.pam.store.IntegrityException;
import org.mitre.pam.interfaces.*;
import org.mitre.pam.store.SourceSinkInfo;

/**
* This level of the exec implements an api with which
* to change/manage configurations.  The code that implements
* this package assumes that there is only one editor.  There
* is no synchronization.<p>
*
* @author Nazario Irizarry, Jr.
*/
class Exec04 extends Exec02 implements ExecToEditor{
	java.util.Vector listeners;
	
	private static final String FALSE = "false";
	private static final String LOG_CONNECTION_EVENTS = "logConnectionLifeEvents";
	private static final String LOG_NODE_EVENTS = "logNodeLifeEvents";
	
	/**
	* Create a node if it does not exist, modify it if it does. 
	*
	* @param nodeType kind of iPAM agent
	* @param nodeId of agent
	* @param attributes may be null to start with default properties
	*
	* @exception IntegrityException may be generated if a node of the same id already exists
	*/
	public void declareNode(String nodeType, String nodeId, Hashtable attributes) 
			throws IntegrityException
	{
		if (nodeType.equals("main"))
			return;
		if (nodeType.equals(SourceSinkInfo.PIM_STR))
			return;
		if (nodeType.equals(SourceSinkInfo.DIM_STR))
			return;
		if (nodeType.equals(SourceSinkInfo.JUGGLER_STR))
			return;
		if (nodeType.equals(SourceSinkInfo.DAEMON_STR))
			return;
		if (nodeType.equals(SourceSinkInfo.MODULE_STR))
			return;
			
		if (nodeType.equals("exec")) {
			if (attributes != null)
				setAttributesFromHashtable(attributes);
			return;
		}
		
		String className = sourceSinkInfo.getClassNameFromCommonName(nodeType);
		if (className == null) {
			reporter.warning("Ignoring creation of kind="+nodeType+",  id="+nodeId+" because can't resolve item class name");
			return;
		}
	/*
	* Look for existing instance of the item
	*/
		try {
			boolean newModule = false;
			
			Module module = (Module)getActiveAgent(nodeId);
			if (module == null) {
					Class itemClass = Class.forName(className);
					module = (Module)itemClass.newInstance();
					module.setKind(nodeType);
					module.setId(nodeId); // Module assumes that setId() happens after setKind()
					newModule = true;
					
					
					if (module instanceof DIM) {
						DIMList.put(nodeId, module);
					}
					else if (module instanceof PIM) {
						module.setFlowDispatcher(queueDispatcher);
						PIMList.put(nodeId, module);
					}
					else if (module instanceof Juggler) {
						module.setFlowDispatcher(queueDispatcher);
						jugglerList.put(nodeId, module);
						jugglerScheduler.add( (Schedulable) module);
					}
					else if (module instanceof Handler) {
						module.setFlowDispatcher(queueDispatcher);
						jugglerList.put(nodeId, module);
					}
					else if (module instanceof Daemon) {
						daemonList.put(nodeId, module);
						jugglerScheduler.add((Schedulable) module);
					}
					else {
						reporter.warning(" exec can't create "+className+" because it's not PIM, DIM, Juggler, or Daemon"); 
						newModule = false;
					}
			}
			if (attributes != null && module != null) {
				module.setAttributesFromHashtable(attributes);
				}
			if (newModule) {
				if (module instanceof AsynchronousAgent) {
					asynchronousAgentList.put(nodeId,module);
					synchronized (asynchronousAgentList) {
						if (asyncAgentsBulkStarted)
							((Module)module).execSetAsync(true);
					}
				}
				notifyListenersOfNewNode(module);
			}
		}
		catch (ClassNotFoundException e) {
			reporter.reportException(e);
		}
		catch (IllegalAccessException e) {
			reporter.reportException(e);
		}
		catch (InstantiationException e) {
			reporter.reportException(e);
		}
	}
			
	/**
	* Delete a node with the specified id.  Uses deleteNode(Agent item).
	* no exception thrown.
	* @exception IntegrityException reserved for future, currently, none is thrown
	*/
	public void deleteNode(String nodeId) 
			throws IntegrityException
	{
//		Agent item = findAgentWithIntegrity(nodeId);
		Agent item = getActiveAgent(nodeId);
		try {
			if (item != null) 
				deleteNode(item);
		}
		catch (IntegrityException e) {
			reporter.reportException(e);
		}
		if (1==0)
			throw new IntegrityException("no way");
	}

	/**
	* Connect nodes.
	* @exception IntegrityException if source or sink can't be found
	*/
	public void connectNodes(String fromId, String toId) 
			throws IntegrityException
	{
		connectNodes(findSourceWithIntegrity(fromId), findSinkWithIntegrity(toId));
	}


	/**
	* Disconnect two nodes (modules).<p>
	*
	* @exception IntegrityException thrown if the fromId is not a Source or toId is not a Sink
	*/
	public void disconnectNodes(String fromId, String toId) 
			throws IntegrityException
	{
		Agent from = getActiveAgent(fromId);
		Agent to =  getActiveAgent(toId);
		
		if (!(from instanceof Source))
			throw new IntegrityException("Request to disconnect "+fromId+" fails because it is not a Source");

		if (!(to instanceof Sink))
			throw new IntegrityException("Request to disconnect "+toId+" fails because it is not a Sink");


		if (from != null && to != null)
			disconnectNodes((ConnectedAgent)from, (ConnectedAgent)to);
		
	}


	/**
	* Delete all nodes (modules) and their connections.  The nodes are first disabled and then
	* allowed to go idle. <p>
	*
	* @exception IntegrityException is currently a placeholder (not really generated)
	*/
	public void deleteAll() 
			throws IntegrityException
	{
		for (int i = 0; i<moduleList.length; i++) {
			Enumeration enum = moduleList[i].elements();
			while (enum.hasMoreElements()) {
				Agent item = (Agent) enum.nextElement();
				deleteNode(item);
			}
		}
	}

	/**
	* Delete all nodes (modules) and their connections except for daemons.
	* The nodes are first disabled and then
	* allowed to go idle. <p>
	*
	* @exception IntegrityException is currently a placeholder (not really generated)
	*/
	public void deleteAllButDaemons() 
			throws IntegrityException
	{
		for (int i = 0; i<moduleList.length; i++) {
			Enumeration enum = moduleList[i].elements();
			while (enum.hasMoreElements()) {
				Agent item = (Agent) enum.nextElement();
				if (item instanceof Daemon)
					break;
				deleteNode(item);
			}
		}
	}

	/**
	* Delete all nodes (modules) that have the specified attribute with the specified
	* value.  This allows the editor to set up logical groups of various kinds.
	* The nodes are first disabled and then
	* allowed to go idle. <p>
	*
	* @exception IntegrityException is currently a placeholder (not really generated)
	*/
	public void deleteGroup(String attribute, String value) 
			throws IntegrityException
	{
		if (attribute == null || value == null)
			throw new IntegrityException("null argument to Exec04.deleteGroup");

		for (int i = 0; i<moduleList.length; i++) {
			Enumeration enum = moduleList[i].elements();
			while (enum.hasMoreElements()) {
				Agent item = (Agent) enum.nextElement();
				
				String match = item.getProperty(attribute);
				if (match != null && match.equals(value)) {
					deleteNode(item);
				}
			}
		}
	}

	/**
	* Enable or disable all nodes (modules) that have the specified 
	* attribute with the specified
	* value.  This allows the editor to enable or disable all members of a group. <p>
	*
	* @param attribute the name of the module attribute that defines the group
	* @param value the value of the module attribute that defines the group
	* @param enableState true to enable; false to disable
	*
	*/
	public void setGroupEnabled(String attribute, String value, boolean enableState)
	{
		for (int i = 0; i<moduleList.length; i++) {
			Enumeration enum = moduleList[i].elements();
			while (enum.hasMoreElements()) {
				Agent item = (Agent) enum.nextElement();
				
				String match = item.getProperty(attribute);
				if (match != null && match.equals(value)) {
					item.setEnabled(enableState);
				}
			}
		}
	}

	/**
	* Enable or disable a list of nodes, probably in preparation for deletes.
	* All nodeIds must exist in order for any to be enabled or disabled.
	*
	* @exception IntegrityException if one of the nodeIds does not exist.
	*/
	public void setNodesEnabled(String nodeIds[], boolean enabled) 
			throws IntegrityException
	{
		if (nodeIds.length <= 0)
			return;
			
		Agent[] agents = new Agent[nodeIds.length];
		
		for (int i = 0; i<nodeIds.length; i++) {
			agents[i] = findAgentWithIntegrity(nodeIds[i]);
		}

		for (int i = 0; i<nodeIds.length; i++) {
			agents[i].setEnabled(enabled);
		}
		
		
	}


/* ************ *********** */

	/**
	* Connect nodes.
	* @exception IntegrityException is currently a placeholder (not really generated)
	*/
	protected void connectNodes(Agent source, Agent sink) 
			throws IntegrityException
	{
		if (source == null || sink == null) {
			reporter.assert(false, "bad args, sink=",sink.toString()," source=",source.toString());
			throw new IntegrityException("Can't connect, Exec.connectNodes bad args, sink="+sink+" source="+source);
		}
		if (!(source instanceof Source)) {
			reporter.assert(false, "bad args, first arg must be of type Source - ",source.toString(),null,null);
			throw new IntegrityException("Can't connect, "+source+" was expected to be a product Source");
		}
		if (!(sink instanceof Sink)) {
			reporter.assert(false, "bad args, second arg must be of type Sink - ",sink.toString(),null,null);
			throw new IntegrityException("Can't connect, "+sink+" was expected to be a product sink");
		}
		
		ConnectedAgent cSink = (ConnectedAgent)sink;
		ConnectedAgent cSource = (ConnectedAgent)source;

		Object otherEnd1 = cSink.getSourceList().get(source.getId());
		Object otherEnd2 = cSource.getSinkList().get(sink.getId());
		
		/*
		* don't add connection if it already exists
		*/
		if (otherEnd1 == source && otherEnd2 == sink)
			return;

		cSink.getSourceList().put(source.getId(), source);
		cSource.getSinkList().put(sink.getId(), sink);
		if (cSink instanceof WantsConnectionChanges) {
			try {
				((WantsConnectionChanges)cSink).aSourceWasAdded(cSource);
			} catch (Exception e) {
				reporter.reportException(e);
			}
		}
		if (cSource instanceof WantsConnectionChanges) {
			try {
				((WantsConnectionChanges)cSource).aSinkWasAdded(cSink);
			} catch (Exception e) {
				reporter.reportException(e);
			}
		}
		
		notifyListenersOfConnectionMade(cSource, cSink);
	}

	/**
	* Disconnect a source to sink connection. <p>
	*
	* @exception IntegrityException is currently a placeholder (not really generated)
	*/
	protected void disconnectNodes(ConnectedAgent source, ConnectedAgent sink) 
			throws IntegrityException
	{
		if (source == null || sink == null) {
			reporter.assert(false, "Exec.connectNodes bad args, sink=",sink.toString()," source=",source.toString());
			throw new IntegrityException("Can't connect, Exec.connectNodes bad args, sink="+sink+" source="+source);
		}
			
//		boolean sourceEnable = (Module) source.isEnabled();
//		boolean sinkEnable = (Module) sink.isEnabled();
		
		ConnectedAgent cSink = (ConnectedAgent) sink;
		ConnectedAgent cSource = (ConnectedAgent) source;
		cSink.getSourceList().remove(source.getId());
		cSource.getSinkList().remove(sink.getId());

		if (cSink instanceof WantsConnectionChanges) {
			try {
				((WantsConnectionChanges)cSink).aSourceWasRemoved(cSource);
			} catch (Exception e) {
				reporter.reportException(e);
			}
		}
		if (cSource instanceof WantsConnectionChanges) {
			try {
				((WantsConnectionChanges)cSource).aSinkWasRemoved(cSink);
			} catch (Exception e) {
				reporter.reportException(e);
			}
		}

		notifyListenersOfConnectionBroken(source, sink);
	}
	/**
	* Delete a node (module).  First disable node, then wait for it to go idle
	* before deleting and disconnecting it.
	* @exception IntegrityException if there is no such node
	*/
	public void deleteNode(Agent item) 
			throws IntegrityException
	{
		item.setEnabled(false);
		item.waitTillIdle();	// wait for it to go idle
		
		if (item instanceof AsynchronousAgent) {
			try {
				((Module)item).execSetAsync(false);
			}
			catch (Exception e) {
				reporter.reportException(e);
			}
			asynchronousAgentList.remove(item.getId());
		}

		if (item instanceof Source) {
			ConnectedAgent source = (ConnectedAgent) item;
			
			Enumeration sinks = ((Hashtable)source.getSinkList()).elements();
			while (sinks.hasMoreElements()) {
				disconnectNodes(source, (ConnectedAgent) sinks.nextElement());
			}
		}

		if (item instanceof Sink) {
			ConnectedAgent sink = (ConnectedAgent) item;
			
			Enumeration sources = ((Hashtable)sink.getSourceList()).elements();
			while (sources.hasMoreElements()) {
				disconnectNodes((ConnectedAgent) sources.nextElement(), sink);
			}
		}
		

		if (item instanceof PIM) {
			PIMList.remove(item.getId());
		}
		else if (item instanceof DIM) {
			DIMList.remove(item.getId());
		}
		else if (item instanceof Juggler) {
			jugglerList.remove(item.getId());
			jugglerScheduler.remove((Juggler)item);
		}
		else if (item instanceof Handler) {
			jugglerList.remove(item.getId());
		}
		else if (item instanceof Daemon) {
			daemonList.remove(item);
			jugglerScheduler.remove((Daemon)item);
		}
		
		if (item instanceof WantsConnectionChanges) {
			try {
				((WantsConnectionChanges)item).youHaveBeenDiscarded();
			} catch (Exception e) {
				reporter.reportException(e);
			}
		}

		notifyListenersOfDeletedNode(item);

	}
/**
* Disconnect two items, if they are not connected then nothing happens.  NOT IMPLEMENTED
*/
	public void disconnectItems(String fromKind, String fromId, String toKind, String toId) 
	{
/*
		if (getState() == STARTING)
			connectItemsInitially(fromKind, fromId, toKind, toId);
*/
	}
/**
* Delete an item, but only if it has no connections.  If it has connections
* an integrity exception is thrown.  NOT IMPLEMENTED
*/
	public void deleteItem(String kind, String id, Memento memento) throws IntegrityException{
/*
		if (getState() == STARTING)
			createNodeInitially(kind, id, memento);
		else
			createItemLater(kind, id, memento);
*/
	}
/**
* Allows a controller to define aliases for pims, dims, or jugglers
*
* @param alias the name to be used as an alias (has pdif visibility)
* @param javaClassName must match the javaClassName, used for dynamic loading
*/
	public void defineModuleAlias(String alias, String javaClassName){
		try {
			sourceSinkInfo.defineModuleAlias(alias, javaClassName);
		}
		catch (Exception e) {
			reporter.reportException(e);
		}
	}
/**
* Add listener to monitor config events
*/
	public void addConfigurationListener(ConfigurationListener listener){
		if (listeners == null) {
		    listeners = new java.util.Vector();
		}
		if (!listeners.contains(listener))
			listeners.addElement(listener);
	}
/**
* Remove listener 
*/
	public void removeConfigurationListener(ConfigurationListener listener){
		if (listeners == null) {
		    return;
		}
		listeners.removeElement(listener);
		if (listeners.size() == 0)
			listeners = null;
	}
/**
* Notify listeners
*/
	private void notifyListenersOfNewNode(Agent agent) {
		reporter.debugMsg("created node -",agent,null,null,LOG_NODE_EVENTS);
		java.util.Vector targets;
		synchronized (this) {
		    if (listeners == null) {
		    	return;
		    }
		    targets = (java.util.Vector) listeners.clone();
		}
		ConfigurationEvent ce = new ConfigurationEvent(this, ConfigurationEvent.NODE_CREATED, agent);
	
		for (int i = 0; i < targets.size(); i++) {
		    ConfigurationListener target = (ConfigurationListener)targets.elementAt(i);
		    try {
		    	target.nodeCreated(ce);
		    }
		    catch (Exception e) {
		    	reporter.reportException(e);
		    }
		}
	}
/**
* Notify listeners
*/
	private void notifyListenersOfDeletedNode(Agent agent) {
		reporter.debugMsg("removed node -",agent,null,null,LOG_NODE_EVENTS);
		java.util.Vector targets;
		synchronized (this) {
		    if (listeners == null) {
		    	return;
		    }
		    targets = (java.util.Vector) listeners.clone();
		}
		ConfigurationEvent ce = new ConfigurationEvent(this, ConfigurationEvent.NODE_DELETED, agent);
	
		for (int i = 0; i < targets.size(); i++) {
		    ConfigurationListener target = (ConfigurationListener)targets.elementAt(i);
			try {
		    	target.nodeDeleted(ce);
		    }
		    catch (Exception e) {
		    	reporter.reportException(e);
		    }
		}
	}
/**
* Notify listeners
*/
	private void notifyListenersOfConnectionMade(ConnectedAgent agentFrom, ConnectedAgent agentTo) {
		reporter.debugMsg("created connection\tfrom\t",agentFrom,"\tto\t",agentTo,LOG_CONNECTION_EVENTS);
		java.util.Vector targets;
		synchronized (this) {
		    if (listeners == null) {
		    	return;
		    }
		    targets = (java.util.Vector) listeners.clone();
		}
		ConfigurationEvent ce = new ConfigurationEvent(this, ConfigurationEvent.CONNECTION_MADE, agentFrom, agentTo);
	
		for (int i = 0; i < targets.size(); i++) {
		    ConfigurationListener target = (ConfigurationListener)targets.elementAt(i);
		    try {
		    	target.connectionMade(ce);
		    }
		    catch (Exception e) {
		    	reporter.reportException(e);
		    }
		}
	}
/**
* Notify listeners
*/
	private void notifyListenersOfConnectionBroken(ConnectedAgent agentFrom, ConnectedAgent agentTo) {
		reporter.debugMsg("removed connection\tfrom\t",agentFrom,"\tto\t",agentTo,LOG_CONNECTION_EVENTS);
		java.util.Vector targets;
		synchronized (this) {
		    if (listeners == null) {
		    	return;
		    }
		    targets = (java.util.Vector) listeners.clone();
		}
		ConfigurationEvent ce = new ConfigurationEvent(this, ConfigurationEvent.CONNECTION_BROKEN, agentFrom, agentTo);
	
		for (int i = 0; i < targets.size(); i++) {
		    ConfigurationListener target = (ConfigurationListener)targets.elementAt(i);
		    try {
		    	target.connectionBroken(ce);
		    }
		    catch (Exception e) {
		    	reporter.reportException(e);
		    }
		}
	}
}