/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.executive;

import java.util.Stack;
import java.util.Vector;
import java.util.Hashtable;
import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.interfaces.Doer;
import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.interfaces.Task;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.executive.ExecGlobal;


/**
* Has a fixed number of threads from which to execute commands. <p>
*
* @author Nazario Irizarry, Jr.
*/
final class Executer extends Thread {
/**
* a count of threads available to process commands
*/
	private int freeThreadCount;
/**
* flag used to synchronize threads waiting for a command
*/
	private CommandProxy commandAvailableFlag = new CommandProxy();
/**
* flag used to synchronize threads waiting for a command
*/
	private Object commandTakenFlag = new Object();
/**
* a reference to the exec
*/
	private ExecCallbacks exec;
/**
* A reference to an agent with log callbacks that we can use to report things
*/
	private Reporter reporter;
/**
* Name of executor
*/
	private String name;
/**
* the state of the executor, mirrors the state of the exec
*/
	private String state = ExecGlobal.STARTING;
	private long sleepInterval = 1003;	
/**
* The current threadList
*/
	private Vector threadList = new Vector(10,10);
/**
* Desired thread count
*/
	private int desiredThreadCount = 2;
/**
* Temporary additional thread count
*/
	private int temporaryAdditionalThreadCount;
/**
* Used to assign unique ids to threads of this executor
*/
	private static int threadId = 100;
/**
* Fow now this is just a simple vector.  As the numbers grow
* a more sophisticated scheme may be required.
*/
	private Vector commandList = new Vector(100,100);
/**
* Update flag
*/
	int updateFlag = 0;
/**
* Limit on the number of doers that may run concurrently
*/
	private int concurrencyLimit;
/**
* A table to keep track of concurrent execution
*/
	private Hashtable concurrencyTable;
	
	private boolean keepRunning;
/**
* On creation create a starting number of threads
*/
	public Executer(ExecCallbacks exec, String name, int desiredThreadCount) {
		super(name);
		this.reporter = exec.getExecAsAgent().getReporter();
		this.exec = exec;
		this.name = name;
		this.desiredThreadCount = desiredThreadCount;
		keepRunning = true;
		concurrencyLimit = 1;
		concurrencyTable = new Hashtable();
		start();
	}
	
/**
* Add an item to execute
*/
	public void add(Task cmd) {
	/*
	* if adding a Task.SEND_PRODUCT Command then increase the reference count.
	* it will be decremented after the command is performed 
	* in ExecuterThread.executeCommandChain()
	*/
		if (cmd.getCommand() == Task.SEND_PRODUCT) {
			cmd.getProduct().incrementUsage(this);
		}
		
		int reportIterations = 0;
	/*
	* only allow a certain number of outstanding elements
	*/
		int min = 5;
		int upperLimit = temporaryAdditionalThreadCount + 3*desiredThreadCount;
		if (upperLimit > min)
			min = upperLimit;
			
		while (commandList.size() > min) {
			if (reportIterations-- <= 1) {
				reporter.log("waiting for que slots to open");
				reportIterations = 3;
			}
			try {
				sleep(5031);
			}
			catch (InterruptedException e) {}
		}
		
		reporter.debugMsg(this,"added",cmd,null,ExecGlobal.TRACE_MSG_FLOW);		

		commandList.addElement(cmd);
		updateFlag++;
	}

/**
* Set the desired number of threads.  
*/
	public void setDesiredThreadCount(int desiredThreadCount) {
		this.desiredThreadCount = desiredThreadCount;
	}
/**
* Increase the temporary thread count
*/
	public synchronized void incrementTempThreadCount() {
		temporaryAdditionalThreadCount++;
		reporter.debugMsg("+ temp thread count = ",Integer.toString(temporaryAdditionalThreadCount),
				" desired thread count = ",Integer.toString(desiredThreadCount),
				 ExecGlobal.TRACE_MSG_FLOW);
	}
/**
* Decrease the temporary thread count
*/
	public synchronized void decrementTempThreadCount() {
		temporaryAdditionalThreadCount--;
		reporter.debugMsg("- temp thread count = ",Integer.toString(temporaryAdditionalThreadCount),
				" desired thread count = ",Integer.toString(desiredThreadCount),
				 ExecGlobal.TRACE_MSG_FLOW);
	}
/**
* Increase or decrease thread count to achieve goals
*/
	void adjustThreadCount() {
		int totalThreadCount = temporaryAdditionalThreadCount + desiredThreadCount;
		if (totalThreadCount == threadList.size())
			return;
		reporter.debugMsg("#, total thread count = ",Integer.toString(totalThreadCount),
				null,null, ExecGlobal.TRACE_MSG_FLOW);
			
	/*
	* Remove threads
	*/
		while ( threadList.size() > totalThreadCount) {
			int indexOfThread = threadList.size() -1;
			ExecuterThread t = (ExecuterThread) threadList.elementAt(indexOfThread);
			threadList.removeElementAt(indexOfThread);
			t.executeNoMore();
		}
	/*
	* Add more threads.
	* Each thread has a name that starts with P or D
	* assuming the name of the executer is either "PIM Executer" or "DIM Executer"
	*/
		while ( threadList.size() < totalThreadCount) {
			ExecuterThread t = new ExecuterThread(this, name.charAt(0)+Integer.toString(threadId++), exec, commandAvailableFlag);
			threadList.addElement(t);
		}
		updateFlag++;
	}
/**
* Get active thread count
*/
	public int getCurrentThreadCount() {
		return threadList.size();
	}
/**
* The run thread wakes up periodically and
* pulls items off of the queue for execution.
*/
	public void run() {
//		reporter.log(name+" running");
		while ( (exec.getState() != ExecGlobal.SHUTTING_DOWN) && keepRunning ) {
		/*
		* Adjust thread count
		*/
			adjustThreadCount();
		/*
		* Pause and do maintenance
		*/
			try {
				setState(ExecGlobal.IDLE);
				if (updateFlag > 0) {
					updateFlag = 0;
					exec.update(null, name);
					}
					
				sleep(sleepInterval);
			}
			catch (InterruptedException e) {}
			
			
		/*
		* Execute items unless we are shutting down.
		* We continue to execute even if the Exec is idle so that
		* any outstanding jugglers may complete if they were waiting for
		* PIMS or DIMS on this executer.  Most likely the exec is idle
		* because it wants to checkpoint and it is just waiting for
		* everything to go idle.  Therefore we must let the PIMs
		* and DIMs complete so that the jugglers will go idle.  Otherwise
		* the juggler would be waiting, not idle.
		*/
			String execState;
			if ( ((execState = exec.getState()) != ExecGlobal.SHUTTING_DOWN) && (execState != ExecGlobal.STARTING) ) {
				Task nextTask;
				while ( (freeThreadCount > 0) && ( commandList.size() > 0) && ((nextTask = getNextCommand()) != null) ) {	
					reporter.debugMsg(this,"about to issue command",nextTask,null,ExecGlobal.TRACE_MSG_FLOW);		
					synchronized (commandTakenFlag) {
						synchronized (commandAvailableFlag) {
							commandAvailableFlag.command = nextTask;
							commandAvailableFlag.notify();
						}
						try {
							commandTakenFlag.wait();
						} catch (InterruptedException ignore) {}
					}
					setState(ExecGlobal.WORKING);
					updateFlag++;
				}
			}
			
		}
		setState(ExecGlobal.SHUTTING_DOWN);
		commandList = null;
		desiredThreadCount = 0;
		temporaryAdditionalThreadCount = 0;
		adjustThreadCount();
		reporter.log(name+" stopped");
	}
/**
* Get and remove next command atomically.
* The target Doer is checked.  If it is already executing
* it is bypassed for this check.
*/
	private Task getNextCommand(){
		int n = commandList.size();
		for (int i = 0; i < n; i++) {
			Task c = (Task) commandList.elementAt(i);
			Doer doer = c.getDoer();
			if (doer.getTaskingPermission(c)) {
				commandList.removeElementAt(i);
				return c;
			}
			else {
				reporter.debugMsg(this," command delayed to avoid concurrency conflict: ",c,null, ExecGlobal.TRACE_MSG_FLOW);
			}		
		}
		return null;
	}
/**
* Used by ExecuterThreads to signal that they got the command.
* The body is internally synchronized so only one
* thread can do this at a time.
*/
	protected void commandTaken(){
		synchronized (commandTakenFlag) {
			decrementFreeCount();
			commandTakenFlag.notify();
		}
	}
/**
* set stateof the executer
*/
	protected void setState(String state) {
		this.state = state;
	}
/**
* get state of the executer
*/
	public String getState() {
		return state;
	}
/**
* set the limit on the number of doers that may run concurrently
*/
	public void setConcurrencyLimit(int n) {
		this.concurrencyLimit = concurrencyLimit;
	}

	public boolean isIdle(){
		
		return (state == ExecGlobal.IDLE || state == ExecGlobal.SHUTTING_DOWN);
	}
/**
* Are you idle?
*/
	public void waitTillIdle(){
		
		if (isIdle())
			return;
		
		while (!isIdle()) {
			try {
				Thread.sleep(2171);
			}
			catch (Exception ignore) {}
		}
	}

/**
* get a list of what is currently running
*
*@param whatsRunning is an empty input vector which is filled on return
*@param whatsHolding is an empty input vector which is filled on return
*/
	public void getQueueLists(Vector whatsRunning, Vector whatsHolding) {
//		synchronized (commandList) {
			for (int i=0; i<commandList.size(); i++) {
				Task cmd = (Task) commandList.elementAt(i);
				whatsHolding.addElement(cmd.getAsDisplayString());
			}
			for (int i=0; i<threadList.size(); i++) {
				String s = ((ExecuterThread) threadList.elementAt(i)).getWhatsRunning();
				if ( s != ExecGlobal.IDLE ) {
					whatsRunning.addElement(s);
				}
			}
			
//		}
	}
/*
* For debug use
*/
	public String toString() {
		return name;
	}
/**
* killThreads should only be called during executive shutdown.  Once it
* is called it cannot be restarted.
*/
	public  void killThreads() {
		keepRunning = false;
	}
/**
* Increment the free thread count,
* This method should only be called by
* ExecuterThreads because they take care to synchronize access
*/
	void incrementFreeCount() {
		freeThreadCount++;
		reporter.debugMsg(this,"+ free count",null,null,ExecGlobal.TRACE_MSG_FLOW);		
	}
/**
* Decrement the free thread count
* This method should only be called by
* ExecuterThreads because they take care to synchronize access
*/
	void decrementFreeCount() {
		reporter.debugMsg(this,"- free count",null,null,ExecGlobal.TRACE_MSG_FLOW);		
		if (freeThreadCount <= 0)
			reporter.assert(false,"free thread count already 0",null,null,null);
		freeThreadCount--;
		if (freeThreadCount <= 0)
			reporter.debugMsg(this,"0 free count=",Integer.toString(freeThreadCount),null,ExecGlobal.TRACE_MSG_FLOW);	
	}

	static class CommandProxy {
		Task command;
	}
}

