/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.executive;

import java.util.Stack;

import org.mitre.pam.interfaces.Doer;
import org.mitre.pam.interfaces.Tasker;
import org.mitre.pam.interfaces.Task;
import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.interfaces.ExecThread;
import org.mitre.pam.executive.ExecGlobal;


/**
* Knows how to run/dispatch commands. When
* the command is complete this thread goes to sleep
* for a little while. <p>
*
* @author Nazario Irizarry, Jr.
*/

final class ExecuterThread extends Thread implements ExecThread{

	private Executer monitor;
	private ExecCallbacks exec;
	private String name;
	private boolean keepRunning;
	private boolean busy;
	private String whatsRunning;
	private Executer.CommandProxy waitFlag;
	private Reporter reporter;


	ExecuterThread (Executer monitor, String name, ExecCallbacks exec, Executer.CommandProxy waitFlag) {
		super(name);
		this.waitFlag = waitFlag;
		this.monitor = monitor;
		this.exec = exec;
		this.keepRunning = true;
		this.reporter = exec.getExecAsAgent().getReporter();
		setWhatsRunning(ExecGlobal.IDLE);
		start();
	}
	
/**
* When we run we send a command to a doer.
* When done we check for housekeeping chores.
*/
	public void run() {
		Task mainCommand;
		do {
			mainCommand = null;
			clearTasks();
			synchronized (waitFlag) {
				monitor.incrementFreeCount();
				try {
					waitFlag.wait();
					mainCommand = waitFlag.command;
					waitFlag.command = null;
				} catch (InterruptedException ignore) {}
				monitor.commandTaken(); // also decrementsFreeCount as side effect
			}
			reporter.debugMsg(getName(), "accepted cmd", mainCommand, null, ExecGlobal.TRACE_MSG_FLOW);		
			executeCommandChain(mainCommand);
		} while (keepRunning && (exec.getState() != ExecGlobal.SHUTTING_DOWN));
	/*
	* On shutdown do a bit of cleanup
	*/
		setWhatsRunning(ExecGlobal.SHUTTING_DOWN);
		monitor.updateFlag++;
		monitor = null;
		exec = null;
	}
/**
* This used to execute a sequence of commands
* but now it only executes one command since commands are
* no longer chained.
*/
	private void executeCommandChain(Task command) {
		if (command == null)
			return;
		try {
			setWhatsRunning(command.getAsDisplayString());
			reporter.debugMsg(getName(),"will run",command.getAsDisplayString(),null,ExecGlobal.TRACE_MSG_FLOW);	
			command.perform();
		/*	this is where we would recursively executeCommandChain() */
		}
		catch (Error e) {
			reportException(e);
			command.getDoer().releaseTaskingPermission(command);
			command.signalDone();
		}
		/*
		* if a Task.SEND_PRODUCT Command then decrease the reference count
		* that was previously incremented by Executor.add()
		*/
		finally {
			if (command.getCommand() == Task.SEND_PRODUCT) {
				command.getProduct().decrementUsage(monitor);
			}
		}
	}
/**
* Cause thread to stop after it completes whatever is currently running
*/
	void executeNoMore() {
		keepRunning = false;
	}

/**
* reset state on task completion
*/
	public synchronized void clearTasks() {
		setWhatsRunning(ExecGlobal.IDLE);
		monitor.updateFlag++;
	}
/**
* Get the description of what is currenty running
*/
	public String getWhatsRunning() {
		return this.whatsRunning;
	}
/**
* Set the description of what is currenty running
*/
	private void setWhatsRunning(String what) {
		this.whatsRunning = what;
	}
	
	public String toString() {
		return "ExecutorThread "+getName();
	}
/**
* Report on an exception
*/
	private void reportException(Throwable e) {
		reporter.reportException(e);
	}
/**
* are we busy?
*/
	public synchronized boolean isBusy() {
		return busy;
	}
/**
* Set busy state
*/
	private synchronized void setBusy(boolean busy) {
		this.busy = busy;
	}
/**
* If this thread is put to sleep for more than
* 2 seconds then the Executor is
* told to temporarily increase its thread allocation so
* that the system does not come to a stop because
* all of the currently processing thread decide to pause. <p>
*
* Use this instead of Thread.sleep()
*/
	public void goToSleep(long milliseconds) throws InterruptedException {
		if (milliseconds > 2000) {
			try {
				monitor.incrementTempThreadCount();
				Thread.sleep(milliseconds);
			}
			finally {
				monitor.decrementTempThreadCount();
			}
		}
		else {
			Thread.sleep(milliseconds);
		}
	}
}
