/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.executive;

import java.io.*;
import java.util.Enumeration;
import java.util.Hashtable;

import org.mitre.pam.store.IntegrityException;
import org.mitre.pam.store.SourceSinkInfo;
import org.mitre.pdif.eventParser.ParseEvent3;
import org.mitre.pdif.eventParser.ParseEvent2;
import org.mitre.pdif.eventParser.ConnectorEnd;
import org.mitre.pdif.eventParser.BadPDIFException;
import org.mitre.pam.util.pdifHashtables.PDIFRead;
import org.mitre.pam.executive.PropertySet;
import org.mitre.pam.product.view.ProductStore;
import org.mitre.pam.interfaces.Memento;
import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.interfaces.ExecToEditor;
import org.mitre.pam.interfaces.ExecToMonitor;
import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.interfaces.Reporter;

import org.mitre.pam.interfaces.Product;



/**
* Saves and restores all iPAM state. <P>
*
* @author Nazario Irizarry, Jr.
*/
final class PAMLoad {
	private ExecToEditor exec;
	private Reporter reporter ;
	private Hashtable idsLoaded = new Hashtable(150);
	private String javaClassName;

	public PAMLoad (ExecToEditor exec) {
		this.exec = exec;
		this.reporter = exec.getExecAsAgent().getReporter();
	}

/**
* In order to load we need an external callback
* to hand off instances of properly prepared mementos.
* Load calls back for every new memento. <p>
*
*/
	public void load(Reader fin) 
	throws IntegrityException{
//		long startTime = System.currentTimeMillis();
//		System.out.println("start time: "+startTime);
		try {
						
			ParseEvent3 pEvent = new ParseEvent3(fin);
			
			try {
				while ( true ) {
					try {
						int type = pEvent.next();
						
						if (type == ParseEvent3.EOF)
							break;
											
						if (type == ParseEvent3.NODE_ENTER) {
//							System.out.println("node: "+pEvent.node.id);
							String kind = pEvent.node.kind.intern();
							String id = pEvent.node.id;
						/*
						* Define a new template
						*/
							if ((kind == SourceSinkInfo.PIM_STR) ||
								(kind == SourceSinkInfo.DIM_STR) ||
								(kind == SourceSinkInfo.JUGGLER_STR) ||
								(kind == SourceSinkInfo.DAEMON_STR) ||
								(kind == SourceSinkInfo.MODULE_STR)
								) {
								Hashtable initialAttributes = new Hashtable();
								javaClassName = null;
								loadHashtable(pEvent, initialAttributes, null);
								idsLoaded.put(id, initialAttributes);
								if (javaClassName == null || javaClassName.length() == 0) {
									throw new BadPDIFException("Attempt to define new "+pEvent.node.id+" without a java class name");
								}
								exec.defineModuleAlias(id, javaClassName);
							}
						/*
						* Define a new agent based on a template
						*/
							else {
								if (id == null  || id.equals(""))
									id = PropertySet.generateUid();	// ensure uid to keep memento's happy
								
								Hashtable initialAttributes = new Hashtable();
								Hashtable parentAttributes = (Hashtable) idsLoaded.get(kind);
								loadHashtable(pEvent, initialAttributes, parentAttributes);
								idsLoaded.put(id, initialAttributes);
								exec.declareNode(kind, id, initialAttributes);
							}
						}
						else if (type == ParseEvent3.CONNECTOR) {
							ConnectorEnd c1 = pEvent.connectorEnds[0];
							ConnectorEnd c2 = pEvent.connectorEnds[1];
							if (c1.role.equalsIgnoreCase("to")) {
								exec.connectNodes( c2.id, c1.id);
							}
							else if (c2.role.equalsIgnoreCase("to")) {
								exec.connectNodes( c1.id,  c2.id);
							}
						}
						else if (type == ParseEvent3.NON_NODE) {
							exec.deleteNode(pEvent.node.id);
						}
						else if (type == ParseEvent3.NON_CONNECTOR) {
							ConnectorEnd c1 = pEvent.connectorEnds[0];
							ConnectorEnd c2 = pEvent.connectorEnds[1];
							if (c1.role.equalsIgnoreCase("to")) {
								exec.disconnectNodes( c2.id, c1.id);
							}
							else if (c2.role.equalsIgnoreCase("to")) {
								exec.disconnectNodes( c1.id,  c2.id);
							}
						}
					}
					catch (IntegrityException e) {
						reporter.warning(e.toString());
					}
				}
			}
			catch (BadPDIFException e) {
				reporter.reportException(e);
			}
			
		}
		catch (IOException e) {
			reporter.reportException(e);
		}

		long endTime = System.currentTimeMillis();
//		System.out.println(" end time: "+endTime+",   elapsed: "+(startTime-endTime)/1000);

		idsLoaded.clear();
	}

/**
* Assumes that a PDIF input stream has been defined.
* Ignores MEMENTO_UID, MEMENTO_KIND, and any property that starts with "meta_".
*/
	
	private void loadHashtable(ParseEvent3 pEvent, Hashtable list, Hashtable parentAttributes){
		
		try {
			int type = pEvent.next();
			while ( type == ParseEvent3.NODE_ATTRIBUTE ) {
				if ( pEvent.attribute.name.startsWith("meta_") ) {
					if ( pEvent.attribute.name.equals(PDIFRead.META_NOT_REDUNDANT)) {
						if (parentAttributes != null) {
							Enumeration attrs = parentAttributes.keys();
							while (attrs.hasMoreElements()) {
								String attr = (String) attrs.nextElement();
								Object val = parentAttributes.get(attr);
								if (val != null) {
									list.put(attr, val);
//									System.out.println(" fillin attr="+attr+",   val="+val);
								}
							}
						}
					}

				} 
				else {
					String name = pEvent.attribute.name.intern();
					if ( name == ExecGlobal.STATE) {
					}
					else if ( name == Memento.MEMENTO_UID) {
					}
					else if ( name == Memento.MEMENTO_KIND) {
					}
					else if ( name == SourceSinkInfo.JAVA_CLASS_STR) {
						javaClassName = pEvent.attribute.getValue();
					}
					else {
						if (pEvent.attribute.getValueList().size() > 1) {
							list.put(name, pEvent.attribute.getValueList());
						}
						else
							list.put(name, pEvent.attribute.getValue());
					}
				}
				type = pEvent.next();
			}
		
		}
		catch (BadPDIFException e) {
			reporter.reportException(e);
		}
		catch (IOException e) {
			reporter.reportException(e);
		}		
	}
	
	public void validatePDIF(Reader in) throws BadPDIFException, IOException {
//		System.out.println("start val");
		long start = System.currentTimeMillis();
		
		ParseEvent2 pEvent = new ParseEvent2(in);
		pEvent.validatePDIF();
//		System.out.println(" end val, time sec = "+(System.currentTimeMillis() - start)/1000);
/*		while ( true ) {
			int type = pEvent.next();
			
			if (type == ParseEvent2.EOF )
				return;
		}
*/
	}

/**
* Load saved product, and purge unused files from the persistence directory.
*/
	void loadSavedProduct(File productFile) throws IntegrityException{
		ProductStore prodStore = new ProductStore();
		
		if (!productFile.exists())
			return;
		
		try {
			Reader productStream = new BufferedReader(new FileReader(productFile),500);
			prodStore.load(productStream);
		}
		catch (FileNotFoundException e) {}
		catch (IOException io) {
			reporter.reportException(io);
			return;
		}
	/*
	* First for jugglers, then for DIMs
	*/
		for (int i = 0; i<=3; i++) {
			Enumeration thingsWithProduct = null;
			
			if (i==0)
				thingsWithProduct = exec.getListOfActiveModules(ExecToMonitor.JUGGLER);
			else if (i==1)
				thingsWithProduct = exec.getListOfActiveModules(ExecToMonitor.DIM);
			else if (i==2)
				thingsWithProduct = exec.getListOfActiveModules(ExecToMonitor.PIM);
			else 
				thingsWithProduct = exec.getListOfActiveModules(ExecToMonitor.DAEMON);
				
			while (thingsWithProduct.hasMoreElements()) {
				Agent mb = (Agent) thingsWithProduct.nextElement();
				String productId = mb.getProperty(ExecGlobal.PERSISTENT_PRODUCT);
			/*
			* some agents may not persist a product
			*/
				if (productId != null) {
					Product itsProduct = prodStore.getProduct(productId);
					if (itsProduct != null)
						mb.setPersistentProduct(itsProduct);				
				}
			}
		}
	}
 
}
