/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.executive;

import java.util.*;
import java.io.InputStream;
import java.io.IOException;
import java.io.Serializable;
import java.util.Enumeration;

import org.mitre.pam.interfaces.Memento;
import org.mitre.pam.interfaces.HasProperties;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.interfaces.IDictionary;
import org.mitre.pam.interfaces.MementoEvent;
import org.mitre.pam.interfaces.MementoListener;

import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pdif.eventParser.PDIFWriter;
import org.mitre.pdif.eventParser.ParseEvent3;
import org.mitre.pdif.eventParser.BadPDIFException;
import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.util.AttributeKey;

import org.mitre.pam.util.SortedHash;

/**
* Contains the persistent information
* for an object.  This memento can save and restore
* itself from a parenthesized stream.  However, it
* is designed to be used by PAMStore and therefore
* does not independently support the Load method.
* Note that property names are interned before they get
* held making it possible to do == on enumerations of property names.
*
* @author Nazario Irizarry, Jr.
* @author John R. Fantry - added Java serialization
*/
public class PropertySet implements Memento, Serializable {
    private transient java.util.Vector listeners;
    /*
    * Once upon a time we used to have a reference to our parent.
    * That posed the problem that if the parent was a Product
    * the products would not get garbage collected because of circular
    * reference between the product and its property set.
    * So instead we just hold the kind and id of the parent
    * which is all we really need
    */
    private transient String parentKind;
    private transient String parentId;
    private SortedHash list = new SortedHash();
    private transient IDictionary dictionary;
/**
* This is a constant value used to indicate that
* a property has been deleted.  It is a bit of a kludge
* but it does provide a way to positively identify properties to be deleted.
*/
    public final static String DELETED = "*1*~DELETED~*2*-*@kJd*"; 
    private final static ValueList emptyStringValueList = new ValueList("");
	protected static final ValueList DELETED_VALUE_LIST = new ValueList(DELETED);
/**
* used for generating unique ids
*/
	static Random rand = new Random(System.currentTimeMillis());
	private static Reporter reporter;

/**
* The additional hashtable is used only by the JoinMemento subclass.
* It is here so that the PDIF serialization
* can work across package boundaries.
* This isn't clean, but it is the price of evolutionary software.
*/
	protected Hashtable addOn;

	private PropertySet () {
	}
/**
* @param suggestedUid is a string to use as a memento uid
* @param temporary identifies this memento as temporary so it will not
* check for uid conflicts
*/
	public PropertySet (String parentKind, String parentId, Reporter reporter) {
		this.reporter = reporter;
		setParentInfo(parentKind, parentId);
	}
/**
* Allow change of parent kind and id
*/
	final public void setParentInfo(String parentKind, String parentId) {
		if (parentKind == null || parentId == null || parentKind.length() == 0 || parentId.length() ==0)
			throw new IllegalArgumentException("kind="+parentKind+", id="+parentId);
		this.parentKind = parentKind;
		this.parentId = parentId;
		deriveUidFromParent();
	}
/**
* Get parent's kind
*/
	final public String getParentKind() {
		return parentKind;
	}
/**
* Get parent's id
*/
	final public String getParentId() {
		return parentId;
	}
/**
* A dictionary is used to support keyed access to properties
*/
	public void useDictionary(IDictionary dictionary) {
		if (dictionary == this.dictionary)
			return;
		if (this.dictionary != null)
			throw new RuntimeException("Attempt to replace dictionary \""+this.dictionary.getDictionaryName()+
					"\" with dictionary \""+dictionary.getDictionaryName()+"\"");
		list.useDictionary(dictionary);
		this.dictionary = dictionary;
	}
/**
* Update uid from parent
*/
	private void deriveUidFromParent() {
		String suggestedUid = makeId(parentKind, parentId);
		String existingUid = getUid();
		
		if (existingUid == null || !existingUid.equals(suggestedUid)) {
			setUid(suggestedUid);
		}
		
	}
/*
* used internally
*/
	static public String generateUid() {
		long id = rand.nextLong();
		return Long.toString(id>0 ? id : -id);
	}
/**
* Set this memento's uid if it is being restored from persistent storage. 
*/
	public void setUid(String uid) {
		list.put(MEMENTO_UID, uid);
		firePropertyChange(MEMENTO_UID, null, uid);
	}
	

/**
* Get the uid
*/
	public String getUid() {
		return getProperty(MEMENTO_UID);
	}
/**
* Add a property-value pair
*/
	public void setProperty(String property, String value) {
		if (value == null || property == null) {
			String errs = "offensive (property, value) = ("+property+", "+value+")";
			String memUid = getProperty(Memento.MEMENTO_UID);
			if (memUid != null)
				errs += "\n related "+Memento.MEMENTO_UID+" = "+memUid;
			throw new IllegalArgumentException(errs);
		}
		if (property == MEMENTO_UID)
			setUid(value);
		else {
			if (!hasListeners()) {
				list.put(property.intern(), value);
			}
			else {
				Object oldValue = list.get(property);
				list.put(property.intern(), value);
				firePropertyChange(property, oldValue, value);
			}
		}
	}
	private Object internGet(String property) {
		Object prop = list.get(property);
		if (prop == null || prop == DELETED)
			return null;
		return prop;
	}
/**
* Get a generic object that holds the value of the key.
* @return an object of String, Numeric, or ValueList type, or null
*/
	public Object getPropertyAsObject(String property) {
		return internGet(property);
	}
/**
* Get a property.  If the property is multivalued it will be converted to String.
*/
	public String getProperty(String property) {
		Object prop = internGet(property);
		if (prop == null)
			return null;
		return prop.toString();
	}
/**
* Add a property-value pair
*/
	public void setPropertyValues(String property, ValueList value) {
		if (value == null || property == null)
			throw new IllegalArgumentException();
		Object oldValue = list.get(property);
		if (property == MEMENTO_UID)
			reporter.assert(false,"Illegal attempt to add multivalued uid",null,null,null);
		else {
			if (value.size() == 1) {
				String sval = (String) value.elementAt(0);
				list.put(property.intern(), sval);
				firePropertyChange(property, oldValue, sval);
			}
			else {
				list.put(property.intern(), value);
				firePropertyChange(property, oldValue, value);
			}
		}
	}
/**
* Get a multivalued property.
* @return always of type ValueList, never null
*/
	public ValueList getPropertyValues(String property) {
		Object val = list.get(property);
		if (val == null)
			return ValueList.EMPTY_VALUE_LIST;
			
		if (val == DELETED)
			return DELETED_VALUE_LIST;

		if (val instanceof ValueList)
			return (ValueList) val;
		
		if (val instanceof String) {
			String sval = (String) val;
			if (sval.length() > 0)
				return new ValueList(sval);
			return emptyStringValueList;
			}
			
		reporter.assert(false,"Request for ",property.toString()," resulted in illegal value type - ",val.toString());
		return ValueList.EMPTY_VALUE_LIST;
	}
/**
* @return 0 if property does not exist, has been deleted, or is an empty valuelist
* @return 1 if string or valuelist with 1 object
* @return >1 if valuelist with more than one value
*/
	public int getCurrentDimension(String property){
		Object val = list.get(property);
		if (val == null)
			return 0;
			
		if (val == DELETED)
			return 0;

		if (val instanceof String) 
			return 1;

		if (val instanceof ValueList)
			return ((ValueList) val).size();
		
			
		reporter.assert(false,"Request for ",property.toString()," resulted in illegal value type - ",val.toString());
		return 0;
	}
/**
* Copy the property and value potentially with a different name
*/
	public void copyPropertyFromAs(String fromName, HasProperties fromOther, String newName){
		if (fromOther.getCurrentDimension(fromName) > 1)
			setPropertyValues(newName, fromOther.getPropertyValues(fromName));
		else
			setProperty(newName, fromOther.getProperty(fromName));
	}
/**
* Get enumeration of all properties;
* @return Enumeration of Strings
*/
	public Enumeration getProperties() {
//		return list.keys();
		return new PSEnumerator();
	}
/**
* Remove a property
*/
	public void removeProperty(String property) {
		firePropertyChange(property, list.remove(property), null);
	}
/**
* Remove a property but tag it so that remote, dependent
* data copies can synchronize with this deletion
*/
	public void removePropertyAndRemoteCopies(String property){
		Object oldValue = list.put(property.intern(), DELETED);
		firePropertyChange(property, oldValue, null);
	}
/**
* Creates a string that has no funny characters in it
* and that is not too long, yet is hopefully as unique
* as the kind/id pair.
*/
	private String makeId(String kind, String id) {
		StringBuffer str= new StringBuffer(20);
		
		if (kind.length() < 6) {
			str.append(kind);
		}
		else {
			str.append(kind.charAt(0));
			str.append(kind.charAt(2));
			str.append(kind.charAt(4));
			str.append(kind.charAt(5));
		}

		int idLength = id.length();
		if (idLength < 6) {
			str.append(id);
		}
		else {
		/* we want approx 6 chars from id part */
			int skip = idLength / 6;
			if (skip < 1)
				skip = 1;
			for (int i = 0; i< idLength-3; i+= skip) {
				char c = id.charAt(i);
				c = Character.toLowerCase(c);
				if (Character.isLetterOrDigit(c))
					str.append(c);
				else
					str.append('x');
			}
		/* we also want last two chars */
			for (int i = idLength-3; i< idLength; i++) {
				char c = id.charAt(i);
				c = Character.toLowerCase(c);
				if (Character.isLetterOrDigit(c))
					str.append(c);
				else
					str.append('x');
			}
		}
		
		return str.toString();
	}
/**
* Report a bound property update to any registered listeners.
* No event is fired if old and new are non-null and .equal.
*
*/
    private void firePropertyChange(String propertyName, 
					Object oldValue, Object newValue) {

		Vector targets;
		synchronized (this) {
		    if (!hasListeners()) {
		    	return;
		    }
		    targets = (java.util.Vector) listeners.clone();
		}
		if (oldValue != null && oldValue.equals(newValue)) {
		    return;
		}
	
		MementoEvent evt = new MementoEvent(this, propertyName, oldValue, newValue);
	
		for (int i = 0; i < targets.size(); i++) {
		    MementoListener target = (MementoListener)targets.elementAt(i);
		    target.propertyChange(evt);
		}
    }
/**
* Add a PropertyChangeListener to the listener list.
*/

    public synchronized void addMementoListener( MementoListener listener) {
		if (listeners == null) {
		    listeners = new java.util.Vector();
		}
		if (!listeners.contains(listener))
			listeners.addElement(listener);
    }
/**
* Remove a MementoListener to the listener list.
*
*/
    public synchronized void removeMementoListener(MementoListener listener) {
		if (listeners == null) {
		    return;
		}
		listeners.removeElement(listener);
    }
/**
* Get a boolean-valued property.  Internally this does a .equalsIgnoreCase("true") on the property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public boolean getBooleanProperty(String property, boolean defaultValue){
		Object prop = internGet(property);
		if (prop == null)
			return defaultValue;
		if (prop instanceof Boolean)
			return ((Boolean)prop).booleanValue();
		if (prop instanceof String)
			return prop.toString().equalsIgnoreCase("true");
		return defaultValue;
	}
/**
* Get a single string-valued property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public String getStringProperty(String property, String defaultValue){
		Object prop = internGet(property);
		if (prop == null)
			return defaultValue;
		return prop.toString();
	}
/**
* @return true if the property exists
*/
	public boolean isProperty(String propertyName){
		return (null != internGet(propertyName));
	}
/**
* Get an integer-valued property.  If there is a parse exception
* then the default value is returned.
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public int getIntegerProperty(String property, int defaultValue){
		Object prop = internGet(property);
		if (prop == null)
			return defaultValue;
		if (prop instanceof Number)
			return ((Number)prop).intValue();

		try {
			return Integer.parseInt(prop.toString());
		}
		catch (NumberFormatException e) {
			return defaultValue;
		}
	}
/**
* Get an long-valued property.  If there is a parse exception
* then the default value is returned.
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public long getLongProperty(String property, long defaultValue){
		Object prop = internGet(property);
		if (prop == null)
			return defaultValue;
		if (prop instanceof Number)
			return ((Number)prop).longValue();

		try {
			return Long.parseLong(prop.toString());
		}
		catch (NumberFormatException e) {
			return defaultValue;
		}
	}
/**
* Get an double-valued property.  If there is a parse exception
* then the default value is returned.
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public double getDoubleProperty(String property, double defaultValue){
		Object prop = internGet(property);
		if (prop == null)
			return defaultValue;
		if (prop instanceof Number)
			return ((Number)prop).doubleValue();

		try {
			return Double.valueOf(prop.toString()).doubleValue();
		}
		catch (NumberFormatException e) {
			return defaultValue;
		}
	}
/**
* Set a boolean-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public void setProperty(String property, boolean value){
		setProperty(property, value?"true":"false");
	}
/**
* Set a single integer-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(String property, int value){
		setProperty(property, Integer.toString(value));
	}
/**
* Set a single double-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(String property, double value){
		setProperty(property, Double.toString(value));
	}
/**
* Set a single long-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(String property, long value){
		setProperty(property, Long.toString(value));
	}

/**
* return the number of elements
*/
	public int size() {
		return list.size();
	}
/**
* A list of keys
*/
	public Enumeration keys() {
		return list.keys();
	}
/**
* do we contain specified key
*/
	public boolean containsKey(Object key) {
		return list.get(key) != null;
	}
/*
* This memento has listerners of some kind
*/
	final protected boolean hasListeners() {
		return listeners != null;
	}

/* +++++++ Start keyed accessors  +++++++++ */

/**
* Remove a property without regard for other remote
* copies of the data
*
*/
	public void removeProperty(AttributeKey propertyKey){
		removeProperty(propertyKey.toString());
	}
/**
* Remove a property but tag it so that remote, dependent
* data copies can synchronize with this deletion
*
*/
	public void removePropertyAndRemoteCopies(AttributeKey propertyKey){
		removePropertyAndRemoteCopies(propertyKey.toString());
	}
/**
* Get a property
*/
	public ValueList getPropertyValues(AttributeKey propertyKey){
		return getPropertyValues(propertyKey.toString());
	}
/**
* Set a property
*/
	public void setPropertyValues(AttributeKey propertyKey, ValueList values){
		setPropertyValues(propertyKey.toString(), values);
	}
/**
* Get a boolean-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public Boolean getProperty(AttributeKey propertyKey, Boolean defaultValue){
		Object prop = internGet(propertyKey.toString());
		if (prop == null)
			return defaultValue;
		if (prop instanceof Boolean)
			return (Boolean)prop;
		if (prop instanceof String)
			return prop.toString().equalsIgnoreCase("true") ? Boolean.TRUE : Boolean.FALSE ;
		return defaultValue;
	}
/**
* Get a single string-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public String getProperty(AttributeKey propertyKey, String defaultValue){
		Object prop = internGet(propertyKey.toString());
		if (prop == null)
			return defaultValue;
		return prop.toString();
	}
/**
* Get a single integer-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public Integer getProperty(AttributeKey propertyKey, Integer defaultValue){
		Object prop = internGet(propertyKey.toString());
		if (prop == null)
			return defaultValue;
		if (prop instanceof Integer)
			return (Integer) prop;
		if (prop instanceof Number)
			return new Integer(((Number)prop).intValue());
		try {
			return new Integer(prop.toString());
		}
		catch (NumberFormatException e){
			return defaultValue;
		}
	}
/**
* Get a single long-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public Long getProperty(AttributeKey propertyKey, Long defaultValue){
		Object prop = internGet(propertyKey.toString());
		if (prop == null)
			return defaultValue;
		if (prop instanceof Long)
			return (Long) prop;
		if (prop instanceof Number)
			return new Long(((Number)prop).longValue());
		try {
			return new Long(prop.toString());
		}
		catch (NumberFormatException e){
			return defaultValue;
		}
	}
/**
* Get a single double-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public Double getProperty(AttributeKey propertyKey, Double defaultValue){
		Object prop = internGet(propertyKey.toString());
		if (prop == null)
			return defaultValue;
		if (prop instanceof Double)
			return (Double) prop;
		if (prop instanceof Number)
			return new Double(((Number)prop).doubleValue());
		try {
			return new Double(prop.toString());
		}
		catch (NumberFormatException e){
			return defaultValue;
		}
	}

/**
* Set a property
*/
	public void setProperty(AttributeKey propertyKey, String value){
		setProperty(propertyKey.toString(), value);
	}
/**
* Set a Boolean-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, Boolean value){
		setProperty(propertyKey.toString(), value.booleanValue());
	}
/**
* Set a Integer-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, Integer value){
		setProperty(propertyKey.toString(), value.intValue());
	}
/**
* Set a Double-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, Double value){
		setProperty(propertyKey.toString(), value.doubleValue());
	}
/**
* Set a Long-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, Long value){
		setProperty(propertyKey.toString(), value.longValue());
	}



/**
* Set a boolean-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, boolean value){
		setProperty(propertyKey.toString(), value);
	}
/**
* Set a int-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, int value){
		setProperty(propertyKey.toString(), value);
	}
/**
* Set a double-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, double value){
		setProperty(propertyKey.toString(), value);
	}
/**
* Set a long-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, long value){
		setProperty(propertyKey.toString(), value);
	}

/**
* @return true if the property exists
*/
	public boolean isProperty(AttributeKey propertyKey){
		return isProperty(propertyKey.toString());
	}
/* ------- End keyed accessors  -------- */

	class PSEnumerator implements Enumeration {
		Enumeration e;
		Object next;
		PSEnumerator() {
			e = list.keys();
			advance();
		}
		public boolean hasMoreElements() {
			return next != null;
		}
		public Object nextElement() {
			Object ret = next;
			advance();
			return ret;
		}
		private void advance() {
			do {
				if (e.hasMoreElements())
					next = e.nextElement();
				else
					next = null;
			} while ( (next != null) && (internGet((String)next) == null) );
		}
	}
	
/**
* A package-only access to the internal list
*/
	final SortedHash getSortedHash() {
		return list;
	}
}

