/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.executive;

import java.util.*;

import org.mitre.pam.interfaces.*;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.executive.Module;
import org.mitre.pam.executive.Command;
import org.mitre.pam.util.EnabledTable;

/**
 * A helper class for agents that need to manage

 * their own queues
 *
 * @author Nazario Irizarry
 */

public class QueueModuleHelper implements QueueAgentAttributes {

	protected RunBody deliveryThread = null;
	protected boolean logDrops;
	protected boolean dropData = true;
	protected boolean useSalience = false;
	
	protected long droppedCount = 0;
	protected long dropReportTime = 0;
	protected long reportedDrop = 0;
	protected int maxQueueSize = 20;
	protected Vector queue = new Vector(50);
	protected Object haveStuff = new Object(); // flag to sync on
	protected Object takenStuff = new Object(); // flag to sync on
	int errorCount;

	private Module parentModule;
	private static int threadCounter;
	private boolean flushOnStop = true;
	private boolean queuePausesAllowed = true;

	public QueueModuleHelper (Module parentModule) {

		this.parentModule = parentModule;
		initializeQueueModuleHelper();
	}

	public void setFlushOnStop(boolean flushOnStop) {
		this.flushOnStop = flushOnStop;
	}
	
	public void initializeQueueModuleHelper() {
	/*
	* The followint setProperties CANNOT invoke the parentModule
	* setProperty() because this happens during the construction
	* at which time the parent still has no valid reference to 
	* the instance of this class
	*/
		setProperty(THREAD_PRIORITY, "6");
		parentModule.metaIntegerInfo(THREAD_PRIORITY, 6, "Java standard priority in the range of 1 - 10 ");

		setProperty(MAX_QUEUE_SIZE, "20");
		parentModule.metaIntegerInfo(MAX_QUEUE_SIZE, 20, "Size of the FIFO queue");

		setProperty(Module.LOG_ACTIVITY, "false");

		setProperty(LOG_DROPS, "true");
		parentModule.metaBooleanInfo(LOG_DROPS, true, "When set will cause a log message to be generated when data is dropped");

		setProperty(DROP_USING_SALIENCE, "false");
		parentModule.metaBooleanInfo(DROP_USING_SALIENCE, false, "If selected and dropping is enabled, products with lowest numerical attibute 'salience' will get dropped if necessary");

		setProperty(DROP_IF_FULL, "true");
		parentModule.metaBooleanInfo(DROP_IF_FULL, true, "Data should be dropped if the queue is full");

		setProperty(QUEUE_PAUSES_ALLOWED, "true");
		parentModule.metaBooleanInfo(QUEUE_PAUSES_ALLOWED, true, "Queue can hold during housekeeping tasks such as checkpointing");
		
		parentModule.stateMemento.setProperty(DROPPED_DATA_COUNT, 0); // don't call getMemento here
		parentModule.metaIntegerInfo(DROPPED_DATA_COUNT, 0, "Number of items dropped because the queue was full");

		parentModule.stateMemento.setProperty(ITEMS_ON_QUEUE, 0); // don't call getMemento here
		parentModule.metaIntegerInfo(ITEMS_ON_QUEUE, 0, "Number of items on the queue(read only)");

	}


/**
* @return true if this class recognizes and handles the property and the
* parentModule should not handle the property
*/
	public boolean setProperty(String property, String value) {

		if (property.equals(DROPPED_DATA_COUNT)) {
			return true;  // readonly; tell parent not to handle this one
		}
		else if (property.equals(ITEMS_ON_QUEUE)) {
			return true;  // readonly; tell parent not to handle this one
		}
		else if (property.equals(DROP_IF_FULL)) {
			dropData = value.equalsIgnoreCase("true");
		}
		else if (property.equals(QUEUE_PAUSES_ALLOWED)) {
			queuePausesAllowed = value.equalsIgnoreCase("true");
		}
		else if (property.equals(LOG_DROPS)) {
			logDrops = value.equalsIgnoreCase("true");
		}
		else if (property.equals(DROP_USING_SALIENCE)) {
			useSalience = value.equalsIgnoreCase("true");
		}
		else if (property.equals(THREAD_PRIORITY) && deliveryThread != null) {
			int prio=0;
			try {
				prio = Integer.parseInt(value);
			} catch (NumberFormatException e) {
				prio = Thread.NORM_PRIORITY;
			}
			if (prio < Thread.MIN_PRIORITY) {
				prio = Thread.MIN_PRIORITY;
			}
			if (prio > Thread.MAX_PRIORITY) {
				prio = Thread.MAX_PRIORITY;
			}
			value = Integer.toString(prio);
			deliveryThread.setPriority(prio);
		}
		else if (property.equals(MAX_QUEUE_SIZE)) {
			maxQueueSize = 20;                // Default
			if (value.length() != 0) {
				try {
					maxQueueSize = Integer.parseInt(value);
				} catch (NumberFormatException e) {
					value = "20";
				}
			}
			if (maxQueueSize < 1) {
				maxQueueSize = 1;
				value = "1";
			}
		}
		else {
			return false; // tell parent to handle the set
		}
		parentModule.stateMemento.setProperty(property, value);
		return true;  // we are done with this; tell parent not to handle this one
	}
	
/**
* Queue a product if there is space in the queue.
* Report if something gets dropped, but no more than
* once every ten seconds.
*/

	public void queueProduct(Product p){
		if (p == null)
			return;
		if (parentModule.logActivity) {
			parentModule.reporter.log("input, queueSize=="+queue.size()+"/"+maxQueueSize);
		}
	/*
	* decide whether we need to drop or block
	*/
		synchronized(takenStuff) {
			while (queue.size() >= maxQueueSize) {
				if (dropData) {
					if (parentModule.logActivity) {
						parentModule.reporter.log("dropping input");
					}
					if (useSalience) {
						if (p != null) {
							p.incrementUsage(this);
							queue.addElement(p);
						}
						dropOneProductFromQueue(queue);
					}
					droppedCount++;
					p = null;
					break;
				}
				else  {
					try {
						takenStuff.wait();
					} catch (InterruptedException ignore) {}
				}
			}
		}
		
	/*
	* Add to queue and signal
	*/
		if (p != null) {
			synchronized(haveStuff) {
				p.incrementUsage(this);
				queue.addElement(p);
				haveStuff.notify();
			}
		}
	/*
	* Report status if appropriate
	*/
		long t;	
		if ( (droppedCount > reportedDrop) && 
				((t = System.currentTimeMillis()) > dropReportTime)) {
			String drops = Long.toString(droppedCount);
			parentModule.stateMemento.setProperty(DROPPED_DATA_COUNT,drops);
			if (logDrops)
				parentModule.reporter.warning("Dropped data count = "+drops);
			dropReportTime = t + 10000;  
			reportedDrop = droppedCount;
		}
		updateNItemsOnQueueProperty(false);
	}

	/**
	* update the memento if the named attributes
	* is one of the ones we report on.
	* if attribute == null, then update regardless
	*/
	public void updateStats(String attribute) {
		if (attribute == null || attribute.equals(DROPPED_DATA_COUNT))
			parentModule.stateMemento.setProperty(DROPPED_DATA_COUNT,droppedCount);
	}
	
	protected Product deQueueProduct(){
		Product p;
		synchronized (takenStuff) {
			synchronized(queue) {
				p = (Product) queue.elementAt(0);
				queue.removeElementAt(0);
			}
			takenStuff.notify();
		}
		updateNItemsOnQueueProperty(false);
		return p;
	}

	synchronized public void enableService() {
		if (parentModule.logActivity) {
			parentModule.reporter.log("enabled");
		}
		startIfAppropriate();
	}

	synchronized public void disableService() {
		stopIfAppropriate();
		if (parentModule.logActivity) {
			parentModule.reporter.log("disabled");
		}
	}

	synchronized public void pauseService() {
		if (!queuePausesAllowed)
			return;
		if (deliveryThread != null)
			deliveryThread.suspend();
		updateNItemsOnQueueProperty(true);
	}

	synchronized public void resumeService() {
		if (deliveryThread != null)
			deliveryThread.resume();
		updateNItemsOnQueueProperty(true);
	}  

	protected void finalize() {
		stopIfAppropriate();
		updateNItemsOnQueueProperty(true);
	}
	
	protected void startIfAppropriate() {
		if (deliveryThread == null) {
			errorCount = 0;
			droppedCount = 0;
			deliveryThread = new RunBody("QHelper-"+Integer.toString(++threadCounter)+"-"+toString());
			deliveryThread.setPriority(parentModule.getIntegerProperty(THREAD_PRIORITY,Thread.NORM_PRIORITY));
			deliveryThread.start();
		}
		updateNItemsOnQueueProperty(true);
	}

	protected void stopIfAppropriate() {
		if(deliveryThread != null) {
			deliveryThread.keepRunning = false;
			deliveryThread = null;
			if(flushOnStop)
				clearQueue();
		}
		updateNItemsOnQueueProperty(true);
	}


	class RunBody extends Thread {	
		boolean keepRunning = true;
		
		public RunBody(String name) {
			super(name);
		}
		
		public void run() {
			while (keepRunning) {
				try {
					while(queue.size() > 0) {
						Product p = deQueueProduct();
						if (p != null) {
							try {
								parentModule.putQueuedProduct(p);
								p.decrementUsage(this);
							}
							catch (Exception e) {
								parentModule.reporter.reportException(e);
								parentModule.reporter.warning("Error count is "+errorCount+"/10");
								if (++errorCount > 10) {
									parentModule.reporter.warning("Too many errors, autodisabling (may be manually reenabled)");
									parentModule.setEnabled(false);
									clearQueue();
								}
							}
						}
					}
				/*
				* The test for queue.size() in the next synchronized block is NOT
				* redundant with the above since we don't know
				* if something was added to the queue in the small
				* interval of time between when we last performed
				* the test and when we test queue.size() again
				*/
					synchronized(haveStuff) {
						if (queue.size() == 0)
							haveStuff.wait(15000); // allow time to update the queue count property
					}
					updateNItemsOnQueueProperty(false);
				}
				catch (Exception e){
					parentModule.reporter.reportException(e);
					++errorCount;
				}
			}
		}
	}
	
/**
* Runs down the queues contents gracefully
*/
	protected void clearQueue() {
		while (queue.size()>0) {
			Product plost = deQueueProduct();
			plost.decrementUsage(this);
		}
	}
/**
* If the queue is full and the option for salience based dropping is implemented
* then do so.
*
* @param queue a Vector with products, one of which needs to be removed
*/
	public void dropOneProductFromQueue(Vector queue) {
		synchronized (queue) {
			int lowestSalience = Integer.MAX_VALUE;
			int nItems = queue.size();
			int productIndex = nItems-1;
			Product removedProd=null;
			for (int i = 0; i< nItems; i++) {
				Product p = (Product) queue.elementAt(i);
				int salience = p.getIntegerProperty(SALIENCE, 0);
				if (salience <= lowestSalience) { // favors removal of last item
					lowestSalience = salience;
					productIndex = i;
					removedProd = p;
				}
			}
			queue.removeElementAt(productIndex);
			removedProd.decrementUsage(this);
		}
	}

/**
* @return current size of the queue
*/

	public int getNumberOfItemsOnQueue() {
		return queue.size();
	}
/**
* Updates the property associated with number of items on the queue.
* The update will typically be made
* if 3059 milliseconds have passed since the last update
* if the queue size has changed since the last update.  This is to
* minimize the number of changes to the property in case there are
* listeners on the memento.
*
* @param updateIsMandatory is true to force an update regardless of the last update time or count
*/
	private long nextUNIOQPTime;
	private int lastUNIOQPCount=-1;
	synchronized public void updateNItemsOnQueueProperty(boolean updateIsMandatory) {
		long curTime = System.currentTimeMillis();
		if (updateIsMandatory || ((curTime > nextUNIOQPTime) &&(lastUNIOQPCount != queue.size()))) {
			lastUNIOQPCount = queue.size();
			nextUNIOQPTime = curTime + 3059;
			parentModule.stateMemento.setProperty(ITEMS_ON_QUEUE, lastUNIOQPCount);
		}
	}
/**
* Nice debug string
*/
	public String toString() {
		return "Queue helper for "+parentModule;
	}
}
