/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.executive;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Observable;
import java.util.Observer;
import java.util.Properties;
import java.util.Enumeration;
import java.io.PrintStream;

import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.interfaces.Doer;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.interfaces.Schedulable;
import org.mitre.pam.interfaces.ScheduledCommand;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.interfaces.ScheduleAttributes;
import org.mitre.pam.util.SortedHashOnElements;
import org.mitre.pam.util.SortedLink;
import org.mitre.pam.util.BinaryCompare;


/**
* Executes a set of agents on a schedule.  The agents
* themselves identify their execution schedule. Each
* Executer runs on a slightly different sleep interval.
* <p>
* This class is public because it is referred to outside
* of the executive from SimpleSSI class.
*
* @author Nazario Irizarry, Jr.
* @author Jason Mathews PAM 2.1.62 Added "none" schedule value to 
*         static buildSchedule method and added private printScheduleValue
*         method to perform repetitive printing in one place.
*/
final public class Scheduler extends Thread implements BinaryCompare{
	private static final String DEBUG_VERBOSE_TIMING_CHANGE = "traceScheduleChanges";
	private static final String SALIENCE = "salience";

/**
* Keep track of how many exceptions there have been within the scheduling code.
* Used to autoshutdown if too many scheduling errors.
*/
	int exceptionCount;
/**
* a ref to the exec
*/
	ExecCallbacks exec;
/**
* the state of the scheduler, mirrors the state of the exec
*/
	String state = ExecGlobal.STARTING;
/*
* executerCount is used in conjunction with the sleepInterval algorithm
*/
	static int executerCount = 0;
/*
* No two executors have identical intervals
*/
	long maxSleepInterval = 2000+executerCount*450;	
	long sleepInterval = maxSleepInterval;
	Reporter reporter;
	Vector scheduleChangedList = new Vector(50);
	
/**
* Time window within which all items compete based on salience
*/
	int timeWindow = 5000; // 5 seconds
/**
* Update flag
*/
	int updateFlag = 0;
/*
* A list of doers
*/
	SortedHashOnElements doers = new SortedHashOnElements(this);
	String name;
	boolean keepRunning;
	
	public Scheduler (String name, ExecCallbacks exec) {
		super(name);
		this.name = name;
		this.exec = exec;
		this.reporter = exec.getExecAsAgent().getReporter();
		this.keepRunning = true;
		executerCount++;
		start();
	}
/**
* Add a schedulable item to this queue.
* The queue is maintained in order based on activation time.
* Items with null activation time are placed at the end.
*/
	public void add(Schedulable item) {
		reporter.debugMsg(" add: ", item, null,null,DEBUG_VERBOSE_TIMING_CHANGE);
		ScheduledCommand sc = item.getNextScheduleCommand();
		addDoerByLong(item, sc == null ? Long.MAX_VALUE : sc.getTime());
		updateFlag++;
	}

	private void moveToTail(Schedulable item) {
		addDoerByLong(item, Long.MAX_VALUE);
		reporter.debugMsg("Moved to tail: ",item,null,null,DEBUG_VERBOSE_TIMING_CHANGE);
	}

	private void addDoerByLong(Schedulable item, long sortOrder) {
		doers.put(item, new Long(sortOrder));
	}
	
	public void remove(Schedulable item) {
		reporter.debugMsg(" remove: ", item, null,null,DEBUG_VERBOSE_TIMING_CHANGE);
		doers.remove(item);
		updateFlag++;
	}

/**
* The run loop checks to see if it has anything to execute
*/
	public void run() {
		while ( keepRunning && 
				(exec.getState() == ExecGlobal.STARTING) ) {
			try {
				sleep(5000); // wait in increments of 5 seconds
			}
			catch (InterruptedException e) { }
		}
		ScheduledCommand nextCommand;
		while ( keepRunning && 
				(exec.getState() != ExecGlobal.SHUTTING_DOWN)) {
			try {
			/*
			* First a sleep period
			*/
				try {
					setState(ExecGlobal.IDLE);
	
					if (updateFlag > 0) {
						updateFlag = 0;
						exec.update(null, name);
						}
					else
						exec.update(null, ExecGlobal.TICK);
					long currentTime = System.currentTimeMillis();
					long nextTime = identifyNextPotentialTime();
					sleepInterval = nextTime - currentTime;
					if ( sleepInterval < 0 ) 
						sleepInterval = 100;				
					else if ( sleepInterval > maxSleepInterval ) 
						sleepInterval = maxSleepInterval;
						
					sleep(sleepInterval);
				}
				catch (InterruptedException e) {
				}
			/**
			* A little debug?
			*/
/*
				if (reporter.isDebugKey(DEBUG_VERBOSE_TIMING_CHANGE)){
					Enumeration agents = doers.buckets();
					System.out.println();
					while (agents.hasMoreElements()) {
						System.out.println(agents.nextElement());
					}
				}
*/
			/*
			* Second some cleaning
			*/
				processScheduleChanges();
			/*
			* Third an activity period
			*/
				while ( (exec.getState() == ExecGlobal.WORKING) &&
						(nextCommand = scheduleNext()) != null ) {
					setState(ExecGlobal.WORKING);
					try {
						nextCommand.perform();
					}
					catch (Exception e) {
						reporter.reportException(e);
					}
					processScheduleChanges(); // could happen as side effect of nextCommand.perform();
				}
			}
			catch (Exception all) {
				reporter.reportException(all);
				reporter.warning("After 5 exceptions scheduler will turn off, currently at "+Integer.toString(++exceptionCount));
				if (exceptionCount >= 5){
					reporter.warning("Too many scheduler exceptions, shutting off");
					keepRunning = false;
				}
			}
		}
		state = ExecGlobal.SHUTTING_DOWN;
		doers = new SortedHashOnElements(this);
		scheduleChangedList.setSize(0);
		reporter.log("Scheduler "+name+" stopped");
	}
/**
* Scans list of doers to see who is going to be next.
*/
	private synchronized ScheduledCommand scheduleNext() {
		if ( doers.size() <= 0 ) 
			return null;

		Candidate candidate = null;
		Candidate betterCandidate = null;
		long nextTime = 0;
		long currentTime = System.currentTimeMillis();
		long timeLimit =  currentTime + timeWindow;
		int highestSalience = 0;
		ScheduledCommand nextCommand = null;
		
		candidate = findCandidate(currentTime, timeLimit); // also reserves
		
		if (candidate == null)
			return null;
			
		reporter.debugMsg("Candidate: ",candidate.command.getDoer(),null,null,DEBUG_VERBOSE_TIMING_CHANGE);
	/*
	* Find something else with higher salience
	*/
/*
		if (candidate != null)
			betterCandidate = findHigherSalienceCandidate(candidate, timeLimit); 
		if (betterCandidate != null) {
			candidate = betterCandidate;
			reporter.debugMsg("Better candidate: ",betterCandidate.command.getDoer(),null,null,DEBUG_VERBOSE_TIMING_CHANGE);
		}
*/
		Schedulable schedulableAgent = (Schedulable) candidate.command.getDoer();
		moveToTail(schedulableAgent);
	
		return candidate.command;
	}
/**
* Find higher salience candidate.
* If one is found the current candidate is released and the
* higher salience candiate is locked for tasking.
*/
	private Candidate findHigherSalienceCandidate(Candidate theCandidate, long timeLimit){
		int highestSalience = theCandidate.salience;
		/*
		* Iterate looking for candidate of higher salience within the time limit
		*/
			Enumeration enum = doers.buckets();
			while (enum.hasMoreElements()) {
				SortedLink ce = (SortedLink)enum.nextElement();
				long time = ((Long)ce.getValue()).longValue();
				if (time > timeLimit)
					break;
				Schedulable agent = (Schedulable) ce.getKey();
				int salience = agent.getIntegerProperty(SALIENCE,0);

				if (salience > highestSalience) {
					ScheduledCommand command = agent.getNextScheduleCommand();
					if (command != null && 
						command.getDoer().getTaskingPermission(command)) {
					/*
					* release old
					*/
						theCandidate.command.getDoer().releaseTaskingPermission(theCandidate.command);
						theCandidate = new Candidate( time, salience, command );
						highestSalience = salience;
					}
				}					
			}
		return theCandidate;
	}
/**
* What is the next earliest time that something might run.
*/
	private long identifyNextPotentialTime() {
		Enumeration enum = doers.buckets();
		while (enum.hasMoreElements()) {
			SortedLink ce = (SortedLink)enum.nextElement();
			long time = ((Long)ce.getValue()).longValue();
		/*
		* Scan up to the agents that have a time
		* of Long.MAX_VALUE.  Ignore these
		*/
			if (time == Long.MAX_VALUE)
				return Long.MAX_VALUE;
			Schedulable agent = (Schedulable) ce.getKey();
			if (agent.isIdle() && agent.isEnabled()) {
				return time;
			}
		}
		return Long.MAX_VALUE;
	}
/**
* find the first available candidate for scheduling
*/
	private Candidate findCandidate(long currentTime, long timeLimit) {
		Candidate theCandidate = null;
		/*
		* Iterate looking for next candidate who's scheduled time has passed.
		* Make a list of things that need delaying but do not delay them in this loop
		* so that the side effects of delaying them don't mess up the list we
		* are traversing (doers list)
		*/
			Enumeration enum = doers.buckets();
			while (enum.hasMoreElements()) {
				SortedLink ce = (SortedLink)enum.nextElement();
				long time = ((Long)ce.getValue()).longValue();
				if (time > timeLimit)
					return null;
					
				Schedulable agent = (Schedulable) ce.getKey();
				if (agent.isEnabled() && agent.isIdle()) {
					int salience = agent.getIntegerProperty(SALIENCE,0);
					ScheduledCommand command = agent.getNextScheduleCommand();
					if (command != null && time < currentTime) {
						if (command.getDoer().getTaskingPermission(command) ){
							theCandidate = new Candidate( time, salience, command);
							break;
						}
					}
				}
				else {
					if (!agent.isEnabled())
						reporter.debugMsg("agent is not enabled:",agent,null,null,DEBUG_VERBOSE_TIMING_CHANGE);
					if (!agent.isIdle())
						reporter.debugMsg("agent is not idle   :",agent,null,null,DEBUG_VERBOSE_TIMING_CHANGE);
				}
			}
		return theCandidate;
	}
/**
* Check when an scheduled command notifies us of change
*/
	public void update(Observable o, Object arg) {
	/*
	* ignore command unless it is the current one
	*/
/*		if (o != nextCommand) {
			o.deleteObserver(this);
			return;
		}
*/
	/*
	* determine new schedule and wake up the thread
	*/
/*		scheduleNext();
		this.interrupt();
*/
	}

/**
* set stateof the scheduler
*/
	synchronized void setState(String state) {
		this.state = state;
	}
/**
* get state of the scheduler
*/
	public synchronized String getState() {
		return state;
	}
/**
* Get a list of what is currently on queue.
* For now the list is all shown as holding since
* what will actually run is getting complicated
* with the idea of salience.
*
*@param whatsRunning is an empty input vector which is filled on return
*@param whatsHolding is an empty input vector which is filled on return
*/
	public void getQueueLists(Vector whatsRunning, Vector whatsHolding) {
		Enumeration enum = doers.keys();
		boolean first = false; // used to be true
		while (enum.hasMoreElements()) {
/*
			ScheduledCommand cmd = ((Schedulable) enum.nextElement()).getNextScheduleCommand();
			if (cmd != null) {
				if (first){
					whatsRunning.addElement(cmd.toString());
					first = false;
				}
				else
					whatsHolding.addElement(cmd.toString());
			}
*/
			Schedulable agent = (Schedulable) enum.nextElement();
			if (agent.getState() == ExecGlobal.IDLE) {
				ScheduledCommand cmd = (ScheduledCommand) agent.getNextScheduleCommand();
				if (cmd != null)
					whatsHolding.addElement(cmd.toString());
			} else
				whatsRunning.addElement(agent.getId());
		}

	}
/**
* A notification that an existing agents schedule has changed.
* Schedules tend to change a lot by small amounts so we will
* remember who changed and then batch process it later.
*/
	public synchronized void scheduleChanged(Schedulable whoChanged) {
		if (scheduleChangedList.contains(whoChanged)) {
//			reporter.debugMsg("Change already on queue for ",whoChanged.getId(),null,null,DEBUG_VERBOSE_TIMING_CHANGE);
			return;
		}
		reporter.debugMsg("Time change queued for ",whoChanged,null,null,DEBUG_VERBOSE_TIMING_CHANGE);
		scheduleChangedList.addElement(whoChanged);
	}
/**
* Process schedule changes all at once
*/
	private synchronized void processScheduleChanges() {
		if (scheduleChangedList.size() == 0)
			return;
		updateFlag++;
		reporter.debugMsg("Begin batch process of time changes ",null,null,null,DEBUG_VERBOSE_TIMING_CHANGE);
	/*
	* Run through the loop once.
	* This causes any schedulable items that may need to recompute
	* to do so.  A byproduct of recomputing is that they will probably
	* call Scheduler.scheduleChanged indirectly via the exec since
	* if they recompute their schedule will change.
	* This way when we get to the next loop we won't cause the item
	* to be added again to the scheduleChangedList since that loop
	* will not cause a recompute.
	*/
		for (int i = 0; i<scheduleChangedList.size(); i++) {
			Schedulable changer = (Schedulable)scheduleChangedList.elementAt(i);
			changer.getNextScheduleCommand();
		}
		while (scheduleChangedList.size() > 0) {
			int indexToKill = scheduleChangedList.size() -1;
			Schedulable whoChanged = (Schedulable)scheduleChangedList.elementAt(indexToKill);
			scheduleChangedList.removeElementAt(indexToKill);
//			remove(whoChanged);
			add(whoChanged);
			reporter.debugMsg("Time changed for ",whoChanged.getId(),null,null,DEBUG_VERBOSE_TIMING_CHANGE);
		}
	}
/**
* killThreads should only be called during executive shutdown.  Once it
* is called it cannot be restarted.
*/
	public  void killThreads() {
		keepRunning = false;
	}

  /**
   * This method takes a PrintStream and a Properties Object and writes
   * an html table for editing a juggler schedule to the stream.  The properties
   * should contain the appropriate juggler schedule properties such as "schedule"
   * and "period".  This method is called by the SSIParser used by the 
   * SrvAgentEditor servlet.
   */
  public static void buildSchedule(PrintStream out, Properties substitutionVar) 
  {    
    // get the schedule type and set the appropriate variables
    String schedtype = substitutionVar.getProperty(ScheduleAttributes.SCHEDULE_VERB);
    if (schedtype == null)
      return;

    String day, period;
    day = new String("");
    period = new String("0");

    int hour, minute;
    hour = minute = 0;

    boolean isPeriodic, isWeekly, isDaily, isHourly, isOnce, isNone;
    isOnce=isPeriodic=isWeekly=isDaily=isHourly=isNone = false;

    if (schedtype.equals(ScheduleAttributes.SCHEDULE_ONCE)) {
      isOnce = true;
    }
    else if (schedtype.equals(ScheduleAttributes.SCHEDULE_PERIODIC)) {
      isPeriodic = true;
      period = substitutionVar.getProperty(ScheduleAttributes.PERIOD);
    }
    else if (schedtype.equals(ScheduleAttributes.SCHEDULE_WEEKLY)) {
      isWeekly = true;
      day = substitutionVar.getProperty(ScheduleAttributes.DAY_OF_WEEK);
      hour = Integer.parseInt(substitutionVar.getProperty(ScheduleAttributes.HOUR_OF_DAY));
      minute = Integer.parseInt(substitutionVar.getProperty(ScheduleAttributes.MINUTE_OF_HOUR));
    }
    else if (schedtype.equals(ScheduleAttributes.SCHEDULE_DAILY)) {
      isDaily = true;
      hour = Integer.parseInt(substitutionVar.getProperty(ScheduleAttributes.HOUR_OF_DAY));
      minute = Integer.parseInt(substitutionVar.getProperty(ScheduleAttributes.MINUTE_OF_HOUR));
    }
    else if (schedtype.equals(ScheduleAttributes.SCHEDULE_HOURLY)) {
      isHourly = true;
      minute = Integer.parseInt(substitutionVar.getProperty(ScheduleAttributes.MINUTE_OF_HOUR));
    }
    else if (schedtype.equals(ScheduleAttributes.SCHEDULE_NONE)) {
	isNone = true;
    }
   else {
	// unknown value
	ExecGlobal.getExec().getExecAsAgent().getReporter().warning("schedule has an unknown or unsupported value: " + schedtype);
   }
      
    // start printing the table
    out.println("<table>");
    out.println("<tr>");
    if (isWeekly)
      out.println("<td><input type=\"radio\" name=\"schedule\" value=\"weekly\" CHECKED>Weekly</td>");
    else
      out.println("<td><input type=\"radio\" name=\"schedule\" value=\"weekly\">Weekly</td>");

    out.println("<td rowspan=3 align=CENTER>");
    out.println("<table>");
    out.println("<tr><td>Day</td><td>Hour</td><td>Minute</td></tr>");
    out.println("<tr><td><select name=\"dayOfWeek\">");
    if (!isWeekly) {
      out.println("<option selected><option>Sunday<option>Monday<option>Tuesday<option>Wednesday<option>Thursday<option>Friday<option>Saturday</select>");
    }
    else {
      if (day.equals("Monday"))
	out.println("<option>Sunday<option SELECTED>Monday<option>Tuesday<option>Wednesday<option>Thursday<option>Friday<option>Saturday</select>");
      else if (day.equals("Tuesday"))
	out.println("<option>Sunday<option>Monday<option SELECTED>Tuesday<option>Wednesday<option>Thursday<option>Friday<option>Saturday</select>");
      else if (day.equals("Wednesday"))
	out.println("<option>Sunday<option>Monday<option>Tuesday<option SELECTED>Wednesday<option>Thursday<option>Friday<option>Saturday</select>");
      else if (day.equals("Thursday"))
	out.println("<option>Sunday<option>Monday<option>Tuesday<option>Wednesday<option SELECTED>Thursday<option>Friday<option>Saturday</select>");
      else if (day.equals("Friday"))
	out.println("<option>Sunday<option>Monday<option>Tuesday<option>Wednesday<option>Thursday<option SELECTED>Friday<option>Saturday</select>");
      else if (day.equals("Saturday"))
	out.println("<option>Sunday<option>Monday<option>Tuesday<option>Wednesday<option>Thursday<option>Friday<option SELECTED>Saturday</select>");
      else 
	out.println("<option SELECTED>Sunday<option>Monday<option>Tuesday<option>Wednesday<option>Thursday<option>Friday<option>Saturday</select>");
    }
    
    // print hour of day select
    out.print("</td><td><select name=\"hourOfDay\">");
    for(int i=0; i<24; i++) {
      if (i == hour)
	out.print("<option SELECTED>" + i);
      else
	out.print("<option>" + i);
    }
    out.println("</select></td>");
    
    // print minuteOfHour select
    out.print("<td><select name=\"minuteOfHour\">");
    for (int i=0; i<60; i++) {
      if (i < 10) {
	if (i == minute)
	  out.print("<option SELECTED>0" + i);
	else
	  out.print("<option>0" + i);
      }
      else {
	if (i == minute)
	  out.print("<option SELECTED>" + i);
	else
	  out.print("<option>" + i);
      }
    }
    out.println("</select></td></tr></table>");
    out.println("</td></tr>");

    // print isDaily radio button
    printScheduleValue(out, "daily", "Daily", isDaily);
    out.println("</tr>");

    // print isHourly radio button
    printScheduleValue(out, "hourly", "Hourly", isHourly);
    out.println("</tr>");

    // print isPeriodic radio button
    printScheduleValue(out, "periodic", "Periodic", isPeriodic);
    out.println("<td><input type=\"text\"  name=\"period\" size=30 value=" + period + "> seconds</td>");
    out.println("</tr>");

    // print isOnce radio button
    printScheduleValue(out, "once", "Once", isOnce);
    out.println("</tr>");

    // print isNone radio button
    printScheduleValue(out, "none", "None", isNone);

    out.println("</tr></table>");
  }

  private static void printScheduleValue(PrintStream out, String schedValue, String schedName, boolean isSet) {
    out.print("<tr><td><input type=\"radio\" name=\"schedule\" value=\"");
    out.print(schedValue);

    if (isSet)
      out.print("\" CHECKED>");
    else
      out.print("\">");

    out.print(schedName);
    out.print("</td>");
  }

	public boolean isBefore( Object first, Object second ) {
		return ((Long) first).longValue() < ((Long) second).longValue();
	}

	class Candidate {
		long time;
		int salience;
		ScheduledCommand command;
		Candidate(long time, int salience, ScheduledCommand command) {
			this.time = time;
			this.salience = salience;
			this.command = command;
		}
	}
}
