/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.interfaces;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Date;
import java.util.Vector;
import java.util.Observable;

/**
* Generic interface describing a standard product. 
* This is the interface for all products, however,
* ContainerProducts have more interfaces. <p>
*
* @see ContainerProduct
*
* @author Nazario Irizarry, Jr.
*/
public interface Product extends Summary, Nameable, HasKeyedProperties{
/*
* The name of the property for a product that represents an entry point
* or index of other items.  This property has never been used so support
* for index's was removed from the objects that implement products.
*/
//	public static String PROPERTY_INDEX = "index";
/**
* The name of the property for a product that represents a root
* of a hierarchical product.  This is commonly used to tag
* the index .html page for a collection 
*/
	public static final String PROPERTY_ROOT = "root";
/**
* The name of the property for a product that is a component
* of a larger product and does not stand by itself
*/
	public static final String PROPERTY_SUPPORT_COMPONENT = "supportComponent";
/**
* The lifetime of the product in days.
*/
	public static final String PROPERTY_LIFETIME = "lifeTimeLong";
/**
* The topic that a product falls under
*/
	public static final String PROPERTY_TOPIC = "topic";
/**
* The prefix used to flag product attributes that are attached to modules.
* @see Module.copyAttributesToProduct
*/
	public static final String PRODUCT_PREFIX = "product_";
/**
* A property put on products during garbage collection to help
* identify the Agent that uses this product.  Intended for
* for debugging.  Do not use this as part of the production.
*/
	public static final String PROPERTY_USED_BY_AGENT = "dbgUsedByAgent";
/**
* The constant the defines the attribute for product "id"
*/
	public static final String PROPERTY_ID = "id";
/**
* The constant the defines the attribute for product "description"
*/
	public static final String PROPERTY_DESCRIPTION = "description";
/**
* The constant the defines the attribute for product "modificationDate"
*/
	public static final String PROPERTY_MODIFICATION_DATE = "modificationDate";
/**
* Return the last modification or augmentation date of this product.
*/
	public Date getModificationDate();
/**
* Set the last modification or augmentation date of this product.
*/
	public void setModificationDate(Date date);
/**
* Set the last modification or augmentation date of this product.
*/
	public void setModificationDate(long date);
/**
* Get the lifetime of the product in seconds.
*/
	public long getLifetime();
/**
* Return the content of this product.  
* Valid contents are: <p>
* - java.util.Enumeration - for a list of Product (only a list of Product) <p>
* - java.lang.String - for immediate text <p>
* - java.lang.URL - for a url <p>
* - org.mitre.pam.store.FileRef - for a local file <p>
* - null <p>
*/
	public Object getContent();
/**
* Get unique id, used for persistence
*/
	public String getUid();
/**
* Get the products topic.
*/
	public String getTopic();
/**
* Get the size of the product
*/
	public double getSizeInBytesAsDouble();
/**
* Set the products topic.
*/
	public void setTopic(String topic);
/**
* Makes a clone of this product with all of the current products attributes
* including id and description.  Also clones any references to content.
*/
	public Product cloneWithFirstLevelReferences();
/**
* Makes a clone of this product with all of the current products attributes
* excluding id and description.  Also clones any references to content.
*
* @param newId the new id to use for the product
* @param newDescription the new description to use for the product
*/
	public Product cloneWithFirstLevelReferences(String newId, String newDescription);
/**
* Makes a clone of this product with all of the current products attributes
* excluding id and description.  The resulting product is empty of content or content references.
*
* @param newId the new id to use for the product
* @param newDescription the new description to use for the product
*/
	public Product cloneShallow(String newId, String newDescription);
/**
* Makes a clone of this product with all of the current products attributes
* including id and description.  The resulting product is empty of content or content references.
*/
	public Product cloneShallow();
/**
* Make a clone of a product all the way down to it's content.  <br>
* With String and URL, the reference is copied but a duplicate 
* String or URL or Fileref are not created.
* With Vector and Hashtable the structure is duplicated.  All product
* attributes encountered along the way are copied.
*
*/
	public Product cloneDeeper(String newId, String newDescription);
/**
* Make a clone of a product all the way down to it's content.  <br>
* With String and URL, the reference is copied but a duplicate 
* String or URL or Fileref are not created.
* With Vector and Hashtable the structure is duplicated.  All product
* attributes encountered along the way are copied.
*
*/
	public Product cloneDeeper();
/**
* Apply the filter to this product.  If this product is a container 
* then also apply to all the contained products. Iteration should stop
* if the filter returns false.
* @return false if the filter returns false
*/
	public void apply(ProductFilter filter);
/**
* Set the products description.
*/
	public void setDescription(String description);
/**
* Increment the reference count for this product.
* Any agent that holds a reference to a product across
* its putProduct boundary should use this.
*/
	public void incrementUsage(Object user);

/**
* Decrement the reference count for this product.
* Any agent that holds a reference to a product across
* its putProduct boundary should use this.
*/
	public void decrementUsage(Object user);
/**
* A product that may contain other products
*/
	public boolean canContainProducts();
/**
* Return the root of the product if it exists. Simple
* products would not have a root but this makes life easier.
* @return null if there is no root/index page
*/
	public Product getRoot();
	
}
