/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.product.view;

import java.util.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.File;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.interfaces.ProductPool;

import org.mitre.pam.store.FileRef;
import org.mitre.pdif.eventParser.PDIFWriter;
import org.mitre.pam.executive.ExecGlobal;

/**
* A basic product can be anything almost.<p>
*
* @author Nazario Irizarry, Jr.
* @author Jason Mathews, added digest members and methods
*/
public class BasicProduct extends ProductShell {


    public static final int DIGEST_NONE = 0;
    public static final int DIGEST_OTHER = 1;
    public static final int DIGEST_ID = 2;
    public static final int DIGEST_DESCRIPTION = 3;
    public static final int DIGEST_CONTENT = 4;
/**
* Helps keep track of whether agents are actually using this product
*/
	transient protected int usageInPool;
/**
* Identifies the product pool that we belong to
*/
	transient BasicProductPoolImpl  poolParent;
/**
* the contents can be anything, it is up to a renderer to understand
*/
	protected Object content = "";

/**
* Message digest or signature of the contents to determine equality
*/
    protected byte[] digest;

/**
* Level of digest computed: content, description, etc.
*/
    protected int digestType = DIGEST_NONE;

/**
* only need id and description to have a product of some sort
*/
	public BasicProduct (String id, String description) {
		super(id, description);
	}
/**
* used only by persistence subsystem
*/
	protected BasicProduct () {
		super();
	}

/**
* Return the content of this product.  
*/
	public Object getContent(){
		return content;
	}

/**
 * Computes message digest on the content of this product
 *
 * @return digest as a byte array if computable, null if it cannot be determined
*/
    public byte[] getDigest() {
      if (digest == null)
        computeDigest();
      return digest;
    }

/**
* Returns a string representation of this product's message digest
*/
    public String getDigestString() {
	if (getDigest() == null) return "";
	return hexDigits(digest);
    }

   /**
     * Return message digest bytes as a hex string
     */
    private static String hexDigits (byte[] bytes) {
        char c;	
	StringBuffer buf = new StringBuffer(40); // hold 40-hex digits
	for (int i=0; i < bytes.length; i++) {
	  c = (char) ((bytes[i] >> 4) & 0xf);
	  if (c > 9) {
	    c = (char) ((c - 10) + 'a');
	  } else {
	    c = (char) (c + '0');
	  }
	  buf.append(c); // high-order bits
	  c = (char) (bytes[i] & 0xf);
	  if (c > 9) {
	    c = (char)((c - 10) + 'a');
	  } else {
	    c = (char)(c + '0');
	  }
	  buf.append(c); // low-order bits
	}
	return buf.toString();
    }

    protected int getDigestType() {
	return digest != null ? digestType : DIGEST_NONE;
    }

    protected void resetDigest() {
      digest = null;
    }

    protected void computeDigest() {
      digest = null; // first clear the digest
      try {
        if (content != null) {
          if (content instanceof String) {
            computeDigest( (String) content);
          } else if (content instanceof URL) {
            computeDigest( ((URL) content).toString() );
          } else if (content instanceof FileRef) {
            computeDigest( (FileRef) content );
          } else if (content instanceof byte[]) {
            computeDigest( (byte[]) content );
          } else {
            assert(false,"unknown product content type: "+content.getClass()+" on product id="+getId());
            return;
          }
          if (digest != null) {
	      digestType = DIGEST_CONTENT;
	      return;
	  }
	}

	String description = getDescription();
	if (description != null && description.length() != 0) {
	    computeDigest(description);
	    if (digest != null) {
		digestType = DIGEST_DESCRIPTION;
		return;
	    }
	}

	// use product id as last resort which cannot be null
	computeDigest(getId());
	digestType = DIGEST_ID;

      } catch (NoSuchAlgorithmException e) {
	  reportException(e);
      }
    }

    private void computeDigest(String s) throws NoSuchAlgorithmException {
	MessageDigest md = MessageDigest.getInstance("SHA-1");
	byte[] b = s.getBytes();
	digest = md.digest(b);
    }

    private void computeDigest(byte[] b) throws NoSuchAlgorithmException {
	MessageDigest md = MessageDigest.getInstance("SHA-1");
	digest = md.digest(b);
    }

    private void computeDigest(FileRef fileRef) throws NoSuchAlgorithmException
    {
	MessageDigest md = MessageDigest.getInstance("SHA-1");
	FileInputStream in = null;
	try {
	    byte[] b = new byte[4096];
	    in = new FileInputStream(fileRef.getFile());
	    int len;
	    while (true) {
		len = in.read(b, 0, 4096);
		if (len <= 0) break;
		md.update(b, 0, len);
	    }
	    digest = md.digest();
	}
	catch (IOException e) {
	    reportException(e);
	} finally {
	    if (in != null) {
		try {
		    in.close();
		} catch (IOException ignore) {}
	    }
	}
    }

/**
* Compares two basic products for equality.
*/
    public boolean equals(Object obj) {
	if (this == obj) return true; // same object
	if (! (obj instanceof BasicProduct)) return false;
	BasicProduct bp = (BasicProduct) obj;
	if (getDigest() == null) return false; // cannot compute digest
	byte[] objDigest = bp.getDigest();
	if (objDigest == null) return false; // cannot compute digest
	if (digestType != bp.getDigestType()) return false;
	return MessageDigest.isEqual(digest, objDigest);
    }

    public void setId(String id) {
	super.setId(id);
	// if the digest is affected by a change in ID
	// then it must be invalidated.
	if (digest != null && digestType <= DIGEST_ID)
	    resetDigest();
    }

    public void setDescription(String description) {
	super.setDescription(description);
	// if the digest is affected by a change in description
	// then it must be invalidated.
	if (digest != null && digestType <= DIGEST_DESCRIPTION)
	    resetDigest();
    }

/**
* Set the content of this product to be a String
*/
	public void setContent(String content){
		if (content == null)
			this.content = "";
		else
			this.content = content;
		resetDigest();
	}
/**
* Set the content of this product to be byte array
*/
	public void setContent(byte[] content){
		if (content == null)
			this.content = "";
		else
			this.content = content;
		resetDigest();
	}
/**
* Set the content of this product to be a list of other products.
*/
	public void setContent(Vector content){
		throw new RuntimeException("BasicProduct no longer supports setContent(Vector), try using VectorProduct");
	}
/**
* Set the content of this product to be a URL
*/
	public void setContent(URL content){
		if (content == null)
			this.content = "";
		else
			this.content = content;
		resetDigest();
	}
/**
* Set the content of this product to be a FileRef
*/
	public void setContent(FileRef content){
		if (content == null)
			this.content = "";
		else
			this.content = content;
		resetDigest();
	}
/**
* Add to vector contents.
*/
	public void addElement(Object element) {
		throw new RuntimeException("BasicProduct no longer supports addElement(Object), try using VectorProduct");
	}
	
/**
* @return Enumeration of PDIFItems or null
*/
	public Enumeration getRelatedPDIFItems(){
		return null;
	}

/**
* Store the contents of this product.
*/
	protected void storeContent(PDIFWriter pdifOut) throws IOException {
		if (content == null) {
			assert(false, "Null content discovered on: "+toString());
			return;
		}
		if (content instanceof FileRef) {
			FileRef f = (FileRef) content;
			if (f.getFile() != null)
				pdifOut.writeAttributeValue(PDIF_CONTENT+"."+nameFor(content), f.serialOut());
			else {
				reporter.warning( "Missing File for "+getId()+" "+f.getSourceName());
				pdifOut.writeAttributeValue(PDIF_CONTENT+"."+"String", "Missing file: "+f.getSourceName());
			}
		}
		else if (content instanceof byte[]) {
			pdifOut.writeAttributeValue(PDIF_CONTENT+"."+nameFor(content), "\"\"");
		}
		else
			pdifOut.writeAttributeValue(PDIF_CONTENT+"."+nameFor(content), content.toString());
	}
	
/**
* load content
*/
	protected void loadContent(String attribute, String value) {
		String type = attribute.substring(PDIF_CONTENT.length()+1);
		if (value == null)
			value = "";
	/*
	* String content
	*/
		if ( type.equals("String") ) {
			setContent(value);
		}
	/*
	* Byte content
	*/
		else if ( type.equals("ByteArray") ) {
			setContent(new byte[1]);
		}
	/*
	* FileRef content
	*/
		else if ( type.equals("FileRef") ) {
			FileRef fr = new FileRef(value);
			setContent(fr);
		}
	/*
	* URL content
	*/
		else if ( type.equals("URL") ) {
			try {
				setContent(new URL(value));
			}
			catch (MalformedURLException e) {
				reportException(e);
			}
		}
		else {
			assert(false,"unknown product content type: "+type+" on product id="+getId());
		}

	}
	
/**
* Makes a clone of this product with all of the current products attributes
* excluding id and description.  The resulting product is empty of content or content references.
*
* @param newId the new id to use for the product
* @param newDescription the new description to use for the product
*/
	public Product cloneShallow(String newId, String newDescription){
		return cloneWithFirstLevelReferences(newId, newDescription);
	}
/**
* Makes a clone of this product with all of the current products attributes
* excluding id and description.  Also clones any references to content.
*
* @param newId the new id to use for the product
* @param newDescription the new description to use for the product
*/
	public Product cloneWithFirstLevelReferences(String newId, String newDescription){
		BasicProduct clone = new BasicProduct(newId, newDescription);
		cloneTo(clone);
		clone.content = this.content;
		return clone;
	}

/**
* Make a clone of a product all the way down to it's content.  <br>
* With String and URL, the reference is copied but a duplicate 
* String or URL or Fileref are not created.
* With Vector and Hashtable the structure is duplicated.  All product
* attributes encountered along the way are copied.
*
*/
	public Product cloneDeeper(String newId, String newDescription) {
		return cloneWithFirstLevelReferences(newId, newDescription);
	}
/**
* Simply returns true for a simple "basic" product
*/
	protected boolean mapply(ProductFilter filter){
		return true;
	}
/*
*
*/
	String nameFor(Object foo) {
		String name = foo.getClass().getName();
		if (foo instanceof byte[]) {
			return "ByteArray";
		}
		else if (foo instanceof String) {
			return "String";
		}
		else if (foo instanceof FileRef) {
			return "FileRef";
		}
		else if (foo instanceof URL) {
			return "URL";
		}
		else {
			int period = name.lastIndexOf('.');
			if ( period < 0 ) {
				return name;
			}
			return name.substring(period+1);
		}
	}
/**
* Get the size of the product
*/
	public double getSizeInBytesAsDouble(){
		if (content == null)
			return 0.0;
		double size = 0.0;
		if (content instanceof String) {
			return content.toString().length();
		}
		else if (content instanceof URL) {
			return content.toString().length();
		}
		else if (content instanceof byte[]) {
			return ((byte[])content).length;
		} 
		else if (content instanceof FileRef) {
			FileRef fr = (FileRef)content;
			if (fr.getFile() == null)
				return 0.0;
			File f = fr.getFile();
			if (f != null && f.exists() && f.isFile())
				return f.length();
		}
		return 0.0;
	}
/*
*
*/
	public String toString() {
		return "("+getId()+" "+super.toString()+")";
	}	

/**
* Increment the reference count for this product in case
* the product is part of a pool. If the product is
* a container product then it increments the reference count for
* its contents. This currently is only used for products that
* are placed on a queue.
*/
	public void incrementUsage(Object user){
	/*
	* testing for pool parent here could save time
	*/
		if (poolParent == null)
			return;
		synchronized (this) {
			usageInPool++;
			if (BasicProductPoolImpl.reporter.isDebugKey(BasicProductPoolImpl.DEBUG))
				BasicProductPoolImpl.reporter.debugMsg("++usage="+usageInPool,this,"by",user,BasicProductPoolImpl.DEBUG);
		}
	}

/**
* Decrement the reference count for this product. 
* When the count drops to zero the product is free for reuse.
* This currently is only used for products that
* are placed on a queue. <p>
*/
	public void decrementUsage(Object user){
	/*
	* testing for pool parent here could save time
	*/
		if (poolParent == null)
			return;
		synchronized (this) {
			usageInPool--;
			if (BasicProductPoolImpl.reporter.isDebugKey(BasicProductPoolImpl.DEBUG))
				BasicProductPoolImpl.reporter.debugMsg("--usage="+usageInPool,this,"by",user,BasicProductPoolImpl.DEBUG);
			if (usageInPool < 0) {
				reporter.assert(false,"BasicProduct usage went below zero! - this should not happen -",this,null,null);
			}
			if (usageInPool == 0)
				poolParent.returnToPool(this);
		}
			
	}

/**
* A product that may ontain other products
*/
	public boolean canContainProducts(){
		return false;
	}
/**
* Return the root of the product if it exists. Simple
* products would not have a root but this makes life easier.
* @return null if there is no root/index page
*/
	public Product getRoot(){
		return null;
	}
}
