/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.product.view;

import java.util.Hashtable;
import org.mitre.pam.interfaces.ProductPool;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.util.alarm.AlarmReporter;
/**
* A set of products that can be recycled
* so we can reduce the reliance on Java GC
* for near realtime applications.
*
* @author Nazario Irizarry, Jr.
*/
public class BasicProductPoolImpl implements ProductPool {
	Object key;
	BasicProduct[] members;
	BasicProduct nextAvailable;
	static private BufferEmptyException BEE = new BufferEmptyException();
	static Reporter reporter = new AlarmReporter("product pool");
	static String DEBUG = "product pool";
/*
* Indicates that the pool has been disposed of
*/
	boolean wasDisposed;
/**
* The constructor allocates a fixed array for the members of the pool
* but does not actually create the members.
* This allows the pool to be generic.  The
* factory that creates the pool must create the members of the pool.
* @param key used as a generic reference to the identity of this product pool
* @param nItemsInPool number of items in the pool
*/
	BasicProductPoolImpl(Object key, Product[] members) {
		if (members.length < 1)
			throw new IllegalArgumentException("Number of products in pool should be >= 1, request was: "+members.length);
		this.key = key;
		this.members = new BasicProduct[members.length];
		for (int i = 0; i< members.length; i++) {
			BasicProduct p = (BasicProduct) members[i];
			p.poolParent = this;
			this.members[i] = p;
		}
	}

	
	public String toString() {
		return "BasicProductPool "+key.toString()+" with "+members.length+" members";
	}

/**
* Get an unused product from the pool.  The actual type of the product
* depends on the Product class that implements this interface method.
* If an empty product is not available then the call blocks until one
* is available but not more than about 1 minute.
* This call also increments the usage count on the allocated product - so it must
* be decremented when done with this product with an explicit
* call to doneWithProduct() @see doneWithProduct;
*
* @param originator a reference to the agent that gets the product from the pool
*
* @return an unused Product
* @throws InterruptedException if it was blocking and the block was interrupted
*/
	synchronized public Product getBlocking(Object originator) 
			throws InterruptedException 
	{
		if (wasDisposed) {
			throw new RuntimeException("Call to getBlocking() after pool was disposed");
		}
		int countDown = 6;
		while (nextAvailable == null) {
			countDown--;
			scanForNextAvailable();
			if (nextAvailable != null)
				break;
			if (countDown == 5)
				reporter.debugMsg("blocking for product availability for",originator,null,null,DEBUG);
			if (countDown <= 0) {
				reporter.error("Blocked too long on request from "+originator);
				throw new InterruptedException("couldn't wait any longer");
			}
			try {
				wait(10000); // wait 10 seconds
			}
			catch (InterruptedException e){
				if (ExecGlobal.getExec().getState() == ExecGlobal.SHUTTING_DOWN)
					throw e;
			}
		}
		
		BasicProduct hold = nextAvailable;
		nextAvailable = null;
		cleanIt(hold);
		reporter.debugMsg(" get nextAvail, product:",hold,"for",originator, DEBUG);
		hold.usageInPool = 1;
		return hold;
	}

/**
* Get an unused product from the pool. 
* This call also increments the usage count on the allocated product - so it must
* be decremented when done with this product with an explicit
* call to doneWithProduct() @see doneWithProduct;
*
* @param originator a reference to the agent that gets the product from the pool
*
* @return an empty BasicProduct 
* @throws BufferEmptyException if nothing is available, ignore the stacktrace
* on this exception
*/
	synchronized public Product getWithoutBlocking(Object originator)
			throws BufferEmptyException
	{
		if (wasDisposed) {
			throw new RuntimeException("Call to getWithoutBlocking() after pool was disposed");
		}
		if (nextAvailable == null) {
			scanForNextAvailable();
			if (nextAvailable == null)
				throw BEE;
		}

		BasicProduct hold = nextAvailable;
		nextAvailable = null;
		cleanIt(hold);
		reporter.debugMsg(" got nonblocking, product:",hold,"for",originator,DEBUG);
		hold.usageInPool = 1;
		return hold;
	}
/**
* Look for an available product in the pool
*/
	private void scanForNextAvailable() {
		for (int i = 0; i< members.length; i++) {
			if (members != null && members[i].usageInPool == 0) {
				nextAvailable = members[i];
				return;
			}
		}
	}

/**
* Make the recycled product "new"
*/
	private void cleanIt(Product p) {
		p.setModificationDate(System.currentTimeMillis());
	}
/**
* Allows a member Product to add itself back to the available list.
*/
	synchronized void returnToPool(BasicProduct p) {
		if (wasDisposed) {
			p.poolParent = null; // allow gc to kick in on product
			for (int i = 0; i<members.length; i++) {
				if (members[i] == p) {
					members[i] = null;
				}
			}
		}
		else {
			nextAvailable = p;
			notify();
		}
	}
/**
* Flags that the agent that allocated the product has handed off the product
* and is now done with this product.
*
* @param originator a reference to the agent that got the product from the pool
* @param product that originator is done with
*
*/
	public void doneWithProduct(Object originator, Product p){
		p.decrementUsage(originator);
	}
/**
* Free up all space used by the product pool.
* If products are currently in use they will be freed when they
* they are no longer being used by any agents.
*/
	public void dispose() {
		wasDisposed = true;
		ProductFactoryImpl.productPools.remove(key);
	}
}