/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.product.view;

import java.util.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.io.IOException;
import java.io.StringWriter;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;

import org.mitre.pam.daemon.Dict;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.interfaces.ProductPool;

import org.mitre.pam.store.FileRef;
import org.mitre.pam.util.AttributeKey;
import org.mitre.pdif.eventParser.PDIFWriter;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.interfaces.IDictionary;
import org.mitre.pdif.eventParser.ParseEvent3;

/**
* Has a list of indexed contents.<p>
*
* This Product uses a Dictionary object which specifies the array index and datatype
* associated with each field name. <p>
*
* Field names should be unique. <p>
*
* Null will be returned when retrieving data from a field in which no value has been stored. <p>
*
* Automatically allocates more memory if necessary to store
* additional elements. <p>
*
* @author Jones Leung
* @author John R. Fantry - added Java serialization support
* @author Nazario Irizarry - eliminate integer access, generalize the access
*
* @see org.mitre.pam.interfaces.ExecCallbacks
* @see org.mitre.pam.interfaces.Dictionary
* @see org.mitre.pam.daemon.Dict
*/
public class DictionaryProduct extends BasicProduct implements java.io.Serializable {
    /** The dictionary name for this instance of DictionaryProduct */
    public static final String DICTIONARY_NAME = "iPAMDictionaryName";
    /** The dictionary name for this instance of DictionaryProduct */
    public static final String DEFAULT_DICTIONARY_NAME = "defaultDictionary";

    /** The dictionary used by this Product (retrieved using the getDictionary() exec callback) */
    protected transient IDictionary dictionary = null;

    // used during deserialization to flag whether or not this product is valid
    // if the referenced dictionary object is not available then the product is invalid
    protected transient boolean valid = false;


/**
* Gets invoked via Class.forName() when we restore products from
* persistent store in ProductStore.
*
*/
    public DictionaryProduct () {
    }
/**
* Deserialize from pdif stream.  
* Invoke super class behavior to restore properties.
* Then restore the dictionary.
*/
	void PDIFSerialIn(ParseEvent3 pEvent) {
		super.PDIFSerialIn(pEvent);
		valid = true;
        restoreDictionary();
	}
/**
* Deserialize from Java serialization.
* Invoke default deserialization to restore properties.
* Then restore the dictionary.
*/
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		valid = true;
		restoreDictionary();
	}
/**
* Build a dictionary product.
* The dictionary name should not be the same as that of an existing agent.
* Multiple products can share a dictionary. The dictionary can be created
* via ExecCallbacks.getDictionary(String).  @see org.mitre.pam.interfaces.ExecCallbacks#getDictionary
*
* @param id             the id of this product
* @param description    the product's description
* @param dictionary 	the dictionary to be used for this product
*/
    public DictionaryProduct (String id, String description, IDictionary dictionary) {
        super (id, description);
        setDictionary((Dict) dictionary);
        properties.useDictionary(dictionary);
    }
/**
* Return the content of this product.
* @return   a formatted String listing all of the fieldnames and their values
*/
	public Object getContent(){
		reporter.assert(dictionary != null, "null dictionary",null,null,null);

		int i;
		StringWriter sw = new StringWriter ();
		Enumeration e = getProperties();
		while (e.hasMoreElements()) {
	        String fieldname = (String)e.nextElement();
			if (fieldname == null || (fieldname.length() == 0))
				continue;
			
			String value = getProperty(fieldname);
			
        	if (value != null) {
				sw.write(fieldname);
			/*
				sw.write("(");
				sw.write(dictionary.getDataTypeAt(i));
				sw.write(")");
			*/
				sw.write(": ");
				sw.write(value.toString());
				sw.write("\n");
        	}
		}
		return sw.toString();
	}
/**
* Gets dictionary used by this product
* @return the Dictionary object used by this Product
*/
    public IDictionary getDictionary() {
		return dictionary;
    }

/**
* Makes a clone of this product with all of the current products attributes
* excluding id and description.  Also clones any references to content.
*
* @param newId the new id to use for the product
* @param newDescription the new description to use for the product
*/
	public Product cloneWithFirstLevelReferences(String newId, String newDescription){
		DictionaryProduct clone = new DictionaryProduct(newId, newDescription, dictionary);
		cloneTo(clone);
		return clone;
	}

	public boolean isValid() {
		return valid;
	}

/**
* Restores this DictionaryProduct's reference to its Dictionary object
*/
	private void restoreDictionary() {
        setDictionaryName(getStringProperty (DICTIONARY_NAME, DEFAULT_DICTIONARY_NAME));
        properties.useDictionary(dictionary);
    }


	private void setDictionaryName(String name) {
		dictionary = Dict.makeDict(ExecGlobal.getExec(), name) ;
		super.setProperty(DICTIONARY_NAME, dictionary.getDictionaryName());
	}
	private void setDictionary(Dict dict) {
		dictionary = dict;
		super.setProperty(DICTIONARY_NAME, dictionary.getDictionaryName());
	}

}