/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.product.view;

import java.util.*;
import java.net.URL;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.Observable;

//import org.mitre.pam.executive.PropertySet;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.interfaces.Memento;
import org.mitre.pam.interfaces.HasProperties;
import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.interfaces.Identifiable;
import org.mitre.pam.interfaces.Reporter;

import org.mitre.pam.store.FileRef;
import org.mitre.pdif.eventParser.PDIFItem;

import org.mitre.pdif.eventParser.ParseEvent3;
import org.mitre.pdif.eventParser.PDIFWriter;
import org.mitre.pdif.eventParser.BadPDIFException;
import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.util.AttributeKey;
import org.mitre.pam.executive.PropertySet;

/**
* A basic product can be anything almost.<p>
*
* @author Nazario Irizarry, Jr.
* @author John R. Fantry - added support for Java serialization
*/
abstract public class ProductShell extends Observable implements Product, PDIFItem, Serializable {
/**
* This will be the exec's reporter
*/
	static protected Reporter reporter;
/**
* Used by the exec to define itself
*/
	static protected ExecCallbacks exec;
/**
* Used as the arg for notification that this products ID has changed
*/
	static String CHANGED_ID = "idChanged";
/**
* Used as the arg for notification that this products description has changed
*/
	static String CHANGED_DESCRIPTION = "descriptionChanged";
/**
* The constant the defines the PDIF tag for product "content"
*/
	static final String PDIF_CONTENT = "content";
/**
* The constant the defines the PDIF tag for "product"
*/
	static final String PDIF_PRODUCT = "product";

/**
* the character used to separate topics when concatenated to form the id
* it is of String type to make it easy to use in the StringTokenizer argument list.
*/
	public static final String TOPIC_SEPARATOR = ",";
/**
* used for generating unique ids
*/
	static Random rand = new Random(System.currentTimeMillis());
/**
* uid
*/
  protected transient String uid = generateUid();

/**
* modification date refers to its contents, not its attributes
*/
	protected Date modificationDate;
/**
* properties are defined by visitors
*/
	protected PropertySet properties;
	
/**
* Only need id and description to have a product of some sort.
* Default lifetime is 3 days.
*/
	public ProductShell (String id, String description) {
		this();
		setDescription(description);
		setId(id);
		setLifetime(60*60*24*3); // default lifetime is 3 days
		this.modificationDate = new Date();
	}
/**
* used only by persistence subsystem
*/
	public ProductShell () {
		createBagOfProperties();
	}
/**
* Allows subclasses to make specialized properties 
*/
	void createBagOfProperties() {
//		properties = exec.makeNewMemento(this);
		properties = new PropertySet(getKind(), "generic", reporter);
	}
/*
* used internally
*/
	static private String generateUid() {
		long uid = rand.nextLong();
		return "bp"+Long.toString(uid>0 ? uid : -uid);
	}
/**
* The identification of this item.
*/
	public String getId() {
		String id = properties.getProperty(PROPERTY_ID);
		return id == null ? "" : id;
	}
/**
* The topic of this product, may be null. 
*/
	public String getTopic() {
		return getProperty(PROPERTY_TOPIC);
	}
/**
* Set the products topic.  The topic
* is actually stored as a property.
*/
	public void setTopic(String topic){
		setProperty(Product.PROPERTY_TOPIC, topic);
	}
/**
* Typically this is the first sentence of the long summary.
* This is intended for lists of one-liners.
*/
	public String getDescription(){
		return properties.getStringProperty(PROPERTY_DESCRIPTION,"");
	}
/**
* Typically this is the first sentence of the long summary.
* This is intended for lists of one-liners.
*
* @param description if null then a description of "" is used
*/
	public void setDescription(String description){
		if (description == null)
			description = "";
		properties.setProperty(PROPERTY_DESCRIPTION, description);
		setChanged();
		notifyObservers(CHANGED_DESCRIPTION);
	}
/**
* Return the last modification or augmentation date of this product.
*/
	public Date getModificationDate(){
		return modificationDate;
	}
/**
* Set the last modification or augmentation date of this product.
*/
	public void setModificationDate(Date date){
		modificationDate = date;
	}
/**
* Set the last modification or augmentation date of this product.
*/
	public void setModificationDate(long date){
		modificationDate = new Date(date);
	}
/**
* Return the content of this product.  
*/
	abstract public Object getContent();
/**
* Get this products property list
*/
	public String getProperty(String propertyName){
		return properties.getProperty(propertyName);
	}
/**
* Get multiple values
*/
	public ValueList getPropertyValues(String property) {
		return properties.getPropertyValues(property);
	}
/**
* Get a boolean-valued property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public boolean getBooleanProperty(String property, boolean defaultValue){
		return properties.getBooleanProperty(property, defaultValue);
	}
/**
* Get a integer-valued property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public int getIntegerProperty(String property, int defaultValue){
		return properties.getIntegerProperty(property, defaultValue);
	}
/**
* Get a long-valued property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public long getLongProperty(String property, long defaultValue){
		return properties.getLongProperty(property, defaultValue);
	}
/**
* Get a generic object that holds the value of the key.
* @return an object of String, Numeric, or ValueList type, or null
*/
	public Object getPropertyAsObject(String property){
		return properties.getPropertyAsObject(property);
	}
/**
* Get a double-valued property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public double getDoubleProperty(String property, double defaultValue){
		return properties.getDoubleProperty(property, defaultValue);
	}
/**
* Get a single string-valued property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public String getStringProperty(String property, String defaultValue){
		return properties.getStringProperty(property, defaultValue);
	}
/**
* @return true if the property exists
*/
	public boolean isProperty(String propertyName){
		return properties.isProperty(propertyName);
	}

/**
* Set this products property list
*/
	public void setProperty(String propertyName, String value) {
		
	    if (value == null) {
			removeProperty(propertyName);
			return;
		}
		if (propertyName.equals(PROPERTY_ID))
			setId(value);
	    else if (propertyName.equals(PROPERTY_DESCRIPTION))
			setDescription(value);
		else
			properties.setProperty(propertyName,value);
	    
	}

	public void setPropertyValues(String property, ValueList value) {
		if (value == null) {
			removeProperty(property);
			return;
		}

		properties.setPropertyValues(property, value);
	}
/**
* Set a boolean-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public void setProperty(String property, boolean value){
		setProperty(property, value?"true":"false");
	}
/**
* Set a single integer-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(String property, int value){
		setProperty(property, Integer.toString(value));
	}
/**
* Set a single double-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(String property, double value){
		setProperty(property, Double.toString(value));
	}
/**
* Set a single long-valued property
* @param property name of the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(String property, long value){
		setProperty(property, Long.toString(value));
	}

/**
* Remove a property from the property list
*/
	public void removeProperty(String propertyName){
		properties.removeProperty(propertyName);
	}
/**
* Remove a property but tag it so that remote, dependent
* data copies can synchronize with this deletion
*/
	public void removePropertyAndRemoteCopies(String property){
		properties.removePropertyAndRemoteCopies(property);
	}

/**
* @return 0 if property does not exist, has been deleted, or is an empty valuelist
* @return 1 if string or valuelist with 1 object
* @return >1 if valuelist with more than one value
*/
	public int getCurrentDimension(String property){
		return properties.getCurrentDimension(property);
	}
/**
* Copy the property and value potentially with a different name
*/
	public void copyPropertyFromAs(String fromName, HasProperties fromOther, String newName){
		if (fromOther.getCurrentDimension(fromName) > 1)
			setPropertyValues(newName, fromOther.getPropertyValues(fromName));
		else
			setProperty(newName, fromOther.getProperty(fromName));
	}
	
/* +++++++++  Implementation of HasKeyedProperties interface +++++ */
/**
* Remove a property without regard for other remote
* copies of the data
*
*/
	public void removeProperty(AttributeKey propertyKey){
		properties.removeProperty(propertyKey);
	}
/**
* Remove a property but tag it so that remote, dependent
* data copies can synchronize with this deletion
*
*/
	public void removePropertyAndRemoteCopies(AttributeKey propertyKey){
		removePropertyAndRemoteCopies(propertyKey.toString());
	}
/**
* Get a property
*/
	public ValueList getPropertyValues(AttributeKey propertyKey){
		return getPropertyValues(propertyKey.toString());
	}
/**
* Set a property
*/
	public void setPropertyValues(AttributeKey propertyKey, ValueList values){
	    if (values == null) {
			removeProperty(propertyKey);
			return;
		}
		properties.setPropertyValues(propertyKey, values);
	}
/**
* Get a boolean-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public Boolean getProperty(AttributeKey propertyKey, Boolean defaultValue){
		return properties.getProperty(propertyKey, defaultValue);
	}
/**
* Get a single string-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public String getProperty(AttributeKey propertyKey, String defaultValue){
		return properties.getProperty(propertyKey, defaultValue);
	}
/**
* Get a single integer-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public Integer getProperty(AttributeKey propertyKey, Integer defaultValue){
		return properties.getProperty(propertyKey, defaultValue);
	}
/**
* Get a Long-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public Long getProperty(AttributeKey propertyKey, Long defaultValue){
		return properties.getProperty(propertyKey, defaultValue);
	}
/**
* Get a Double-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public Double getProperty(AttributeKey propertyKey, Double defaultValue){
		return properties.getProperty(propertyKey, defaultValue);
	}

/**
* Set a String-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, String value){
	    if (value == null) {
			removeProperty(propertyKey);
			return;
		}
		if (propertyKey.getName().equals(PROPERTY_ID))
			setId(value);
	    else if (propertyKey.getName().equals(PROPERTY_DESCRIPTION))
			setDescription(value);
		else
			properties.setProperty(propertyKey, value);
	}
/**
* Set a Boolean-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, Boolean value){
		properties.setProperty(propertyKey, value);
	}
/**
* Set a Integer-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, Integer value){
		properties.setProperty(propertyKey, value);
	}
/**
* Set a Long-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, Long value){
		properties.setProperty(propertyKey, value);
	}
/**
* Set a Doublee-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, Double value){
		properties.setProperty(propertyKey, value);
	}


/**
* Set a boolean-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return true or false, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, boolean value){
		properties.setProperty(propertyKey, value);
	}
/**
* Set a int-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, int value){
		properties.setProperty(propertyKey, value);
	}
/**
* Set a long-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, long value){
		properties.setProperty(propertyKey, value);
	}
/**
* Set a double-valued property
* @param propertyKey key for the property
* @param defaultValue value to use if the property does not exist
* @return the value, if the property does not exist false is returned
*/
	public void setProperty(AttributeKey propertyKey, double value){
		properties.setProperty(propertyKey, value);
	}



/**
* @return true if the property exists
*/
	public boolean isProperty(AttributeKey propertyKey){
		return properties.isProperty(propertyKey);
	}

/* ------------ End of HasKeyedProperties implementation ------------ */

	
/**
* Get unique id, used for persistence
*/
	public String getUid() {
		return uid;
	}
/**
* Get the pdif kind of the product
*/
	public String getKind() {
		return getClass().getName();
	}
/**
* Get list of properties
*/
	public Enumeration getProperties() {
		return properties.getProperties();
	}
/**
* Set unique id, for use in deserialization
*/
	protected void setUid(String uid) {
		this.uid = uid;
		properties.setParentInfo(getKind(), uid);
	}
/**
* Set the lifetime of the product in seconds
*/
	public void setLifetime(long life) {
		setProperty(PROPERTY_LIFETIME, Long.toString(life));
	}
/**
* Get the lifetime of the product in seconds.
*/
	public long getLifetime() {
		return Long.parseLong(getProperty(PROPERTY_LIFETIME));
	}
/**
* Set id, for use in deserialization
*/
	public void setId(String id) {
		if (id == null || id.length() == 0) {
			assert(false, "product set with null id");
			id = "<no id>";
		}
		properties.setProperty(PROPERTY_ID, id);
		
		setChanged();
		notifyObservers(CHANGED_ID);
	}
/**
* Load from pdif stream.  When called the pdif stream
*/
	void PDIFSerialIn(ParseEvent3 pEvent) {
	
		setUid(pEvent.node.id);
		try {
			int type = pEvent.next();
			while ( type == ParseEvent3.NODE_ATTRIBUTE ) {
				String value = pEvent.attribute.getValue();
				if (value == null)
					value = "";
				loadPart(pEvent.attribute.name, value);
				type = pEvent.next();
			}
		}
		catch (IOException e) {
			reportException(e);
		}
		catch (BadPDIFException e) {
			reportException(e);
		}
		catch (Exception e) {
			reportException(e);
		}

	}
/**
* Allows exec to define itself
*/
	static public void setExec(ExecCallbacks exec, Reporter reporter) {
		ProductShell.exec = exec;
		ProductShell.reporter = reporter;
	}
/**
* report an exception
*/
	protected void reportException(Throwable e) {
		reporter.reportException(e);
	}
/**
* report an exception
*/
	protected void assert(boolean cond, String msg) {
		reporter.assert(cond, msg, null, null, null);
	}
/**
* Build product attributes, but not vector content
*/
	private void loadPart( String attribute, String value){
	/*
	* restore id
	*/
		if ( attribute.equals(PROPERTY_ID) ) {
			setId(value);
		}
	/*
	* restore description
	*/
		else if ( attribute.equals(PROPERTY_DESCRIPTION) ) {
			setDescription(value);
		}
	/*
	* restore modification date
	*/
		else if ( attribute.equals(PROPERTY_MODIFICATION_DATE) ) {
			try {
				setModificationDate(Long.parseLong(value));
			}
			catch (NumberFormatException e) {}						
		}
	/*
	* restore immediate content
	*/
		else if ( attribute.startsWith(PDIF_CONTENT, 0) ) {
			loadContent(attribute, value);
		}
	/*
	* restore visitor attribute
	*/
		else  {
			setProperty(attribute, value);
		}
	}
/**
*
*/
	abstract protected void storeContent(PDIFWriter pdifOut) throws IOException;
/**
* load content
*/
	abstract protected void loadContent(String attribute, String value);
	
/**
*
*/
	public void storeRelations(PDIFWriter pdifOut) throws IOException{
	}

/**
* Go through the body of the product and recursively
* output its parts without relations.
*
* 
*/
	public void storeParts(PDIFWriter pdifOut) throws IOException {
			
//		pdifOut.addPartAlreadyOutput(getUid(), this);

	/*
	* uid
	*/
		pdifOut.writeNode( getKind(), getUid() );
	/*
	* id
	*/
//		pdifOut.writeAttributeValue( PROPERTY_ID, getId());
	/*
	* description
	*/
//		pdifOut.writeAttributeValue( PROPERTY_DESCRIPTION, getDescription());
	/*
	* modification date as long
	*/
		pdifOut.writeAttributeValue( PROPERTY_MODIFICATION_DATE, getModificationDate().getTime());
	/*
	* properties
	*/
		Enumeration e=getProperties();

		while ( e.hasMoreElements() ) {

			String property = (String) e.nextElement();
			if (property == null)
				assert(false, "ProductShell.storeParts with null property "+getId());
			ValueList values = getPropertyValues(property);
			if (values == null) // || size() == 0 ??
				assert(false, "ProductShell.storeParts with null property value "+getId());

			if (values != null) {
				pdifOut.writeAttributeValue(property, values);
			}
		}
	/*
	* immediate content
	*/
		storeContent(pdifOut);
				
	}
/**
* Initializes the product with this objects 
* product properties, id, topic, 
* modification date, description. Does not clone the uid.
*/
	protected void cloneTo(ProductShell clone){
		
		clone.modificationDate = this.modificationDate;
		
		if (properties.size() > 0) {
			Enumeration attrs = properties.getProperties();
			while (attrs.hasMoreElements()) {
				String attr = (String) attrs.nextElement();
				String value = properties.getProperty(attr);
								
				if (!attr.equals(Memento.MEMENTO_UID))
					clone.setProperty(attr,value);
			}
		}
		
	}
/**
* Makes a clone of this product with all of the current products attributes
* including id and description.  Also clones any references to content.
*/
	public Product cloneWithFirstLevelReferences(){
		return cloneWithFirstLevelReferences(getId(), getDescription());
	}
/**
* Makes a clone of this product with all of the current products attributes
* excluding id and description.  Also clones any references to content.
*
* @param newId the new id to use for the product
* @param newDescription the new description to use for the product
*/
	abstract public Product cloneWithFirstLevelReferences(String newId, String newDescription);
/**
* Makes a clone of this product with all of the current products attributes
* excluding id and description.  The resulting product is empty of content or content references.
*
* @param newId the new id to use for the product
* @param newDescription the new description to use for the product
*/
	abstract public Product cloneShallow(String newId, String newDescription);
/**
* Makes a clone of this product with all of the current products attributes
* including id and description.  The resulting product is empty of content or content references.
*/
	public Product cloneShallow(){
		return cloneShallow(getId(), getDescription());
	}
/**
* Make a clone of a product all the way down to it's content.  <br>
* With String and URL, the reference is copied but a duplicate 
* String or URL or Fileref are not created.
* With Vector and Hashtable the structure is duplicated.  All product
* attributes encountered along the way are copied.
*
*/
	abstract public Product cloneDeeper(String newId, String newDescription);
/**
* Make a clone of a product all the way down to it's content.  <br>
* With String and URL, the reference is copied but a duplicate 
* String or URL or Fileref are not created.
* With Vector and Hashtable the structure is duplicated.  All product
* attributes encountered along the way are copied.
*
*/
	public Product cloneDeeper() {
		return cloneDeeper(getId(), getDescription());
	}
/**
* Apply the filter to ourselves and to anything we contain
*/
	public void apply(ProductFilter filter){
		if (!filter.filter(null, this))
			return;
		mapply(filter);
	}
/**
* Apply the filter to anything we contain, breadth first, so as
* to allow filters that may be deleting children during
* the application
*/
	abstract protected boolean mapply(ProductFilter filter);

  private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
    in.defaultReadObject();

    // get a new unique id
    uid = generateUid();
  }
 
	public String toString() {
		String id = getId();
		if (id == null)
			return super.toString();
		return "Product \""+id+"\"";
	}
}