/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.product.view;

import java.util.*;

import java.io.IOException;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pdif.eventParser.PDIFWriter;
import org.mitre.pdif.eventParser.PDIFItem;



/**
* A vector product has contents which are a vector of products.<p>
*
* @author Nazario Irizarry, Jr.
*/
public class VectorProduct extends EnumeratedProduct implements ContainerProduct {
	Vector vector = new Vector(10);
/**
* only need id and description to have a product of some sort
*/
	public VectorProduct (String id, String description) {
		super(id, description);
	}
/**
* used only by persistence subsystem
*/
	protected VectorProduct () {
		super();
	}
/**
* Add to contents.
*/
	public void addElement(Product element) {
		if (element == null)
			throw new IllegalArgumentException( "argument cannot be null");
		vector.addElement(element);
	}

/**
* Add to contents, role is ignored
*/
	public void addElement(String role, Product element) {
		addElement(element);
	}
/**
* Remove from content
*/
	public void removeElement(Product element) {
		if (element == null)
			throw new IllegalArgumentException( "argument cannot be null");
		vector.removeElement(element);
	}
/**
* @return true if element is already contained
*/
	public boolean containsElement(Product element) {
		return vector.contains(element);
	}
/**
* @return true if element is already contained
*/
	public boolean isEmpty() {
		return vector.size() <= 0;
	}

/*
* Add to contents, and stamp element with a PROPERTY_INDEX attribute
* Never used
*/
/*
	public void addIndexElement(Product element) {
		assert(element != null, "VectorProduct.addIndexElement(null) call");
		element.setProperty(PROPERTY_INDEX, PROPERTY_INDEX);
		vector.addElement(element);
	}
*/
/*
* Add to contents, and stamp element with a PROPERTY_ROOT attribute
*/
	public void addRootElement(Product element) {
		assert(element != null, "VectorProduct.addRootElement(null) call");
		element.setProperty(PROPERTY_ROOT, PROPERTY_ROOT);
		vector.addElement(element);
	}
/**
* Get the enumeration of vector contents
*/
	public Object getContent() {
		return vector.elements();
	}
/**
* Set the contents to a vector that you have, by default this
* product creates its vector at construction time so you can directly
* use addElement().
*/
	public void setContent(Vector content) {
		if (content == null) {
			throw new IllegalArgumentException( "argument cannot be null");
		}
		this.vector=content;
	}
/**
* Get the size of the product
*/
	public double getSizeInBytesAsDouble(){
		double size = 0.0;
		Enumeration e = vector.elements();
		while (e.hasMoreElements()) {
			size += ((Product)e.nextElement()).getSizeInBytesAsDouble();
		}
		return size;
	}
/**
* Get size of vector
*/
	public int size() {
		return vector.size();
	}
/**
* Get n'th element
*/
	public Product elementAt(int n) {
		return (Product) vector.elementAt(n);
	}

/**
* Get size of vector
*/
	public void setElementAt(Product p, int n) {
		vector.setElementAt(p, n);
	}

/**
* Used by persistence system
*/
	public void storeRelations(PDIFWriter pdifOut) throws IOException{
		Enumeration cv = (Enumeration) getContent();
		
		if (!cv.hasMoreElements())
			return;

		String uidOut = getUid();

		while (cv.hasMoreElements()) {
			PDIFItem toProduct = (PDIFItem) cv.nextElement();
			pdifOut.writeRelation(
				getKind(), "from", getUid(),  
				toProduct.getKind(), "to", toProduct.getUid());
		}
	}
/**
* Returns a copy of this product with String or URL or FileRef
* contents, but without enumerated contents.  
* Also clones a copy of product properties, id, topicVector, 
* modification date, description
*/
	public Product cloneShallow(String newId, String newDescription){
		VectorProduct clone = new VectorProduct(newId, newDescription);
		
		cloneTo(clone);
		
		return clone;
	}
/**
* Invokes cloneShallow() and then copies the first level references
* to other products
*/
	public Product cloneWithFirstLevelReferences(String newId, String newDescription){
		VectorProduct clone = (VectorProduct) cloneShallow(newId, newDescription);
		Enumeration enum = vector.elements();
		while(enum.hasMoreElements()) {
			clone.addElement((Product)enum.nextElement());
		}
		
		return clone;
	}
/**
* Make a clone of a product all the way down to it's content.  <br>
* With String and URL, the reference is copied but a duplicate 
* String or URL or Fileref are not created.
* With Vector and Hashtable the structure is duplicated.  All product
* attributes encountered along the way are copied.
*
*/
	public Product cloneDeeper(String newId, String newDescription) {
		VectorProduct clone = (VectorProduct) cloneShallow(newId, newDescription);
		Enumeration enum = vector.elements();
		while(enum.hasMoreElements()) {
			clone.addElement(((Product)enum.nextElement()).cloneDeeper());
		}
		
		return clone;
	}
/**
* Tells containers to eliminate branches that have no children
* @return true if this ContainerProduct should be kept; false if it should be removed
*/
	public boolean prune(){
		if (vector.size() == 0)
			return false;
			
		for (int i = vector.size()-1; i>=0; i--) {
			Product p = (Product) vector.elementAt(i);
			if (p instanceof ContainerProduct) {
				if (!((ContainerProduct)p).prune())
					vector.removeElementAt(i);
			}
		}

		if (vector.size() == 0)
			return false;
		return true;
	}
/**
* Apply the filter to everything the vector contains. Breadth first application.
* This is done from the end for a vector, in case the filter is trying to delete.
*/
	protected boolean mapply(ProductFilter filter){			
		for (int i = vector.size()-1; i>=0; i--) {
			if (!filter.filter(this, (Product) vector.elementAt(i)))
				return false;
		}
		for (int i = vector.size()-1; i>=0; i--) {
			if (!((ProductShell) vector.elementAt(i)).mapply(filter))
				return false;
		}
		return true;
	}
/**
* Increment the reference count for all products
* in this container
*/
	public void incrementUsage(Object user){
		if (ProductFactoryImpl.productPools == null)
			return;
		for (int i = vector.size()-1; i>=0; i--) {
			((Product) vector.elementAt(i)).incrementUsage(user);
		}
	}

/**
* Decrement the reference count for all products
* in this container
*/
	public void decrementUsage(Object user){
		if (ProductFactoryImpl.productPools == null)
			return;
		for (int i = vector.size()-1; i>=0; i--) {
			((Product) vector.elementAt(i)).decrementUsage(user);
		}
	}
}
