/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.putter;

import java.util.*;

import org.mitre.pam.interfaces.*;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.executive.Module;
import org.mitre.pam.executive.Command;
import org.mitre.pam.util.EnabledTable;
import org.mitre.pam.executive.QueueModuleHelper;
import org.mitre.pdif.eventParser.ValueList;

/**
 * A base class for DIMs that maintain a queue to isolate
 * the input flow from the output flow.
 *
 * @author Nazario Irizarry
 */

abstract public class QueuedDIMShell extends Module implements DIM, ManagesOwnInputQueue, AsynchronousAgent, QueueAgentAttributes {

	protected EnabledTable sourceList = new EnabledTable();
	protected QueueModuleHelper queueHelper;

	public QueuedDIMShell () {
		setProperty(HAS_INPUTS,"true");
		setProperty(Schedulable.HAS_SCHEDULE, "false");
		initializeQueuedDIM();
	}

	protected void initializeQueuedDIM() {
		queueHelper = new QueueModuleHelper(this);
	}

	public void setProperty(String property, String value) {
		if (queueHelper != null && queueHelper.setProperty(property, value))
			return;  // don't handle if the queueHelper handled it

		super.setProperty(property, value);
	}
	

/**
* Callback to receive a command.  There
* is only one thing we can do - send product.
*/
	public synchronized void doCommittedTask(){
		/* assume Command.SEND_PRODUCT */
		declareStart(ExecGlobal.WORKING);
		Task task = taskHelper.getCommittedTask();
		putProduct(task.getProduct());
	}
/**
* Accept a new product.
*/
	public void putProduct(Product p) {
		if (enabled)
			queueHelper.queueProduct(p);
		declareEnd(false); // clears committed task
	}

/**
* identify that we are starting
*/
	final public void declareStart(String state) {
		setState(state);
	}

/**
* set the conditions at end of update
*/
	final public void declareEnd(boolean ignored) {
		setState(ExecGlobal.IDLE); // must do this to clear committed task
	}

  //From AsynchronousAgent interface:
	public void enableService() {
		queueHelper.enableService();
	}

	public void disableService() {
		queueHelper.disableService();
	}

	public void pauseService() {
		queueHelper.pauseService();
	}

	public void resumeService() {
		queueHelper.resumeService();
	}

	public Product getPersistentProduct(){
		return null;
	}
  
	public void setPersistentProduct(Product p){
	}
  	
		
/**
* get the list of enabled Sources
*/
	public Hashtable getSourceList(){
		return sourceList;
	}

/**
* get enumeration of all Sources including disabled items
*/
	public Enumeration getFullSourceElementEnumeration(){
		return sourceList.allElements();
	}

	public Object getActivator() {
		return null;
	}
	
	abstract protected void putQueuedProduct(Product p);
/**
* Checks the specified putToken
* to see if it is valid, before putting the product.
* This prevents asyncrhonouse sources from sneaking product
* in while the sink is in the middle of another put or task.
*
* @param product to put
* @param the token that grants authorization to put
*/
	final public void pleasePutProduct(Product product, Task token){
		taskHelper.pleasePutProduct(product, token);
	}
	
	public String getProperty(String property) {
		queueHelper.updateStats(property);
		return super.getProperty(property);
	}
	public Memento getMemento() {
		queueHelper.updateStats(null);
		return super.getMemento();
	}
	public ValueList getPropertyValues(String property) {
		queueHelper.updateStats(property);
		return super.getPropertyValues(property);
	}
/**
* Old call, should not be used anymore
*/
	final public void startService() {
		super.startService();
	}

/**
* Old call, should not be used anymore
*/
	final public void stopService() {
		super.stopService();
	}
}
