/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
/*--------------------------------------------------------------------------
  Copyright (c) 1998 by The MITRE Corporation.  All rights reserved.
--------------------------------------------------------------------------*/
package org.mitre.pam.rmi.editor;

import java.rmi.RemoteException;
import java.util.Hashtable;

import org.mitre.pam.store.IntegrityException;

/**
* Defines methods for adding, deleting, connecting, and disconnecting
* agents.
*
* @author Michael E. Los, MIT3449RE, D510
* @author Nazario Irizarry, Jr., MITRE
*/
public interface RemoteExecToEditor extends RemoteExecToMonitor
{
  /** value agentType for defining a PIM */
	public static final String PIM_STR = "pimType";
  /** value agentType for defining a DIM */
	public static final String DIM_STR = "dimType";
  /** value agentType for defining a juggler */
	public static final String JUGGLER_STR = "jugglerType";
  //public static final String DAEMON_STR = "daemonType";

  public static final String AGENT_TYPE = "agentType";
  public static final String AGENT_ID = "agentId";

	/**
   * Create an agent if it does not exist within the iPAM runtime state;
   * modify it if it does. A agent may represent a PIM, juggler, or
   * DIM.
   * <P>
   * Notes:
   * <OL>
   * <LI>The following attributes must be specified for every agent:
   *           agentType, agentID
   * <LI>agentType is a string the denotes the type of agent. 
   *     For example: "passThru"
   * <LI>agentID is a String that uniquely identifies the agent.  The
   *           agent creator is responsible for ensuring the identifier
   *           is unique.
   * </OL>
   * <P>
   *
   * @param attributes name-value pairs that comprise the attributes
   *                   of the agent.
   *
   * @exception IntegrityException is generated if a agent of the same id already exists
   * @exception RemoteException
  */
  public void declareAgent(Hashtable attributes)
      throws IntegrityException, RemoteException;

	/**
   * Delete the agent associated with the given identifier.
   * <P>
   * @exception IntegrityException reserved for future, currently, none is thrown
   * @exception RemoteException
	*/
  public void deleteAgent(String agentId)
      throws IntegrityException, RemoteException;

	/**
   * Connect two agents.  The source agent must be a PIM or juggler.
   * The destination agent must be a juggler or DIM.
   * <P>
   * @param fromId    agent Id of the source agent
   * @param toId      agent Id of the destination (sink) agent
   *
   * @exception IntegrityException if either agent cannot be found
   * @exception RemoteException
	*/
  public void connectAgents(String fromId, String toId)
      throws IntegrityException, RemoteException;

	/**
   * Delete the connection between the two specified agents.
   * <P>
   * @param fromId    agent Id of the source agent
   * @param toId      agent Id of the destination (sink) agent
   *
   * @exception IntegrityException thrown if the fromId is not a Source or toId is not a Sink
   * @exception RemoteException
	*/
  public void disconnectAgents(String fromId, String toId)
      throws IntegrityException, RemoteException;

  /**
   * Delete all agents that have the specified attribute with the specified
   * value.  This allows the editor to set up logical groups of various kinds.
   * <P>
   * @param attribute name of the attribute on which to base the group
   * @param value     value of the attribute on which the group is based
   *
   * @exception IntegrityException is currently a placeholder (not really generated)
   * @exception RemoteException
	*/
	public void deleteGroup(String attribute, String value)
      throws IntegrityException, RemoteException;

	/**
   * Enable or disable all agents that have the specified attribute
   * with the specified value.  This allows the editor to enable
   * or disable all members of a logical group en masse.
   * <P>
   *
   * @param attribute name of the attribute on which to base the group
   * @param value     value of the attribute on which the group is based
   * @param enableState true to enable; false to disable
   *
   * @exception RemoteException
	*/
  public void setGroupEnabled(String attribute, String value, boolean enableState)
      throws RemoteException;

	/**
   * Enable or disable a list of agents identified by their identifiers.
   * All agentIds must exist in order for any to be enabled/disabled.
   * <P>
   * @param agentIDs   String array of agent identifiers
   * @param enabled   true to enable all agents; false to disable
   *
   * @exception IntegrityException if one of the agentIds does not exist.
   * @exception RemoteException
	*/
  public void setAgentsEnabled(String agentIds[], boolean enabled)
      throws IntegrityException, RemoteException;

	/**
   * Set attribute-value pair for the specified agent.  If the agent
   * contains the attribute-value pair, the value will be modified;
   * otherwise, the attribute-value pair will be associated with the agent.
   * <P>
   * @param  agentID    agent identifier
   * @param attribute  name of the attribute
   * @param value   object to be associated with the attribute
   *
   * @exception RemoteException
	*/
  public void setAgentAttributeValue(String agentId, String attribute, Object value)
      throws RemoteException;


	/**
   * Deletes the attribute-value pair from the specified agent.
   * <P>
   * @param  agentID    agent identifier
   * @param attribute  name of the attribute
   *
   * @exception RemoteException
	*/
  public void removeAgentAttribute(String agentId, String attribute)
      throws IntegrityException, RemoteException;
}