/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
/*--------------------------------------------------------------------------
  Copyright (c) 1998 by The MITRE Corporation.  All rights reserved.
--------------------------------------------------------------------------*/
package org.mitre.pam.rmi.editor;

import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.interfaces.ExecToEditor;
import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.interfaces.Memento;
import org.mitre.pam.store.IntegrityException;
import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.interfaces.Reporter;

public class RmtExecToEditorImpl extends RmtExecToMonitorImpl
                                    implements RemoteExecToEditor
{
  //--------------------------------------------
  // PRIVATE DATA MEMBERS
  //--------------------------------------------

  //--------------------------------------------
  // CONSTRUCTORS
  //--------------------------------------------
  public RmtExecToEditorImpl(ExecToEditor exec, Reporter reporter) throws RemoteException
  {
  	super(exec,reporter);
  }//RemoteExecToEditorImpl()



	/**
   * Create a agent if it does not exist within the iPAM runtime state;
   * modify it if it does. A agent may represent a PIM, juggler, or
   * DIM.
   * <P>
   * Notes:
   * <OL>
   * <LI>The following attributes must be specified for every agent:
   *           agentType, agentID
   * <LI>agentType is a string the denotes the type of agent. 
   *     For example: "passThru"
   * <LI>agentID is a String that uniquely identifies the agent.  The
   *           agent creator is responsible for ensuring the identifier
   *           is unique.
   * </OL>
   * <P>
   *
   * @param attributes name-value pairs that comprise the attributes
   *                   of the agent.
   *
   * @exception IntegrityException is generated if a agent of the same id already exists
   * @exception RemoteException
  */
  public void declareAgent(Hashtable attributes)
      throws IntegrityException, RemoteException
  {
	try {
		debug("declareAgent begin");
	    // Caller must provide agentType and agentID properties
	    String agentType = (String) attributes.get(AGENT_TYPE);
	    if (agentType == null)
	      throw new IllegalArgumentException("AGENT_TYPE attribute is not defined");
	
	    String agentID = (String) attributes.get(AGENT_ID);
	    if (agentID == null)
	      throw new IllegalArgumentException("AGENT_ID attribute is not defined");
		
	    // Direct the executive to create/modify the agent
	    exec.declareNode(agentType, agentID, attributes);
	}
	catch (Exception e) {
		reporter.reportException(e);
	}
	debug("declareAgent end");
  }//declareAgent()

	/**
   * Delete the agent associated with the given identifier.
   * <P>
   * @exception IntegrityException reserved for future use
   * @exception RemoteException
	*/
  public void deleteAgent(String agentId)
      throws IntegrityException, RemoteException
  {
	try {
		debug("deleteAgent begin - ",agentId);
		exec.deleteNode(agentId);
	}
	catch (Exception e) {
		reporter.reportException(e);
	}
	debug("deleteAgent end");
  }

	/**
   * Connect two agents.  The source agent must be a PIM or juggler.
   * The destination agent must be a juggler or DIM.
   * <P>
   * @param fromId    agent Id of the source agent
   * @param toId      agent Id of the destination (sink) agent
   *
   * @exception IntegrityException if either agent cannot be found
   * @exception RemoteException
	*/
  public void connectAgents(String fromId, String toId)
      throws IntegrityException, RemoteException
  {
	try {
		debug("connectAgents ",fromId,toId);
    	exec.connectNodes(fromId, toId);
	}
	catch (Exception e) {
		reporter.reportException(e);
	}
	debug("connectAgents end");
  }

	/**
   * Delete the connection between the two specified agents.
   * <P>
   * @param fromId    agent Id of the source agent
   * @param toId      agent Id of the destination (sink) agent
   *
   * @exception IntegrityException thrown if the fromId is not a Source or toId is not a Sink
   * @exception RemoteException
	*/
  public void disconnectAgents(String fromId, String toId)
      throws IntegrityException, RemoteException
  {
	try {
		debug("disconnectAgents ",fromId,toId);
		exec.disconnectNodes(fromId, toId);
	}
	catch (Exception e) {
		reporter.reportException(e);
	}
	debug("disconnectAgents end");
  }

  /**
   * Delete all agents that have the specified attribute with the specified
   * value.  This allows the editor to set up logical groups of various kinds.
   * <P>
   * @param attribute name of the attribute on which to base the group
   * @param value     value of the attribute on which the group is based
   *
   * @exception IntegrityException is currently a placeholder (not really generated)
   * @exception RemoteException
	*/
	public void deleteGroup(String attribute, String value)
      throws IntegrityException, RemoteException
  {
	try {
		debug("deleteGroup ",attribute,value);
		exec.deleteGroup(attribute, value);
	}
	catch (Exception e) {
		reporter.reportException(e);
	}
	debug("deleteGroup end");
  }

	/**
   * Enable or disable all agents that have the specified attribute
   * with the specified value.  This allows the editor to enable
   * or disable all members of a logical group en masse.
   * <P>
   *
   * @param attribute name of the attribute on which to base the group
   * @param value     value of the attribute on which the group is based
   * @param enableState true to enable; false to disable
   *
   * @exception RemoteException
	*/
  public void setGroupEnabled(String attribute, String value, boolean enableState)
      throws RemoteException
  {
	try {
		debug("setGroupEnabled ",attribute,value);
		exec.setGroupEnabled(attribute, value, enableState);
	}
	catch (Exception e) {
		reporter.reportException(e);
	}
	debug("setGroupEnabled end");
  }

	/**
   * Enable or disable a list of agents identified by their identifiers.
   * All agentIds must exist in order for any to be enabled/disabled.
   * <P>
   * @param agentIDs   String array of agent identifiers
   * @param enabled   true to enable all agents; false to disable
   *
   * @exception IntegrityException if one of the agentIds does not exist.
   * @exception RemoteException
	*/
  public void setAgentsEnabled(String agentIds[], boolean enabled)
      throws IntegrityException, RemoteException
  {
	try {
		debug("setAgentsEnabled begin");
		exec.setNodesEnabled(agentIds, enabled);
	}
	catch (Exception e) {
		reporter.reportException(e);
	}
	debug("setAgentsEnabled end");
  }

	/**
   * Set attribute-value pair for the specified agent.  If the agent
   * contains the attribute-value pair, the value will be modified;
   * otherwise, the attribute-value pair will be associated with the agent.
   * <P>
   * @param  agentID    agent identifier
   * @param attribute  name of the attribute
   * @param value   Object to be associated with the attribute, either a Vector, String[], or String
   *
   * @exception RemoteException
	*/
  public void setAgentAttributeValue(String agentId, String attribute, Object value)
      throws RemoteException
  {
	try {
		debug("setAgentAttributeValue -",agentId,attribute);
	    Agent agent = exec.getActiveAgent(agentId);
	    if (agent == null) {
	    	reporter.log("Request to set attribute for nonexisting agent ignored, agent="+agentId);
	    	return;
	    }
	    
	    if (value instanceof String)
	      agent.setProperty(attribute, (String) value);
	    else if ((value instanceof String []) || (value instanceof Vector))
	    {
	      ValueList vals = null;
	      if (value instanceof Vector)
	        vals = new ValueList((Vector) value);
	      else
	        vals = new ValueList((String []) value);
	
	      agent.setPropertyValues(attribute, vals);
	    }
	}
	catch (Exception e) {
		reporter.reportException(e);
	}
	debug("setAgentAttributeValue end");
  }

	/**
   * Deletes the attribute-value pair from the specified agent.
   * <P>
   * @param  agentID    agent identifier
   * @param attribute  name of the attribute
   *
   * @exception RemoteException
	*/
  public void removeAgentAttribute(String agentId, String attribute)
      throws IntegrityException, RemoteException
  {
	try {
		debug("removeAgentAttribute -",agentId,attribute);
	    Agent agent = exec.getActiveAgent(agentId);
	    if (agent == null) {
	    	reporter.log("Request to remove attribute for nonexisting agent ignored, agent="+agentId);
	    	return;
	    }
	    agent.removeProperty(attribute);
	}
	catch (Exception e) {
		reporter.reportException(e);
	}
	debug("removeAgentAttribute end");
  }
}