/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.store;

import java.io.*;
import java.util.*;

import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.interfaces.Reporter;

/**
* The static part of this object keeps a list of files that are used. Used by FileRef.
* Provides a way to support SweepAndGC.  There is one instance of this object
* for every referenced file on the disk.
*
* @author Nazario Irizarry, Jr.
* @author John R. Fantry - added suppport for Java serialization
*/
public class FileRefList implements Serializable {

	transient private static ExecCallbacks exec;
	
	private static Hashtable fileList = new Hashtable(500);
	
/**
* A flag to indicate a file that must always be kept around
*/
	protected boolean keepFlag;
/**
* The instance hold token must match the static hold token
* to disable automatic file delete on finalize
*/
	transient protected long instanceHoldToken = 0;
/**
* The instance hold token must match the static hold token
* to disable automatic file delete on finalize
*/
	transient protected static long staticHoldToken = 1;
/**
* A debug key word
*/
	static final String FILEREF_LIFE_CREATE_DELETE_LOGGING = "fileRefDbg";
/**
* The reference to the file that we proxy
*/
	transient File thisFile;
/**
* A reference count
*/
	transient private int referenceCount = 0;
/**
 * Our own reporter.  We have one to enhance debug key caching.
 */
	transient private static Reporter reporter;

	public FileRefList(File file) {
		thisFile = file;
		if (reporter.isDebugKey(FILEREF_LIFE_CREATE_DELETE_LOGGING)) {
			reporter.log("created\t"+file.toString());
		}
	}
	
	
	static public FileRefList getPeer(File file) {
		String key = file.getAbsolutePath();
		FileRefList existing = (FileRefList)fileList.get(key);
		if (existing != null)
			return existing;
		
		FileRefList peer = new FileRefList(file);
		fileList.put(key, peer);
		return peer;
	}
/**
* Change the parent on this file ref.  Used during product
* reconstruction when the product is "sent" to another
* site.
*/
	void resetParent(File parent){
		File newFile = new File(parent, thisFile.getName());
		String key = thisFile.getAbsolutePath();
		fileList.remove(key);
		
		fileList.put(newFile.getAbsolutePath(), this);
		thisFile = newFile;
	}
/**
* Delete this file
* if the exec is not shutting down and if the keepFlag is not set.
*/	
	synchronized private void delete() {
		boolean debug = reporter.isDebugKey(FILEREF_LIFE_CREATE_DELETE_LOGGING);
		if (thisFile == null) {
			if (debug) {
				reporter.log("file appears to have been purged already\t"+this);
			}
			return;
		}

		if (!thisFile.exists()) {
			if (debug) {
				reporter.log("file does not exist\t"+this);
			}
		}
		else if (!keepFlag ) {
			thisFile.delete();
			if (debug) {
				reporter.log("deleted\t"+this);
			}
			else if (exec.getExecAsAgent().getBooleanProperty(ExecGlobal.LOG_DELETED_FILES,false)){
				reporter.log("deleted\t"+thisFile.getAbsolutePath());
			}
		} else {
			if (debug) {
				reporter.log("did not delete dereferenced file\t"+this);
			}
		}
		fileList.remove(thisFile.getAbsolutePath());
		thisFile = null;
	}

/**
* When the reference count > 0 then we may not delete this file
*/	
	synchronized void incrementRefCount() {
		referenceCount++;
		instanceHoldToken = staticHoldToken;
/*
		if (reporter.isDebugKey(FileRefList.FILEREF_LIFE_CREATE_DELETE_LOGGING)) {
			reporter.log("+1\t"+this);
		}
*/
	}

/**
* When the reference count == 0 then we may delete this file
* if the exec is not shutting down and if the keepFlag is not set.
*/	
	synchronized void decrementRefCount() {
		referenceCount--;
/*
		if (reporter.isDebugKey(FileRefList.FILEREF_LIFE_CREATE_DELETE_LOGGING)) {
			reporter.log("-1\t"+this);
		}
*/
		if (referenceCount <= 0)
			delete();
	}
	
/**
* Checks to see if we have a current mark and calls delete() if we don't
*
*@param distance the minimum number of cycles that the product must have missed in
* order to be discarded. This should be a small number but > 0.  Zero
* would cause the file to be deleted even if it had the current mark.
*/
	private void purgeOne(int distance) {

		if (reporter.isDebugKey(FILEREF_LIFE_CREATE_DELETE_LOGGING)) {
			reporter.log(" test "+this);
		}

		if (staticHoldToken - instanceHoldToken >= distance) {
/*
			if (reporter.isDebugKey(FILEREF_LIFE_CREATE_DELETE_LOGGING)) {
				reporter.log("purge\t"+this);
			}
*/
			delete();
		}
	}
	
/**
* Checks all instances of FileRefList to see if they need to be purged
*
*@param distance the minimum number of cycles that the product must have missed in
* order to be discarded. This should be a small number but > 0.  Zero
* would cause even current files to be deleted.
*/
	static public synchronized void purgeAll(int distance) {
		if (distance <= 0)
			distance = 1;

		if (reporter.isDebugKey(FILEREF_LIFE_CREATE_DELETE_LOGGING)) {
			reporter.log(" purging at current token= "+staticHoldToken);
		}

		Enumeration e = fileList.elements();
		while(e.hasMoreElements()) {
			((FileRefList)e.nextElement()).purgeOne(distance);
		}
	}
/**
* Get number of files currently "alive"
*/
	static public int getLivingFileCount() {
		return fileList.size();
	}
/**
* Set our exec
*/
	public static void setExec(ExecCallbacks execRef) {
		exec = execRef;
		reporter = execRef.makeReporter("execFileMgr");
	}
/**
* Identifies whether a specified file is known to have a live FileRef
*/
	public static boolean fileIsKnown(File f) {
		return fileList.containsKey(f.getAbsolutePath());
	}
/**
* Sweeps a directory - deletes unused files.
* Used by the exec.
*
* @returns true if the directory has files when this method completes
*/
	public static boolean doSweep(File rootDir) {
		int fileCount=0;
		
		String [] fileList = rootDir.list();
		for (int i = 0; i<fileList.length; i++) {
			File fileToTest = new File(rootDir, fileList[i]);
			
			if (!fileToTest.exists()) {
				fileCount++;
			}
			else if (fileToTest.isFile()) {
				if (fileIsKnown(fileToTest))
					fileCount++;
				else {
					fileToTest.delete();
					reporter.log("deleted unused file\t"+fileToTest);
				}
			}
			else if (fileToTest.isDirectory()) {
				if (doSweep(fileToTest))
					fileCount++;
				else {
					fileToTest.delete();
					reporter.log("deleted empty dir\t"+fileToTest);
				}
			}
		}
		return fileCount > 0;
	}

/**
* Mark this with the current hold token
*/
	void mark()
	{
		instanceHoldToken = staticHoldToken;

		if (reporter.isDebugKey(FILEREF_LIFE_CREATE_DELETE_LOGGING)) {
			reporter.log("mark "+this);
		}

	}
/**
* Advance hold token
*/
	static public void advanceHoldToken()
	{
		staticHoldToken++;
/*
		if (reporter.isDebugKey(FILEREF_LIFE_CREATE_DELETE_LOGGING)) {
			reporter.log("advance token "+staticHoldToken);
		}
*/
	}
/**
* Converts this FileRefList to a <code>String</code>.
*
* @return  a string representation of this FileRefList.
*/
  public String toString() {
    return "FileRefList[keepFlag="+keepFlag
      +",file="+thisFile+",refs="+referenceCount+",token="+instanceHoldToken+"]";
  }
}