package org.mitre.pam.util;

import java.util.*;
import java.io.*;
import java.net.*;

import org.mitre.pam.executive.Exec;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.util.alarm.*;
import org.mitre.pam.interfaces.*;
import org.mitre.pam.product.view.*;
import org.mitre.pam.store.*;

/**
* This class contains static methods used to maintain an archive of product. <p>
*
* @author Jones Leung
*/
public class Archiver {
	static protected Reporter reporter = ExecGlobal.getExec().getExecAsAgent().getReporter();
    // What is the lifetime of the archived product (in days)?
    // (configured by user)
    public static final String LIFETIME = "productLifetime";
    
    // Should expired products be deleted?
    // (configured by user)
    public static final String DELETE_EXPIRED = "deleteExpired";
    
    // What is the unique date archived tag stored in an agent's memento.
    public static final String DATE_ARCHIVED_TAG = "WhatIsMyUniqueDateArchivedTag?";
    
    // debug key to turn on debugging messages
    public static final String DEBUG_KEY = "ArchiverUtils";    
    
/**
* ProductFilter which deletes expired elements from a product archive.
*/
public static class ExpireFilter implements ProductFilter {
    protected String UNIQUE_TAG = null;  // unique date archived tag
    protected int LIFETIME = 7;  // product lifetime in days

/**
* Constructor for product expiration filter.
*/ 
    public ExpireFilter (String uniqueTag, int productLifetime) {
        UNIQUE_TAG = uniqueTag;
        LIFETIME = productLifetime;
    }
/**
* Deletes expired products.<p>
*
* Note: This function determines if a product is expired based solely on a unique
* date archived tag, a date (in milliseconds) added to the product by this juggler,
* and the product lifetime which this juggler is configured with.<p>
*
* Also, retags products which have invalid date archived values, so that the date
* archived values are valid.<p>
*/
    public boolean filter (ContainerProduct parent, Product productToTest) {
        
        // this is to get by the test in apply() so that mapply() will be called
        if (parent == null) {
            return true;
        }
        
        if (System.currentTimeMillis() > (getDateArchived(productToTest, UNIQUE_TAG) + LIFETIME*1000*60*60*24)) {
/*
            reporter.debugMsg("Product is expired: "+new Date (getDateArchived(productToTest, UNIQUE_TAG)
                + LIFETIME*1000*60*60*24),null,null,null, DEBUG_KEY);
*/
            parent.removeElement (productToTest);
//            printProductInfo (productToTest, UNIQUE_TAG);           
        } 
/*        else {
            reporter.debugMsg("Product is not expired: " + new Date (getDateArchived(productToTest, UNIQUE_TAG) 
                + LIFETIME*1000*60*60*24), null,null,null, DEBUG_KEY);  
            printProductInfo (productToTest, UNIQUE_TAG);
        }
*/
        // do not stop iteration until we have gone through the entire product
        return true;  
    }
}
    
/**
* What is the date archived of this product?
*/
    public static long getDateArchived (Product p, String uniqueTag) {
        long dateArchivedLong = System.currentTimeMillis();
        try {
            dateArchivedLong = Long.parseLong (p.getProperty (uniqueTag));
        } catch (NumberFormatException e) {
            // just retag it
            addDateArchivedTag (p, true, uniqueTag);
        }
        return (dateArchivedLong);
    }
/**
* Add my own unique date archived tag and date to a product, if it doesn't already
* have the unique date archived tag and date.  if 'retag' is true, then 
* any existing date archived tag will be updated and overwritten.
*/
    public static void addDateArchivedTag (Product prod, boolean retag, String uniqueTag) {
        reporter.debugMsg("adding date archived tag",null,null,null, DEBUG_KEY);       
        
        String value = prod.getProperty (uniqueTag);
        Long dateArchivedLong = new Long (System.currentTimeMillis());
                
        if (value == null || value.equals ("") || retag) {            
            prod.setProperty (uniqueTag, dateArchivedLong.toString());
        }  
    }               
/**
* Compare products using Product's built-in comparison function, which
* overrides Object's default comparison function, and determines equality
* of product content.
*/
	public static boolean isEqual (Product a, Product b, String uniqueTag) {	    
	    if (a instanceof BasicProduct && b instanceof BasicProduct) {
	        // For BasicProducts, we take into account the id and description
	        // as well as content to determine equality.
	        return (a.getId().equals (b.getId()) && a.getDescription().equals (b.getDescription()) &&
	            a.equals (b));
	    } else if (a instanceof ContainerProduct && b instanceof ContainerProduct) {
	    	// if it is a rooted product then we want to do the compare based on
	    	// the root page only
	    	ContainerProduct ac = (ContainerProduct) a;
	    	ContainerProduct bc = (ContainerProduct) b;
	    	Product rootA = ac.getRoot();
	    	Product rootB = bc.getRoot();
	    	if (rootA != null && rootB != null) {
				return (rootA.getId().equals(rootB.getId()) && rootA.getDescription().equals(rootB.getDescription())
						&& rootA.equals(rootB));
	    	}
	    	else {
				return (ac.getId().equals (bc.getId()) && ac.getDescription().equals (bc.getDescription()));
	    	}
	    } else {
	        return false;
	    }
	}    
/**
* Merge products and eliminate duplicates.
*/
    public static Product mergeProduct (Object jugglersProduct, Object archivedProduct, String uniqueTag) {
//        printProductInfo ((Product)jugglersProduct, uniqueTag);
//        printProductInfo ((Product)archivedProduct, uniqueTag);
        if (jugglersProduct instanceof ContainerProduct 
            && archivedProduct instanceof ContainerProduct) {
                
    	        return (mergeProduct ((ContainerProduct)jugglersProduct,
    	        (ContainerProduct)archivedProduct, uniqueTag));
    	        
        } else if (jugglersProduct instanceof BasicProduct
            && archivedProduct instanceof ContainerProduct) {
                
                return (mergeProduct ((BasicProduct)jugglersProduct,
                (ContainerProduct)archivedProduct, uniqueTag));
                
        } else if (jugglersProduct instanceof ContainerProduct 	        
            && archivedProduct instanceof BasicProduct) {               
                reporter.assert (false, "archivedProduct is always either null or a ContainerProduct.",null,null,null);
                return (null);
        
        } else if (jugglersProduct instanceof BasicProduct
            && archivedProduct instanceof BasicProduct) {
                
                reporter.assert (false, "archivedProduct is always either null or a ContainerProduct.",null,null,null);
                return (null);
            
        } else {
                // don't know how to handle this product type
                reporter.assert (false, "Encountered unknown Product type(s) in Archiver Juggler.",jugglersProduct.toString(),null,null);   
                return (null);
        }	           
    }
/**
* Merge the new Product from the sources stored in jugglersProduct with the old archived
* Product stored in archivedProduct.  <br>Handles case when both jugglersProduct and 
* archivedProduct are ContainerProducts.<br>
* Note: This mergeProduct() method calls itself recursively to handle merging of ContainerProducts.
*/
	public static Product mergeProduct(ContainerProduct jugglersProduct, ContainerProduct archivedProduct, String uniqueTag) {
	    reporter.debugMsg("Merging two ContainerProducts into a new jugglersProduct",null,null,null, DEBUG_KEY); 

	    Product interimProduct = null;
	    
	    // examine new jugglersProduct and old archivedProduct to find where 
	    // products may be merged with duplicates eliminated	    
	    Enumeration jugglersEnum = (Enumeration) jugglersProduct.getContent();
        
        	    
	    while (jugglersEnum.hasMoreElements()) {
      	    Enumeration archivedEnum = (Enumeration) archivedProduct.getContent();

	        Product jp = (Product) jugglersEnum.nextElement();    
	        
	        // add our special date archived tag to new subproducts
	        addDateArchivedTag (jp, false, uniqueTag);
	        
	        // do not merge products which have a root, just add them
	        if (containsRoot(jp)) {
	            // just add it, do not attempt to merge
	            archivedProduct.addElement (jp);
	            
	            // update this container's date archived
                addDateArchivedTag (archivedProduct, true, uniqueTag);
	            continue;
	        }
	        
	        boolean foundMatch = false;
	        
	        while (archivedEnum.hasMoreElements()) {
    	        Product ap = (Product) archivedEnum.nextElement();
	        /*
	        * If the two product are the same at this level then merge them
	        */
	            if (isEqual (jp, ap, uniqueTag)) {
/*
					reporter.debugMsg("Merging jugglersProduct with archivedProduct",null,null,null, DEBUG_KEY);
					printProductInfo (jp, uniqueTag);
					printProductInfo (ap, uniqueTag);
*/
	                foundMatch = true;
				/*
				* If two basic products are equal we don't do anything
				*/
	                if ((jp instanceof ContainerProduct)
	                    && (ap instanceof ContainerProduct)) {	                        
	                        ContainerProduct jpc = (ContainerProduct) jp;
	                        ContainerProduct apc = (ContainerProduct) ap;
                            archivedProduct.removeElement (apc);
	                        if (jpc.getRoot() != null && apc.getRoot() != null) {
	        	                archivedProduct.addElement (jpc);
	                        }
	                        else {
	        	                archivedProduct.addElement (mergeProduct ((ContainerProduct) jp, 
	        	                    (ContainerProduct) ap, uniqueTag));
	                        }
	                        addDateArchivedTag(archivedProduct, true, uniqueTag);
                    }  	           	     
                    
                    break;
            	     
        	    }
        	}
    	/*
    	* When there is not match, then just add the product
    	*/
    	    if (!foundMatch) {
/*
       	        reporter.debugMsg("Adding Products",null,null,null, DEBUG_KEY);
                printProductInfo (jp, uniqueTag);
                printProductInfo (archivedProduct, uniqueTag);                
*/
                archivedProduct.addElement (jp);
                addDateArchivedTag (archivedProduct, true, uniqueTag);
    	    }
	            
	    }
	    
	    interimProduct = archivedProduct;	   
	    
	    // finally, return merged product
	    return (interimProduct);
	}
/**
* Merge the new Product from the sources stored in jugglersProduct with the old archived
* Product stored in archivedProduct.  <br>Handles case when jugglersProduct is a BasicProduct and
* archivedProduct is a ContainerProduct.<br>
* Note: Never called recursively by other mergeProduct() methods.  Only called directly by
* buildProduct(). 
*/
	public static Product mergeProduct(BasicProduct jugglersProduct, ContainerProduct archivedProduct, String uniqueTag) {
	    reporter.debugMsg("Merging BasicProduct and ContainerProduct into a new jugglersProduct", null,null,null,DEBUG_KEY); 
        
        // add our special date archived tag to new Product
        addDateArchivedTag (jugglersProduct, false, uniqueTag);
        
	    Product interimProduct = null;
	    
	    // examine new jugglersProduct and old archivedProduct to find where 
	    // products may be merged with duplicates eliminated	    
        Enumeration archivedEnum = (Enumeration) archivedProduct.getContent();
        
        boolean foundMatch = false;
        
        while (archivedEnum.hasMoreElements()) {
            Product prod = (Product) archivedEnum.nextElement();    

            if (isEqual (prod, jugglersProduct, uniqueTag)) {                
/*
                reporter.debugMsg("Merging jugglersProduct with archivedProduct", null, null, null, DEBUG_KEY);
                printProductInfo (prod, uniqueTag);
                printProductInfo (archivedProduct, uniqueTag);
*/
                // same product, so do nothing and leave existing product untouched
                foundMatch = true;
                break;
            }
        }
        
        if (!foundMatch) {
            // update this container's date archived tag 
            addDateArchivedTag (archivedProduct, true, uniqueTag);
/*            
            reporter.debugMsg("Adding BasicProduct to ContainerProduct",null,null,null, DEBUG_KEY);
            printProductInfo (jugglersProduct, uniqueTag);
            printProductInfo (archivedProduct, uniqueTag);
*/       
            // different product, so add it
            VectorProduct compositeProduct = (VectorProduct) archivedProduct;            
            compositeProduct.addElement (jugglersProduct);
            interimProduct = compositeProduct;
        }

   	    // finally, return merged product
	    return (interimProduct);
	}    
/**
* Delete expired products from jugglersProduct before passing it along
*/
    public static void deleteExpiredProducts (Product jugglersProduct, String uniqueTag, int productLifetime) {
        // instantiate archived product expiration filter
	    ExpireFilter expirationFilter = new ExpireFilter (uniqueTag, productLifetime);	        

        if (jugglersProduct != null) {
            Product jp = (Product) jugglersProduct;
            jp.apply (expirationFilter);
        }
    }   
/**
* Does this product contain a root?
*/
    public static boolean containsRoot (Object prod) {
        if (prod instanceof ContainerProduct) {
            ContainerProduct cp = (ContainerProduct) prod;
            return (cp.getRoot() != null);
        } else {
            return false;
        }
    }    
/**
* Product information -- for debugging purposes
*/
	public static void printProductInfo (Product prod, String uniqueTag) {
	    String dateArchived = prod.getProperty (uniqueTag);
	    
	    if (prod instanceof ContainerProduct) {
	        reporter.debugMsg("ContainerProduct",null,null,null, DEBUG_KEY);
	    } else if (prod instanceof BasicProduct) {
	        reporter.debugMsg("BasicProduct",null,null,null, DEBUG_KEY);
	    }
        reporter.debugMsg("ID: ",prod.getId(),null,null, DEBUG_KEY);
        reporter.debugMsg("Description: ", prod.getDescription(),null,null, DEBUG_KEY);        
        if (dateArchived != null) {
            reporter.debugMsg("Date archived: ",new Date (Long.parseLong (dateArchived)).toString(),null,null, DEBUG_KEY);            
        } else {
            reporter.debugMsg("Date archived: " ,dateArchived, null,null, DEBUG_KEY);
        }
	}	
/**
* Restore or create unique date archived tag for this juggler instance to use.<br>
* Sets UNIQUE_TAG for this instance of the juggler.<br> 
* Returns UNIQUE_TAG.
*/
    public static String getUniqueTag (Memento memento) {
   	    // look for the unique date archived tag in the memento
		String uniqueTag = memento.getProperty(DATE_ARCHIVED_TAG);
		String UNIQUE_TAG = null;
		
		if (uniqueTag != null && !uniqueTag.equals ("")) {
		    reporter.debugMsg("Restoring value of unique date archived tag", null,null,null,DEBUG_KEY);
		    UNIQUE_TAG = uniqueTag;
		    memento.setProperty(DATE_ARCHIVED_TAG, UNIQUE_TAG);
		} else {
		    reporter.debugMsg("Creating a new unique date archived tag",null,null,null, DEBUG_KEY);
		    UNIQUE_TAG = FileRef.getNewFileId("DateArchived");
		    memento.setProperty(DATE_ARCHIVED_TAG, UNIQUE_TAG);
		}		
		return (UNIQUE_TAG);
    }	
/**
* What is the lifetime of the archived product? (default = 7 days) 
*/
    public static int getProductLifetime (Memento memento) {
	    return memento.getIntegerProperty(LIFETIME,7);
    }    
/**
* Should expired products be deleted from the archive?
*/
    public static boolean deleteExpired(Memento memento) {
        return memento.getBooleanProperty(DELETE_EXPIRED, false);    
    }     
/**
* Sorts a ContainerProduct such that all of its elements appear in reverse chronological order.
* @param    cp              The ContainerProduct to be sorted in reverse chronological order
*/
    public static void sortReverseChronological (ContainerProduct cp) {
// TODO            
    }
}