/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util;

import java.io.*;

/**
 * This class is similar to java.io.PrintWriter except that it
 * used CRLF pairs on output (for HTML output).  Ideally we would subclass
 * PrintWriter but the Sun folks in their wisdom have made lineSeparator, 
 * newLine, etc. private so that we have to do more work.
 */

public class PrintCRLFWriter extends PrintWriter {

    protected boolean autoFlush = false;
    /**
     * Line separator string.  This is the value of the line.separator
     * property at the moment that the stream was created.
     */
    protected String lineSeparator = "\r\n";

    /**
     * Create a new PrintCRLFWriter, without automatic line flushing.
     *
     * @param  out        A character-output stream
     */
    public PrintCRLFWriter (Writer out) {
	this(out, false);
    }

    /**
     * Create a new PrintCRLFWriter, without automatic line flushing, from an
     * existing OutputStream.  This convenience constructor creates the
     * necessary intermediate OutputStreamWriter, which will convert characters
     * into bytes using the default character encoding.
     *
     * @param  out        An output stream
     *
     * @see java.io.OutputStreamWriter#OutputStreamWriter(java.io.OutputStream)
     */
    public PrintCRLFWriter(OutputStream out) {
	this(out, false);
    }

    /**
     * Create a new PrintCRLFWriter from an existing OutputStream.  This
     * convenience constructor creates the necessary intermediate
     * OutputStreamWriter, which will convert characters into bytes using the
     * default character encoding.
     *
     * @param  out        An output stream
     * @param  autoFlush  A boolean; if true, the println() methods will flush
     *                    the output buffer
     *
     * @see java.io.OutputStreamWriter#OutputStreamWriter(java.io.OutputStream)
     */
    public PrintCRLFWriter(OutputStream out, boolean autoFlush) {
	this(new BufferedWriter(new OutputStreamWriter(out)), autoFlush);
    }


    /**
     * Create a new PrintCRLFWriter.
     *
     * @param  out        A character-output stream
     * @param  autoFlush  A boolean; if true, the println() methods will flush
     *                    the output buffer
     */
    public PrintCRLFWriter(Writer out,  boolean autoFlush) {
		super(out, autoFlush);
		this.autoFlush = autoFlush;
    }


  /**
   * Override PrintWriter newLine to print CRLF and not
   * the native line separator.
   */
    private void newLine() {
		super.write(lineSeparator);
		if (autoFlush)
		    flush();
    }


    /* Methods that do terminate lines.
	Override all of PrintWriter's println methods to
	call our newLine method instead of the one in PrintWriter.
     */

    /** Finish the line. */
    public void println() {
	synchronized (lock) {
	    newLine();
	}
    }

    /** Print a boolean, and then finish the line. */
    public void println(boolean x) {
	synchronized (lock) {
	    print(x);
	    newLine();
	}
    }

    /** Print a character, and then finish the line. */
    public void println(char x) {
	synchronized (lock) {
	    print(x);
	    newLine();
	}
    }

    /** Print an integer, and then finish the line. */
    public void println(int x) {
	synchronized (lock) {
	    print(x);
	    newLine();
	}
    }

    /** Print a long, and then finish the line. */
    public void println(long x) {
	synchronized (lock) {
	    print(x);
	    newLine();
	}
    }

    /** Print a float, and then finish the line. */
    public void println(float x) {
	synchronized (lock) {
	    print(x);
	    newLine();
	}
    }

    /** Print a double, and then finish the line. */
    public void println(double x) {
	synchronized (lock) {
	    print(x);
	    newLine();
	}
    }

    /** Print an array of characters, and then finish the line. */
    public void println(char x[]) {
	synchronized (lock) {
	    print(x);
	    newLine();
	}
    }

    /** Print a String, and then finish the line. */
    public void println(String x) {
	synchronized (lock) {
	    print(x);
	    newLine();
	}
    }

    /** Print an Object, and then finish the line. */
    public void println(Object x) {
	synchronized (lock) {
	    print(x);
	    newLine();
	}
    }

}
