/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import org.mitre.pam.interfaces.IDictionary;
import org.mitre.pam.daemon.Dict;

/**
* A hashtable that returns a sorted Enumeration.
* This was written pre java 1.2. 
* We used to extend a com.objectspace.jgl.OrderedMap but
* that library was too big to include and it was too hard
* to pull apart.
* Sort order is preserved using a binary search to keep
* insert time to a reasonable amount.
* Lookup by key is the same as Hashtable lookup speed.
* Has additional logic to support AttributeKeys and IDictionaries are used.
* Using an AttributeKey as the key will cause that key to be looked up
* once if an IDictionary is also being used.
*
* @author Nazario Irizarry, Jr.
*/
public class SortedHash implements  java.io.Serializable, BinaryCompare {
	protected Hashtable hash = new Hashtable();
	protected Vector orderedList; /*  = new Vector(30); */
	protected transient Vector indexedValueList;

	protected BinaryCompare compareClass;
	protected boolean sortOnElements;
	protected transient Dict dictionary;
	static private int nextNum = 0;
	private int idNum = nextNum++;
	private int nPuts, nGets, nSorts;
/**
* Constructs a hashtable that is sorted on keys
* using the .toString() method to compare
* the values of keys alphanumerically.
*/
	public SortedHash() {
		sortOnElements = false;
		compareClass = this;
	}

/**
* Constructs a hashtable that is sorted on keys
* using the specified object to compare keys to determine
* sort order
*/
	public SortedHash(BinaryCompare compareClass) {
		sortOnElements = false;
		this.compareClass = compareClass;
	}
/**
* If we use a dictionary then we put values into slots with known indexes
*/
	public synchronized void useDictionary(IDictionary dictionary) {
		if (dictionary == null)
			throw new RuntimeException("Attempt to use null value for dictionary");
		indexedValueList = new Vector(20,20);
		this.dictionary = (Dict) dictionary;
		
	/*
	* scan any existing attributes and make sure they are in the dictionary
	* and the indexed list
	*/
		Enumeration e = hash.elements();
		while (e.hasMoreElements()) {
			SortedLink link = (SortedLink) e.nextElement();
			setIndexedValue(link.key, link);
		}
	}
/*
* Method assumes that dictionary != null
*/
	private void setIndexedValue(Object key, SortedLink sortedLink) {
		int index = -1;
		if (key instanceof AttributeKey) {
			AttributeKey attributeKey = (AttributeKey) key;
		/*
		* Perform fast lookup if the attribute key was last used on our dictionary
		*/
			if (attributeKey.referenceObject1 == dictionary) {
				index = attributeKey.index1;
			}
		/*
		* Do a slow lookup.
		*/
			else {
				attributeKey.referenceObject1 = dictionary;
				attributeKey.index1 = index = dictionary.getIndexByName(key.toString());
			}
		}
		else {
			index = dictionary.getIndexByName(key.toString());
		}
		if (indexedValueList.size() <= index)
			indexedValueList.setSize(index+10);
		indexedValueList.setElementAt(sortedLink, index);
	}

/**
* Just like java.util.Hashtable
* @return old value if there was one
*/
	public synchronized Object put(Object argKey, Object value) {
		nPuts++;
		SortedLink link = getLink(argKey);
		Object oldValue=null;
	/*
	* for existing key just replace the value
	*/
		if (link != null) {
			oldValue = link.value;
			link.value = value;
			if (sortOnElements) {
				if (orderedList != null)
					orderedList.removeElement(link);
			}
			else {
				return oldValue; // no need to resort since the key which the order did not change
			}
		}
	/*
	* A new key must be added to hashtable and the dictioanry, then sorted
	*/
		else {
			Object key = (argKey instanceof AttributeKey) ? argKey.toString() : argKey;
			link = new SortedLink(key, value);
			hash.put(key, link);
		}
	/*
	* update a dictionary entry
	*/
		if (dictionary != null && oldValue == null)
			setIndexedValue(argKey, link);			
	/*
	* find position in table
	*/
		if (orderedList != null) {
			orderedList.insertElementAt(link, findPosInOrder(link));
		}
		return oldValue;
	}

	protected int findPosInOrder(SortedLink link) {
//	System.out.println("findPos "+link);
		int len = orderedList.size();
		if (len == 0)
			return 0;
		int mask = 1 << 14;  // Assume a limit on number of entries
		int addr = 0;
		int testAddr = 0;
		int retAddr = 0;
		while (mask >= len)
			mask >>= 1;
		testAddr = addr | mask;
	/*
	* Here we do a binary search on the existing sorted list
	*/		
		boolean linkIsLess = false;
		retAddr = 0;
		while (mask > 0) {
			
			testAddr = addr | mask;
			if (testAddr < len) {
				SortedLink testLink = (SortedLink) orderedList.elementAt(testAddr);
				linkIsLess = 
							sortOnElements 
							? compareClass.isBefore(link.value, testLink.value)
							: compareClass.isBefore(link.key, testLink.key);
//		System.out.println("  m="+mask+"  a="+addr+"  t="+testAddr+"  c="+linkIsLess+"   tl="+testLink);
				if (linkIsLess) {
					retAddr = testAddr;
				}
				else {
					addr = testAddr;
					retAddr = testAddr + 1;
				}
			}
			
			mask >>= 1;
		}
	/*
	* address zero is the only address that will never get tested in the above loop
	*/
		if (addr == 0) {
				SortedLink testLink = (SortedLink) orderedList.elementAt(0);
				linkIsLess = 
							sortOnElements 
							? compareClass.isBefore(link.value, testLink.value)
							: compareClass.isBefore(link.key, testLink.key);
//		System.out.println(" fm="+mask+"  a="+addr+"  t="+testAddr+"  c="+linkIsLess+"   tl="+testLink);
				if (linkIsLess) 
					retAddr = 0;
				else 
					retAddr = 1;
		}
//		System.out.println("     p="+retAddr);
		return retAddr;
	}
	
	private void removeIndexedValue(Object key) {
		int index = -1;
		if (key instanceof AttributeKey) {
			AttributeKey attributeKey = (AttributeKey) key;
		/*
		* Perform fast lookup if the attribute key was last used on our dictionary
		*/
			if (attributeKey.referenceObject1 == dictionary) {
				index = attributeKey.index1;
			}
		/*
		* Do a slow lookup.
		*/
			else {
				attributeKey.referenceObject1 = dictionary;
				attributeKey.index1 = index = dictionary.getIndexByName(key.toString());
			}
		}
		else {
			index = dictionary.getIndexByName(key.toString());
		}
		if (indexedValueList.size() <= index)
			return;
		indexedValueList.setElementAt(null, index);
	}
/**
* Just like java.util.Hashtable.
*/
    public synchronized Object remove(Object argKey) {
		Object key = (argKey instanceof AttributeKey) ? argKey.toString() : argKey;
    	SortedLink entry = (SortedLink) hash.remove(key);
		if (dictionary != null)
			removeIndexedValue(argKey);
			
    	if (entry == null) {
    		return null;
    	}
    	else {
    		if (orderedList != null)
    			orderedList.removeElement(entry);
			return entry.value;
    	}
    }
/**
* Just like java.util.Hashtable.
*/
    public int size() {
		return hash.size();
    }
/**
* Just like java.util.Hashtable.
* A copy is made of the ordered list to support deletions during traversal.
*/
    public synchronized Enumeration keys() {
//    	reportStats("keys");
    	orderIf();
    	return new SortedEnumerator((Vector)orderedList.clone(), true);
    }
/**
* Just like java.util.Hashtable.
* A copy is made of the ordered list to support deletions during traversal.
*/
    public synchronized Enumeration elements() {
//    	reportStats("elements");
    	orderIf();
    	return new SortedEnumerator((Vector)orderedList.clone(), false);
    }
/**
* Allows us to simultaneously traverse the keys and values.
* A copy is made of the ordered list to support deletions during traversal.
*/
    public synchronized Enumeration buckets() {
//    	reportStats("buckets");
    	orderIf();
    	return new SortedEnumerator((Vector)orderedList.clone(), false, true);
    }
    
/**
* Report a few statistics
*/
	public void reportStats(String s) {
		System.out.println(this +" "+s+ "  nPuts="+nPuts+",  nGets="+nGets+",  nSorts="+nSorts);
	}
/**
* @return a SortedLink
*/
    private SortedLink getLink(Object key) {
    	if (key instanceof AttributeKey) {
    		AttributeKey attributeKey = (AttributeKey) key;
    		if (dictionary == null) {
		    	return (SortedLink)hash.get(attributeKey.name);
    		}
    		else {
    			if (attributeKey.referenceObject1 != dictionary) {
					attributeKey.referenceObject1 = dictionary;
					attributeKey.index1 = dictionary.getIndexByName(key.toString());
    			}
				return (SortedLink) indexedValueList.elementAt(attributeKey.index1);
    		}
    	}
    	else {
	    	return (SortedLink)hash.get(key);
	    }
    }
/**
* Similar to java.util.Hashtable.
* If the key is of type AttributeKey then every lookup after
* the first on that key is fast because we use the AttributeKey
* to "cache" the lookup index.  <p>
*
* If the key is anything else then this boils down to a
* conventional hash lookup.
*/
    public synchronized Object get(Object key) {
    	nGets++;
    	SortedLink entry = getLink(key);
    	return entry == null ? null : entry.value;
    }
/**
* Clear of  all entries
*/
	public synchronized void clear() {
		nGets = nPuts = nSorts = 0;
		hash.clear();
		if (orderedList != null) {
			orderedList.setSize(0);
			orderedList = null;
		}
		if (indexedValueList != null)
			indexedValueList.setSize(0);
	}
/**
* Internal method to order the existing entries in batch form.
* This allows an initial set of puts (at system startup)
* without incurring the overhead for ordering until after much
* of the system has started.
*/
	synchronized private void orderIf() {
		nSorts++;
		if (orderedList != null)
			return;
//		reportStats("orderIf");
		orderedList = new Vector(hash.size() + 10);
		Enumeration e = hash.elements();
		while (e.hasMoreElements()) {
			SortedLink link = (SortedLink) e.nextElement();
			orderedList.insertElementAt(link, findPosInOrder(link));
		}
//		dumpLists();
	}
/**
* A debug method
*/
	public void dumpLists() {
		orderIf();
		System.out.println("Dump list for "+this);
		System.out.println(" ordered list");
		for (int i = 0; i<orderedList.size(); i++) {
			System.out.println("   "+orderedList.elementAt(i));
		}
		System.out.println(" keyed list");
		Enumeration e = hash.keys();
		while (e.hasMoreElements()) {
			Object key = e.nextElement();
			Object value = hash.get(key);
			System.out.println("   "+key+":  "+value);
		}
	}
	
/**
*
*/
	public String toString() {
		return "SortedHash "+idNum;
	}
/**
* Has same signature as com.objectspace.jgl.BinaryPredicate.
* Does a comparison based on lowercase keys.  If
* these are equal then it does a comparison based on regular case.
*/

/* old and slow
	public boolean isBefore( Object first, Object second ) {
		String lc1 = first.toString().toLowerCase();
		String lc2 = second.toString().toLowerCase();
		int comp = lc1.compareTo(lc2);
		if (comp == 0)
			return first.toString().compareTo( second.toString() ) < 0;
		return comp < 0;
	}
*/
// should be faster because it avoids new string creation
	public boolean isBefore( Object first, Object second ) {
		
		String cs1 = first.toString();
		int l1 = cs1.length();
		String cs2 = second.toString();
		int l2 = cs2.length();
		
		int n = l1;
		if (l2 < n)
			n = l2;
		if (n == 0)
			return (l2 > 0);
		int eqCase = 0;
		for (int i = 0; i < n; i++) {
			char c1 = cs1.charAt(i);
			char c2 = cs2.charAt(i);
			int diff = Character.toLowerCase(c1) - Character.toLowerCase(c2);
			if (diff != 0) {
				return diff < 0;
			}
			
			if (eqCase == 0) {
				eqCase = c1-c2;
			}
		}
		
		if (l1 == l2) {
			return eqCase < 0;
		}

		return l1 < l2;
	}
/*
	public static void main(String[] args) {
		Enumeration e=null;
		System.out.println("begin\n");
		SortedHash sh = new SortedHash();
		sh.put("key33","33");
		sh.put("key11","11");
		e = sh.elements();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");
		
		sh.put("key22","22");
		e = sh.elements();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");
		
		sh.put("key00","00");
		e = sh.elements();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");

		sh.put("key44","44");
		e = sh.elements();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");

		sh.put("key32","32");
		e = sh.elements();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");

		sh.put("zkey32","z32");
		e = sh.elements();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");

		sh.put("akey32","a32");
		e = sh.keys();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");

		e = sh.buckets();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");
		

		AttributeKey ak1 = new AttributeKey("key11");
		AttributeKey ak2 = new AttributeKey("key22");
		AttributeKey ak3 = new AttributeKey("key33");

		sh.put(ak1,"111a");
		sh.put(ak3,"333a");
		sh.put(ak2,"222a");
		e = sh.elements();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}

		System.out.println("\nend\n");
	}
*/

/*
	public static void main(String[] args) {
		Enumeration e=null;
		System.out.println("begin\n");
		SortedHash sh = new SortedHashOnElements();
		sh.put("key33","33");
		sh.put("key11","33");
		e = sh.buckets();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");
		
		sh.put("key22","33");
		e = sh.buckets();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");
		
		sh.put("key01","01");
		e = sh.buckets();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");

		sh.put("key22","00");
		e = sh.buckets();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");

		sh.put("key32","32");
		e = sh.buckets();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");

		sh.put("zkey32","z32");
		e = sh.buckets();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");

		sh.put("akey32","a32");
		e = sh.buckets();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");

		e = sh.buckets();
		while(e.hasMoreElements()) {
			System.out.println(e.nextElement());
		}
		System.out.println("\n");
		

		System.out.println("\nend\n");
	}
*/
}

