/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util.alarm;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileWriter;              // JF
import java.io.RandomAccessFile;        // JF
import java.text.ParseException;        // JF
import java.text.SimpleDateFormat;      // JF
import java.util.Date;                  // JF
import java.util.Random;                // JF
import java.util.StringTokenizer;
import java.util.TimeZone;              // JF
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

import org.mitre.pam.util.PAMDate;

/**
* Used to flag an assertion exception so that it is easy to
* print the stack trace
*/
	class AssertionException extends Exception {
		public AssertionException(String s) {
			super(s);
		}
	}

/**
 * The Alarm class is used to ease error/warning message maintenance by having
 * one class manage all the messages in an application.
 *
 * <p>Alarm messages are embedded in the code where they have the best chance
 * of being current.  Alarm messages can have multiple parameters.  To provide
 * multiple parameters, provide them in a single string separated by '\|'
 * characters.  Every occurence of %s in the alarm comment will be replaced
 * with the next string from the list of parameters.
 *
 * @author Tim Taylor
 */
public final class Alarm {

	private static String DEBUG = "debug";
	private static String LOG = "log";
	private static String WARN = "warn";
	private static String ERROR = "error";

  /** Indicates a change in the state of debug keys **/
  static int debugKeyActivityNumber=1;
  
  /** Line ending for files on this platform. */
  private static final String lineEnd = System.getProperty("line.separator");

  /** Field separator to separate out fields of the diagnostic message. */
  private static final String FIELD_SEPARATOR = "\t";

  /** This is the output stream we can optionaly log messages to. */
  private static PrintWriter logFile = null;

  /** This is the handle of the file that the logFile stream writes to. */
  private static File logFileHandle = null;  // JF

  /**
   * This is the set of listeners that want to be reported when a message is
   * posted.
   */
  private static Vector listeners = null;

  /**
  * A table of categories that have been activated for terseDebug messages
  */
  private static Hashtable debugKeyList = new Hashtable(100);

  /**
  * A table of available keys.
  */
  private static Hashtable availableDebugList = new Hashtable(100);

	static {
		availableDebugList.put("showThreadNames","showThreadNames");
	}
/**
* A list of registered debug keys
*/
	public static Enumeration getAvailableDebugKeyList() {
		return availableDebugList.keys();
	}
/**
* A list of enabled debug keys
*/
	public static Enumeration getEnabledDebugKeyList() {
		return debugKeyList.keys();
	}
/**
* Write a message to the log
*/
	public static void log(String category, String msg) {
		reportInternal(category, LOG,  msg);
	}
/**
* Write a message to the log
*/
	public static void warn(String category, String msg) {
		reportInternal(category, WARN,  msg);
	}
/**
* Write a message to the log
*/
	public static void error(String category, String msg) {
		reportInternal(category, ERROR,  msg);
	}
/**
* Check if the assert condition is false in which case we output the message
*
* @param cond
* @param category
* @param msg1 the message/part 1
* @param msg2 the message/part 2 - may be null
* @param msg3 the message/part 3 - may be null
* @param msg4 the message - part 4 - may be null
*/
	static void assert(String category, Object msg1, Object msg2, Object msg3, Object msg4) {
		try {
		/*
		*  throw exception on purpose so we can process the
		*  assertion failure as if a genuine exception.
		*  to send to the log file or System.err
		*/
			StringBuffer bf = new StringBuffer(80);
			if (msg1 != null)
				bf.append(msg1.toString());
			if (msg2 != null) {
				bf.append(" ");
				bf.append(msg2.toString());
			}
			if (msg3 != null) {
				bf.append(" ");
				bf.append(msg3.toString());
			}
			if (msg4 != null) {
				bf.append(" ");
				bf.append(msg4.toString());
			}
			throw new AssertionException(bf.toString()); 
		}
		catch (AssertionException e) {
			reportInternal(category, WARN, e);
		}
	}

/**
* Report a debugging message.
* The strings are not concatenated unless the category is active, for
* performance reasons.
* @param msg1 the message/part 1
* @param msg2 the message/part 2 - may be null
* @param msg3 the message/part 3 - may be null
* @param msg4 the message - part 4 - may be null
* @param category is the category flag that may be used to 
* disable/enable entire sections of debug messages
*/
	static void debugMsg(Object msg1, Object msg2, Object msg3, Object msg4, String category){
		if (!availableDebugList.containsKey(category)) {
			availableDebugList.put(category,category);
		}
		if (!debugKeyList.containsKey(category))
			return;
		StringBuffer bf = new StringBuffer(80);
		if (msg1 != null)
			bf.append(msg1.toString());
		if (msg2 != null) {
			bf.append(" ");
			bf.append(msg2.toString());
		}
		if (msg3 != null) {
			bf.append(" ");
			bf.append(msg3.toString());
		}
		if (msg4 != null) {
			bf.append(" ");
			bf.append(msg4.toString());
		}
		reportInternal(category, DEBUG, bf.toString());
	}

/**
* Report an exception
*/
	static void reportException(String category, Throwable e) {
		reportInternal(category, ERROR, e);
	}
/**
* Is the specified key a terse debug key
*/
	public static boolean isDebugKey(String key) {
		if (!availableDebugList.containsKey(key)) {
			availableDebugList.put(key,key);
		}
		return debugKeyList.containsKey(key);
	}
/**
* Specifies a debug category to be activated
* 
*/
	public static void setDebug(String name, boolean on) {
		if(name  == null) 
			return;
		debugKeyActivityNumber++;
		if (on)
			debugKeyList.put(name, name);
		else
			debugKeyList.remove(name);
	}

	
	private static synchronized 
	void reportInternal( String category, String level, Throwable exception ) {
		if (exception == null)
			return;	
	/*
	* print to the log file
	*/
		PAMDate now = new PAMDate();
		if(logFile != null) {
		/*
		* show thread names?
		*/
			String threadName = null;
			if (isDebugKey("showThreadNames")) {
				threadName = Thread.currentThread().getName();
			}
			logFile.print(now.asNumberString());
			logFile.print(FIELD_SEPARATOR);
			logFile.print(level);
			if (threadName != null) {
				logFile.print(FIELD_SEPARATOR);
				logFile.print(threadName);
			}
			logFile.print(FIELD_SEPARATOR);
			logFile.print(category);
			logFile.print(FIELD_SEPARATOR);
			logFile.println(exception.getMessage());
			logFile.print(FIELD_SEPARATOR);
	    	exception.printStackTrace(logFile);
			logFile.flush();
	    }
	    else {
			System.err.print(now.asNumberString());
			System.err.print(FIELD_SEPARATOR);
			System.err.print(level);
			System.err.print(FIELD_SEPARATOR);
			System.err.print(category);
			System.err.print(FIELD_SEPARATOR);
			System.err.println(exception.getMessage());
			System.err.print(FIELD_SEPARATOR);
	    	exception.printStackTrace(System.err);
	    }
	/*
	* let listeners know
	*/
		notifyAlarmListeners(new AlarmEvent(category, level, exception));
	}

	private static synchronized 
	void reportInternal( String category, String level,  String msg ) {
		if (msg == null)
			return;	
	/*
	* print to the log file
	*/
		PAMDate now = new PAMDate();
		if(logFile != null) {
		/*
		* show thread names?
		*/
			String threadName = null;
			if (isDebugKey("showThreadNames")) {
				threadName = Thread.currentThread().getName();
			}

			logFile.print(now.asNumberString());
			logFile.print(FIELD_SEPARATOR);
			logFile.print(level);
			if (threadName != null) {
				logFile.print(FIELD_SEPARATOR);
				logFile.print(threadName);
			}
			logFile.print(FIELD_SEPARATOR);
			logFile.print(category);
			logFile.print(FIELD_SEPARATOR);
			logFile.println(msg);
			logFile.flush();
	    }
	    else {
			System.err.print(now.asNumberString());
			System.err.print(FIELD_SEPARATOR);
			System.err.print(level);
			System.err.print(FIELD_SEPARATOR);
			System.err.print(category);
			System.err.print(FIELD_SEPARATOR);
			System.err.println(msg);
	    }
		
		if(! msg.endsWith(lineEnd))
		  msg = msg + lineEnd;
		
	/*
	* let listeners know
	*/
		notifyAlarmListeners(new AlarmEvent(category, level, msg));
	}


  /**
   * Sets the log file to the filename specified.  If we already have a log
   * file, the current one is closed.
   * @param filename   The name of the log file to use on subsequent calls to
   *  <i>Alarm.report</i>.
   */
  public static void setLogFile(File file) {
    try {
      // If we have a log file, close it
      if(logFile != null) {
        logFile.close();
      }

      logFile = new PrintWriter(new FileOutputStream(file));
    }
    catch(IOException exp) {
      System.err.println("Unable to open logfile '" + file + "'");
      logFile = null;
    }
    catch(SecurityException exp) {
      System.err.println("Security restrictions prevent creation of logfile '" + file + "'");
      logFile = null;
    }
    logFileHandle = file;               // JF
  }


  /**
   * This is our abort method.  We try to do as graceful a system shutdown as
   * we can by stopping all active threads, and waiting for them to complete
   * before we shut down the Java virtual machine.
   */
  private static void exitGracefully() {
    // Get the list of all threads
    Thread myThrd = Thread.currentThread();
    ThreadGroup thisGroup = myThrd.getThreadGroup();
    int count = thisGroup.activeCount();
    Thread[] thrds = new Thread[count + 20];  // +20 for slop
    thisGroup.enumerate(thrds);
/*
 * abort the hard way
 * so we don't have to rely on the "outlaws stop method of java 1.2
 *
    // Stop all threads
    for(int i = 0; i < thrds.length; i++) {
      if(thrds[i] != null && thrds[i] != myThrd)
        thrds[i].stop();
    }

    // Wait for all threads to complete
    for(int i = 0; i < thrds.length; i++) {
      if(thrds[i] != null && thrds[i] != myThrd) {
        try {
          thrds[i].join();
        } catch(InterruptedException e) {
          // Just skip this thread
        }
      }
    }
*/
    // Now we can exit
    System.exit(-1);
  }


  private static void notifyAlarmListeners(AlarmEvent e) {
    if(listeners == null) {
      System.err.println(e.getCategory()+"\t"+e.getMessage());
    } else {
      java.util.Enumeration myListeners = listeners.elements();
      while(myListeners.hasMoreElements()) {
        AlarmListener al = (AlarmListener) myListeners.nextElement();
        al.alarmMessageReport(e);
      }
    }
  }

  public static void addAlarmListener(AlarmListener al) {
    if(listeners == null) {
      listeners = new Vector(2, 2);
    } else if(listeners.contains(al))
      return;
    listeners.addElement(al);
  }

  public static void removeAlarmListener(AlarmListener al) {
    listeners.removeElement(al);
    if(listeners.size() == 0) listeners = null;
  }


/**
* Performs two layer filtering of the log file in order to
* keep the file manageable.  Two intervals (in units of days)
* are provided as arguments in order to perform
* the filtering.  Any entry whose date is before the eraseAllBeforeDays interval is disgarded, and any
* entry that is before the eraseAllButExceptionsBeforeDays interval is also disgarded UNLESS the entry
* is for an exception or assertion.  All entries that are made after the eraseAllButException
* date are saved.
*/
  public static void filterLog(int eraseAllBeforeDays, int eraseAllButExceptionsBeforeDays) {
    Date today = new Date();
    Date d1 = new Date(today.getTime() - eraseAllBeforeDays*24*60*60*1000);
    Date d2 = new Date(today.getTime() - eraseAllButExceptionsBeforeDays*24*60*60*1000);
    filterLog(d1, d2);
  }

/**
* Looks through the log file to see if there are
* any new exceptions.  An exception is considered new if its time is recent
* based on the milliSecondsBack. Exceptions that are new are recorded in file whose handle
* is also supplied as an argument.
*/
  public static void filterExceptions(long milliSecondsBack, File file) {
    Date today = new Date();
    Date d = new Date(today.getTime() - milliSecondsBack);
    filterExceptions(d, file);
  }

/**
* Looks through the log file to see if there are
* any new exceptions.  An exception is considered new if it was reported after
* the supplied date argument.  Exceptions that are new are recorded in file whose handle
* is also supplied as an argument.
*/
  private static synchronized void filterExceptions(Date date, File file) {
    FileWriter out = null;
    try {
      out = new FileWriter(file);
    } catch(IOException ioe) {
      ioe.printStackTrace(System.err);
    }

    SimpleDateFormat sdt = new SimpleDateFormat("yyyyMMdd kkmmss");
    sdt.setTimeZone(TimeZone.getDefault());

    try {
      RandomAccessFile in = new RandomAccessFile(logFileHandle, "r");

      while(in.getFilePointer() < in.length()) {
        String line = in.readLine();

        while((line.endsWith("\r")) || (line.endsWith("\n"))) {
          line = line.substring(0, line.length() - 1);
        }

        try {
          int i = Integer.parseInt(line.substring(0, 1));
          Date lineDate = sdt.parse(line.substring(0, 15));

          if(lineDate.after(date)) {
            int a = line.indexOf("Assertion");
            int e = line.indexOf("Exception");
            if((a != -1) || (e != -1)) {
              out.write(line + lineEnd);
              writeText(in, out);
            }
          }
        } catch(NumberFormatException nfe) {
          //nfe.printStackTrace(System.err);  // these print statements produce a lot of unnecessary output
        } catch(ParseException pe) {
          //pe.printStackTrace(System.err);
        } catch(StringIndexOutOfBoundsException sie) {
          //sie.printStackTrace(System.err);
        }
      }
      in.close();
      out.flush();
      out.close();
    } catch(IOException ioe) {
      ioe.printStackTrace(System.err);
    }
  }

  /**
  * Performs two layer filtering of the log.txt file in order to
  * keep the file manageable.  Two dates are provided as arguments in order to perform
  * the filtering.  Any entry whose date is before the eraseAll date is disgarded, and any
  * entry that is before the eraseAllButException date is also disgarded UNLESS the entry
  * is for an exception or assertion.  All entries that are made after the eraseAllButException
  * date are saved.
  */
  private static synchronized void filterLog(Date eraseAll, Date eraseAllButException) {
    FileWriter out = null;
    Random generator = new Random();
    File temp = new File(System.getProperty("user.dir") + File.separator + Long.toString(generator.nextLong()));
    while(temp.exists()) {
      temp = new File(System.getProperty("user.dir") + File.separator + Long.toString(generator.nextLong()));
    }
    try {
      out = new FileWriter(temp);
    } catch(IOException ioe) {
      ioe.printStackTrace(System.err);
    }

    SimpleDateFormat sdt = new SimpleDateFormat("yyyyMMdd kkmmss");
    sdt.setTimeZone(TimeZone.getDefault());

    try {
      RandomAccessFile in = new RandomAccessFile(logFileHandle, "r");

      while(in.getFilePointer() < in.length()) {
        String line = in.readLine();

        while((line.endsWith("\r")) || (line.endsWith("\n"))) {
          line = line.substring(0, line.length() - 1);
        }

        try {
          int i = Integer.parseInt(line.substring(0, 1));
          Date lineDate = sdt.parse(line.substring(0, 15));

          int a = line.indexOf("Assertion");
          int e = line.indexOf("Exception");

          if(lineDate.after(eraseAllButException)) {
            if((a != -1) || (e != -1)) {
              out.write(line + lineEnd);
              writeText(in, out);
            } else {
              out.write(line + lineEnd);
            }
          } else if(lineDate.after(eraseAll)) {
            if((a != -1) || (e != -1)){
              out.write(line + lineEnd);
              writeText(in, out);
            }
          }
        } catch(NumberFormatException nfe) {
          //nfe.printStackTrace(System.err);  // these print statements produce a lot of unnecessary output
        } catch(ParseException pe) {
          //pe.printStackTrace(System.err);
        } catch(StringIndexOutOfBoundsException sie) {
          //sie.printStackTrace(System.err);
        }
      }
      in.close();
      out.flush();
      out.close();

      logFile.close();
      logFileHandle.delete();
      logFile = new PrintWriter(new FileOutputStream(logFileHandle));
      RandomAccessFile from = new RandomAccessFile(temp, "r");
      while(from.getFilePointer() < from.length()) {
        String line = from.readLine();

        while((line.endsWith("\r")) || (line.endsWith("\n"))) {
          line = line.substring(0, line.length() - 1);
        }

        logFile.println(line);
      }
      logFile.flush();
      from.close();
      temp.delete();
    } catch(IOException ioe) {
      ioe.printStackTrace(System.err);
    }
  }

  private static void writeText(RandomAccessFile in, FileWriter out) throws IOException {
    String line = null;
    long pointer = in.getFilePointer();
    while(pointer < in.length()) {
      line = in.readLine();

      while((line.endsWith("\r")) || (line.endsWith("\n"))) {
          line = line.substring(0, line.length() - 1);
      }

      try {
        int i = Integer.parseInt(line.substring(0, 1));  // if the line starts with an int we are done
        out.write(lineEnd);
        in.seek(pointer);
        pointer = in.length();
      } catch(NumberFormatException nfe) {
        out.write(line + lineEnd);
        pointer = in.getFilePointer();
      } catch(StringIndexOutOfBoundsException sie) {
        out.write(lineEnd);
        pointer = in.getFilePointer();
      }
    }
    out.flush();
  }
}