/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */

package org.mitre.pam.util.pdifHashtables;


import java.io.InputStream;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import org.mitre.pdif.eventParser.ParseEvent3;
import org.mitre.pdif.eventParser.BadPDIFException;
import org.mitre.pdif.eventParser.PDIFWriter;
import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.executive.PropertySet;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.util.alarm.AlarmReporter;

/**
* A PDIF parsing utility which builds tables of nodes, connections, and the 
* properties associated with each node.  
* @version 1.2, 29 May 1998, Nazario Irizarry to do true state differencing for getting nonitems
* @version 1.1, 28 May 1998, Nazario Irizarry to add factory methods and get rid of string ids and types
* @version 1.01, 19 December 97
* @version 1, 4 December 97
* @author Sam Gerstein
*/
public class PDIFRead extends ParseEvent3 {
	
	static Reporter reporter = new AlarmReporter("bert");
	/*
	 * stores pdif header attributes if any, may be null if there are none
	*/
	protected Hashtable headerAttributes;
	/*
	 * stores the nodes
	*/
	protected Hashtable nodeTable;
	/*
	 * stores the connections
	*/
	protected Vector connectionTable;
	
	/*
	 * stores the baseline nodes
	 */
	protected Hashtable baseNodeTable;
	
	/*
	 * stores baseline connections
	*/
	protected Vector baseConnectionTable;

	protected boolean changedElements;

/**
* Used when stuff is read in to indicate that
* the node has attributes with the same values
* as its "parent" 
*/
	public static final String META_NOT_REDUNDANT = "meta_sameAsParent";
/**
 * 
 * @param inStream java.io.InputStream
 */
public PDIFRead(java.io.Reader inStream) {
	super(inStream);
	connectionTable = new Vector(200);
	nodeTable = new Hashtable(200);
	baseNodeTable = new Hashtable(200);
	baseConnectionTable = new Vector(500);
}
/**
 * Adds a connection using the names and roles passed in.
 * It also adds the connection to the nodes.
 * The nodes must exist.  entries in the roles array should
 * not have the type attached.
 * returns the connection created.
 */
	public void addConnection(PDIFConn conn) {
		// add the connection to each of the involved nodes
		for (int i = 0; i < conn.nodeList.length; i++)
			conn.nodeList[i].putConn(conn);
			
		connectionTable.addElement(conn);
		changedElements = true;
	}
	
	public void addNode(PDIFNode pn) throws DuplicateIDException {
		if (pn == null) {
			System.err.println ("PDIFRead: asked to add null node.  Ignoring.");
			return;
		}
		if (nodeTable.get(pn.name) != null)
			throw new DuplicateIDException(pn.name + " already exists.");
		
		// put the node in.
		nodeTable.put(pn.name, pn);
		pn.changed = true;
		changedElements = true;
	}	
/**
 * Clears the changed flag from all nodes and connections.
 * Makes a complete,independent copy of the nodes and connections.
 */
	public void clearChangedFlags() {
		changedElements = false;
		baseNodeTable.clear();
		baseConnectionTable.setSize(0);
	/*
	* First make a copy of the existing node and its attributes
	*/
		Enumeration enum = nodeTable.elements();
		while (enum.hasMoreElements()) {
			PDIFNode node = (PDIFNode)enum.nextElement();
			node.changed = false;
			PDIFNode nodeCopy = makeNewBaseNode(node.getType(), node.getName());
			nodeCopy.copyAttributesNotConnections(node);
			baseNodeTable.put(nodeCopy.name, nodeCopy);
		}
	/*
	* Then make a master list of all the connections.
	* It is not necessary for each node to know of its connections.
	* We just need the master composite list so we can write
	* differences later.
	*/
		enum = nodeTable.elements();
		while (enum.hasMoreElements()) {
			PDIFNode existingNode = (PDIFNode)enum.nextElement();
			Vector existingConns = existingNode.connections;
			for (int i = existingConns.size()-1; i>= 0; i--) {
				PDIFConn existingConn = (PDIFConn) existingConns.elementAt(i);
				/*
				* only get half of the connections
				* so that we don't have duplicate connections
				* as we scan both endpoints
				*/
				if (existingConn.nodeList[0] == existingNode) {
					PDIFConn baseConn = makeNewBaseConn(2);
					for (int endPoint = existingConn.nodeList.length-1; endPoint>= 0; endPoint--) {
						baseConn.roleList[endPoint] = existingConn.roleList[endPoint];
						String nodeName = existingConn.nodeList[endPoint].name;
						baseConn.nodeList[endPoint] = (PDIFNode) baseNodeTable.get(nodeName);
						if (baseConn.nodeList[endPoint] == null)
							System.err.println("null node during PDIFRead.clearChangedFlags(): "+nodeName);
					}
					baseConnectionTable.addElement(baseConn);
				}
			}
		}
		
	}	
/**
* passes the request on to connections
*/
	public Enumeration connections() {
		return connectionTable.elements();
	}	
/**
* passes the request on to nodeTable
*/
	public Enumeration elements() {
		return nodeTable.elements();
	}
/**
* passes the request on to nodeTable
* (gets elements of type s)
*/
	public Enumeration elements(String s) {
		return new NodeHashtableEnumerator(nodeTable.elements(), s);
	}		

/**
* get a node of specific id
*/
	public PDIFNode get(String name) {
		return (PDIFNode)nodeTable.get(name);
	}	
	
/**
 * does the actual processing of the PDIF token stream.
*/
public void go() throws DuplicateIDException, BadPDIFException, java.io.IOException {
		int ptoken = next();
		PDIFNode tmpNode;
			
		while (ptoken != EOF) {
			if (ptoken == NODE_ENTER) { // we just found a node
				String kind = node.kind.intern();
				String id;
				if (node.id == null || node.id.length() == 0)
					throw new BadPDIFException("Node of type "+kind+" without an id");
				else
					id = node.id.intern();
				tmpNode = makeNewNode(kind, id);
				ptoken = next();
				while (ptoken != NODE_LEAVE) {
					// find all the properties
					if (ptoken == NODE_ATTRIBUTE) {
						String name = attribute.name.intern();
						if (name == META_NOT_REDUNDANT) {
							PDIFNode oldNode = (PDIFNode) nodeTable.get(kind);
							if (oldNode != null) {
								Enumeration e = oldNode.getProperties();
								while (e.hasMoreElements()) {
									String attr = (String) e.nextElement();
									tmpNode.copyPropertyFromAs(attr, oldNode, attr);
								}
							}
						}
						else if (attribute.getValueList().size() > 1) {
							tmpNode.setPropertyValues(name, attribute.getValueList());
						}
						else
							tmpNode.setProperty(name, attribute.getValue());
					}	
					ptoken = next();
				}	

				if (nodeTable.put(tmpNode.name, tmpNode) != null)
					throw new DuplicateIDException(tmpNode.name);
//				nodeTable.put(tmpNode.name, tmpNode);
				
			}	else if (ptoken == CONNECTOR) { // just found a connection
				PDIFConn tmpConn = makeNewConn(2);
				
				for (int i = 0; i < 2; i++) {
					String id = connectorEnds[i].id.intern();
					tmpNode = (PDIFNode)nodeTable.get(id);
					if (tmpNode == null)
						throw new BadPDIFException("Invalid connection: node \"" + connectorEnds[i].id + "\" does not exist.");
					tmpNode.putConn(tmpConn);
					String role = connectorEnds[i].role.intern();
					tmpConn.roleList[i] = role;
					tmpConn.nodeList[i] = tmpNode;
				}
				connectionTable.addElement(tmpConn);
			}	else if (ptoken == HEADER_ATTRIBUTE) { // save the attributes 
					if (headerAttributes == null)
						headerAttributes = new Hashtable(20);
					headerAttributes.put(attribute.name, attribute.getValue());
			}	
			
			ptoken = next();
		}
// undo the effects of putting the nodes together.
		clearChangedFlags(); 
}	
/**
 * checks all nodes and connections for the changed flag
 */
	public boolean hasChangedElements() {
		if (changedElements)
			return true;
		Enumeration enum = nodeTable.elements();
		while (enum.hasMoreElements())
			if (((PDIFNode)enum.nextElement()).changed)
				return true;
		
		return false;
	}	
	public void removeConnection (PDIFConn conn) {
		changedElements = true;
		if (!connectionTable.removeElement(conn)) 
			throw new IllegalArgumentException("removeConnection: Connection " + conn + " is not in the list.");
		for (int i = 0; i < conn.nodeList.length; i++) {
			conn.nodeList[i].removeConn(conn);
		}	
	}	

	public void removeNode(PDIFNode deadNode) {
		if (deadNode == null) {
			throw new IllegalArgumentException ("removeNode: null isn\'t a valid node..");
		}
		changedElements = true;
		nodeTable.remove(deadNode.name);
//		deadNode.changed = true;
	/*
	* walk through all the connections
	*/
		Enumeration enum = deadNode.connections();
		for (int i = deadNode.connections.size()-1; i>= 0; i--) {
			PDIFConn conn = (PDIFConn) deadNode.connections.elementAt(i);
		/*
		* Tell each node at the end of the connection to remove the connection.
		*/
			for (int j = 0; j< conn.nodeList.length; j++) {
				conn.nodeList[j].removeConn(conn);
			}
		/*
		* delete the connection from the list of connections
		*/
			connectionTable.removeElement(conn);
		}
	}	


	public int size() {
		if (nodeTable == null)
			return 0;
		else
			return nodeTable.size();
	}	

	private void writeConns(PDIFWriter pOut, boolean writeUnchangedElements) throws java.io.IOException{
		Enumeration connEnum = connectionTable.elements();
		while (connEnum.hasMoreElements()) {
			PDIFConn c = (PDIFConn)connEnum.nextElement();
			/*
			* a connection must be written if it does not exist in the base connection table
			*/
			boolean exist = false;
			if (!writeUnchangedElements) {
				for (int i = baseConnectionTable.size()-1; i >= 0; i--) {				
					if (c.equalsBasedOnNameAndRole((PDIFConn)baseConnectionTable.elementAt(i))) {
						exist = true;
						break;
					}
				}
			}

			if (writeUnchangedElements || !exist ) {
				pOut.writeRelation(c.nodeList[0].type, c.roleList[0], c.nodeList[0].name, 
						c.nodeList[1].type, c.roleList[1], c.nodeList[1].name);
//				System.out.println(" + conn "+c.roleList[0]+" " +c.nodeList[0].name+" "+c.roleList[1]+" " +c.nodeList[1].name);
			}
		}		


	}


	private void writeNonConns(PDIFWriter pOut) throws java.io.IOException{
		
		Enumeration connEnum = baseConnectionTable.elements();
		while (connEnum.hasMoreElements()) {
			PDIFConn c = (PDIFConn)connEnum.nextElement();
			String endPoint1 = c.nodeList[0].name;
			String endPoint2 = c.nodeList[1].name;
			/*
			* a connection ceases to exist if either of its endpoints no longer exist
			* or if the endpoints exist buy they are no longer connected
			*/
			boolean exist = false;
			PDIFNode oneEnd = (PDIFNode)nodeTable.get(endPoint1);
			if (oneEnd != null && nodeTable.containsKey(endPoint2) ) {
				for (int i = oneEnd.connections.size()-1; i>=0 ; i--) {
					PDIFConn conn = (PDIFConn)oneEnd.connections.elementAt(i);
					if (conn.equalsBasedOnNameAndRole(c)) {
						exist = true;
						break;
					}
				}
			}

			if ( !exist ) {
				pOut.writeNonRelation(c.nodeList[0].type, c.roleList[0], c.nodeList[0].name, 
						c.nodeList[1].type, c.roleList[1], c.nodeList[1].name);
//				System.out.println(" - conn "+c.roleList[0]+" " +c.nodeList[0].name+" "+c.roleList[1]+" " +c.nodeList[1].name);
			}
		}		
	}

	private void writeNonNodes(PDIFWriter pOut) throws java.io.IOException{
		Enumeration nodeEnum = baseNodeTable.elements();

		while (nodeEnum.hasMoreElements()) {
			PDIFNode n = (PDIFNode)nodeEnum.nextElement();
			if (!nodeTable.containsKey(n.name)) {
				pOut.writeNonNode(n.type, n.name);
//				System.out.println(" - node: "+n.name);
			}
		}	
	}	

	private void writeNodes(PDIFWriter pOut, boolean writeUnchangedElements) throws java.io.IOException{
		Enumeration nodeEnum = nodeTable.elements();

		while (nodeEnum.hasMoreElements()) {
			PDIFNode n = (PDIFNode)nodeEnum.nextElement();
			if (writeUnchangedElements || n.changed || (!baseNodeTable.containsKey( n.name))) {
				pOut.writeNode(n.type, n.name);
//				System.out.println(" + node: "+n.name);
				Enumeration attribEnum = n.getAttributes().getProperties();
				while (attribEnum.hasMoreElements()) {
					Object attribObj;
					attribObj = attribEnum.nextElement();
					if (attribObj instanceof String) {
						ValueList vl = n.getPropertyValues((String)attribObj);
						if (vl.size() == 0) {
							pOut.writeAttributeValue((String)attribObj, "");
						}
						else if (vl.size() == 1) {
							pOut.writeAttributeValue((String)attribObj, vl.elementAt(0));
						}
						else {
							pOut.writeAttributeValue((String)attribObj, vl.elements());
						}
					}
				}	
			}	
		}	
	}	
	public void writeState(java.io.PrintWriter PWout, boolean writeNonElements, boolean writeUnchangedElements) throws java.io.IOException {
		PDIFWriter pOut = new PDIFWriter(PWout, "PAM", "GUI");

		if (writeNonElements) {
			writeNonConns(pOut);
			writeNonNodes(pOut);
		}

		writeNodes(pOut, writeUnchangedElements);
		writeConns(pOut, writeUnchangedElements);
		pOut.close();
	}	
	
	public PDIFNode makeNewNode(String type, String id) {
		if (type == null)
			throw new IllegalArgumentException("type == null");
		if (id == null)
			throw new IllegalArgumentException("id == null");
		PDIFNode pn = new PDIFNode(type, id, new PropertySet(type, id, reporter));
		newNodeMade(pn);
		return pn;
	}
	public PDIFConn makeNewConn(int len) {
		PDIFConn c = new PDIFConn(len);
		newConnMade(c);
		return c;
	}

	public PDIFNode makeNewBaseNode(String type, String id) {
		return makeNewNode(type, id);
	}
	public PDIFConn makeNewBaseConn(int len) {
		return makeNewConn(len);
	}

	public void newNodeMade(PDIFNode node) {
	}
	public void newConnMade(PDIFConn conn) {
	}
/**
* When you rename a node you also have to update the hashtable
* that contains it and you have to ensure that the new 
* name is in the dictionary
*/
	public void renameNode(PDIFNode existingNode, String newName) {
		String oldName = existingNode.getName();
		existingNode.rename(newName);
		nodeTable.remove(oldName);
		nodeTable.put(newName, existingNode); // new name = new key
	}	
}