/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util.smtp;

/*********************************************************************

@version 1.00 7/17/96
@author James Driscoll maus@io.com

Usage -

Version History:
0.8 5/15/96  - First version
0.9 5/16/96  - fixed date code, added localhost to HELO,
               fixed Subject bug
0.91 7/10/96  - Yet another date fix, for European TimeZones.  Man, they
                gotta fix that code...
1.00 7/17/96  - renamed to Qsmtp, as I have plans for the SMTP code,
                and I want to get this out and announced.  Also cleaned it
                up and commented out the DEBUG code (for size, just in case
                the compiler didn't optimize it out on your machine - mine
                didn't (Symantec Cafe Lite, you get what you pay for, and
                I paid for a book)).
1.01 9/18/96  - Fixed the call to getLocalHost local, which 1.02 JDK didn't
                like (Cafe Lite didn't mind, though).  Think I'll be using
                JDK for all compliations from now on.  Also, added a close
                method, since finalize() is not guarenteed to be called(!).
1.1 12/26/96 -  Fixed problem with EOL, I was using the Unix EOL, not the
                network end of line.  A fragile mail server was barfing.
                I can't believe I wrote this - that's what half a year will do.
                Also, yanked out the debug code.  It annoyed me.
1.11 12/27/97 - Forgot to flush(), println used to do that for me...


***********************************************************************/


/**
* 6-2-98
* This class has been rewritten to support JDK 1.1.6 and to use the MailComposer
* object.  Responsibility for writing properly formatted mail messages has been
* removed from this class and placed in the MailComposer object.  This class
* is now only responsible for the connection with the server and the SMTP
* protocol.
*
* @author John R. Fantry
* @author Jason Mathews 10/11/99 Added private readLine method to return non-null strings
*         and added check for non-null values in close(), both to prevent null-ptr-exceptions.
*         Added localhost string to lookup localhost name only once.
*         Added connected flag to check for a live connection.
*/

import java.net.*;
import java.io.*;
import java.util.*;

import org.mitre.pam.util.PrintCRLFWriter;

public class Qsmtp {

    protected static final int DEFAULT_PORT = 25;

    //protected BufferedWriter out;
    protected PrintCRLFWriter out;
    protected BufferedReader in;

    protected Socket sock;

    protected String host;

    protected int port = DEFAULT_PORT;

    protected String localhost;

    private boolean connected;

    /**
    *   Create a Qsmtp object pointing to the specified host
    *   @param hostid The host to connect to.
    */
    public Qsmtp( String hostid) {
        this(hostid, DEFAULT_PORT);
    }

    public Qsmtp( String hostid, int port) {
      host = hostid;
      this.port = port;
    }

    public Qsmtp( InetAddress address ) {
        this(address, DEFAULT_PORT);
    }

    public Qsmtp( InetAddress address, int port ) {
      host = address.getHostName();
      this.port = port;
    }

    public void connect() throws IOException, ProtocolException {

      if (connected) return; // already connected

      sock = new Socket( host, port );
      in = new BufferedReader(new InputStreamReader(sock.getInputStream()));
      //out = new BufferedWriter(new OutputStreamWriter(sock.getOutputStream()));
      out = new PrintCRLFWriter(new OutputStreamWriter(sock.getOutputStream()));
      String rstr = readLine();
      if (!rstr.startsWith("220")) throw new ProtocolException(rstr);
      while (rstr.indexOf('-') == 3) {
        rstr = readLine();
        if (!rstr.startsWith("220")) throw new ProtocolException(rstr);
      }

      if (localhost == null) {
      InetAddress local;
      try {
        local = InetAddress.getLocalHost();
	} catch (UnknownHostException ioe) {
        System.err.println("No local IP address found - is your network up?");
        throw ioe;
      }
	localhost = local.getHostName();
      }

      //out.write("HELO " + localhost);
      out.println("HELO " + localhost);
      //out.newLine();
      out.flush();
      rstr = readLine();
      if (!rstr.startsWith("250")) throw new ProtocolException(rstr);

      connected = true;
    }

    protected void finalize() throws Throwable {
        this.close();
        super.finalize();
    }

    public final boolean isConnected() {
	return connected;
    }

  /**
  * Use this function to send a piece of email.
  *
  * @param message  The message to send.
  *
  * @throws IOException  Thrown if there is an error communicating with the server.
  * @throws ProtocolException  Thrown if there is a violation of the SMTP protocol.
  */
  public void sendmsg(MailComposer message) throws IOException, ProtocolException {
    if (!connected)
	throw new IOException("Not connected");

    String rstr;
    String sstr;
    String from = message.getFrom();
    if(from.indexOf('<') == -1) {
      from = "<" + from + ">";
    }
    sstr = "MAIL FROM: " + from;
    //out.write(sstr);
    out.println(sstr);
    //out.newLine();
    out.flush();
    rstr = readLine();
    if (!rstr.startsWith("250")) throw new ProtocolException(rstr);
    String to = message.getTo();
    if(to.indexOf('<') == -1) {
      to = "<" + to + ">";
    }
    sstr = "RCPT TO: " + to;
    //out.write(sstr);
    out.println(sstr);
    //out.newLine();
    out.flush();
    rstr = readLine();
    if (!rstr.startsWith("250")) throw new ProtocolException(rstr);
    //out.write("DATA");
    out.println("DATA");
    //out.newLine();
    out.flush();
    rstr = readLine();
    if (!rstr.startsWith("354")) throw new ProtocolException(rstr);

    message.writeMessage(out);
    out.flush();

    rstr = readLine();
    if (!rstr.startsWith("250")) throw new ProtocolException(rstr);
  }

  /*
   * Returns non-null string input from input stream
   */
  private String readLine() throws IOException {
    String line = in.readLine();
    if (line == null) line = "";
    return line;
  }

  public void close() {
    connected = false;
    if (sock != null) {
      try {
	if (out != null) {
	    //out.write("QUIT");
	    out.println("QUIT");
	    //out.newLine();
	    out.flush();
	}
	sock.close();
      } catch (IOException ioe) {
	// As though there's anything I can do about it now...
      }
      sock = null;
    }
  }

}
