/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util.smtp;

import java.util.*;
import java.io.*;
import java.net.*;

import org.mitre.pam.executive.ExecGlobal;

/**
* This class allows an agent to send mail with attachments via SMTP.
*
* Copyright (c)1997 The Mitre Corporation, All rights reserved
*
* @author John Fantry, Technical Co-op (D510)
*/
public class SendMail {

  protected String host;
  protected Qsmtp smtp;

/**
* A given subclass may wish to change the timeout and retry strategy.
* The length of timeouts[] must be the same as the length of retries[]
*/
  protected int timeouts[] = {5000, 15000, 60000};
/**
* A given subclass may wish to change the timeout and retry strategy.
* The length of timeouts[] must be the same as the length of retries[]
*/
  protected int retries[] = {4, 3, 2};

  /**
  * Creates a new <tt>SendMail</tt> object.
  * @parm host  The address of the SMTP mail server (i.e. mbunix.mitre.org).
  */
  public SendMail(String host) {
    this.host = host;
    smtp = new Qsmtp(this.host);
  }

  /**
  * Establishes an SMTP connection with the host.
  * @exception  SocketException
  *             Occurs if a socket connection can not
  *             be established with the host.
  */
  protected void connectToSMTP() throws SocketException {

	for (int i = 0; i < retries.length; i++) {
	    int tries = 0;

	    while ((! smtp.isConnected()) && (tries < retries[i])) {
	      tries++;
	      try {
	        smtp.connect();                     // Establish connection with SMTP server
		return;	                            // connection established
	      }
	      catch (Exception e) {
	        try {
	          ExecGlobal.getExec().getExecAsAgent().getReporter().log("mail retry, delay "+timeouts[i]/1000+" seconds");
	          Thread.sleep(timeouts[i]);
	        }
	        catch(Exception ex) { }
	      }
	    }
	}
    if (! smtp.isConnected()) {
      throw new SocketException("Can't establish connection with " + host);
    }
  }

  /**
  * Closes the SMTP connection.
  */
  public void closeSMTP() {
      smtp.close();
  }

  /**
  * Closes the connection with the host if it is still open.
  */
  protected void finalize() {
    closeSMTP();
  }

  /**
  * This function sends a mail message.
  *
  * @param message  The mail message to send.
  *
  * @throws IOException  Thrown if there is an error communicating with the server.
  * @throws ProtocolException  Thrown if there is a violation of the SMTP protocol.
  */
  public void sendMail(MailComposer message) throws IOException, ProtocolException {
    if (! smtp.isConnected()) {
      connectToSMTP();
    }
	try {
    	smtp.sendmsg(message);
    }
    catch (ProtocolException e) {
    	String msg = e.getMessage();
    	if (msg.startsWith("550")) {
    		ExecGlobal.getExec().getExecAsAgent().getReporter().log("Unknown user at SMTP host "+host+"("+msg+")");
    	}
    	else {
    		throw e;
    	}
    }
  }

  /**
  * This function will construct a mailing list from a given userList string.
  * The default host should be the name of an SMTP server which will also become our
  * initial default host.  The "userList" will contain the names of users, full
  * email addresses, or combinations of both.  If an element in the "userList"
  * is a user name, we'll append the name of the current default host. If an element
  * in the "userList" is a full email address, then that host will become
  * the new default host.
  *
  * @return  A vector of email addresses (user@host pairs).
  *
  */
  public static Vector buildMailingList(String userList, String defHost) {
    Vector mailingList = new Vector();
    if (userList == null || userList.length() == 0)
	return mailingList; 				// empty list

    StringTokenizer recipiantList = new StringTokenizer(userList, ", ");
    String host = "@" + defHost;  // default value

    while (recipiantList.hasMoreElements()) {
      String recipiant = (String) recipiantList.nextElement();

      /**
      * If the recipiant includes a host name then that host will become the
      * new default host for all the recipiants that do not have host names
      * until a new default is found.
      */
      int ind = recipiant.indexOf('@');
      if (ind >= 0) {
        host = recipiant.substring(ind);
        mailingList.addElement(recipiant);
      }
      else {
        // The recipient is just a user name, we'll supply the host
        mailingList.addElement(recipiant + host);
      }
    }
    return mailingList;
  }


}