/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util.smtpIn;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.text.SimpleDateFormat;

import java.util.Date;
import java.util.Vector;


/**
* This abstract class is intended to be used as a base class for any class
* that is used to decompose mail messages.  This class implements the MailDecomposer
* interface and provides function implementations that may be common to all
* mail decomposers.
*
* @author John R. Fantry
*/
public abstract class GenericMaild implements MailDecomposer {

  protected static final String DATE_FIELD = "date";
  protected static final String FROM_FIELD = "from";
  protected static final String SUBJECT_FIELD = "subject";
  protected static final String TO_FIELD = "to";
  protected static final String DATE_FORMAT = "EEE, dd MMM yyyy hh:mm:ss zzzzz";

  protected SimpleDateFormat dateFormatter = new SimpleDateFormat(DATE_FORMAT);

  // Header fields
  protected Date date;
  protected String from;
  protected String subject;
  protected String field;
  protected String value;

  /**
  * A message can be addressed to multiple recipiants so we need a vector
  * for this header field
  */
  protected Vector to = new Vector();

  protected Vector attachmentNames = new Vector();

  protected GenericMaild() {
  }

  /**
  * This function mimicks the behavior of the trim function contained in the
  * java.lang.String class except that it seems to work better.  I found that
  * the original trim function was returning strings that still had white space
  * at the beginnig.
  *
  * @return  A string where all the white space has been removed from the
  *          begining and end of the source string.
  *
  * @param src  The string to trim.
  */
  protected static String trim(String src) {
    if(src.length() < 1) {
      return src;
    }

    int start = 0;
    int stop = src.length() - 1;
    while(start < stop && Character.isWhitespace(src.charAt(start))) {
      start++;
    }
    while(stop > start && Character.isWhitespace(src.charAt(stop))) {
      stop--;
    }

    if(Character.isWhitespace(src.charAt(start)))
      return new String();
    else
      return src.substring(start, stop + 1);
  }

  /**
  * This function determines whether or not a given string contains a header
  * field.
  *
  * @return True if the line contains a header field, false otherwise.
  *
  * @param line  A string that may contain a header field.
  */
  protected boolean containsField(String line) {

    // All header fields are of the format "field:value".
    int indexOfColon = line.indexOf(":");

    // If the line does not contain a ":" than it can't be a field
    if(indexOfColon < 0) {
      return false;
    }
    else {
      field = line.substring(0, indexOfColon);

      // Fields may contain empty values
      if(line.length() > indexOfColon + 1) {
        value = line.substring(indexOfColon + 1, line.length());
      } else {
        value = new String();
      }
      field = trim(field);
      value = trim(value);
      return true;
    }
  }

  /**
  * This function attempts to extract an attachment with the given name.
  * The attachment is saved into the dst file.
  *
  * @return A String representing the MIME type of the attachment that
  * is extracted.
  *
  * @param attachmentName  The name of the attachment to extract.
  * @param dst             The destination file for the attachment.
  *
  * @exception FileNotFoundException  Thrown if the attachment name can't be
  *                                   found in the mail message.
  * @exception IOException            Thrown if there is an error reading the file.
  */
  public String extractFileAttachment(String attachmentName, File dst) throws FileNotFoundException, IOException  {
    if(!attachmentNames.contains(attachmentName)) {
      throw new FileNotFoundException(attachmentName + " not found");
    }
    int attachmentNum = attachmentNames.indexOf(attachmentName);
    return extractFileAttachment(attachmentNum, dst);
  }

  /**
  * This function attempts to extract an attachment that is identified within
  * the mail message by index.  The index is zero based.
  * The attachment is saved into the dst file.
  *
  * @return A String representing the MIME type of the attachment that
  * is extracted.
  *
  * @param attachmentNum  The index of the attachment to extract.
  * @param dst            The destination file for the attachment.
  *
  * @exception IOException  Thrown if there is an error reading the file.
  */
  public abstract String extractFileAttachment(int attachmentNum, File dst) throws IOException;

  public Date getDate() {
    return date;
  }

  public String getFrom() {
    return from;
  }

  public String getSubject() {
    return subject;
  }

  public Vector getTo() {
    return to;
  }

  public boolean attachmentsExist() {

    // If the vector contains any elements then an attachment exists
    if(attachmentNames.size() > 0)
      return true;
    else
      return false;
  }

  public Vector getAttachmentNames() {
    return attachmentNames;
  }

  public int numOfAttachments() {
    return attachmentNames.size();
  }
}