/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.util.smtpIn;

import java.io.File;
import java.io.IOException;
import java.io.FileNotFoundException;

import java.util.Hashtable;
import java.util.Enumeration;


import java.net.ServerSocket;

import java.util.Vector;


/**
* The main purpose of this class is to start and stop the SMTP server.
* The server starts by creating a ServerSocket and spawning a number
* of connection threads to listen for incoming connections.  The server stops
* by setting a stop flag in the connection threads.  Once the stop flag is
* set the threads will finish processing any open connections and will
* then shut down.  Once all threads have stopped the ServerSocket is closed.
* <P>
* If you are unfamiliar with the Simple Mail Transfer Protocol (SMTP) please
* see RFC 821.  This document can be found at "ftp://ds.internic.net/rfc/rfc821.txt".
*
* @author John R. Fantry
*/
public class SMTPserver {

  protected static final int SMTPport = 25;  // well known SMTP port
  protected static final int defaultThreadNum = 2;
  protected static final int defaultTimeout = 5000;  // miliseconds

  protected int threadNum;
  protected int timeout;
  protected String hostname;
  protected File workingDir;
  protected ServerSocket server;
  protected Vector threads = new Vector();

  /**
  * At a minimum you must provide the hostname of the machine that the
  * server is running on and a working directory for the server to use
  * for mail transfers.  The hostname is needed so that the proper
  * human readable responses can be generated by the server during an
  * SMTP session.  The working directory is needed so the server can save
  * all incoming mail messages.  Once a mail message has been delivered
  * it is deleted.
  *
  * @param hostname    The hostname that the server will be running on.
  * @param workingDir  Directory used to store temporary files.
  */
  protected SMTPserver(String hostname, File workingDir) {
    this(hostname, workingDir, defaultThreadNum, defaultTimeout);
  }

  /**
  * Use this constructor if you want to override the default values for
  * threadNum(2) and timeout(5000).  threadNum represents the total
  * number of connection threads that will be spawned by the server and
  * timeout represents how long (in miliseconds) a connection thread will
  * listen for an incoming connection before stopping to see if its stop
  * flag has been set to true.  If the stop flag is set to false the
  * thread will start listening again, otherwise it will shut down.
  *
  * @param hostname    The hostname that the server will be running on.
  * @param workingDir  Directory used to store temporary files.
  * @param threadNum   The number of connection threads to create.
  * @param timeout     The time (in miliseconds) a connection thread
  *                    listens before checking the stop flag.
  */
  protected SMTPserver(String hostname, File workingDir, int threadNum, int timeout) {
    if(threadNum > 0)
      this.threadNum = threadNum;
    else
      this.threadNum = defaultThreadNum;

    if(timeout > 0)
      this.timeout = timeout;
    else
      this.timeout = defaultTimeout;

    this.hostname = hostname;

    if(workingDir.isDirectory())
      this.workingDir = workingDir;
    else
      this.workingDir = new File(System.getProperty("user.dir"));
  }

  /**
  * This function is responsible for creating a ServerSocket and
  * spawning the connection threads.  Once this function returns
  * the server will be up and running.
  *
  * @exception IOException  Thrown if there is an error creating the
  *                         ServerSocket.
  */
  public void start() throws IOException {
    server = new ServerSocket(SMTPport);
    server.setSoTimeout(timeout);

    for(int i=0; i<threadNum; i++) {
      SMTPconnection connection = new SMTPconnection(hostname, workingDir, server);
      threads.addElement(connection);

      /**
      * Not exactly sure what affect this next statement has.  In theory
      * all connection threads should be daemon threads because they only
      * live to service other threads.
      */
      connection.setDaemon(true);
      connection.start();
    }
  }

  /**
  * This function is responsible for setting the stop flag on the
  * connection threads, waiting for them to die, and then closing
  * the ServerSocket.  Once this function returns the server will
  * be down.
  *
  * @exception IOException  Thrown if there is an error closing the
  *                         ServerSocket.
  */
  public void stop() throws IOException {

    /**
    * The stop flag is a static member variable of the SMTPconnection
    * class.  This makes it easy to stop all the connection threads
    * at once, however it is impossible to stop individual connection
    * threads.  The ability to select an individual connection
    * thread for shutdown does not currently exist so this design should
    * be sufficient for now.
    */
    SMTPconnection.stopConnection();
    for(int i=0; i<threads.size(); i++) {
      try {
      ((Thread)threads.elementAt(i)).join();
      } catch(InterruptedException ie) {
        /**
        * Not exactly sure what to do if this happens.  For now I'll
        * consider the thread object that created the exception to be
        * dead and move on to the next thread.  Basically this exception
        * is ignored.
        */
	SMTPconnection.reportException(ie);
      }
    }
    server.close();
  }

/*
* This methods is used to maintain a central registry of addresses and the
* SMTPlisteners that want to receive mail messages for those addresses.
* This class was not designed to be instantiated because it is a shared
* resource.  All public functions are declared static.
*
* SMTPlistners must register here if they want to receive mail from the server.
* They must provide a mailing list which represents the names of the users
* for which they want to receive mail.
*
* SMTPconnection threads look in the registry to find out if a particular
* recipient is registered.  This determines whether or not that thread
* will accept or reject an incoming mail message.
*
*
* @author John R. Fantry
*/
  

  /**
  * This hashtable represents the entire registry.  The keys in this hashtable
  * are email addresses and the values are vectors of SMTPlisteners that want
  * to be notified if email is received for that address.
  */
  protected static Hashtable registry = new Hashtable();


  /**
  * This function allows an SMTPlistener to register to receive mail.
  * An SMTPlistener (registrant) must provide a reference to itself for
  * callback and a mailing list.  The mailing list should contain a list
  * of email addresses for which it wants to receive mail.
  *
  * @param registrant   An SMTPlistener object.
  * @param mailingList  A list of email addresses that the registrant wants
  *                     to receive mail for.
  */
  public void addSMTPeventListener(Object registrant, Vector mailingList) {
    for(int i=0; i<mailingList.size(); i++) {

      /**
      * The String objects (email addresses) in the mailing list will be
      * used as the keys in the registry.
      */
      String recipient = (String)mailingList.elementAt(i);

      /**
      * If the recipient is already a key in the registry then a different
      * SMTPlistener has already registerd this address.  In this case the
      * new registrant will be added onto the existing list of SMTPlisteners.
      * All the SMTPlisteners in the list will receive mail events for this
      * address.
      *
      * If the recipient is not in the registry then a new list of SMTPlisteners
      * will be created with the registrant as the only entry.
      */
      if(registry.containsKey(recipient)) {
        Vector subscriberList = (Vector)registry.get(recipient);
        subscriberList.addElement(registrant);
        registry.put(recipient, subscriberList);
      }
      else {
        Vector subscriberList = new Vector();
        subscriberList.addElement(registrant);
        registry.put(recipient, subscriberList);
      }
    }
  }

  /**
  * This function allows an SMTPlistener to unregister.  When this function
  * is complete the SMTPlistener will no longer receive any mail events until
  * they register again.
  *
  * @param registrant  An SMTPlistener that no longer wants to receive any email.
  */
  public void removeSMTPeventListener(Object registrant) {
    Enumeration recipients = registry.keys();
    while(recipients.hasMoreElements()) {
      String recipient = (String)recipients.nextElement();
      Vector subscriberList = (Vector)registry.get(recipient);
      if(subscriberList.contains(registrant)) {
        subscriberList.removeElement(registrant);
        registry.put(recipient, subscriberList);
      }
    }
  }

  /**
  * This function is used to verify whether or not an email address is
  * registered.  If the address is registered then there is at least one
  * SMTPlistener that wants to be notified.  If the address is not registered
  * then the mail should not be accepted.
  *
  * The SMTPconnection class uses this function when deciding whether or not
  * to accept a new mail message.
  *
  * @return A boolean that tells whether or not the recipient is registered.
  *
  * @param recipient  The address that will be looked up as a key.
  */
  public static boolean registered(String recipient) {
    if(registry.containsKey(recipient))
      return true;
    else
      return false;
  }

  /**
  * The value associated with the provided key will be returned.
  * The key in this case should be a registered email address and the
  * Vector that gets returned will be a list of SMTPlisteners that have
  * registered to receive mail events for that address.
  *
  * @return A Vector that contains a list of SMTPlisteners that are registered
  * to receive mail events for the provided address.
  *
  * @param key  Key in registry.
  */
  public static Vector getSubscriberList(Object key) {
    return (Vector)registry.get(key);
  }

  /**
  * This function is responsible for delivering SMTPevent objects.  The first
  * part of the function creates customized mailing lists for the SMTPlisteners.
  * The event object is then delivered to those SMTPlisteners.
  *
  * @param event  An SMTPevent object to deliver.
  */
  public static void fireSMTPevent(SMTPevent event) {
	Hashtable subscriberMailingList = new Hashtable();

    Vector mailingList = event.getMailingList();

    // time to build the customized mailing lists
    for(int i=0; i<mailingList.size(); i++) {
      String recipient = (String)mailingList.elementAt(i);
      Vector subscriberList = getSubscriberList(recipient);
      for(int x=0; x<subscriberList.size(); x++) {
        SMTPlistener listener = (SMTPlistener)subscriberList.elementAt(x);
        if(subscriberMailingList.containsKey(listener)) {
          Vector list = (Vector)subscriberMailingList.get(listener);
          list.addElement(recipient);
          subscriberMailingList.put(listener, list);
        }
        else {
          Vector list = new Vector();
          list.addElement(recipient);
          subscriberMailingList.put(listener, list);
        }
      }
    }

    /**
    * Now that all the customized mailing lists have been created, it's
    * time to deliver the mail event.
    */
    Enumeration listeners = subscriberMailingList.keys();
    while(listeners.hasMoreElements()) {
      SMTPlistener listener = (SMTPlistener)listeners.nextElement();
      Vector list = (Vector)subscriberMailingList.get(listener);
      try {
        listener.SMTPmailEvent(new SMTPevent(event.getMailMessage(), list));

        /**
        * These exceptions would be thrown by the SMTPevent constructor.
        * I don't know why this would ever happen, but if it does I guess all
        * I can do is ignore it, consider the message dead, and move on.
        */
      } catch(FileNotFoundException fnfe) {
	SMTPconnection.reportException(fnfe);
      } catch(IOException ioe) {
	SMTPconnection.reportException(ioe);
      }
    }
  }

}
