/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pdif.eventParser;

import java.io.PrintWriter;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Date;
import java.io.Writer;

/**
* An output utility class that complements the structure of ParseEvent3.
* Isolates user from many details of PDIF data format.
*
* @see ParseEvent3
* @author Nazario Irizarry, Jr.
*/

public class PDIFWriter {
/**
* The print stream that PDIF is written to.
*/
	PrintWriter prnt;
	
	private int outputLevel = 0;
	private static int NODES = 1;
	private static int CONNECTORS = 2;
    private static int NONNODES = 3;
    private static int NONCONNECTORS = 4;
	private int inWhere = 0;
	private int lineLength;
	
/**
* Creates an instance of this class and writes the header domain, source, and date
* to the output.
*/
	public PDIFWriter(PrintWriter prnt, String domain, String source, long date) throws IOException {
		this.prnt = prnt;
		writeHeader(domain, source, date);
	}

/**
* Creates an instance of this class and writes the header domain, source, and current date
* to the output.
*/
	public PDIFWriter(PrintWriter prnt, String domain, String source) throws IOException {
		this.prnt = prnt;
		writeHeader(domain, source, System.currentTimeMillis());
	}
	
/*
*  1       2      3
* (pdif (node (attribute
*/
	private void closeToLevel(int level) {
		if (outputLevel <= level)
			return;
		
		while (outputLevel > level) {
			print(")");
			outputLevel--;
			if (outputLevel < 2)
				println();
		}
	}

/*
*  1       2      3
* (pdif (node (attribute
*/
	private void writeHeader(String domain, String source, long date) throws IOException{
		closeToLevel(0);
		print("(pdif (header ");
		
		outputLevel =2 ;
		writeHeaderAttribute("domain", domain);
		writeHeaderAttribute("source", source);
		writeHeaderAttribute("date", (new PDIFDate().toPDIF()));

		
	}

/**
* Header attributes can only be written at level 1
*/
	public void writeHeaderAttribute(String attr, String value) throws IOException{
		if (outputLevel != 2) {
			throw new IOException("attempt to write PDIF header attribute when level was "+outputLevel+" instead of 1");
		}
		if (attr == null || attr.length()==0 || value == null || value.length()==0) {
			throw new IOException("attempt to null or empty attribute header ");
		}
		print("(");
		print(attr);
		print(" ");
		print(ParseEvent2.selectiveQuote(value));
		print(")");
	}

/**
* Begins out put of new PDIF node.  Automatically closes out definition
* of any prior nodes.
*/
	public void writeNode(String kind, String id) throws IOException {
		if (kind == null || kind.length() == 0) {
			throw new IOException("attempt to write null or empty kind");
		}

		
		if (outputLevel <= 0) {
			throw new IOException("attempt to write Node to pdif stream before writing PDIF header");
		}
		else if (inWhere != NODES) {
			closeToLevel(1);
			println();
			println("\t(node");
			outputLevel = 2;
			inWhere = NODES;
		}

		closeToLevel(2);

		println();
		print("\t\t(");
		print(kind);
		if (id != null && id.length() > 0) {
			print(' ');
			print(ParseEvent2.selectiveQuote(id));
		}
		
		outputLevel = 3;

	}
/**
* Writes out a PDIF nonnode declaration.  
* Automatically adjusts parenthesis as needed to keep PDIF integrity.
*/
	public void writeNonNode(String kind, String id) throws IOException {
		if (kind == null || kind.length() == 0) {
			throw new IOException("attempt to write null or empty kind");
		}

		
		if (outputLevel <= 0) {
			throw new IOException("attempt to write Node to pdif stream before writing PDIF header");
		}
		else if (inWhere != NONNODES) {
			closeToLevel(1);
			println();
			println("\t(nonnode");
			outputLevel = 2;
			inWhere = NONNODES;
		}

		closeToLevel(2);

		println();
		print("\t\t(");
		print(kind);
		if (id != null && id.length() > 0) {
			print(' ');
			print(ParseEvent2.selectiveQuote(id));
		}
		
		outputLevel = 3;

	}

/**
* Writes out an attribute of any type, single or multi-valued.
* Understands Date, PDIFDate, ValueList.  The rest
* have their values extracted via toString().
*/
	public void writeAttributeValue(String attribute, Object value) throws IOException {
		if (value instanceof ValueList) {
			writeAttributeValue(attribute, ((ValueList) value).elements());
			return;
		}
		
		writeAttributeValuePreamble(attribute);
		if (value != null) {
			print(' ');
			String val;
			if (value instanceof PDIFDate) {
				val = ((PDIFDate)value).toPDIF();
			}
			else if (value instanceof Date) {
				val = PDIFDate.toPDIF((Date)value);
			}
			else {
				val = value.toString();
			}
			print(ParseEvent2.selectiveQuote(val));
		}
		print(')');
		convenientEOL();
	}
/**
* Writes out a multi-valued attribute.
*/
	public void writeAttributeValue(String attribute, String[] values) throws IOException {
		writeAttributeValuePreamble(attribute);
		for (int i = 0; i<values.length; i++) {
			String value = values[i];
			print(' ');
			if (value != null ) {
				print(ParseEvent2.selectiveQuote(value));
			}
			else {
				print("\"\"");
			}
		}
		print(')');
		convenientEOL();
	}

/**
* Writes out a multi-valued attribute.  The Enumerated object is tested.  If it
* is of type Date it is properly output.  Otherwise its toString() method is used.
*/
	public void writeAttributeValue(String attribute, Enumeration values) throws IOException {
		writeAttributeValuePreamble(attribute);
		while (values.hasMoreElements()) {
			Object value = values.nextElement();
			if (value != null) {
				print(' ');
				String val;
				if (value instanceof PDIFDate) {
					val = ((PDIFDate)value).toPDIF();
				}
				else if (value instanceof Date) {
					val = PDIFDate.toPDIF((Date)value);
				}
				else {
					val = value.toString();
				}
				print(ParseEvent2.selectiveQuote(val));
			}
			else {
				print("\"\"");
			}
		}
		print(')');
		convenientEOL();
	}

/**
* Writes out a multi-valued attribute.
* Automatically adjusts parenthesis as needed to keep PDIF integrity.
*/
	private void writeAttributeValuePreamble(String attribute) throws IOException {
		if (attribute == null || attribute.length() == 0) {
			throw new IOException("attempt to write null or empty attribute");
		}

		closeToLevel(3);
		
		if (outputLevel <= 2) {
			throw new IOException("attempt to write attibute value to PDIF stream before writing node");
		}

		print('(');
		print(ParseEvent2.fancyQuote(attribute));
	}

/**
* Writes a PDIF comment at whatever level the output stream happens to be.
*/
	public void writeComment(String comment) {
		print("(- ");
		print(comment);
		println(')');
	}
	
/**
* Writes out a single-valued attribute.
* Automatically adjusts parenthesis as needed to keep PDIF integrity.
*/
	public void writeAttributeValue(String attribute, long value) throws IOException {
		writeAttributeValue(attribute, Long.toString(value));
	}

/**
* Writes out a single-valued attribute.
* Automatically adjusts parenthesis as needed to keep PDIF integrity.
*/
	public void writeAttributeValue(String attribute, int value) throws IOException {
		writeAttributeValue(attribute, Integer.toString(value));
	}

/**
* Writes out a single-valued attribute.
* Automatically adjusts parenthesis as needed to keep PDIF integrity.
*/
	public void writeAttributeValue(String attribute, boolean value) throws IOException {
		writeAttributeValue(attribute, value?"true":"false");
	}


/*
*  1       2             3            4
* (pdif (connector (kind.role  id (kind.role id)
*/
	private String lastContainerKind = null;
	private String lastContainerRole = null;
	private String lastContainerId = null;
	
/**
* Writes out a binary relation.  If several calls are made to this
* method where the first kind and role are the same then a "factored"
* form of PDIF output is used.
* Automatically adjusts parenthesis as needed to keep PDIF integrity.
*/
	public void writeRelation(
			String containerKind1, 
			String containerRole1,  
			String containerId1, 
			String containedKind2, 
			String containedRole2, 
			String containedId2
				) throws IOException {

		if (containerKind1 == null || containerKind1.length() == 0) {
			throw new IOException("attempt to write null or empty container kind in relation");
		}
		if (containerId1 == null || containerId1.length() == 0) {
			throw new IOException("attempt to write null or empty container id in relation");
		}
		if (containedKind2 == null || containerKind1.length() == 0) {
			throw new IOException("attempt to write null or empty contained kind in relation");
		}
		if (containedId2 == null || containerId1.length() == 0) {
			throw new IOException("attempt to write null or empty contained id in relation");
		}

		if (outputLevel <= 0) {
			throw new IOException("attempt to write Node to pdif stream before writing PDIF header");
		}

		if (inWhere != CONNECTORS) {
			closeToLevel(1);
			println();
			println("\t(connector");
			outputLevel = 2;

			lastContainerKind = null;
			lastContainerRole = null;
			lastContainerId = null;

			inWhere = CONNECTORS;
		}
		
	/*
	* See if the new container is same as old container
	*/
		if ( (containerRole1 != lastContainerRole) ||
				(containerKind1 != lastContainerKind) ||
				(containerId1 != lastContainerId) ) {
		
			closeToLevel(2);
			println();
			print("\t\t(");
			print(containerKind1);
			
			if (containerRole1 != null && containerRole1.length() > 0) {
				print('.');
				print(ParseEvent2.selectiveQuote(containerRole1));
			}
			print(' ');
			print(ParseEvent2.selectiveQuote(containerId1));
			
			outputLevel = 3;
		}
		
		lastContainerRole = containerRole1;
		lastContainerKind = containerKind1;
		lastContainerId = containerId1;

	/*
	* Output info for contained item
	*/
		print(" (");
		print(containedKind2);
		
		if (containedRole2 != null && containedRole2.length() > 0) {
			print('.');
			print(ParseEvent2.selectiveQuote(containedRole2));
		}
		print(' ');
		print(ParseEvent2.selectiveQuote(containedId2));
		print(')');
		convenientEOL();
	}
/**
* Writes out a binary nonrelation.  If several calls are made to this
* method where the first kind and role are the same then a "factored"
* form of PDIF output is used.
* Automatically adjusts parenthesis as needed to keep PDIF integrity.
*/
	public void writeNonRelation(
			       String containerKind1, 
			       String containerRole1,  
			       String containerId1, 
			       String containedKind2, 
			       String containedRole2, 
			       String containedId2
			       ) throws IOException 
	{
		if (containerKind1 == null || containerKind1.length() == 0) {
		  throw new IOException("attempt to write null or empty container kind in relation");
		}
		if (containerId1 == null || containerId1.length() == 0) {
		  throw new IOException("attempt to write null or empty container id in relation");
		}
		if (containedKind2 == null || containerKind1.length() == 0) {
		  throw new IOException("attempt to write null or empty contained kind in relation");
		}
		if (containedId2 == null || containerId1.length() == 0) {
		  throw new IOException("attempt to write null or empty contained id in relation");
		}

		if (outputLevel <= 0) {
		  throw new IOException("attempt to write Node to pdif stream before writing PDIF header");
		}
    
		if (inWhere != NONCONNECTORS) {
			closeToLevel(1);
			println();
			println("\t(nonconnector");
			outputLevel = 2;

			lastContainerKind = null;
			lastContainerRole = null;
			lastContainerId = null;

			inWhere = NONCONNECTORS;
		}
    
	/*
	* See if the new container is same as old container
	*/
	    if ( (containerRole1 != lastContainerRole) ||
			(containerKind1 != lastContainerKind) ||
			(containerId1 != lastContainerId) ) {
	      
			closeToLevel(2);
			println();
			print("\t\t(");
			print(containerKind1);

			if (containerRole1 != null && containerRole1.length() > 0) {
				print('.');
				print(ParseEvent2.selectiveQuote(containerRole1));
			}
			print(' ');
			print(ParseEvent2.selectiveQuote(containerId1));

			outputLevel = 3;
		}
	    
	    lastContainerRole = containerRole1;
	    lastContainerKind = containerKind1;
	    lastContainerId = containerId1;
	    
	/*
	* Output info for contained item
	*/
	    print(" (");
	    print(containedKind2);
	    
		if (containedRole2 != null && containedRole2.length() > 0) {
			print('.');
			print(ParseEvent2.selectiveQuote(containedRole2));
	    }
	    print(' ');
	    print(ParseEvent2.selectiveQuote(containedId2));
	    print(')');
	    convenientEOL();
	}
/**
* On close just balance parenthesis
*/
	public void close() {
		closeToLevel(0);
		prnt.flush();
		prnt.close();
	}
/**
* On close just balance parenthesis
*/
	public void flush() {
		closeToLevel(0);
		prnt.flush();
	}
/*
	public static void main(String args[]) {
		PDIFOutStream ps = new PDIFOutStream(new PrintWriter(System.out));
		
		try {
			ps.writeHeader("private","pam");
			ps.writeNode("kind1","id1111");
			ps.writeNode("kind2","id2222");
			ps.writeNode("kind3","id3333");
			ps.writeAttributeValue("a1","v1");
			ps.writeAttributeValue("a2","v2");
			ps.writeAttributeValue("a3","v3");
			ps.writeRelation("kind1","from","id1111", "kind2","to","id2222");
			ps.writeRelation("kind1","from","id1111", "kind3","indise","id3333");

			ps.writeNode("kind4","id4444");
			ps.writeAttributeValue("a4","v4");
			ps.writeNode("kind5","id55555");
			ps.writeNode("kind6","id66666");

			ps.writeRelation("kind9","from","id9999", "kind10","to","id1010");

			ps.close();
		}
		catch (IOException e) {
			e.printStackTrace();
		}
	}
*/
	
	private void print(String p) {
		prnt.print(p);
		lineLength += p.length();
	}

	private void print(char c) {
		prnt.print(c);
		lineLength += 1;
	}

	private void println() {
		prnt.println();
		lineLength = 0;
	}

	private void println(String p) {
		print(p);
		println();
	}

	private void println(char p) {
		print(p);
		println();
	}
	
	private void convenientEOL() {
		if (lineLength > 45) {
			println();
		
			for (int i = 0; i<=outputLevel; i++) {
				lineLength += 8;
				prnt.print('\t');
			}
		}
	}

}