/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pdif.eventParser;

import java.io.IOException;
import java.io.Reader;
import java.io.FileReader;
import java.io.FileNotFoundException;
import java.io.File;

/**
* This is a streamlined replacement for java.io.StreamTokenizer.
* This is more efficient as it directly implements our parse rules.
*
* @author Nazario Irizarry, Jr.
*/
public class PStreamTokenizer {

    private Reader reader = null;
/**
* Identifies the type of token
*/
	public int ttype;

/** 
* The stream end has been read. 
*/
    public static final int TT_EOF = -1;
/** 
* Not used here but allows subclasses to be more compatible with
* StreamTokenizer
*/
    public static final int TT_EOL = '\n';
/**
* A word token has been read. 
*/
    public static final int TT_WORD = -3;

/* A constant indicating that no token has been read, used for
* initializing ttype.  FIXME This could be made public and
* made available as the part of the API in a future release.
*/
    private static final int NOTHING = -4;
/**
* String value of the token
*/
    public String sval;
    private StringBuffer token = new StringBuffer(80);
    
    private int lastc;
    private boolean validc;
    private int prevc; // used for # processing

    public PStreamTokenizer(Reader reader) {
    	this.reader = reader;
    }
/**
* Get the next token in the stream.
*/
	public int nextToken() throws IOException {
		ttype = NOTHING;
		skipWhiteSpace();
		while (isLineComment()) {
			eatOneLineComment();
		}
		
		if (lastc == -1) {
			ttype = TT_EOF;
			sval = "";
		}
		else if (lastc == '(') {
			sval = "(";
			ttype = '(';
		}
		else if (lastc == ')') {
			sval = ")";
			ttype = ')';
		}
		else if (lastc == '\"') {
			eatQuoted();
			ttype = TT_WORD;
			sval = token.toString();
		}
		else {
			eatRegularToken();
			ttype = TT_WORD;
			sval = token.toString();
		}
		return ttype;
	}
	
	private void eatQuoted() throws IOException{
		token.setLength(0);
		readc();
		while ((lastc != -1) && (lastc != '\"')){
			if (lastc == '\\')
				readc();
			token.append((char)lastc);
//		System.out.println("  eq -"+(char)lastc+"-");
			readc();
		}
	}

	private void eatRegularToken() throws IOException {
		token.setLength(0);
		while ((lastc != -1) && !isBreak(lastc) && isAlpha(lastc)){
			token.append((char)lastc);
//		System.out.println("  et -"+(char)lastc+"-");
			readc();
		}
		if (isBreak(lastc))
			pushc();
		ttype = TT_WORD;
	}

/*
* Eat from "#" to the next break or alpha on the next line
*/
	private void eatOneLineComment() throws IOException {
		token.setLength(0);
		boolean sawEnd = false;
		while (lastc != -1) {
			if (sawEnd && (isBreak(lastc) || isAlpha(lastc))) {
				break;
			}
			sawEnd = sawEnd || (lastc == '\n');
//		System.out.println("  ec -"+(char)lastc+"-");
			readc();
		}
	}
	
	private void skipWhiteSpace() throws IOException {
		do {
			readc();
//			if (lastc != -1)
//				System.out.println("  over -"+(char)lastc+"-");
		}
		while ((lastc != -1) && isWhiteSpace(lastc));
	}
	
	private boolean isWhiteSpace(int c) {
		return !isBreak(c) && !isAlpha(c);
	}

	private boolean isBreak(int c) {
		return "()\"".indexOf(c) >= 0;
	}

	private boolean isAlpha(int c) {
		if ((c >= 0x21) && (c <= 0x7E))
			return true;
		return false;
	}
	
	private boolean isLineComment() {
		return (prevc == '\n') && (lastc == '#');
	}
	
	private void readc() throws IOException{
		if (validc) {
			validc = false;
		}
		else {
			prevc = lastc;
			lastc = reader.read();
			if (lastc == '\r')
				lastc = '\n'; // only remember newline
		}
	}
	
	private void pushc() {
		validc = true;
	}


	public static void main(String args[]) {
		try {
		Reader fr = new FileReader(new File("test.txt"));
		PStreamTokenizer pst = new PStreamTokenizer(fr);
			for (int i = 0; i< 40; i++) {
				pst.nextToken();
				if (pst.ttype == TT_EOF)
					break;
				System.out.println("ttype: "+pst.ttype+" "+(char)pst.ttype+"  token: "+pst.sval);
			}
			System.out.println(" ***** ");
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		System.out.println("done");
	}
}