/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pdif.eventParser;

import java.util.Vector;
import java.io.*;

/** 
* A parser that is to be used in a "parse event" loop.
* The following description describes an "ideal" PDIF parser.  The actual
* description for THIS pdif parser follows.
*
*<p> The ideal PDIF parser can be decomposed to four levels.  The first two levels are:
*
*<p> Level 1 is basic lexical parsing.  Whitespace is used to delimit tokens.
*    Excess whitespace is ignored.  Strings are delimited with double quotes.
*    Binary sequences (future) are delimited with %%.
*    Carriage return, line-feed, 
*    and other non-printing ASCII characters are ignored.
*
* <p> Level 2 primarily keeps track of parenthesis nesting level.
*     Based on entering and leaving nesting level one of three
*     events can be generated:  Enter, Leave, End-of-file.
*     A source may potentially contain more than one PDIF block.  The client
*     can know when it is at the end of a PDIF block by checking the depth of the
*     returned event.
*     Level 2 should
*     also ignore any parenthesized section that starts with a pdif comment "(-".  Case
*     is important in all places.
*
*
*<p> This parser is a Level 2 parser.
* It generates one of 3 parse events: EOF, ENTER, LEAVE. It explicitly looks for
* embedded HTML tags delimited by "<" and ">" and ignores them unless they are part
* of a PDIF string delimited by quotes.  If you want to parse PDIF it is recommended
* that you use ParseEvent3 since it is smarter about PDIF structure. <p>
*
* As the parser scans the following:
*<pre>
*(pdif
*    (node
*       (point 39a (latitude 45))
*       (leg eg9887)
*    )
*    (connector
*        (leg.container eg9887 (point.from 39a))
*    )
*)
*</pre>
*<p>
* It generates ENTER events at the left paren for pdif, node, point, latitude, leg, etc.  
* It generates LEAVE events at the corresponding right parens.  
* Finally it generates an EOF event at the end of the stream. 
*
*
* @version  4, 12 Dec 1997  break up parser into two parts
* @version  3, 12 Aug 1997  partial support for "nonnode" and "nonconnector"
* @version  2, 17 Sept 1996
* @author   Nazario Irizarry, Jr.
*/
public class ParseEvent2 extends PStreamTokenizer {
/**
* Working space for fancy unquote.
* This eliminates the need to do a "new Stringbuffer" every
* time we need to parse an input value
*/
	private StringBuffer funqSpace = new StringBuffer(200);

/** 
 * Indicates end of parse stream.  This parser leaves things up to the client to break
 * parse when parenthesis gets back to nesting depth 0.
 * Thus, EOF should be treated as an error since we should never reach EOF.
 */
	public static final int EOF = 1;
/** 
 * Returned as the parse event when we enter the next parenthesized item.  Not
 * generated when we first enter a "(pdif (connector" block.  
 * 
 */
	public static final int ENTER = 2;
/** 
 * Returned as the parse event when we leave a parenthesized level
 */
	public static final int LEAVE = 3;
/** 
 * The maximum depth of the PDIF tag nesting.  Determines the size of nameList
 */
	private static final int MAX_DEPTH = 7;

/** 
 * The current depth of PDIF tags corresponding to ENTER or LEAVE.  Starts at depth=1
 */
	public int depth;	
/** 
 * The first token after the parenthesis used as the kind of token.  
 * Guaranteed to be correct when return type==ENTER.  Not guaranteed when type==LEAVE
 */
	public String arg1;
/** 
 * The optional second token after the left parenthesis.
 * Guaranteed to be correct when return type==ENTER or when type==LEAVE
 */
	public String arg2;
/** 
 * The optional second, third, etc tokens after the left parenthesis.
 * Guaranteed to be correct when return type==ENTER or when type==LEAVE
 */
	public Vector multiValues;
/** 
 * The name vector.  Index 0 is not used.  
 */
	public String[] kindList = new String[MAX_DEPTH];
/** 
 * Value vector keeps track of the value part of pairs. Index 0 is not used.
 */
	private Vector valuesList[] = new Vector[MAX_DEPTH];
/** 
 * This is the same as the returned event type from the last
 * call to next().  Valid for any return type;
 */
	public int eventType;
/** 
 * This holds depth state across calls
 */
	int currentDepth;
/**
* keep track of commend level
*/
	int commentLevel=0;
/** 
 * Attaches to an input stream to parse PDIF streams.  
 * Ignores all tokens prior to first parenthesis.
 *
 * @deprecated
 */
	public ParseEvent2 (InputStream inStream) {
		super(new InputStreamReader(inStream));
		init();
	}
/** 
 * Attaches to an input stream to parse PDIF streams.  
 * Ignores all tokens prior to first parenthesis.
 *
 */
	public ParseEvent2 (Reader inStream) {
		super(inStream);
		init();
	}
	private void init() {
		try {
			prepareNext();
		}
		catch (IOException e) { }
		
		currentDepth = 0;
		depth = 0;
	}
/*
	private void init() {
		resetSyntax();
		lowerCaseMode(false);
		whitespaceChars(0,32);
		wordChars(0x21,0x7E);
		whitespaceChars('#','#');
		whitespaceChars(128,255);
		ordinaryChar('(');
		ordinaryChar(')');
		ordinaryChar('<');	// used for ignoring html tags
		ordinaryChar('>');
		quoteChar('\"');
		
		try {
			prepareNext();
		}
		catch (IOException e) { }
		
		currentDepth = 0;
		depth = 0;
	}
*/
/**
* Does some simple checking on the syntax of the input stream.
* The entire input stream is read to the end of file.
*/
	public void validatePDIF() throws IOException, BadPDIFException {
		int pCount = 0;
		while (true) {
			switch (ttype) {
				case ')':
					pCount++;
					break;
				case '(':
					pCount--;
					break;
				case TT_EOF:
					if (pCount != 0)
						throw new BadPDIFException("Unbalanced parenthesis");
					return;
			}
			prepareNext();
		}
		
	}
/** 
 * Gets the next PDIF token. 
 * @return ENTER, LEAVE, EOF, or CONNECTOR_TUPLE
 */
	public int next() throws  BadPDIFException,IOException {
		
		while ( true ) {
			switch (ttype) {
				case ')':
					prepareNext();
					if (commentLevel > 0) {
						commentLevel--;
						continue;
					}

					if (currentDepth <= 0) 
						throw (new BadPDIFException(exceptionMsg("Parenthesis mismatch, too many ')'s")));

					depth = currentDepth;
					setKindAndValue(depth);
					currentDepth--;
					
					return (eventType = LEAVE);
										
				case '(':
					prepareNext();
					if (commentLevel >0) {
						commentLevel++;
						continue;
					}
					if (ttype == TT_WORD && sval.charAt(0) == '-') {
						commentLevel = 1;
						prepareNext();
//						System.out.println("Skipping "+sval);
						continue;
					}

					currentDepth++;

					if (currentDepth >= MAX_DEPTH) 
						throw (new BadPDIFException(exceptionMsg("Nesting too deep, suspected grammar error.")));

					
					depth = currentDepth;
					getArgs(depth);

					setKindAndValue(depth);
					return (eventType = ENTER);
					
					
				case TT_EOF:
					if (currentDepth > 0)
						throw (new BadPDIFException(exceptionMsg("Premature EOF, expecting more ')'s")));
					arg1 = null;
					arg2 = null;
					return (eventType = EOF);

			/*
			* Ignore extra words in the stream
			*/
				case TT_WORD:
//					System.out.println("Skipping "+sval);
					prepareNext();
					break;

				case TT_EOL:
					prepareNext();
					break;
					
			/*
			* Ignore extra stuff when in the middle of a comment
			*/
				default:
					if (commentLevel >0) {
						prepareNext();
						continue;
					}
					throw(new BadPDIFException(exceptionMsg("Extra non-word encountered ttype="+ttype+"/"+sval+"/")));
			}
		}
	}
/**
* Set
*/
	private void setKindAndValue(int depth) {
		arg1 = kindList[depth];
		Vector v = valuesList[depth];
		if (v.size() == 0)
			arg2 = null;
		else
			arg2 = (String) v.elementAt(0);
		multiValues = v;
		kindList[depth+1] = null;
		valuesList[depth+1] = null;
		}
/** 
 * Gets kind and value (first two tokens after parenthesis).  Second token is optional. 
 */
	private void getArgs(int depth) throws BadPDIFException, IOException {

		kindList[depth] = null;
		valuesList[depth] = new Vector(2);
				
	/*
	* the first token is mandatory
	*/
		
		if (ttype != TT_WORD) {
			throw (new BadPDIFException(exceptionMsg("Expected kind as token 1")));
		}
		kindList[depth] = fancyUnquote(sval, funqSpace);
		prepareNext();
		
		if (ttype != TT_WORD && ttype != ')' && ttype != '(' && ttype != '\"') {
			throw (new BadPDIFException(exceptionMsg("expected value or lp or rp as token 2")));
		}
		
	/*
	* any other tokens are optional
	*/
		while (ttype != ')' && ttype != '(') {
			if (ttype == TT_WORD ) {
				try {
					valuesList[depth].addElement(fancyUnquote(sval, funqSpace));
				}
				catch (NumberFormatException e) {
					throw (new BadPDIFException(e.getMessage()+" "+exceptionMsg("Bad number format from fancyUnquote in getArgs")));
				}
			}
			else if ( ttype == '\"') {
				valuesList[depth].addElement(sval);
			}
			prepareNext();
		}
	}

/** 
 * Gets next token.  Skips embedded html tags
 */
	private void prepareNext() throws IOException {
		do {
			myNext();
			if (ttype == '<') {
				do {
					myNext();
				} while (ttype != '>' && (ttype != TT_EOF));
			}
		} while (ttype == '<' || ttype == '>');
//		System.out.println(" next sval="+sval+"   ttype="+ttype);

	}

	private void myNext() throws IOException {
		nextToken();
//		System.out.println(" next raw sval="+sval+"   ttype="+ttype);
	}
/**
* Make an exception message
*/
	private String exceptionMsg(String str) {
		StringBuffer sb = new StringBuffer(120);
		
		
		sb.append(str);
		sb.append("\n\t ttype= ");
		sb.append(ttype);
		sb.append("\n\t kindList=[");
		for (int i = 1; i<=depth; i++) {//i==0 intentionally ignored, not used
			String name = kindList[i];
			if (name == null)
				name = "<<null>> ";
			else {
				sb.append("\"");
				sb.append(name);
				sb.append("\" ");
			}
		}
		sb.append("]\n\t valuesList=[");
		for (int i = 1; i<valuesList.length; i++) {
			Vector nameList = valuesList[i];
			if (nameList == null)
				continue;
//			System.out.println(nameList);
			sb.append('[');
			for (int j = 0; j<nameList.size(); j++) {
				String name = (String)nameList.elementAt(j);
//			System.out.println(name);
				if (j > 0)
					sb.append(", ");
				if (name == null)
					name = "<<null>>";
				else {
					sb.append("\"");
					sb.append(name);
					sb.append("\" ");
				}
			}
			sb.append(']');
		}
		sb.append("]");
		return sb.toString();
	}
/**
* Scans the string and quotes it using an apropriate quoting scheme.  Note
* that the quoting scheme is private.  Multiple schemes may be used. 
* You must use this to
* write attribute values.  The values will be automatically unquoted
* on input from the parser.  The PDIFOutStream class automatically does this
* where appropriate. The resulting string
* will never contain "+"  (plus)
*/
	private static final String ok1 = "_.@-?#:,[]/";
	private static final String ok2 = " ~!^&*()$<>;{}";
	
	static public String selectiveQuote(String in) {
		boolean needSimple = false;
		boolean needFancy = false;
		
		if (in.length() == 0)
			return "\"\"";
			
		int index = 0;
		while (index < in.length()) {
			char c = in.charAt(index++);
			
			if ( Character.isLetterOrDigit(c) || ok1.indexOf(c) >= 0) {
			}
			else if ( ok2.indexOf(c) >= 0 ) {
				needSimple = true;
			}
			else {
				needFancy = true;
				break;
			}		
		}
		
		if (needFancy)
			return fancyQuote(in);
		
		if (needSimple)
			return '\"'+in+'\"';
		
		return in;
	}

/**
* Converts non letter and non-digit chars into their hex codes
*/
	private static StringBuffer b = new StringBuffer(200); // trying to "save" memory, having problems with gc no keeping up
	synchronized static public String fancyQuote(String in) {
				
		boolean changed = false;
		b.setLength(0);
		b.append("%%");
		for (int i = 0; i< in.length(); i++) {
			char c = in.charAt(i);

			if ( Character.isLetterOrDigit(c) || 
						ok1.indexOf(c) >= 0 ) 
			{
				b.append(c);
			}
			else {
				c2q(c,b);
				changed = true;
			}
		}

		return changed? b.toString() : in;
	}
/**
* Undoes what selectiveQuote did.
* @param in the input and presumably quoted string
* @param buffer working space to avoid needing to create working memory inside the method
*/

	static public String selectiveUnquote(String in, StringBuffer buffer) {
		if (in == null)
			return "";
		if (in.length() == 0)
			return "";

		if ( in.charAt(0) == '\"') {
			return in.substring(1,in.length()-2);
		}
		
		return fancyUnquote(in, buffer);
	}

/**
* Undoes what fancyQuote did.
* @param in the input and presumably quoted string
* @param buffer working space to avoid needing to create working memory inside the method
*/	
	static synchronized public String fancyUnquote(String in, StringBuffer buffer) {
		if (in == null)
			return "";
		if (in.length() == 0)
			return "";
			
//		System.out.println(" unquoteng "+in);
		if ( in.charAt(0) != '%') {
			return in;
		}
		
		buffer.setLength(0);
		buffer.ensureCapacity(in.length());
		
		for (int i = 2; i< in.length(); i++) {
			if (in.charAt(i) == '%') {
				i++;
				buffer.append(q2c(in.substring(i,i+2)));
				i++;
			}
			else {
				buffer.append(in.charAt(i));
			}
		}

		return buffer.toString();
	}

	static private void c2q(char c, StringBuffer b) {
		if (  c < 16) {
			b.append("%0");
			b.append(Integer.toHexString(c));
		}
		else {
			b.append('%');
			b.append(Integer.toHexString(c));
		}
		return;
	}

	
	static private char q2c(String q) {
			return (char) Integer.parseInt(q,16);
	}


}