/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.daemon.rules;

import java.util.Vector;

import jess.Context;
import jess.Funcall;
import jess.ReteException;
import jess.RU;
import jess.Userfunction;
import jess.Value;
import jess.ValueVector;

import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.interfaces.ExecToMonitor;
import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.daemon.PropertyTimer;

/**
* Jess user function to create an alarm in the named PropertyTimer daemon,
* which takes facts in the following form:
*
*    (addTimer <propertyTimer-agent-id> <timed-element-name> 
*                    <triggerDate> <list-of-elements>)
*
* and upon activation the list of elements (atoms, strings, numbers) are set
* as multivalued memento attribute value in the named timer daemon with
* the attribute name "user_alarm." concatenated with the time-element-name.
* If the timer agent is being watched (e.g. assertAgent) then this property
* will be asserted as a fact and another rule can take action on it.
*
* @author Jason Mathews
*/
class AddTimer implements Userfunction {

    protected ExecToMonitor exec;
    protected Reporter reporter;

    AddTimer(Reporter reporter, ExecToMonitor exec) {
		this.exec = exec;
		this.reporter = reporter;
    }

    public String name() { 
	return "addTimer"; 
    }

    public Value call(ValueVector vv, Context context)
    {
	// Expects the following types of ordered facts:
	// (addTimer "timer" my-timer-name timestamp foo)
	// (addTimer "timer" this-timer-name timestamp disableAgent agentId)

	// System.out.println("AddTimer size=" + vv.size() + " call=" + vv);

	if (vv.size() < 5) {
	    // System.out.println("Bad call to timer");
	    return Funcall.FALSE();
	}

	try {
	    String agentName = vv.get(1).stringValue();

	    Agent agent = exec.getActiveAgent(agentName);

	    if (agent == null || !(agent instanceof PropertyTimer))
		return Funcall.FALSE();

	    String timerName = vv.get(2).stringValue();
	    
	    PropertyTimer timer = (PropertyTimer)agent;

	    // convert seconds (used by Jess) to milliseconds (used by Java)
	    long time = (long) vv.get(3).intValue() * 1000;

	    Vector v = new Vector( vv.size() - 4 );

	    for (int i=4; i < vv.size(); i++) {
		Value value = vv.get(i);
		// Strings have to be treated differently than other
		// types since toString method appends quotes to the value.
		if (value.type() == RU.STRING)
		    v.addElement(value.stringValue());
		else
		    v.addElement(value.toString());
	    }

	    timer.addTimer(timerName, time, v);

	    return Funcall.TRUE();
	    
	} catch (ReteException re) {
	    reporter.reportException( re);
	    return Funcall.FALSE();
	}
    }
}
