/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.daemon.rules;

import java.io.InputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.File;

import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.daemon.DaemonBase;
import org.mitre.pam.interfaces.AsynchronousAgent;
import org.mitre.pam.interfaces.ScheduleAttributes;
import org.mitre.pam.interfaces.ExecToEditor;
import org.mitre.pam.interfaces.Product;

import jess.Rete;
import jess.Jesp;
import jess.ReteException;
import jess.Userpackage;

/**
* A daemon that uses the Jess system to make inferences and
* to control the PAM system
*
* @author Nazario Irizarry, Jr. 3/21/98
* @author Jason Mathews Added onCheckpoint property and getPersistentProduct method,
*                       4/29/99 PAM 2.1.62: Use boolean flag to stop run() method rather
*			than calling deprecated stop() method.
*/
public class Infer extends DaemonBase implements Runnable, AsynchronousAgent {

	public static final String SHOW_INPUT_WINDOW = "showInputWindow";
	public static final String WINDOW_TITLE = "inputWindowTitle";
	public static final String RULE_FILE = "startingRuleFile";
	public static final String ON_CHECKPOINT = "onCheckpoint";

    protected MyDisplay reteDisplay;
    protected Rete rete;
    protected Thread thisThread;
    protected AutoRun autoRunThread;
    protected SyncSemaphore triggerObject = new SyncSemaphore();
    protected ReflectAgent reflectAgent;
    protected boolean running;

    protected static final String [] packages = { "jess.StringFunctions",
                           "jess.PredFunctions",
                           "jess.MultiFunctions",
                           "jess.MiscFunctions",
                           "jess.MathFunctions",
//                           "jess.BagFunctions",
//                           "jess.reflect.ReflectFunctions",
                           "jess.view.ViewFunctions",
//                           "org.mitre.pam.daemon.rules.CopyFile"
//                           "org.mitre.pam.daemon.rules.ReflectAgent"
                         };

	public Infer() {
		setDescription("Control Executive with rules");
		setProperty(SHOW_INPUT_WINDOW, "false");
		setProperty(WINDOW_TITLE, "Rule console input");
		setProperty(AutoRun.JESS_ERRORS, "0");
		setProperty(RULE_FILE, "");
		setProperty(AutoRun.USE_AUTORUN, "true");
		setProperty(AutoRun.ERROR_THRESHOLD, "5");
		setProperty(ON_CHECKPOINT, "");
	}

/**
* Intercepts the showInputWindow property.
* If showInputWindow=true then show the reteDisplay.
* If showInputWindow=false the hide the reteDisplay but only if it has been
* created.  This is to deal with "unused" instances of this class that
* may have been created by the exec for reference purposes.
*/
	public void setProperty(String property, String value) {
		if (property.equals(SHOW_INPUT_WINDOW)) {
			if (value.equals("true")){
				if (reteDisplay == null)
					reteDisplay = new MyDisplay(getProperty(WINDOW_TITLE));
				reteDisplay.setVisible(true);
			}
			else {
				if (reteDisplay != null)
					reteDisplay.setVisible(false);
			}
		} else if (property.equals(AutoRun.JESS_ERRORS)) {
//			if (autoRunThread != null) {
//				autoRunThread.clearErrorCount();
//			}
		} else if (property.equals(AutoRun.ERROR_THRESHOLD)) {
			if (autoRunThread != null) {
				autoRunThread.setErrorThreshold(Integer.parseInt(value));
			}
		} else if (property.equals(AutoRun.USE_AUTORUN)) {
			super.setProperty(property, value);
			updateAutoRun();
			return;
		} else if (property.equals(ENABLED)) {
			super.setProperty(property, value);
			updateAutoRun();
			return;
		}

		super.setProperty(property, value);
	}

	private void updateAutoRun() {
		if (reflectAgent == null)
			return;

		boolean autoRun = getBooleanProperty(AutoRun.USE_AUTORUN, false);
		boolean enabled = getBooleanProperty(ENABLED, false);

		triggerObject.setAutorun(enabled && autoRun);
	}

	public void doDaemonTask() {
		setProperty(ScheduleAttributes.SCHEDULE_VERB, ScheduleAttributes.SCHEDULE_NONE);
	}

	public void run() {
	    log("Jess engine now starting");
		String startFileName = getProperty(RULE_FILE);
		if (startFileName != null && startFileName.length() > 0) {
			File startFile = new File(startFileName);
			if (startFile.exists()) {
				FileInputStream fis = null;
				try {
	    			fis = new FileInputStream(startFile);
					Jesp j = new Jesp(fis, rete);
				        // Argument is 'true' for prompting, false otherwise
					log("engine start from "+startFileName);
					j.parse(false);
					log("engine running");
				}
			    catch (FileNotFoundException e) {
				reportException(e);
			    }
			    catch (ReteException re)  {
			    	reportException(re);
				}
			    catch (ArrayIndexOutOfBoundsException ai) {
					log("Wrong number of args in funcall?");
					reportException(ai);
			//rete.errStream().println("Wrong number of args in funcall?");
			//ai.printStackTrace(rete.errStream());
				}
			    catch (NullPointerException npe)  {
					log("Wrong number of args in funcall?");
					reportException(npe);
			//rete.errStream().println("Wrong number of args in funcall?");
			//npe.printStackTrace(rete.errStream());
				}
			    catch (Exception e) {
					log("Unexpected Exception");
					reportException(e);
			//rete.errStream().println("Unexpected exception:");
			//e.printStackTrace(rete.errStream());
				}
				finally {
					if (fis != null) {
						try {
							fis.close();
						}
					    catch (IOException ignore) {}
					}
	    		}
			}
			else {
				reporter.error("can't access starting rule file "+startFileName);
			}
		}

	    InputStream fis = reteDisplay.sin;

	    if (fis != null) {
			Jesp j = new Jesp(fis, rete);
			while (running && exec.getState() != ExecGlobal.SHUTTING_DOWN) {
			    try {
			        // Argument is 'true' for prompting, false otherwise
			        j.parse(false);
				}
			    catch (ReteException re)  {
			        re.printStackTrace(rete.errStream());
				}
			    catch (ArrayIndexOutOfBoundsException ai) {
			        rete.errStream().println("Wrong number of args in funcall?");
			        ai.printStackTrace(rete.errStream());
				}
			    catch (NullPointerException npe)  {
			        rete.errStream().println("Wrong number of args in funcall?");
			        npe.printStackTrace(rete.errStream());
				}
			    catch (Exception e) {
			        rete.errStream().println("Unexpected exception:");
			        e.printStackTrace(rete.errStream());
				}
			}
		}
	}

/**
* Called after the exec has loaded everything when it is ready
* to start real work or when this agent is created its
* properties have been initialized.
*/
	public void enableService(){
	    /*
	     * Set running state to true
	     */
	    running = true;

	    if (thisThread != null)
		return;

		/*
		* Start the console thread
		*/
			thisThread = new Thread(this, getId());
			thisThread.setPriority(Thread.NORM_PRIORITY -1);
			thisThread.setDaemon(true);

			if (reteDisplay == null)
				reteDisplay = new MyDisplay(getProperty(WINDOW_TITLE));
			rete = new Rete(reteDisplay);

		    for (int i=0; i < packages.length; i++) {
				try {
					rete.addUserpackage((Userpackage)
					Class.forName(packages[i]).newInstance());
				}
				catch (Exception t) {
					reportException(t);
				}
			}

		    try {
				reflectAgent = new ReflectAgent((ExecToEditor)exec, reporter,
						triggerObject);
				rete.addUserpackage(reflectAgent);
			}
			catch (Exception t) {
				reportException(t);
			}

			thisThread.start();
		/*
		* Start the autorun thread
		*/
			autoRunThread = new AutoRun(triggerObject, rete, this);
			autoRunThread.setDaemon(true);
			autoRunThread.start();
		}

/**
* Called when the exec is ready to shutdown or when this agent
* is going to be deleted.
*/
	public void disableService(){
	    /*
	     * Indicate no longer running
	     */
	    running = false;
	    if (thisThread != null) {
			log("Jess engine now stopped");
			thisThread = null;
		}
	}

/**
* called on a checkpoint
*/
	public Product getPersistentProduct(){
		String command = getProperty(ON_CHECKPOINT);
		if (command == null || command.length() == 0 || rete == null)
			return null;
		try {
// log("Infer:getPersistentProduct() -> " + command);
			rete.executeCommand(command);
		}
		catch (ReteException re) {
			reportException(re);
		}

		return null;
	}

}
