/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import java.util.*;
import java.net.URL;

import org.mitre.pam.store.FileRef;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.util.Archiver;

/**
* A Juggler that archives all of the Product which it receives from its sources,
* and creates a single composite product from it by merging all of the products
* which have been retreived over several days and eliminating duplicates. <p>
*
* User Configuration Variables:
*<pre>
* productLifetime  -- the lifetime of each archived product (in days) counting
*                     from the date when the product was first added to the archive
*
* deleteExpired -- set to true or false, whether or not to delete expired products
*                  from the archive
*</pre>
* Note: This juggler takes Products from its sources for the purpose
*       of archiving into a composite product which is formed by merging
*       products while eliminating duplicates.<p>
*
*       The juggler will create a single persistent root which points to the composite
*       product which it creates.<p>
*
*       This juggler will merge newly retrieved Product with the composite product
*       pointed to by the persistent root.  This juggler restores its archived
*       product after a shutdown by overriding setPersistentProduct() and
*       getPersistentProduct().<p>
*
*       This juggler tags incoming products with its own unique date archived tag so that
*       it knows when the product was first added to the archive.  Using the date when the
*       product was first added to the archive and the product lifetime which the juggler is
*       configured with, the date when products should expire can be calculated.<p>
*
*       This juggler only pays attention to its own unique date archived tag on products.
*       All other expiration dates and tags are ignored.  This means that a product
*       can remain in the archive even if its lifetime is over, as long as the
*       juggler determines that it has not outlived its lifetime in the archive.<p>
*
* @author Jones Leung
*/
public class Archive extends BufferBase {
    // debug key to turn on verbose messages
    protected static final String DEBUG_KEY = "archive";

    // Unique expiration date tag used by this instance of this juggler.
    // Do not set.  Juggler will determine the value of this variable.
    protected String UNIQUE_TAG = null;   // value will be set by initialize()

    // Product lifetime (default = 7 days)
    protected int LIFETIME = 7;  // value will be set by initialize()

/*
* Unique channel attributes -- None.
*/
	public Archive () {
		setDescription("Archives all product from input and routes merged products to output.");
		setProperty(Archiver.LIFETIME, "7");  // default product lifetime is 7 days
		setProperty(Archiver.DELETE_EXPIRED, "true"); // default is to delete expired products

		metaStringInfo(Archiver.LIFETIME, "7", "Product lifetime in days");  // default product lifetime is 7 days
		metaBooleanInfo(Archiver.DELETE_EXPIRED, true, "Set to delete expired products"); // default is to delete expired products
	}
/**
* Initialize certain variables and values needed by juggler.
*/
	protected void initialize () {
	    UNIQUE_TAG = getUniqueTag();
	    LIFETIME = Archiver.getProductLifetime (getMemento());
	    reporter.debugMsg("UNIQUE_TAG: ",UNIQUE_TAG, null,null, DEBUG_KEY);
	}
/**
* This channel builds a composite product from the product that its sources provide
* by merging those products and eliminating duplicates.
*/
	protected Product buildProduct(Product inProduct) {
	  if (inProduct == null)
	    return gAgentsProduct;
	  
	  Product archivedProduct = gAgentsProduct;

	  Product ourProduct = inProduct.cloneDeeper();

	  initialize();

	  if (archivedProduct == null) {

	    reporter.debugMsg("Adding product to archivedProduct container after archive was null", null,null,null,DEBUG_KEY);
	    
	    if (ourProduct instanceof BasicProduct) {
	      VectorProduct interimProduct = new VectorProduct (getId(), getDescription());
	      interimProduct.addElement (ourProduct);
	      ourProduct = interimProduct;
	    }
	    archivedProduct = ourProduct;
	  } else {
	    reporter.debugMsg("Merging new products with archived products.",null,null,null, DEBUG_KEY);
	    
	    /*
	     * should never get a mergeProduct (bp, bp) call or a mergeProduct
	     * (cp, bp) call since archivedProduct is always either null or
	     * a ContainerProduct
	     */
	    archivedProduct = mergeProduct (ourProduct, archivedProduct);
	  }

	  // delete expired products, if configured to do so
	  if (deleteExpired()) {
	    deleteExpiredProducts(archivedProduct);
	  }

	  // update the modificationDate on the outer product
	  if (archivedProduct != null)
	    archivedProduct.setModificationDate(System.currentTimeMillis());
	  
	  return archivedProduct;
    }
/**
* Restore or create unique date archived tag for this juggler instance to use.<br>
* Sets UNIQUE_TAG for this instance of the juggler.<br>
* Returns UNIQUE_TAG.
*/
    protected String getUniqueTag () {
        return Archiver.getUniqueTag (getMemento());
    }
/**
* Should expired products be deleted from the archive?
*/
    protected boolean deleteExpired() {
        return Archiver.deleteExpired (getMemento());
    }

/**
* Does this product contain a root?
*/
/*
    protected boolean containsRoot (Product prod) {
        return Archiver.containsRoot (prod);
    }
*/
/**
* What is the date archived of this product?
*/
/*
    private long getDateArchived (Product p) {
        return Archiver.getDateArchived (p, UNIQUE_TAG);
    }
*/
/**
* Add my own unique date archived tag and date to a product, if it doesn't already
* have the unique date archived tag and date.  if 'retag' is true, then
* any existing date archived tag will be updated and overwritten.
*/
/*
    private void addDateArchivedTag (Product prod, boolean retag) {
        Archiver.addDateArchivedTag (prod, retag, UNIQUE_TAG);
    }
*/
/**
* Merge products and eliminate duplicates.
*/
    private Product mergeProduct (Object newProduct, Object archivedProduct) {
        return Archiver.mergeProduct (newProduct, archivedProduct, UNIQUE_TAG);
    }
/**
* Merge the new Product  with the old archive product.
*/
	private Product mergeProduct(ContainerProduct newProduct, ContainerProduct archivedProduct) {
	    return Archiver.mergeProduct (newProduct, archivedProduct, UNIQUE_TAG);
	}
/**
* Merge the new Product with the old archived product stored in archivedProduct.
*/
	private Product mergeProduct(BasicProduct newProduct, ContainerProduct archivedProduct) {
	    return Archiver.mergeProduct (newProduct, archivedProduct, UNIQUE_TAG);
	}
/**
* Delete expired products from archive before passing it along
*/
    protected void deleteExpiredProducts (Product bigProduct) {
        Archiver.deleteExpiredProducts (bigProduct, UNIQUE_TAG, LIFETIME);
    }
/**
* Identifies the persistent product for an item, this
* is the method used by the exec to find out what to save
* on system shutdown.
*/
	public Product getPersistentProduct(){
		return gAgentsProduct;
	}
/**
* Identifies the persistent product for an item, this
* is the method used by the exec to find out what to save
* on system shutdown.
*/
	public Product getProduct(){
		return gAgentsProduct;
	}
/**
* Identifies the persistent product for an item, this
* is the method used by the exec to restore product during
* system restart
*/
	public void setPersistentProduct(Product p){
		gAgentsProduct =  p;
	}
/**
* Product information -- for debugging purposes
*/
	private void printProductInfo (Product prod) {
	    Archiver.printProductInfo (prod, UNIQUE_TAG);
	}
}
