/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import java.util.Enumeration;
import java.io.StringWriter;
import java.io.PrintWriter;
import java.io.File;
import java.io.IOException;
import java.io.BufferedOutputStream;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.productVisitor.AsHTMLIndex;

/**
 * A Juggler that builds an HTML index of its input products which is
 * the same as that created in the PutProductToDirectory DIMs.
 *
 * @author Jason Mathews
 */
	

public class PassAsHTMLIndex extends NoSchedBase 
{
	protected final static String COPY_SUPPORT = "copySupportFiles";

	protected final static String SERVE_REMOTE = "serveRemotely";
	protected final static String INDEX_PAGE_NAME = "fileNameOfIndexPage";
	protected final static String DEFAULTPATH = "putProductToDirectory";

	public PassAsHTMLIndex()
	{
		setDescription("Builds an HTML index page of its products, does not forward the files");
		setProperty(SERVE_REMOTE,"true");
		setProperty(COPY_SUPPORT,"false");
		setProperty(INDEX_PAGE_NAME,"");
		metaStringInfo(INDEX_PAGE_NAME,"","file name to use for the index page");
		metaBooleanInfo(SERVE_REMOTE,true,"serve files remotely?");
		metaBooleanInfo(COPY_SUPPORT,false,"copy templates required for index page");
	}

	/**
	 * Copy supporting files (e.g. buttons, images) required by the index page
	 */
	protected void copyTemplates(VectorProduct container, File targetDir) throws IOException
	{
		File sourceDir = ExecGlobal.getSupportDir(DEFAULTPATH);
		/*
		 * Copy any "template" or general support files first
		 */
		if(sourceDir.exists() && sourceDir.isDirectory())
		{
			String fileList[] = sourceDir.list();
			for(int i = 0;i < fileList.length;i++)
			{
				File source = new File(sourceDir,fileList[i]);
				File target = new File(targetDir,fileList[i]);
				FileRef.copyFile(source,target);
				BasicProduct bp = new BasicProduct(fileList[i],"index support file");
				bp.setContent(new FileRef(target,target.lastModified(),fileList[i]));
				container.addElement(bp);
			}
		}
	}

	//copytemplates()
	protected Product buildProduct(Product jugglersProduct)
	{
		if(jugglersProduct == null)
			return null;
		File path = getWorkingDir();
		boolean copySupport = getBooleanProperty(COPY_SUPPORT,false);
		AsHTMLIndex apt = new AsHTMLIndex(getBooleanProperty(SERVE_REMOTE,true),path,copySupport ? null : ExecGlobal.getPeerProtocolHostAndPort() + ExecGlobal.getSupportDirURLRoot() + "/" + DEFAULTPATH,reporter);
		apt.setBackground("ringbinder.jpg");
		
		File file = null;
		String indexName = getStringProperty(INDEX_PAGE_NAME,"");
		if (indexName.length() > 0) {
			file = new File(path,indexName);
		}
		else {
			file = new File(path,FileRef.getNewFileId("index","html"));
		}
		
		BufferedOutputStream os = null;
		try
		{
			os = new BufferedOutputStream(new java.io.FileOutputStream(file));
			apt.visit(jugglersProduct);
			apt.writeResults(new java.io.PrintStream(os));
			os.close();
			os = null;
			BasicProduct fileProduct = new BasicProduct(getId(),getDescription());
			fileProduct.setContent(new FileRef(file,jugglersProduct.getModificationDate().getTime(),"index.html"));
			/* If we copy support files then they are added to the returned product
			 * as supporting files otherwise the index page will refer to the
			 * support files as a fully qualified URL to the iPAM support directory.
			 */
			if(copySupport)
			{
				VectorProduct prod = new VectorProduct(getId(),getDescription());
				prod.addRootElement(fileProduct);
				copyTemplates(prod,path);
				return prod;
			}
			else
			{
				// Stamp element with a PROPERTY_ROOT attribute
				fileProduct.setProperty(Product.PROPERTY_ROOT,Product.PROPERTY_ROOT);
				return fileProduct;
			}
		}
		catch(Exception e)
		{
			reportException(e);
			return null;
		}
		finally
		{
			if(os != null)
				try
				{
					os.close();
				}
				catch(IOException io)
				{
				}
		}
	}
}

