/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import java.util.Enumeration;
import java.net.URL;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.interfaces.ScheduleAttributes;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pdif.eventParser.ValueList;

/**
* A Juggler that builds an aggregate product and passes only those subproducts
* with string or URL content that match any (or none) of the given keywords.
* If the match criteria is "pass" then a subproduct is passed along if any of the
* keywords matches and if the match is "reject" then a subproduct is rejected if
* any of the keywords match.
*
* Non-string and URL subproducts (e.g. files and byte arrays) are passed blindly without
* any keyword checking.
*
* @author Jason Mathews
*/
public class PassSubIfKeyMatch extends NoSchedBase {

  private static final String KEYS = "wordList";
  private static final String PASS_OR_REJECT = "passOrReject";
  private static final String[] PASS_OR_REJECT_VALUES = {"pass", "reject"};
  private static final String DEBUG_KEY = "PassSubIfKeyMatch";

  public PassSubIfKeyMatch () {
    setDescription("Pass/rejct only those products/subproducts that match any of the keywords");
//    setProperty(ScheduleAttributes.SCHEDULE_VERB, ScheduleAttributes.SCHEDULE_NONE);
    String[] initWords = {"word1","word2"};
    ValueList vl = new ValueList(initWords);
    setPropertyValues(KEYS, vl);
    setProperty(PASS_OR_REJECT, "pass");
    metaEnumerationInfo(PASS_OR_REJECT, "pass","If criteria match then pass or reject the product?", PASS_OR_REJECT_VALUES);
    metaMultiStringInfo(KEYS, vl, "list of words for filter criteria");
  }

   class FilterMatches implements ProductFilter {
      ValueList keys;
      boolean rejectFlag;
      Reporter reporter;
      boolean rejectWholeProduct = false;
      boolean debug = false;
		
      FilterMatches ( ValueList keys, boolean rejectFlag, Reporter reporter ) {
		this.keys = keys;
		this.rejectFlag = rejectFlag;
		this.reporter = reporter;
      }

      public boolean filter ( ContainerProduct parent, Product p ) {
         if ( p instanceof ContainerProduct ) {
            return true; // continue search
         }

         if ( p instanceof BasicProduct ) {
	      Object content = p.getContent();
	      String s;
	      // don't test non-string products
	      if (content instanceof String) {
			s = (String) content;
	      } else if (content instanceof URL) {
			s = ((URL)content).toString();
	      } else {
		  	reporter.assert (false, "PassIfKeyMatch Juggler: skipping non-supported product type: ", content.getClass(),null,null); 
		  	return true; // continue
	      }
	      
	      boolean found = false;
	      for (Enumeration e=keys.elements(); e.hasMoreElements(); ) {
			  String key = (String) e.nextElement();
			  if (s.indexOf(key) >= 0) {
	       reporter.debugMsg("match found key=" , key , " in content=" , s, DEBUG_KEY);
			      found = true;
			      break;
			  }
	      } //for

	   if (!found && debug)
	     reporter.debugMsg("match not found keys=[", keys.toString(), "] in content=", s, DEBUG_KEY);
	   
	   // remove this product from the parent
	   // if found==false && rejectFlag==false OR found==true && rejectFlag==true
	   if (found == rejectFlag) {
	     reporter.debugMsg("reject subproduct " , p.getId(), " content=" , s, DEBUG_KEY);
	     // if we reject this basicProduct and there is no parent then
	     // this must be the root product and nothing is passed along
	     // as output.
	     if (parent == null)
	       rejectWholeProduct = true;
	     else
		  	parent.removeElement(p);
	 }
	 }
	 return true; // continue search
      }
   }    // end of class   FilterMatches

  protected Product buildProduct(Product inProduct) {

      if (inProduct == null) {
	  	return null;
      }

      String sReject = getStringProperty(PASS_OR_REJECT, "pass");
      boolean reject = sReject.equalsIgnoreCase("reject") || sReject.equalsIgnoreCase("false");

      ValueList keys = getPropertyValues(KEYS);

      if (keys.size() == 0 || (keys.size() == 1 && ((String)keys.elementAt(0)).length() == 0)) {
	throw new IllegalArgumentException("Missing value for " + KEYS);
	// return reject ? inProduct : null;
      }

      Product p = inProduct.cloneDeeper();

      FilterMatches filter = new FilterMatches(keys, reject, reporter);

      filter.debug = reporter.isDebugKey(DEBUG_KEY);

      p.apply(filter);

      // check if "product" is basic and rejected then we're done
      if (filter.rejectWholeProduct) return null;

      if (p != null && p instanceof VectorProduct) {
		  // cleanup vector product removing dead branches
		  VectorProduct vp = (VectorProduct) p;
		  vp.prune();
		  if (vp.size() == 0)
		      p = null;
		  else if (vp.size() == 1)
		      p = vp.elementAt(0);
      }

      return p;
  }

}
