/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.aggregator;

import java.awt.Polygon;

import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.util.Codecs;

/**
* Determines if the product point is within a user specified geographical
* region of the world.  The region will be represented by the Java Polygon
* class.  In order to form the polygon and use it to determine if a point
* is within the world region it represents, the user must specify three
* inputs.  The first is a series of latitude and longitude pairs that form
* the boundary of the polygon.  The second is the latitude and longitude
* of the approximate center of the polygon.  Since the Polygon class is
* composed of X/Y pairs, the latitude amd longitude pairs must be
* converted to X and Y coordinates with respect to the center of the
* polygon.  Furthermore these calculations are as accurate as the
* closeness of the center to the boundary of the polygon.  As an example,
* the user might want a polygon that encompasses New England (a relatively
* small section of the world) or the desire might be for all of the world
* except for that area (a relatively large section of the world).  Thus in
* order for the conversion to be accurate the user must always specify the
* center of the smaller region of the world represented by the boundary of
* the polygon.  The third input is an inside or outside of the polygon
* indicator.  If the user wanted the region of the world that was
* specified by the other two inputs, then the indicator should be set to
* inside.  If the user really wanted all the world except for the
* specified region, then the indicator should be set to outside. <p>
*
* The point and polygon are expressed in latitude/longitude in a format of DDD.MM.SSA,
* where DDD is degrees (0 - 90 for latitude and 0 - 180 for longitude), MM is minutes
* (0 - 59), SS is seconds (0 - 59), and A for latitude is N for North, S for South,
* and for longitude A is E for East, and W for West. <p>
*
* @author Henry C. Callihan, Jr.
* @author Jason Mathews, Handle degree string values already in decimal form,
*                        Opitmized some computations,
*                        Act as filter to handle ContainerProducts.
*/
public class PointInsidePolygon extends NoSchedBase {
/**
* The latitude string to used to format the latitude.<br>
*/
	public static final String LATITUDE = "latitude";
/**
* The longitude string to used to format the longitude.<br>
*/
	public static final String LONGITUDE = "longitude";
/**
* The latitudes string to used to format the latitudes.<br>
*/
	public static final String LATITUDES = "latitudes";
/**
* The longitudes string to used to format the longitudes.<br>
*/
	public static final String LONGITUDES = "longitudes";
/**
* The center latitude string to used to format the center latitude.<br>
*/
	public static final String CENTER_LATITUDE = "centerLatitude";
/**
* The center longitude string to used to format the center longitude.<br>
*/
	public static final String CENTER_LONGITUDE = "centerLongitude";
/**
* The region string to used to format the desired region.<br>
*/
	public static final String REGION = "region";
	/*
	* Use default polygon of a small rectangle centered over New England.
	*/
	private static final String[] DEFAULT_LATS = {"40.00.00N", "40.00.00N", "32.00.00N", "32.00.00N"};
	private static final String[] DEFAULT_LONGS = {"75.00.00W", "67.00.00W", "67.00.00W", "75.00.00W"};
	private static final String INSIDE = "inside";
	private static final String OUTSIDE = "outside";
        private static final String[] REGION_CHOICE = {INSIDE, OUTSIDE};

        private static final String DEBUG_KEY = "PointInsidePolygon";

    /* geodetic constants */
    private static final double A = 0.99327733;
    private static final double B = 0.00666251;
    private static final double C = 0.00005909;
    private static final double D = 0.00000056;
    private static final double E = 3962.91422;
    private static final double twoE = 2 * E;

	public PointInsidePolygon () {
	    ValueList latitudesList = new ValueList(DEFAULT_LATS);
	    ValueList longitudesList = new ValueList(DEFAULT_LONGS);
	    setPropertyValues(LATITUDES, latitudesList);
	    setPropertyValues(LONGITUDES, longitudesList);
	    setProperty(CENTER_LATITUDE, "36.00.00N");
	    setProperty(CENTER_LONGITUDE, "71.00.00W");
	    setProperty(REGION, INSIDE);
	    setDescription("Passes a product if it has latitude & longitude metadata inside a specified polygon");
	    metaMultiStringInfo(LATITUDES, latitudesList,"The latitudes of the polygon points");
	    metaMultiStringInfo(LONGITUDES, longitudesList,"The longitudes of the polygon points");
	    metaStringInfo(CENTER_LATITUDE, "34.00.00N","The latitude of the center of the polygon");
	    metaStringInfo(CENTER_LONGITUDE, "74.00.00W","The longitude of the center of the polygon");
	    metaEnumerationInfo(REGION, INSIDE,"The region indicator of either inside or outside the polygon",REGION_CHOICE);
	}

/**
* Get the incoming product and obtain the latitude and longitude of it.
* If that position is within the user supplied polygon, then pass along
* the product.  Otherwise, return null. <p>
*
*/
  protected Product buildProduct(Product inProduct) throws IllegalArgumentException {
      /*
       * If there is no incoming product then return null.
       */
      if (inProduct == null)
	  return null;

/*
	if (inProduct instanceof BasicProduct) {
		if (pointInsideRegion(inProduct)) {
			return inProduct;
		} else
			return null;
	}
*/

      VectorProduct vp = new VectorProduct(getId(), getDescription());
      FilterProducts filter = new FilterProducts(reporter, vp);
      inProduct.apply(filter);
      if (vp.size() == 0) return null;
      else if (vp.size() == 1)
	  return vp.elementAt(0);
      else return vp;
  }

/** This method converts the latitude or longitude string into a numerical equivalent.
 * Handles string as signed decimal degrees or DDD.MM.SSA format.
 *
 * Converts lat/lon DDD.MM.SSA (degrees/min/seconds) to double precison lat/lon values
 * e.g. lat   34.10.32N  => 34.175555555555555
 *      long 118.23.35W  => -118.39305555555556
 */
    public static double convertStringToDegrees(String deg, boolean isLatitude) throws NumberFormatException {
	return Codecs.convertEarthCoordinateToDegrees(deg, isLatitude);
    }

  /**
   * This method calculates radians from degrees <p>
   */
  private static double convertDegreesToRadians(double degrees) {
    return (Math.PI * degrees) / 180.0;
  }

class FilterProducts implements ProductFilter
{
    /*
     * pointLatLong is an array with two elements.  The first
     * element is the latitude and the last one is the longitude.
     */
    double pointLatLong[] = new double[2];
    /*
     * pointxYInt is an array with two elements.  The first element is the
     * X-coordinate and the last one is the Y-coordinate.
     */
    int[] pointXYInt = new int[2];

    Polygon regionPoly;
    boolean inside;
    ContainerProduct container;
    Reporter reporter;

  /*
   * centerLatLong is an array with two elements.  The first
   * element is the latitude and the last one is the longitude.
   */
    private double[] centerLatLong = new double[2];
    double sinCenterLat, cosCenterLat;

  FilterProducts(Reporter reporter, ContainerProduct container) {
    this.reporter = reporter;
    this.container = container;

    /*
     * Get the user supplied region (inside or outside) indicator.
     */
    String region = getProperty("region");
    /*
     * Verify the user specified either "inside" or "outside".
     */
    if (region.equalsIgnoreCase("inside"))
	inside = true;
    else
	{
	    if (!region.equalsIgnoreCase("outside"))
		throw new IllegalArgumentException("region must be 'inside' or 'outside'");
	    inside = false;
	}

    /*
     * Build the polygon from the user supplied latitude and longitude points.
     */
    regionPoly = buildPolygon();
  }

  /**
   * This method changes the geodetic latitude to the confirmal
   * latitude (the longitudes are the same so no need to convert):
   * Lat = arcsin([A + B(sin GLat)(Sin GLat) + C(sin GLat)(Sin GLat)(sin GLat)(Sin GLat)
   *        + D(sin GLat)(Sin GLat)(sin GLat)(Sin GLat)(sin GLat)(Sin GLat)][sin GLat])
   *    where:
   *       GLat = geodetic latitude
   *       Lat = conformal latitude
   *       A = 0.99327733
   *       B = 0.00666251
   *       C = 0.00005909
   *       D = 0.00000056 <p>
   *
   */
  private double convertGeoLatToConLat(double geoLat) {
    double sinGeoLat, sinSquared, sinToTheFourth, sinToTheSixth;
    
    sinGeoLat = Math.sin(geoLat);
    sinSquared = sinGeoLat * sinGeoLat;
    sinToTheFourth = sinSquared * sinSquared;
    sinToTheSixth = sinToTheFourth * sinSquared;
    
    return Math.asin(((A + (B * sinSquared) +
		       (C * sinToTheFourth) + (D * sinToTheSixth)) * sinGeoLat));
  }

/**
* This method changes the conformal latitude/longitude to X/Y with respect
* to the center of the polygon:
* X = (2E(sin [Long - CLong])(cos Lat))/(1 + (sin Lat)(sin CLat) + (cos Lat)(cos CLat)(cos [Long - CLong]))
* Y = (2E((sin Lat)(cos CLat) - (cos [Long - CLong])(cos Lat)(sin CLat)))/(1 + (sin Lat)(sin CLat) + (cos Lat)(cos CLat)(cos [Long - CLong]))
*   where:
*      Lat = conformal latitude
*      Long = conformal longitude
*      CLat = center conformal latitude
*      CLong = center conformal longitude
*      E = earth's radius = 3962.91422 miles
*      X = X-Coordinate
*      Y = Y-Coordinate <p>
*
*/
  private double[] convertLatLongToXY(double[]latLong, double[] centerLatLong) {
    /*
     * Both latLong and origLatLong are arrays with two elements.  The first
     * element is the latitude and the last one is the longitude.
     */
    double denominator, cosLat,  sinLat, cosDeltaLong;
    /*
     * xY is an array with two elements.  The first element is the X-coordinate
     * and the last one is the Y-coordinate.
     */
    double[] xY = new double[2];
    cosLat = Math.cos(latLong[0]);
    sinLat = Math.sin(latLong[0]);
    cosDeltaLong = Math.cos(latLong[1] - centerLatLong[1]);
    denominator = 1 + (sinLat * sinCenterLat) + (cosLat * cosCenterLat * cosDeltaLong);

    xY[0] = (twoE * Math.sin(latLong[1] - centerLatLong[1]) * cosLat) / denominator;
    xY[1] = (twoE * ((sinLat * cosCenterLat) - (cosDeltaLong * cosLat * sinCenterLat))) / denominator;
    return xY;
  }

/**
*This method constructs a polygon from the user supplied latitudes and longitudes by first converting
*them to X/Y coordinates based on the user supplied center of the polygon. <p>
*
*/
  private Polygon buildPolygon() throws NumberFormatException {
    Polygon p = new Polygon();
    ValueList latitudesList, longitudesList;
    String center_lat, center_long, point_lat, point_long;
    double lat_degs, long_degs, lat_rads, long_rads, conformal_lat;
    double center_lat_degs, center_long_degs, center_lat_rads, center_long_rads, conformal_center_lat;
    /*
     * pointxY is an array with two elements. The first element is the
     * X-coordinate and the last one is the Y-coordinate.
     */
    double[] pointXY;

    /*
     * Get the user supplied latitude and longitude points that form the polygon
     * and get the user supplied latiude and longitude center of the polygon.
     */
    latitudesList = getPropertyValues(LATITUDES);
    longitudesList = getPropertyValues(LONGITUDES);
    center_lat = getProperty(CENTER_LATITUDE);
    center_long = getProperty(CENTER_LONGITUDE);
    /*
     * The number of polygon latitude values and polygon longitude values must
     * be the same.  Also the polygon must have at least three points.
     */
    if (latitudesList.size() != longitudesList.size())
      throw new NumberFormatException("unequal number of latitudes and longitudes");
    if (latitudesList.size() < 3)
      throw new NumberFormatException("must have at least 3 latitude/longitude pairs");
    /*
     * Convert the latitude and longitude of the center of the polygon
     * strings to radians.  Then convert that geodetic value to conformal.
     */
    center_lat_degs = convertStringToDegrees(center_lat, true);
    center_long_degs = convertStringToDegrees(center_long, false);
    center_lat_rads = convertDegreesToRadians(center_lat_degs);
    center_long_rads = convertDegreesToRadians(center_long_degs);
    conformal_center_lat = convertGeoLatToConLat(center_lat_rads);
    centerLatLong[0] = conformal_center_lat;
    centerLatLong[1] = center_long_rads;
    sinCenterLat = Math.sin(centerLatLong[0]);
    cosCenterLat = Math.cos(centerLatLong[0]);

    /*
     * For each the latitude and longitude pair of the polygon, convert.
     * those strings to radians.  Then convert that geodetic value to
     * conformal.  Finally convert the conformal latitude and longitude to X and Y.
     */
    for (int i = 0; i < latitudesList.size(); i++) {
      point_lat = latitudesList.elementAt(i);
      point_long = longitudesList.elementAt(i);
      lat_degs = convertStringToDegrees(point_lat, true);
      long_degs = convertStringToDegrees(point_long, false);
      lat_rads = convertDegreesToRadians(lat_degs);
      long_rads = convertDegreesToRadians(long_degs);
      conformal_lat = convertGeoLatToConLat(lat_rads);
      pointLatLong[0] = conformal_lat;
      pointLatLong[1] = long_rads;
      pointXY = convertLatLongToXY(pointLatLong, centerLatLong);
      /*
       * Round the X/Y point to the nearest integer.
       */
      for (int j = 0; j < 2; j++) {
	if (pointXY[j] >= 0.0)
	  pointXYInt[j] = (int)(pointXY[j] + 0.5);
	else
	  pointXYInt[j] = (int)(pointXY[j] - 0.5);
      }
      /*
       * Add the X/Y point to the polygon.
       */
      p.addPoint(pointXYInt[0], pointXYInt[1]);
    }
    return p;
  }

    private boolean pointInsideRegion(Product inProduct) {
      reporter.debugMsg("Check point: ", inProduct.getId(),null,null,DEBUG_KEY);
	/*
	 * Get the latitude and longitude of the incoming product.
	 */
	String point_lat = inProduct.getProperty(LATITUDE);
	String point_long = inProduct.getProperty(LONGITUDE);
      
	/*
	 * If there is no latitude or longitude values then return null.
	 */
	if (point_lat == null || point_long == null) {
	  reporter.debugMsg("no lat/lon in product",null,null,null,DEBUG_KEY);
	  return false;
	}

	/*
	 * Convert the incoming product latitude and longitude strings to
	 * radians. Then convert that geodetic value to conformal.
	 * Finally convert the conformal latitude and longitude to X and Y.
	 */

	double lat_degs, long_degs;
	try {
	  lat_degs = convertStringToDegrees(point_lat, true);
	  long_degs = convertStringToDegrees(point_long, false);
	} catch (NumberFormatException nfe) {
	  reporter.reportException(nfe);
	  return false;
	}

	if (reporter.isDebugKey(DEBUG_KEY)) 
	  reporter.log("point lat="+ lat_degs+ ",lon="+ long_degs);

	double lat_rads = convertDegreesToRadians(lat_degs);
	double long_rads = convertDegreesToRadians(long_degs);
	double conformal_lat = convertGeoLatToConLat(lat_rads);
	pointLatLong[0] = conformal_lat;
	pointLatLong[1] = long_rads;
	double[] pointXY = convertLatLongToXY(pointLatLong, centerLatLong);

	/*
	 * Round the X/Y point to the nearest integer.
	 */
	for (int i = 0; i < 2; i++)
	    {
		if (pointXY[i] >= 0.0)
		    pointXYInt[i] = (int)(pointXY[i] + 0.5);
		else
		    pointXYInt[i] = (int)(pointXY[i] - 0.5);
	    }
	/*
	 * If the point is within the polygon as specified by the region
	 * then forward on the incoming product.  Otherwise, forward a null product.
	 */
	boolean within = regionPoly.contains(pointXYInt[0], pointXYInt[1]);
	// if (inside && within || !inside && !within) then true
	return inside == within;
    }

    public boolean filter(ContainerProduct parent, Product productToTest)
    {
      if (productToTest instanceof BasicProduct) {
	if (pointInsideRegion(productToTest)) {
	  container.addElement(productToTest);
	  reporter.debugMsg("point IN region",null,null,null,DEBUG_KEY);
	} else
	  reporter.debugMsg("point *NOT* near region",null,null,null,DEBUG_KEY);
      }
      return true;
    }
}

/*
  public static void main(String[] args) {
// usage: java org.mitre.pam.aggregator.PointInsidePolygon
    String[] values = {
			"1 -34",
			"1 34.00.30S",
			"1 34.30.00S",
			"1 34.30.30N",
			"1 118",		// invalid lat value > 90
			"1 -91",		// invalid lat value < -90
			"1 78.30.30S",		// invalid latitude sign
			"0 -170",
			"0 -170.125",
			"0 +90",
			"0 178",
			"0 178.30.00E",
			"0 179.00.30E",
			"0 178.30.30E",
			"0 118.30.00W",
			"0 270",		// invalid lat value > 180
			"0 270.125",		// invalid lat value > 180
    };
    double d;
    for (int i=0; i < values.length; i++) {
      String s= values[i];
	System.out.print("value="+s);
    for (int sp=0; sp < 20-s.length(); sp++) System.out.print(' ');
      try {
	d = convertStringToDegrees(s.substring(2), s.charAt(0)=='1');
	System.out.println(d);
      } catch (Exception e) {
	System.out.println(e);
      }
    }//for
  }
*/

}
