/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter;

import org.mitre.pam.util.smtpIn.MailDecomposer;
import org.mitre.pam.util.smtpIn.SMTPevent;
import org.mitre.pam.util.smtpIn.SMTPlistener;
import org.mitre.pam.util.smtpIn.PAMSMTPServer;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Date;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.getter.PIMShell1;
import org.mitre.pam.interfaces.AsynchronousAgent;
import org.mitre.pam.interfaces.PIM;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.interfaces.BasicProductAttributes;
import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.store.FileRef;
		
		
/**
 * This class is a PIM that will listen for incoming electronic mail messages
 * by creating an SMTP server.
 *
 * When a new piece of mail is received by the PIM it is stored in a queue
 * until the juggler fires.  When the juggler fires the contents of the queue
 * become the PIM's product and the queue is cleared.  This PIM receives mail
 * asynchronously but there is a delay between the time the PIM receives the
 * mail and the time when the message propogates through PAM.
 *
 * This PIM has five configurable parameters (defaults in parens):
 *   keepNameOfPoster - true/false (true)
 *   mailingList      - a list of email addresses that this PIM should listen for ("")
 *                      Note that it is not necessary to provide a host name (i.e. @host).
 *                      If no host is provided a default will be used (the host name provided
 *                      in the checkpoint file).
 *                      See the buildMailingList function for a more complete
 *                      understanding of how this parameter is parsed.
 *   useAttachments   - true/false (true)
 *   sourceSpec       - not used because there is no source for this PIM
 *
 * This is a sample configuration to help new users of this PIM set it up:
 *   (keepNameOfPoster true)
 *   (mailingList user1, user2, user3)
 *   (useAttachments true)
 *
 * All GetAsyncMail PIM's will share one SMTPserver.
 *
 * If attachments have an extension in the filename then that extension will be
 * used on the local file if it is valid, but if there is no extension then we
 * don't try to add one.
 *
 * @author John R. Fantry
 * @author Jason Mathews, Added synchronization and persistence
 *                        Added option to push incoming mail to all clients where the default action
 *			  is to queue all mail and forward mail products only when doFetch() is called.
 */
public class GetAsyncMail extends PIMShell1 implements AsynchronousAgent, SMTPlistener
{
	protected final static String KEEP_POSTER = "keepNameOfPoster";
	protected final static String MAILING_LIST = "mailingList";
	protected final static String USE_ATTACHMENTS = "useAttachments";
	protected final static String QUEUE_ALL_MAIL = "queueAllMail";
	protected final static String PUSH_UPON_RECEIPT = "pushMailUponReceipt";
	private final static String DEBUG_KEY = "GetAsyncMail";

	/**
	 * If this flag is true then the PIM will queue all new mail, however the
	 * messages will not become part of the PIM's product unless this flag becomes
	 * false.
	 */
	protected boolean paused = false;


	// All GetAsyncMail PIM's share one SMTPserver
	protected PAMSMTPServer server = null;

	// This queue holds all new messages until doFetch is called
	VectorProduct productQueue = null;

	/**
	 * Creates a new <tt>GetAsyncMail</tt> object.
	 */
	public GetAsyncMail()
	{
		setDescription("Accept SMTP mail");
		setProperty(KEEP_POSTER,"true");
		setProperty(PUSH_UPON_RECEIPT,"false");
		setProperty(MAILING_LIST,"");
		setProperty(USE_ATTACHMENTS,"true");
		metaStringInfo(MAILING_LIST,"","a list of user names to accept mail for");
		metaBooleanInfo(USE_ATTACHMENTS,true,"gets mail and attachments");
		metaBooleanInfo(KEEP_POSTER,true,"append name of mail sender to mail id");
		metaBooleanInfo(PUSH_UPON_RECEIPT,false,"Push all incoming mail to clients upon receipt");
	}

	private void resubscribe()
	{
		if(server == null)
			// called from setProperty before startService()
			return;
		try
		{
			server.removeSMTPeventListener(this);
			server.addSMTPeventListener(this, buildMailingList());
		}
		catch(MalformedURLException mue)
		{
			reportException(mue);
		}
	}

	/**
	 * Catch a change in the received mail list
	 */
	public void setProperty(String property, String value)
	{
		super.setProperty(property,value);
		if(property.equals(MAILING_LIST))
		{
			resubscribe();
		}
	}

	/**
	 * Catch a change in the received mail list
	 */
	public void setPropertyValues(String property, ValueList values)
	{
		super.setPropertyValues(property,values);
		if(property.equals(MAILING_LIST))
		{
			resubscribe();
		}
	}

	/**
	 * This is where the PIM will register with the registry and will start the
	 * SMTPserver if it is not already running.
	 */
	public void enableService()
	{
		try
		{
			if(server == null)
			{
				server = PAMSMTPServer.getServer(exec);
				server.addSMTPeventListener(this, buildMailingList());
			}
		}
		catch(MalformedURLException mue)
		{
			reportException(mue);
		}
		catch(IOException ioe)
		{
			reportException(ioe);
		}
	}

	/**
	 * This is where the SMTPserver will be shut down if it is still running.
	 */
	public void disableService()
	{
		server.removeSMTPeventListener(this);
		server = null;
	}

	/**
	 * This function will set the paused flag to true.  As a result the PIM will
	 * continue to queue all new mail, however the PIM's product will remain null
	 * until the paused flag becomes false.
	 */
	public void pauseService()
	{
		paused = true;
	}

	/**
	 * This function will set the paused flag to false.  As a result mail messages
	 * will be propogated through the system again.
	 */
	public void resumeService()
	{
		paused = false;
	}

	/**
	 * If there is new mail in the queue and the service is not paused, then it will
	 * become the PIM's product, otherwise the PIM's product becomes null.
	 * This is synchronized so that we can't also
	 * try to process mail at the same time (SMTPMailEvent())
	 */
	public synchronized Product doFetch()
	{
		if(!paused && productQueue != null) {
			if(productQueue.size() > 0)
			{
				VectorProduct hold = productQueue;
				productQueue = null;
				if(hold.size() == 1)
					return (Product) hold.elementAt(0);
				else
					return hold;
			}
		}
		// if we get here then we don't have anything to release
		return null;
	}

	/**
	 * Returns true if string is a valid filename extension
	 * which in perl would match this reg. expression: /^(\w+)$/
	 */
	private static boolean isValidFileExtension(String ext)
	{
		int len = ext.length();
		if(len == 0)
			return false;
		for(int i = 0;i < len;i++)
		{
			char c = ext.charAt(i);
			if(!Character.isLetterOrDigit(c) && c != '_')
				return false;
		}
		return true;
	}

	/**
	 * This is the entry point for new mail messages.
	 * This is synchronized so that we can't be processing mail
	 * and also service a doFetch() at the same time.
	 */
	public synchronized void SMTPmailEvent(SMTPevent event)
	{
		// Don't let the garbage collector clean up and wipe out the mail files
		// before they are part of our product -> set working status to true so PIM is
		// not reported as IDLE.
		declareStart(ExecGlobal.WORKING);
		boolean keepPoster = getProperty(KEEP_POSTER).equalsIgnoreCase("true");
		boolean useAttachments = getProperty(USE_ATTACHMENTS).equalsIgnoreCase("true");
		try
		{
			Product mailProduct = buildMailProduct(reporter, event, useAttachments, keepPoster, getWorkingDir());
			if (mailProduct != null) {
				if (paused || !getBooleanProperty(PUSH_UPON_RECEIPT,false)) {
					if (productQueue == null)
						productQueue = new VectorProduct(getId(), getDescription());
					productQueue.addElement(mailProduct);
				}
				else {
					if (productQueue != null && productQueue.size() != 0) {
						productQueue.addElement(mailProduct);
						mailProduct = productQueue;
						productQueue = null;
					}
					pushProductToAllClients(mailProduct);
				}
			}
		// end of synchronization
		}
		catch(FileNotFoundException fnfe)
		{
			reportException(fnfe);
		}
		catch(IOException ioe)
		{
			reportException(ioe);
		}
		finally
		{
			declareEnd(false);
		}
	}

	/**
	 * This function will construct a mailing list for this PIM that can be used
	 * to register with the registry.
	 *
	 * @return  A vector of email addresses.
	 *
	 * @exception MalformedURLException  Thrown if there is an error creating
	 *                                   a URL object.
	 */
	protected Vector buildMailingList() throws MalformedURLException
	{
		Vector mailingList = new Vector();
		ValueList vl = getPropertyValues(MAILING_LIST);
		if(vl == null)
		{
		}
		else
			if(vl.size() == 0)
			{
			}
			else
				if(vl.size() == 1)
				{
					URL url = new URL(ExecGlobal.getProtocolHostNoPort());
					StringTokenizer recipiantList = new StringTokenizer(getProperty(MAILING_LIST),", ");
					while(recipiantList.hasMoreElements())
					{
						String recipiant = trim((String)recipiantList.nextElement());
						/**
						 * If the recipiant includes a host name then that part will be stripped off
						 */
						int ind = recipiant.indexOf('@');
						if(ind > -1)
						{
							mailingList.addElement(recipiant.substring(0,ind));
						}
						else
						{
							mailingList.addElement(recipiant);
						}
					}
					if(mailingList.size() > 1)
					{
						setPropertyValues(MAILING_LIST,new ValueList(mailingList));
					}
				}
				// otherwise we have multiple recipiants
				else
				{
					URL url = new URL(ExecGlobal.getProtocolHostNoPort());
					for(int i = 0;i < vl.size();i++)
					{
						String recipiant = trim(vl.elementAt(i));
						/**
						 * If the recipiant includes a host name then that part will be stripped off
						 */
						int ind = recipiant.indexOf('@');
						if(ind > -1)
						{
							mailingList.addElement(recipiant.substring(0,ind));
						}
						else
						{
							mailingList.addElement(recipiant);
						}
					}
				}
		return mailingList;
	}

	/**
	 * called on a checkpoint
	 */
	public Product getPersistentProduct()
	{
		return productQueue;
	}

	/**
	 * Identifies the persistent product for an item, this
	 * is the method used by the exec to restore product during
	 * system restart
	 */
	public void setPersistentProduct(Product p)
	{
		if(p instanceof VectorProduct)
		{
			productQueue = (VectorProduct)p;
			reporter.debugMsg("Restoring mail queue holding ",Integer.toString(productQueue.size())," msgs",null,DEBUG_KEY);
		}
	}

	/**
	 * This function mimicks the behavior of the trim function contained in the
	 * java.lang.String class except that it seems to work better.  I found that
	 * the original trim function was returning strings that still had white space
	 * at the beginnig.
	 *
	 * @return  A string where all the white space has been removed from the
	 *          begining and end of the source string.
	 *
	 * @param src  The string to trim.
	 */
	protected static String trim(String src)
	{
		if(src.length() < 1)
		{
			return src;
		}
		int start = 0;
		int stop = src.length() - 1;
		while(start < stop && Character.isWhitespace(src.charAt(start)))
		{
			start++;
		}
		while(stop > start && Character.isWhitespace(src.charAt(stop)))
		{
			stop--;
		}
		/**
		 * At the conclusion of the above while loops it is possible that start
		 * and stop might equal one another and that the character they point to
		 * might be white space.  In such an event an empty string will be returned.
		 */
		if(Character.isWhitespace(src.charAt(start)))
			return new String();
		else
			return src.substring(start,stop + 1);
	}

	/**
	 * This is a common method shared with the pop mail PIM
	 */
	public static Product buildMailProduct(Reporter reporter, MailDecomposer event, boolean useAttachments, boolean keepPoster, File workingDir) throws IOException
	{
		String subject = event.getSubject();
		String from = event.getFrom();
		String body = event.getBody();
		Date date = event.getDate();
		if(subject == null || subject.length()==0)
		{
			subject = "<Unspecified subject>";
		}
		if(body == null)
		{
			body = "";
		}
		// don't worry about from, we test below
		if(date == null)
		{
			date = new Date();
		}
		if(from == null)
		{
			from = "";
		}

		// log("Adding mail to queue: " + subject);
		// System.out.println("Adding mail to queue: " + subject);
		// Use a VectorProduct if we need to save attachments, otherwise use a BasicProduct

		if(!event.attachmentsExist() || !useAttachments)
		{
			BasicProduct message = null;
			if(keepPoster && from != null && (from.length() > 0))
				message = new BasicProduct(subject + " (from " + from + ")","");
			else
				message = new BasicProduct(subject,"");
			message.setContent(body);
			message.setModificationDate(date.getTime());
			message.setProperty(BasicProductAttributes.FROM,from);
			return message;
		} else {
			VectorProduct attMessage = null;
			if(keepPoster && from != null && (from.length() > 0))
				attMessage = new VectorProduct(subject + " (from " + from + ")",body);
			else
				attMessage = new VectorProduct(subject,body);
			attMessage.setModificationDate(date.getTime());
			attMessage.setProperty(BasicProductAttributes.FROM,from);
			Vector files = event.getAttachmentNames();
			for(int i = 0;i < files.size();i++)
			{
				File dst = new File(workingDir, FileRef.getNewFileId());
				String MIMEtype = event.extractFileAttachment(i,dst);
				/* Lookup extension associated with mime type and
				 * rename working file (if possible). This may not be a good
				 * idea since octel-stream maps to .exe and multiple extensions
				 * exist for jpeg, etc.
				 *
				 * String ext = org.mitre.pam.util.smtp.MIMEc.findExtension(MIMEtype)
				 */
				String ext = null;
				// try to use extension from original file name
				String s = (String)files.elementAt(i);
				int pos = s.lastIndexOf('.');
				// extension position must be > 0 otherwise
				// it could be a UNIX hidden file (e.g. .cshrc)
				if(pos > 0)
				{
					s = s.substring(pos + 1);
					if(isValidFileExtension(s))
						ext = s;
				}
				if(ext != null)
				{
					// try to rename the file up to 3 times
					// retry just in case the "unique" filename already exists.
					boolean renamed = false;
					for(int retries = 0;retries < 3;retries++)
					{
						File newFile = new File(workingDir,FileRef.getNewFileId("fid",ext));
						renamed = dst.renameTo(newFile);
						if(renamed)
						{
							// log("Renamed to " + dst);
							dst = newFile;
							break;
						}
					}
					if(!renamed)
						reporter.assert(false,"unable to rename file ", dst.toString(), " to use extension ", ext);
				}
				FileRef f = new FileRef(dst,dst.lastModified(),(String)files.elementAt(i));
				BasicProduct attachment = new BasicProduct((String)files.elementAt(i),MIMEtype + " " + Long.toString(dst.length()));
				attachment.setContent(f);
				attMessage.addElement(attachment);
			}
			return attMessage;
		}
	}

}

