/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.SocketException;
import java.util.Vector;
import java.util.Date;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.util.smtpIn.maild;
import org.mitre.pam.util.smtpIn.MailDecomposer;
import org.mitre.pam.util.smtpIn.MIMEd;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.getter.PIMShell1;
import org.mitre.pam.util.popmail.RecvMail;

/**
 * This class creates a PIM for retrieving email off the internet.
 *
 * This PIM has four configurable parameters (defaults in parens):
 *   keepNameOfPoster - true/false (true)
 *   useAttachments   - true/false (true)
 *   password         - the password required to access the email account ("")
 *   sourceSpec       - the email account to retrieve mail from ("")
 *
 * This is a sample configuration to help new users of this PIM set it up:
 *   (keepNameOfPoster true)
 *   (useAttachments true)
 *   (password pw)
 *   (popAccountAndHost johndoe@mitre.org)
 *
 * @author John R. Fantry
 * @author Jones Leung, fixed null product id bug,
 *  fixed null email datestamp bug
 */
        
	
public class GetPOPMail extends PIMShell1 
{
	protected final static String PASSWORD = "password";

	protected final static String KEEP_POSTER = "keepNameOfPoster";

	protected final static String USE_ATTACHMENTS = "useAttachments";

	protected final static String POP_ACCOUNT_HOST = "popAccountAndHost";

	RecvMail mailer;

	/**
	 * Creates a new <tt>GetMail</tt> object.
	 */
	public GetPOPMail()
	{
		setDescription("Get POP Mail");
		setProperty(PASSWORD,"");
		setProperty(KEEP_POSTER,"true");
		setProperty(USE_ATTACHMENTS,"true");
		setProperty(POP_ACCOUNT_HOST,"johndoe@foobar.org");
		metaSetIsPassword(PASSWORD,"password to access the POP mail account");
		metaBooleanInfo(KEEP_POSTER,true,"append name of mail sender to mail id");
		metaBooleanInfo(USE_ATTACHMENTS,true,"gets mail and attachments");
		metaStringInfo(POP_ACCOUNT_HOST,"johndoe@foobar.org","POP account, such as: johndoe@foobar.org");
	}

	private String getPOPSpec()
	{
		return getStringProperty(POP_ACCOUNT_HOST,"");
	}

	/**
	 * Retrieves all mail messages available for a user@host pair identified in the sourceSpec.
	 */
	public Product doFetch()
	{
		VectorProduct container=null;
		boolean keepPoster = getBooleanProperty(KEEP_POSTER,true);
		boolean useAttachments = getBooleanProperty(USE_ATTACHMENTS,true);
		MailDecomposer decomposer;
		File mailMessage = new File("");
		if(mailer == null)
		{
			String url = getPOPSpec();
			String user = url.substring(0,url.indexOf('@'));
			String host = url.substring((url.indexOf('@') + 1));
			String pw = getStringProperty(PASSWORD,"");
			setProgress("creating mail receive handler, user=" + user + ", host=" + host);
			mailer = new RecvMail(host,user,pw,getWorkingDir());
		}
		boolean moreMail = false;
		setProgress("fetching mail for " + getPOPSpec());
		try
		{
			if(mailer.isMailExisting())
			{
				moreMail = true;
				container = new VectorProduct(getId(),getDescription());
				while(moreMail)
				{
					try
					{
						mailMessage = mailer.getNextPost();
					}
					catch(FileNotFoundException fnfe)
					{
						moreMail = false;
					// No more mail
					}
					if(moreMail)
					{
						if(MIMEd.isMIMEencoded(mailMessage))
							decomposer = new MIMEd(mailMessage);
						else
							decomposer = new maild(mailMessage);
						Product mailProduct = GetAsyncMail.buildMailProduct(reporter, decomposer, useAttachments, keepPoster, getWorkingDir());
						container.addElement(mailProduct);
						mailMessage.delete();
						mailer.deleteMail();
					}
				}
			}
		}
		catch(FileNotFoundException e)
		{
			reportException(e);
		}
		catch(SocketException se)
		{
			reportException(se);
		}
		catch(IOException ioe)
		{
			reportException(ioe);
		}
		mailer.closePOP();
		if (container == null)
			return null;
		if(container.size() == 1)
		{
			return (Product) container.elementAt(0);
		}
		return container;
	}

}
