/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter;
	
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.util.Random;
import java.util.StringTokenizer;
import java.util.Vector;

import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.interfaces.AsynchronousAgent;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.ExecCallbacks;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.product.view.HashtableProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.util.smtpIn.PAMSMTPServer;
import org.mitre.pam.util.smtpIn.SMTPevent;
import org.mitre.pam.util.smtpIn.SMTPlistener;
import org.mitre.pam.interfaces.ProductFactory;

/**
 * This juggler stores incoming mail messages, produces a table of contents of all
 * received mail, and allows for the retrieval of individual messages.  Messages
 * are deleted when they time out.
 *
 * Each message is assigned a unique id.  Each id begins with a one letter
 * prefix (A - attachments exist, B - body only) followed by a random four digit number.
 *
 * Below is an example of what the table of contents might look like:
 *
 * Table Of Contents
 * 1. (B5300) From: John Fantry <jfantry@mitre.org>  Subject: test2 [1M]
 * 2. (A9621) From: John Fantry <jfantry@mitre.org>  Subject: test1 [30K]
 *
 * Note that all new mail messages will appear at the top of the table.  All older messages
 * (messages that have already appeared in a previous incarnation of the table) are
 * listed below the new messages.  A single line with a hyphen character splits the new and the
 * old.  The message id appears inside parens () but the '(' and ')' characters are not part of
 * the id.
 *
 * In order to retrieve a specific mail message you must send an email message to the
 * account that the juggler has been set up for that has a body (not the subject line)
 * that consists of the following:
 *
 * get <id>
 *
 * That's it.  Just the word "get" followed by the id (not including parens or brackets)
 * of the message you want to retrieve.
 *
 * Here's a list of all supported commands:
 *   get <message id>    - retrieve message
 *   get -a <message id> - retrieve message and include attachments
 *   get -help           - send a list of commands
 *   get -list           - retrieve all messages contained in the list
 *   get -list -a        - retrieve all messages in the list with their attachments
 *   get -table          - generate a table of contents
 *
 * This Juggler has four configurable parameters (defaults in parens):
 *   includeFromField       - true/false (true).  If this field is true then the name of the person
 *                            who sent the mail message will be included in the table of
 *                            contents (see example above).  If this field is false then only
 *                            the subject of the message will be included.
 *   includeSizeField       - true/false (false).  If this field is true then the size of the mail
 *                            message is included in the table of contents (see example above).
 *                            If this field is false then no size will be included.
 *   emailIn                - the email address that this juggler should collect mail for ("").
 *                            Remember that this class was designed to receive email asynchronously.
 *   messageLifetimeInHours - an integer that represents the number of hours that email
 *                            messages should be saved (24).
 *
 * @author John R. Fantry
 * @author Jason Mathews 10/27/99 Added file size to TOC output
 *                                and modified to use StringBuffer for bulding TOC.
 */
public class TableOfContents extends PIMShell1 implements AsynchronousAgent, SMTPlistener
{
        /* memento properties */
	protected final static String LAST_UPDATE_LONG = "lastUpdateLong";
	protected final static String COMMAND_PREFIX = "commandPrefix";
	protected final static String EMAIL_IN = "emailIn";
	protected final static String INCLUDE_FROM_FIELD = "includeFromField";
	protected final static String INCLUDE_SIZE_FIELD = "includeSizeField";
	protected final static String MESSAGE_LIFETIME = "messageLifetimeInHours";
	protected final static String NUMBER_ON_QUEUE = "numberOfPendingItems";

	private final static String TOC_HEADER = "Table Of Contents\n";

	protected final static String ATTACHMENT_PARAM = "-a";

	protected final static String ATTACHMENT_PREFIX = "a";

	protected final static String BODY_ONLY_PREFIX = "b";

	protected final static String GET_COMMAND = "get";

	protected final static String HELP_PARAM = "-help";

	protected final static String ID = "id";

	protected final static String LIST_PARAM = "-list";

	protected final static String TABLE_PARAM = "-table";

	protected Random generator = new Random(System.currentTimeMillis());

	protected PAMSMTPServer server = null;

	protected VectorProduct messages = null;

	protected final static String DEBUG_KEY = "TableOfContents";
	protected ProductFactory pfab;

	public TableOfContents()
	{		
		metaStringInfo(COMMAND_PREFIX,"","prefix that must prepend commands so they can be recognized");
		metaStringInfo(EMAIL_IN,"receiveMailUserName","account name to use to receive mail, such as \"jeff\"");
		metaIntegerInfo(MESSAGE_LIFETIME,24,"how long (in hours) mail will be held");
		metaBooleanInfo(INCLUDE_FROM_FIELD,true,"Include email from field in output messages");
		metaIntegerInfo(NUMBER_ON_QUEUE,0,"If the count of messages held exceeds this amount then the TOC is send automatically");
		metaBooleanInfo(INCLUDE_SIZE_FIELD,false,"Include the size of the enclosure in the summaries");
	}

	/**
	 * This is the delivery function.  This is the one and only entry point for
	 * SMTPevent objects.
	 *
	 * This function first checks to see if the event is actaully a command (get <id>).
	 * If it is a command then the reply is generated immediately.
	 *
	 * If the event is not a command we will add it to our VectorProduct. Because the SMTPevent
	 * object is deleted by the server when this function returns we need to create a local
	 * copy for ourselves.  After we copy the event a unique id is generated, and then the
	 * FileRef and id are stored in a BasicProduct.  The BasicProduct is then added to the
	 * VectorProduct.
	 *
	 * @param event  The SMTPevent object being delivered.
	 */
	public void SMTPmailEvent(SMTPevent event)
	{
		Product p = null;
		if((p = processCommand(event)) != null)
		{
			pushProductToAllClients(p);
		}
		else
		{
			File message = event.getMailMessage();
			File myCopy = new File(getWorkingDir(),FileRef.getNewFileId());
			try
			{
				FileRef.copyFile(message,myCopy);
				FileRef content = new FileRef(myCopy,myCopy.lastModified(),myCopy.getName());
				Product bp = pfab.makeProduct(getId(),getDescription(), content);
				String prefix;
				if(event.attachmentsExist())
				{
					prefix = ATTACHMENT_PREFIX;
				}
				else
				{
					prefix = BODY_ONLY_PREFIX;
				}
				int id = 0;
				// this loop will hopefully ensure that I always get a unique 4 digit id
				do {
					id = Math.abs(generator.nextInt()) % 10000;
				} while (id < 1000 || id > 9999);
				MatchProductId matcher = new MatchProductId(prefix + id);
				messages.apply(matcher);
				bp.setProperty(ID,prefix + id);
				messages.addElement(bp);
			}
			catch(IOException ioe)
			{
				reportException(ioe);
			}
		}
	}

	public Product doFetch()
	{
		String commandPrefix = getStringProperty(COMMAND_PREFIX,"");
		String id;
		if(commandPrefix.length() > 0)
			id = commandPrefix + " " + GET_COMMAND + " " + LIST_PARAM;
		else
			id = GET_COMMAND + " " + LIST_PARAM;
		Product bp = pfab.makeProduct(id, "", buildTableOfContents());
		return bp;
	}

	/**
	 * This function is responsible for creating the table of contents.
	 *
	 * @return  The table of contents.
	 */
	protected String buildTableOfContents()
	{
		StringBuffer tableOfContents = new StringBuffer();
		tableOfContents.append(TOC_HEADER);
		String commandPrefix = getStringProperty(COMMAND_PREFIX,"");
		if(messages == null)
		{
			tableOfContents.append("\nNo Entries\n");
			tableOfContents.append(commandPrefix);
			return tableOfContents.toString();
		}
		long lastUpdateLong = Long.parseLong(getStringProperty(LAST_UPDATE_LONG,"0"));
		setProperty(LAST_UPDATE_LONG, System.currentTimeMillis());
		boolean includeFrom = getBooleanProperty(INCLUDE_FROM_FIELD,false);
		boolean includeSize = getBooleanProperty(INCLUDE_SIZE_FIELD,false);
		long messageLifetime = ((long)getIntegerProperty(MESSAGE_LIFETIME,1)) * 60 * 60 * 1000;
		boolean blankAdded = false;
		/*
		 * This value will be used to determine whether or not an old message has timed out.
		 * We need to make sure that messages aren't deleted prematurely (before they are
		 * included in the table at least one time).  This could happed if the user set the
		 * period to be 2 hours, but set the timeout to be 1 hour.
		 */
		long expiration = System.currentTimeMillis() - messageLifetime;
		if(expiration > lastUpdateLong)
		{
			expiration = lastUpdateLong;
		}
		for(int i = messages.size() - 1;i > -1;i--)
		{
			Product bp = messages.elementAt(i);
			FileRef fr = (FileRef)bp.getContent();
			File file = fr.getFile();
			setProperty(NUMBER_ON_QUEUE, messages.size());
			if(file.lastModified() < expiration)
			{
				/*
				 * Because the messages in the VectorProduct are sorted by age (older messages
				 * at the head of the vector, newer ones at the end) once we find
				 * one that has expired all the messages afterwards are also expired.  There
				 * is no need to perform individual tests on those files.
				 */
				for(;i > -1;i--)
				{
					messages.removeElement((Product)messages.elementAt(i));
				}
				// check if no entries have been added
				if(tableOfContents.length() == TOC_HEADER.length())
				{
					tableOfContents.append("\nNo Entries");
				}
				tableOfContents.append('\n');
				tableOfContents.append(commandPrefix);
				return tableOfContents.toString();
			}
			setProperty(NUMBER_ON_QUEUE, messages.size());
			try
			{
				SMTPevent message = new SMTPevent(file, new Vector());
				String id = bp.getProperty(ID);
				/*
				 * Every line will have the same general form: "#. (<id>) [From: <from>]  Subject <subject> [<size>]"
				 * The number at the beginning of every line has no meaning.  I just thought it might make
				 * the table easier to navigate.
				 */
				String line = (messages.size() - i) + ". (" + id + ")";
				if(includeFrom)
				{
					line = line + " From: " + message.getFrom();
				}
				line = line + "  Subject: " + message.getSubject();
				if(includeSize)
				{
					long size = file.length();
					String suffix = "K]";
					if (size != 0) {
						// smallest amount shown as 1K
						if (size < 1536) size = 1;
						else {
							// round up to nearest 1K; e.g. 1900 -> 2K
							size = (size + 512) / 1024;
							if (size >= 1000) {
								// round up to nearest 1M; e.g. 1500K -> 2M
								size = (size + 500) / 1000;
								suffix = "M]";
							}
						}
					}
					line = line + " [" + size + suffix;
				}
				if((!blankAdded) && (file.lastModified() < lastUpdateLong))
				{
					tableOfContents.append("- \n");
					blankAdded = true;
				}
				tableOfContents.append(line);
				tableOfContents.append('\n');
			}
			catch(IOException ioe)
			{
				reportException(ioe);
			}
		}

		// check if no entries have been added
		if(tableOfContents.length() == TOC_HEADER.length())
		{
					tableOfContents.append("\nNo Entries");
		}

		tableOfContents.append("\n");
		tableOfContents.append(commandPrefix);
		return tableOfContents.toString();
	}

	/**
	 * This function determines whether or not the event is a command.  An event is
	 * a command if the following criteria are met:
	 *   1. If a COMMAND_PREFIX is defined then the first string in the event must match
	 *      that prefix.  The second string must also match one of the predefined command
	 *      strings.  If a COMMAND_PREFIX is not defined then the first string must
	 *      match one of the predefined command strings (i.e. GET_COMMAND).
	 *   2. The next string must be a valid paramater(s) to the command identified in 1.
	 *
	 * There are some potential problems with this function.  There isn't any way to detect
	 * improperly formatted commands.  As of now if an intended command has a mistyped
	 * parameter it will not be interpreted as a command at all.  Instead it will be considered
	 * an ordinary mail message.  This is bad because the user won't know that they sent a
	 * bad command until they see the next table of contents with their command message at
	 * or near the top of the list.  Don't know what to do about this right now.
	 *
	 * If the event is a command then the jugglersProduct will be filled with a suitable reply.
	 *
	 * @return  True if the event is processed as a command, false otherwise.
	 *
	 * @param event  The SMTPevent to test.
	 */
	protected Product processCommand(SMTPevent event)
	{
		Product jugglersProduct = null;
		try
		{
			String body = event.getBody();
			BufferedReader in = new BufferedReader(new StringReader(body));
			reporter.debugMsg("body:",body,null,null, DEBUG_KEY);
			String command = "";
			String prefix = "";
			String commandPrefix = getStringProperty(COMMAND_PREFIX,"");
			String inputLine = "";
			Vector params;
				inputLine = in.readLine();
				reporter.debugMsg("first line:",inputLine,null,null, DEBUG_KEY);
			if (inputLine == null || inputLine.length() ==0 ) return null;
				StringTokenizer st = new StringTokenizer(inputLine," ");
				if(commandPrefix.length() > 0)
				{
			    if (st.hasMoreElements())
					prefix = (String)st.nextElement();
			    if (!prefix.equalsIgnoreCase(commandPrefix))
			      return null;
			  }

			if (st.hasMoreElements()) {
			  command = (String) st.nextElement();
				reporter.debugMsg(" command:",command,null,null, DEBUG_KEY);
			  if (!command.equalsIgnoreCase(GET_COMMAND))
			    return null;
			  params = new Vector();
				for (int i = 0; i<10 && st.hasMoreElements();i++)
				{
					String param = ((String)st.nextElement()).toLowerCase();
					reporter.debugMsg(" arg:",param,null,null, DEBUG_KEY);
					params.addElement(param);
				}
			} else {
				reporter.warning("Potential input line had bad format, ignoring... \""+inputLine+"\"");
				return null;
			}
					
				reporter.debugMsg("get command",null,null,null, DEBUG_KEY);
			if(params.size() == 1 && params.contains(HELP_PARAM)) {
					reporter.debugMsg("get help",null,null,null, DEBUG_KEY);
					String help = new String("get <message id>\n" + "get -a <message id>\n" + "get -help\n" + "get -list\n" + "get -list -a\n" + "get -table");
					jugglersProduct = pfab.makeProduct("Usage:","",help);
					return jugglersProduct;
				}
			else if(params.size() == 1 && params.contains(TABLE_PARAM)) {
					reporter.debugMsg("get table",null,null,null, DEBUG_KEY);
					if(commandPrefix.length() > 0) {
						jugglersProduct = pfab.makeProduct(commandPrefix + " " + GET_COMMAND + " " + LIST_PARAM,"", buildTableOfContents());
					}
					else {
						jugglersProduct = pfab.makeProduct(GET_COMMAND + " " + LIST_PARAM,"", buildTableOfContents());
					}
					return jugglersProduct;
				} else if((params.contains(LIST_PARAM) && params.size() == 1) || 
							(params.contains(LIST_PARAM) && params.contains(ATTACHMENT_PARAM) && params.size() == 2)) {

					reporter.debugMsg("get list",null,null,null, DEBUG_KEY);
					boolean includeAttachments = false;
					if(params.contains(ATTACHMENT_PARAM)) {
						includeAttachments = true;
					}
					String line = in.readLine();
					while(line != null) {
			    int openParen = line.indexOf('(');
			    int closeParen = line.indexOf(')',openParen);
						if(openParen > -1 && closeParen > openParen) {
							String id = line.substring(openParen + 1,closeParen);
							if(includeAttachments && id.startsWith(ATTACHMENT_PREFIX))
							{
								jugglersProduct = buildReplyProductWithAttachments(id);
							}
							else
							{
								jugglersProduct = buildReplyProduct(id);
							}
							pushProductToAllClients(jugglersProduct);
						}
						line = in.readLine();
					}
					jugglersProduct = null;
					return null;
				} else if(params.contains(ATTACHMENT_PARAM) && params.size() == 2) {
					reporter.debugMsg("get with attachment",null,null,null, DEBUG_KEY);
					int index = params.indexOf(ATTACHMENT_PARAM);
					if(index == 0) {
						jugglersProduct = buildReplyProductWithAttachments((String)params.elementAt(1));
					}
					else if(index == 1) {
						jugglersProduct = buildReplyProductWithAttachments((String)params.elementAt(0));
					}
					return jugglersProduct;
				} else if(params.size() == 1) {
					reporter.debugMsg("get item",null,null,null, DEBUG_KEY);
					jugglersProduct = buildReplyProduct((String)params.elementAt(0));
					return jugglersProduct;
				}
			}
		catch(IOException ioe)
		{
			reportException(ioe);
		}
		return null;
	}

	/**
	 * This function builds a BasicProduct that is suitable for reply.
	 */
	protected Product buildReplyProduct(String id)
	{
		try
		{
			MatchProductId matcher = new MatchProductId(id);
			messages.apply(matcher);
			if(matcher.product != null)
			{
				Product p = matcher.product;
				FileRef content = (FileRef)p.getContent();
				File file = content.getFile();
				SMTPevent event = new SMTPevent(file,new Vector());
				if(event.attachmentsExist())
				{
					Vector attachmentNames = event.getAttachmentNames();
					StringBuffer suffix = new StringBuffer("\n \nAttachments:");
					for(int i = 0;i < attachmentNames.size();i++)
					{
						suffix.append("\n" + attachmentNames.elementAt(i));
					}
					return pfab.makeProduct(event.getSubject(),"",event.getBody() + suffix.toString());
				}
				else
				{
					return pfab.makeProduct(event.getSubject(),"",event.getBody());
				}
			}
		}
		catch(IOException ioe)
		{
			reportException(ioe);
		}
		return null;
	}

	/**
	 * This function builds a VectorProduct that is suitable for reply.  The product will
	 * contain all attachments.
	 */
	protected VectorProduct buildReplyProductWithAttachments(String id)
	{
		try
		{
			MatchProductId matcher = new MatchProductId(id);
			messages.apply(matcher);
			if(matcher.product != null)
			{
				Product p = matcher.product;
				FileRef content = (FileRef)p.getContent();
				File file = content.getFile();
				SMTPevent event = new SMTPevent(file,new Vector());
				VectorProduct replyProduct = new VectorProduct(event.getSubject(),event.getBody());
				Vector files = event.getAttachmentNames();
				for(int i = 0;i < files.size();i++)
				{
					String fileName = (String)files.elementAt(i);
					int index = fileName.lastIndexOf('.');
					String newFileName;
					if(index > -1)
					{
						newFileName = FileRef.getNewFileId(null,fileName.substring(index + 1));
					}
					else
					{
						newFileName = FileRef.getNewFileId();
					}
					File dst = new File(getWorkingDir(),newFileName);
					String MIMEtype = event.extractFileAttachment(i,dst);
					FileRef f = new FileRef(dst,dst.lastModified(),fileName);
					Product attachment = pfab.makeProduct(fileName,MIMEtype + " " + Long.toString(dst.length()), f);
					replyProduct.addElement(attachment);
				}
				return replyProduct;
			}
		}
		catch(IOException ioe)
		{
			reportException(ioe);
		}
		return null;
	}

	/**
	 * Identifies the persistent product for an item, this
	 * is the method used by the exec to find out what to save
	 * on system shutdown.
	 */
	public Product getPersistentProduct()
	{
		return messages;
	}

	/**
	 * Identifies the persistent product for an item, this
	 * is the method used by the exec to restore product during
	 * system restart
	 */
	public void setPersistentProduct(Product p)
	{
		if(p instanceof VectorProduct)
		{
			messages = (VectorProduct)p;
		}
	}

	/**
	 * Need to know immediately when the
	 * emailIn field is changed.
	 *
	 * @param attr   The attribute to set.
	 * @param value  The new value for the attribute.
	 */
	public void setProperty(String attr, String value)
	{
		if(attr.equalsIgnoreCase(EMAIL_IN) && server != null)
		{
			Vector mailingList = new Vector();
			mailingList.addElement(value);
			server.removeSMTPeventListener(this);
			server.addSMTPeventListener(this,mailingList);
		}
		super.setProperty(attr,value);
	}

	/**
	 * Called after the exec has loaded everything when it is ready
	 * to start real work or when this agent is created its
	 * properties have been initialized.
	 *
	 * We need to register with the mail server in order to receive
	 * mail messages.
	 */
	public void startService()
	{
		ExecCallbacks exec = ExecGlobal.getExec();
		if (pfab == null)
			pfab = exec.getProductFactory();
		if(messages == null)
			messages = new VectorProduct(getId(),getDescription());
		String emailIn = getStringProperty(EMAIL_IN,"");
		Vector mailingList = new Vector();
		mailingList.addElement(emailIn);
		server = PAMSMTPServer.getServer(exec);
		server.addSMTPeventListener(this,mailingList);
	}

	/**
	 * Called when the exec is ready to shutdown or when this agent
	 * is going to be deleted.
	 */
	public void stopService()
	{
		if(server != null)
			server.removeSMTPeventListener(this);
	}

	/**
	 * Called to request that the agent stop injecting asynchronous events into the system.
	 * However, the agent does not need to stop listening and responding to
	 * sockets.  An agent should have a queue so that it can process its "stimulus"
	 * without stimulating the system during a pause. Do not assume that pauseService
	 * is called before stopService.  The exec may decide to stop and may not issue a pause.
	 */
	public void pauseService()
	{
	// not implemented because this juggler would appear "broken" to the user
	}

	/**
	 * Called when it is ok to pass stimulus to the exec or other agents.
	 */
	public void resumeService()
	{
	// not implemented
	}

	/**
	 * A ProductFilter that looks to match a specified
	 * approveId and returns/stops iteration when found.
	 *
	 * Copied from Approve.java
	 */
	class MatchProductId implements ProductFilter
	{
		String id;

		Product parent = null;

		Product product = null;

		String key = null;

		MatchProductId(String id)
		{
			this.id = id;
		}

		public boolean filter(ContainerProduct parent, Product p)
		{
			String pid = p.getProperty(TableOfContents.ID);
			if(pid == null)
				return true;
			// continue search
			if(pid.equals(id))
			{
				this.parent = parent;
				this.product = p;
				if(parent instanceof HashtableProduct)
				{
					key = ((HashtableProduct)parent).getKey(p);
				}
				return false;
			// halt search
			}
			return true;
		// continue search
		}

	}

}

