/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter.db;

import org.mitre.pam.daemon.Dict;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.ProductFactory;
import org.mitre.pam.interfaces.ContainerProduct;

import org.mitre.pam.store.FileRef;

import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.getter.PIMShell1;
import org.mitre.pam.util.AttributeKey;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Time;
import java.sql.Types;

import java.text.SimpleDateFormat;

import java.util.Date;
import java.util.Hashtable;
import java.util.Vector;

/**
* This input agent is able to perform SQL commands on JDBC compliant databases.  
* Each row is returned as a product in a container product.  Each
* field in the table becomes an attribute on a product.
*
* You must have the JDBC driver required to access the target database.  The driver classes
* must be in the classpath prior to starting iPAM or they will not be found by the class loader.
*
* This agent can to substitute values into the SQL query based on the user definable escape character.
* It may substitute in the last update time as either a date or a time by using the appropriate "date" and "time"
* keywords.  It can also specify the format that the date and time strings should take on by filling in the
* corresponding fields with strings that will be interpreted by the SimpleDateFormat class.  It is also possible
* to substitute in an agent property by identifying the appropriate property name.
*
* @author John Fantry
*/
public class GetJDBC extends PIMShell1 {
  protected static final String DB_DRIVER = "dbDriver";
  protected static final String DEFAULT_DB_DRIVER = "";

  protected static final String DB_URL = "dbURL";
  protected static final String DEFAULT_DB_URL = "";

  protected static final String QUERY = "SQLStatement";
  protected static final String DEFAULT_QUERY = "";

  protected static final String ESCAPE_CHAR = "escapeChar";
  protected static final String DEFAULT_ESCAPE_CHAR = "@";

  // not a PIM property, used internally only
  protected static final String DATE = "date";

  protected static final String DATE_FORMAT = "dateFormat";
  protected static final String DEFAULT_DATE_FORMAT = "yyyy-mm-dd";

  // not a PIM property, used internally only
  protected static final String TIME = "time";

  protected static final String TIME_FORMAT = "timeFormat";
  protected static final String DEFAULT_TIME_FORMAT = "hh:mm:ss";

  protected static final String LAST_UPDATE_TIME = "lastUpdateTime";
  protected static final String DEFAULT_LAST_UPDATE_TIME = "";


  protected Driver currentJDBCDriver = null;

  //protected VectorProduct container = null;

  public GetJDBC() {
    setDescription("Performs a SQL command");
    setProperty(QUERY, DEFAULT_QUERY);
    setProperty(ESCAPE_CHAR, DEFAULT_ESCAPE_CHAR);
    setProperty(DATE_FORMAT, DEFAULT_DATE_FORMAT);
    setProperty(TIME_FORMAT, DEFAULT_TIME_FORMAT);
    setProperty(DB_URL, DEFAULT_DB_URL);
    setProperty(DB_DRIVER, DEFAULT_DB_DRIVER);
    setProperty(LAST_UPDATE_TIME, DEFAULT_LAST_UPDATE_TIME);

    metaStringInfo(QUERY, DEFAULT_QUERY, "The SQL command to perform on the database");
    metaStringInfo(ESCAPE_CHAR, DEFAULT_ESCAPE_CHAR, "This character identifies a property substitution");
    metaStringInfo(DATE_FORMAT, DEFAULT_DATE_FORMAT, "How the date will be formatted for use in the SQL statement");
    metaStringInfo(TIME_FORMAT, DEFAULT_TIME_FORMAT, "How the time will be formatted for use in the SQL statement");
    metaStringInfo(DB_URL, DEFAULT_DB_URL, "This points to the database you want to access");
    metaStringInfo(DB_DRIVER, DEFAULT_DB_DRIVER, "The java class reference for the JDBC driver needed to access the database");
    metaStringInfo(LAST_UPDATE_TIME, DEFAULT_LAST_UPDATE_TIME, "The last time that this agent updated");
  }

  public Product doFetch() {
    ContainerProduct container = null;

    String DBDriver = getStringProperty(DB_DRIVER, "");
    String DBURL = getStringProperty(DB_URL, "");
    String SQLString = parseSQLStatement();

    if(DBDriver ==  null || DBURL == null || SQLString == null) {
      reportException(new IllegalArgumentException("PIM properties not defined properly"));
      return null;
    }

	ProductFactory fab = exec.getProductFactory();

    try {
      // try to load the driver class
      // JDBC drivers are supposed to register themselves with the DriverManager during construction!!
      Driver JDBCDriver = (Driver)Class.forName(DBDriver).newInstance();

      // take steps to make sure that the same driver is not registered twice
      if(currentJDBCDriver == null || JDBCDriver != currentJDBCDriver) {
        if(currentJDBCDriver != null) {
          DriverManager.deregisterDriver(currentJDBCDriver);
        }

        currentJDBCDriver = JDBCDriver;
      }

      Connection DBConnection = DriverManager.getConnection(DBURL);
      Statement SQLStatement = DBConnection.createStatement();

      if(SQLStatement.execute(SQLString)) {
        container = fab.makeVectorProduct(getId(), getDescription());

        // loop through every table that was generated by the SQL statement
        do {
          try {
            container.addElement(processResultSet(SQLStatement.getResultSet(), fab));
          } catch(SQLException SQLe) {
            reportException(SQLe);
          }
        }
        while(!(!SQLStatement.getMoreResults() && SQLStatement.getUpdateCount() == -1));
      }

      SQLStatement.close();
      DBConnection.close();

    } catch(ClassNotFoundException cnfe) {
      reportException(cnfe);
    } catch(InstantiationException ie) {
      reportException(ie);
    } catch(IllegalAccessException iae) {
      reportException(iae);
    } catch(SQLException SQLe) {
      reportException(SQLe);
    }

    setProperty(LAST_UPDATE_TIME, Long.toString(new Date().getTime()));

    if(container.size() == 1) {
      return container.elementAt(0);
    }

    return container;
  }

  /**
  * This function traverses through a table building one product per row
  * and adding them to a vector product.
  */
  protected Product processResultSet(ResultSet results, ProductFactory fab) throws SQLException {
    ContainerProduct vp = null;
    ResultSetMetaData metaData = results.getMetaData();
    int columnCount = metaData.getColumnCount();
    String tableName = metaData.getTableName(1);

	vp = fab.makeVectorProduct(getId(), (tableName.length() == 0)? "unnamed table" : tableName);
	
    while(results.next()) {
      Product dp = makeNewProduct(getId(), getDescription(), fab);

      for(int i=1; i<=columnCount; i++) {
        addStandardColumn(i, metaData.getColumnName(i), metaData.getColumnType(i), results, dp);
      }

      //container.addElement(dp);
      
      vp.addElement(dp);
    }

    if(vp.size() == 1) {
      return vp.elementAt(0);
    }

    return vp;
  }

/**
* Make a new empty product.  This is implemented
* as a factory method so subclasses can fill in defaults
* or other appropriate product types
*/
	protected Product makeNewProduct(String id, String description, ProductFactory fab) {
      return fab.makeProduct(getId(), getDescription(), "");
	}
	
  /**
  * This function reads in the "query" property and parses all fields that are preceeded by the
  * escape character.
  */
  protected String parseSQLStatement() {
    StringBuffer parsedSQLStatement = new StringBuffer();
    String SQLStatement = getStringProperty(QUERY, "");
    String escapeChar = getStringProperty(ESCAPE_CHAR, "");

    if(escapeChar.equals("") || SQLStatement.equals("")) {
      return SQLStatement;
    }

    int indexOfEChar = SQLStatement.indexOf(escapeChar);
    int currentParseIndex = 0;
    while(indexOfEChar != -1) {
      parsedSQLStatement.append(SQLStatement.substring(currentParseIndex, indexOfEChar));

      String tag;
      int endTagPos = SQLStatement.indexOf(" ", indexOfEChar);
      currentParseIndex = endTagPos;
      if(endTagPos != -1) {
        tag = SQLStatement.substring(indexOfEChar + escapeChar.length(), endTagPos);
      }
      else {
        tag = SQLStatement.substring(indexOfEChar + escapeChar.length());
      }

      if(tag.equals(DATE)) {
        String dateFormat = getStringProperty(DATE_FORMAT, "");
        SimpleDateFormat formatter = new SimpleDateFormat(dateFormat);
        String dateString = formatter.format(new Date(Long.parseLong(getStringProperty(LAST_UPDATE_TIME, ""))));
        parsedSQLStatement.append(dateString + " ");
      }
      else if(tag.equals(TIME)) {
        String timeFormat = getStringProperty(TIME_FORMAT, "");
        SimpleDateFormat formatter = new SimpleDateFormat(timeFormat);
        String timeString = formatter.format(new Date(Long.parseLong(getStringProperty(LAST_UPDATE_TIME, ""))));
        parsedSQLStatement.append(timeString + " ");
      }
      else {
        String property = getStringProperty(tag, "");
        parsedSQLStatement.append(property);
      }

      indexOfEChar = SQLStatement.indexOf(escapeChar, indexOfEChar + 1);
    }

    if(currentParseIndex != -1) {
      parsedSQLStatement.append(SQLStatement.substring(currentParseIndex));
    }

    String SQL = parsedSQLStatement.toString();
    if(SQL.length() == 0) {
      SQL = SQLStatement;
    }

    return SQL.trim();
  }

  /**
  * This function attempts to convert the column SQL type to an equivalent type that is
  * supported by a dictionary product.
  */
  protected void addStandardColumn(int columnNum, String columnName, int columnType,
                                   ResultSet results, Product dp) throws SQLException {
    switch(columnType) {
      case Types.BIGINT : {
        dp.setProperty(columnName, results.getLong(columnNum));
        break;
      }
      case Types.BINARY : {
        warning("SQL type BINARY is unsupported");
        break;
      }
      case Types.BIT : {
        dp.setProperty(columnName, results.getBoolean(columnNum));
        break;
      }
      case Types.CHAR: {
        dp.setProperty(columnName, results.getString(columnNum));
        break;
      }
      case Types.DATE : {
        dp.setProperty(columnName, results.getDate(columnNum).getTime());
        break;
      }
      case Types.DECIMAL : {
        dp.setProperty(columnName, results.getDouble(columnNum));
        break;
      }
      case Types.DOUBLE : {
        dp.setProperty(columnName, results.getDouble(columnNum));
        break;
      }
      case Types.FLOAT : {
        dp.setProperty(columnName, results.getDouble(columnNum));
        break;
      }
      case Types.INTEGER: {
        dp.setProperty(columnName, results.getInt(columnNum));
        break;
      }
      case Types.LONGVARBINARY : {
        warning("SQL type LONGVARBINARY is unsupported");
        break;
      }
      case Types.LONGVARCHAR : {
        dp.setProperty(columnName, results.getString(columnNum));
        break;
      }
      case Types.NULL : {
        warning("SQL type NULL is ignored");
        break;
      }
      case Types.NUMERIC : {
        dp.setProperty(columnName, results.getDouble(columnNum));
        break;
      }
      case Types.OTHER : {
        warning("SQL type OTHER is unsupported");
        break;
      }
      case Types.REAL : {
        dp.setProperty(columnName, results.getDouble(columnNum));
        break;
      }
      case Types.SMALLINT : {
        dp.setProperty(columnName, results.getInt(columnNum));
        break;
      }
      case Types.TIME : {
        dp.setProperty(columnName, (long) results.getTime(columnNum).getTime());
        break;
      }
      case Types.TIMESTAMP : {
        dp.setProperty(columnName, (long) results.getTimestamp(columnNum).getTime());
        break;
      }
      case Types.TINYINT : {
        dp.setProperty(columnName, results.getInt(columnNum));
        break;
      }
      case Types.VARBINARY : {
        warning("SQL type VARBINARY is unsupported");
        break;
      }
      case Types.VARCHAR : {
        dp.setProperty(columnName, results.getString(columnNum));
        break;
      }
    } // end switch
  }


}