/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.getter.webcopy;

/*
  Module:       WebCopy.java

  Description:  This class copies a remote subtree to the local disk
                and returns a vector of File objects, last modified dates,
                and MIME types for the downloaded files.

  Product:      GBS MOIE Common Datacasting Architecture

  Author:       Jones Leung (D510)

  Date:         June 11, 1997

  Modified by Jason Mathews 5/18/99 [iPAM 2.2.63]
  -- Use base URL context if found in base tag
  -- Replace base tag with an empty comment

  Things to do:
  -- Frames ... (currently, frames are not downloaded)

  -- JavaScript and VBScript downloads ... (Currently, JavaScript and VBScript code
     are not downloaded)

  -- Applet code downloads ... (Currently, applet classes are not downloaded)

  -- Redirection responses ... (WebCopy does not handle redirection responses at this time.)

     Fix:
     By using the HttpURLConnection subclass of URLConnection provided with JDK1.1, you
     can choose to follow redirection responses automatically.  This HttpURLConnection
     class is not included in JDK1.0.2.  (this code was built on JDK 1.0.2)

  -- Server errors ... (Currently, WebCopy is not aware of server errors and will return the
     error message from the webserver as product.)

     Fix:
     By using the HttpURLConnection subclass of URLConnection provided with JDK1.1, you
     can determine the responseCode returned by the remote webserver easily, and can then
     retry or perform some other appropriate action.  This HttpURLConnection class
     is not included in JDK1.0.2.

  -- Multithreading ... (Currently, WebCopy only copies one link at a time)

  -- File extensions for downloaded files sometimes incorrect ...
     Sometimes, extensions on the files may be incorrect.  If the remote
     file has an extension, then that extension will be used on the
     local file, but if there is no extension on the remote file, we
     don't try to add one.  In order to make sure we get the correct
     extension on the file, we could make a separate network connection
     to determine the content type of each file referenced by a parsed
     link, but this would be very slow.  Netscape doesn't seem to care
     about the extensions on the files.<p>

  -- Sometimes the same file will be downloaded twice.
     This is because http://spider.osfl.disa.mil/dii/ and
     http://spider.osfl.disa.mil/dii/index.html are actually pointing at the
     same file, but when we compare these URLs using the equals() function
     they are not equal.  There is no way to tell that these URLs are both
     pointing to the same file, so the file gets copied twice.

  -- Sometimes we accidentally copy responses from CGI scripts instead of
     making the link to the CGI script absolute (i.e. point to remote host).
     This is because we automatically do not copy URLs containing "?",
     and we make these URLs containing "?" absolute, but some CGI scripts
     do not have querystrings, and so the responses from these CGI scripts
     without querystrings accidentally get copied.

*/

import java.util.*;
import java.io.*;
import java.net.*;

import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.getter.search.parse.*;

/**
 * This class copies a remote subtree.
 *
 * @author Jones Leung, 6/97
 * @author Nazario Irizarry, many modifications 6/24/97
 * @author Mike DeCorpo, removed HtmlStringTokenizer 8/10/98
 *
 */
public final class WebCopy {
    private static final String DEBUG_KEY = "WebCopy";

    /** local download directory */
    private File localWorkingDir = null;

    /** maximum depth to copy */
    private int maxDepth = 1;

    private String indexFileName = null;

    /** hashtable of current set of URLs to be copied */
    private Hashtable copyingNowLinks;

    /** Hashtable of next set of URLs to be copied */
    private Hashtable linksToCopyNext = new Hashtable();

    /**
    * Hashtable of all FileRefs for all URLs copied
    */
    private Hashtable copiedLinks = new Hashtable();

    /**
    * to keep track of the root of the root of the search
    */
    private LinkObject rootLink = null;
    /**
    * the host site of the root link
    */
    private String rootSite = null;

    /**
    * A reference to our reporter
    */
    private Reporter reporter;
    
    private Hashtable requestProperties = new Hashtable();

    /**
    * An accessible reference to the current input stream for use
    * by a monitor that may timeout a connection if the input stream
    * takes too long.
    */
    InputStream remoteInputStream = null;

    /**
    * Keep track of when we are at final copy depth
    */
    boolean atFinalCopyDepth = false;

    public WebCopy(Reporter reporter) {
    	this.reporter = reporter;
    }
/**
* add request properties to the URL queries
*/
	public void addRequestProperty(String property, String value) {
		requestProperties.put(property, value);
	}
/**
* identifies types that are always remote
*/
    private boolean alwaysRemote(String link){
	String name = link.toLowerCase();
	return (name.startsWith ("gopher:") ||
		name.startsWith ("ftp:") ||
		name.startsWith ("file:") ||
		name.startsWith ("telnet:") ||
		name.startsWith ("news:") ||
		name.startsWith ("mailto:"));
	}
/**
* Returns a new edited HTML tag with the HREF portion edited
* so that it points to a file on the local host.
*/
    private String pointLocal (String tagstr, String link, LinkObject linkObj) {
        int index = tagstr.indexOf (link);
        String beforeLink = tagstr.substring (0, index);
        String afterLink = tagstr.substring (index + link.length());
	///		System.out.println("  tagstr before = "+tagstr);
	int anchorLoc = link.indexOf('#');
	if (anchorLoc >= 0) {
	    tagstr = beforeLink + linkObj.getLocalRelative()+link.substring(anchorLoc) + afterLink;
	}
	else
	    tagstr = beforeLink + linkObj.getLocalRelative() + afterLink;
	//		System.out.println("	tagstr after = "+tagstr);
        return (tagstr);
    }


/**
* Returns a new edited HTML tag with the HREF portion edited
* so that it points back to the remote host.
*/
    private String pointBack (String tagstr, String link, LinkObject linkObj) {
        int index = tagstr.indexOf (link);
        String beforeLink = tagstr.substring (0, index);
        String afterLink = tagstr.substring (index + link.length());
//		System.out.println("  tagstr before = "+tagstr);
        tagstr = beforeLink + linkObj.getRemoteAbsolute() + afterLink;
//		System.out.println("	tagstr after = "+tagstr);
        return (tagstr);
    }

/**
* Returns a String for the file extension of a LinkObject.
*/
    private static String getExtension (URL linkURL) {
	String file = linkURL.getFile();

	// check for form parameters and strip them off the URL
	int paramIndex = file.indexOf('?');
	if (paramIndex != -1)
		file = file.substring(0, paramIndex);

	int slashIndex = file.lastIndexOf ("/");
	if (slashIndex != -1) {
            file = file.substring (slashIndex);
	}

	int index = file.indexOf (".");
	if (index == -1) {
            return "";
	} else {
	    return file.substring (index);
	}
    }
/**
* Set the unique local directory for files to be copied into
*/
    public void setLocalDir (File localDir) {
        localWorkingDir = localDir;
        localWorkingDir.mkdirs();
    }

/**
* Set the name of the index file
*/
  public void setIndexFile(String index) {
    indexFileName = index;
  }

/**
* Set the timeout period for a link
*/
    public void setTimeout (long timeout) {
        timeout = timeout;
    }

/**
* Set the maximum depth to copy
*/
    public void setMaxDepth (int maxdepth) {
        maxDepth = maxdepth;
    }

/**
* Begin copying a remote subtree to the local disk.  The results
* of the copy are in the two attributes rootLink and copiedLinks. <p>
* Note: You should set the localWorkingDir, timeout, and maxDepth options
* before you start copying the remote subtree.
*/
    public void startCopy (URL startingURL) {
	int currentDepth = 0;
	File rootFile = null;

        try {
	  if (indexFileName == null || indexFileName.length()==0)
	    // big assumption of .html!!!!
	    indexFileName = FileRef.getNewFileId("index", "html");

            // create new LinkObject for starting URL
	    rootFile = new File(localWorkingDir, indexFileName);

            rootLink = new LinkObject ( rootFile, startingURL );
            rootLink.isRootComponent = true;

            addNextLink(rootLink);

            atFinalCopyDepth = currentDepth >= maxDepth;
            while (currentDepth < 6 && (linksToCopyNext.size() > 0)) {
				reporter.debugMsg("WebCopy currentDepth = ",
					Integer.toString(currentDepth),
					"  links to copy = ",
					Integer.toString(linksToCopyNext.size()),
					DEBUG_KEY);
                copyingNowLinks = linksToCopyNext;
                linksToCopyNext = new Hashtable();

                copyLinks (copyingNowLinks);
                atFinalCopyDepth = ++currentDepth >= maxDepth;
            }

        } catch (Exception e) {
            reporter.reportException(e);
        }

        return ;
    }
/**
* Add urls to both hashtables
*/
	private void addCopiedLink(LinkObject link) {
	    String name = link.getFullyQualified();
//        copyingNowLinks.put (name, link);
	    copiedLinks.put (name, link);
	}
/**
* Add the next link provided that we are not at final copy depth.
*/
	private void addNextLink(LinkObject link) {
	    String name = link.getFullyQualified();
	    linksToCopyNext.put (name, link);
	//        copiedLinks.put (name, link);
	}
/**
* Get an existing link object
*/
    private LinkObject getExistingLinkObject(LinkObject lo) {
	String name = lo.getFullyQualified();
	LinkObject foundObject = (LinkObject) copiedLinks.get (name);
	if (foundObject == null) {
	    foundObject = (LinkObject) copyingNowLinks.get (name);
	    if (foundObject == null) {
		foundObject = (LinkObject) linksToCopyNext.get (name);
	    }
	}
        return foundObject;
    }

/**
* Return results to client
*
* @return enumeration of FileRef
*/
    public Enumeration getResults() {
	return copiedLinks.elements();
    }
/**
* @return number of items fetched
*/
    public int getResultCount() {
	return copiedLinks.size();
    }

/**
* Gets fetchable links out of the HTML file and stores them in a
* Hashtable.
*/
    private void copyModifyLinks (URL contextURL, InputStream in, PrintStream htmlFile) {

	//exec.log("Starting my pim");
	LoggingHTMLParser parser = new LoggingHTMLParser(in,reporter);
	try {
	    Token tok = parser.getNextToken();
	    while (tok.kind != LoggingHTMLParser.EOF) {
		LinkObject newLinkObject = null;
		if (tok.kind == LoggingHTMLParser.A   ||
		    tok.kind == LoggingHTMLParser.IMG ||
		    tok.kind == LoggingHTMLParser.AREA) {

		    Vector names = new Vector();
		    int index=1;
		    Token temp = parser.getToken(index++);
		    while (temp.kind != LoggingHTMLParser.TAGC) {
			if (temp.kind == LoggingHTMLParser.A_NAME)
			    names.addElement(new String(temp.image));
			temp = parser.getToken(index++);
		    }

		    String value = null;
		    String link = null;
		    Hashtable params = null;
		    if (tok.kind == LoggingHTMLParser.A   ||
			tok.kind == LoggingHTMLParser.AREA) {

			/*
			 * Handle A HREF or AREA HREF
			 */
			params = parser.ProcessParameters();
			value = (String) params.get("HREF");
			if (value != null) {
			    link = value;
			}
		    }
		    else if (tok.kind == LoggingHTMLParser.IMG) {
			/*
			 * Handle IMG SRC
			 */
			params = parser.ProcessParameters();
			value = (String) params.get("SRC");
			if (value != null) {
			    link = value;
			}

		    }
		    /*
		     * check to see if there is an HREF in this tag.
		     * If there's no HREF, then just append the unchanged tag.
		     */
		    if (link == null) {
			// print token, parameters, closetag
			htmlFile.print(tok.image);
			int size = names.size();
			String val = null;
			for (int i=0; i < size; i++) {
			    String key = (String) names.elementAt(i);
			    htmlFile.print(" " + key);
			    if ((val = (String) params.get(key.toUpperCase())) != null)
				htmlFile.print("=\"" + val + "\"");
			}
			htmlFile.print(">");
			tok = parser.getToken(0);
			continue;
		    }

		    /*
		     * create a new LinkObject for this link
		     */
		    try {
			URL newLinkURL = new URL(contextURL, link);
			File f = new File (localWorkingDir,
					   FileRef.getNewFileId() + getExtension (newLinkURL));
			newLinkObject = new LinkObject (f, newLinkURL);
		    }
		    catch (MalformedURLException e) {
				/*
				 * It's probably a CGI script and we couldn't create a LinkObject
				 * for it because we can't create a URL object for it, so make it
				 * point back to the remote host.
				 */
			htmlFile.print(tok.image);
			int size = names.size();
			String val = null;
			for (int i=0; i < size; i++) {
			    String key = (String) names.elementAt(i);
			    htmlFile.print(" " + key);
			    if ((val = (String) params.get(key.toUpperCase())) != null)
				htmlFile.print("=\"" + val + "\"");
			}
			htmlFile.print(">");
			tok = parser.getToken(0);
			continue;

		    }

		    LinkObject lo = getExistingLinkObject(newLinkObject);

		    /*
		     * add link to the copied htmlFile, editing if necessary
		     */

		    String tagstr = new String(tok.image);//tag.toString();
		    int size = names.size();
		    String val = null;
		    for (int i=0; i < size; i++) {
			String key = (String) names.elementAt(i);
			tagstr += " " + key;
			if ((val = (String) params.get(key.toUpperCase())) != null)
			    tagstr += "=\"" + val + "\"";
		    }
		    tagstr += ">";

		    /*
		     * add local link or remote link
		     */
		    if (lo != null ) {
			tagstr = pointLocal (tagstr, link, lo);
		    }
		    else {
			boolean copyIt = false;
			String sourceName = newLinkObject.getURL().toString().toLowerCase();

			if (alwaysRemote(sourceName)) {
			    copyIt = false;
			}
			else if (sourceName.endsWith(".gif") || sourceName.endsWith(".jpg")) {
			    copyIt = true;
			    newLinkObject.isSupportComponent = true;
			}
			else if ( atFinalCopyDepth ) {
			    copyIt = false;
			}
			else if ( sourceName.indexOf ("?") >= 0 ) {
			    copyIt = false;
			}
			else if (isFromSameSite(newLinkObject)){
			    copyIt = true;
			}
			else {
			    copyIt = false;
			}

			if (copyIt) {
			    addNextLink(newLinkObject);
			    tagstr = pointLocal (tagstr, link, newLinkObject);
			}
			else {
			    tagstr = pointBack (tagstr, link, newLinkObject);
			}
		    }
		    htmlFile.print (tagstr);	// append tag
		    tok = parser.getToken(0);
		}
		/*
		 * flush base tags
		 */
		else if (tok.kind == LoggingHTMLParser.BASE) {
		    //			System.out.println ("Removed <BASE> tag ...");

		    /*
		     * Change context to reflect the context set in the <BASE> tag
		     */

		    Hashtable params = parser.ProcessParameters();
			
		    String base = (String) params.get("HREF");
			if (base != null && base.length() != 0) {
				try {
					contextURL = new URL(base);
					reporter.debugMsg("Use BASE URL for context",
						base, null, null, DEBUG_KEY);
				} catch (MalformedURLException e) {
					reporter.debugMsg("Unable to use Base href URL",
						base, contextURL, e, DEBUG_KEY);
					
				}
			}
		    // replace <BASE ...> with empty comment <!>
		    htmlFile.print("!>");
		    tok = parser.getToken(0);

		}
		else {
		    htmlFile.print(tok.image);
		    tok = parser.getNextToken();
		}
	    }
	} catch (Exception ex) {
	    reporter.reportException(ex);
	} catch (Error er) {
	}

    }

/**
* Check to see if the link object is from the site we are interested in
*/
    private boolean isFromSameSite(LinkObject link) {

	if ( rootSite == null) {
	    rootSite = rootLink.getURL().getHost();
	    reporter.assert(rootSite != null,"WebCopy.isFromSameSite gets null root site",null,null,null);
	}

	if (rootSite == null) {
	    return false;
	}

	return rootSite.equalsIgnoreCase(link.getURL().getHost());
    }
    /**
     * Copies file from remote web subtree to local directory under a
     * unique filename.
     */
    private void copyLinks (Hashtable copyingNowLinks) {
	Enumeration enum = copyingNowLinks.elements();

	//	    System.out.println("begin copyLinks");
	while (enum.hasMoreElements()) {
	    FileOutputStream out = null;
	    LinkObject thisLinkObject = null;

	    try {
		remoteInputStream = null;
		out = null;

		thisLinkObject = (LinkObject) enum.nextElement();
		URL contextURL = thisLinkObject.getURL();

		URLConnection urlc = contextURL.openConnection();
		addPropertiesToURL(urlc, requestProperties);
		remoteInputStream = urlc.getInputStream();
		out = new FileOutputStream (thisLinkObject.getFile());

		thisLinkObject.setMIMEType (urlc.getContentType());
		long lastModified = urlc.getLastModified();
		if (lastModified == 0)
		    lastModified = System.currentTimeMillis();
		thisLinkObject.setLastModified (lastModified);


        reporter.debugMsg(" Fetch ", thisLinkObject.getURL(),
		     " to ", thisLinkObject.getFile(), DEBUG_KEY);


		if (urlc.getContentType().equalsIgnoreCase ("text/html")) {
		    reporter.debugMsg("  WebCopy text/html ", null ,null,null,DEBUG_KEY);
			copyModifyLinks(contextURL, remoteInputStream, new PrintStream(out));	// parseout HTML links as we copy to out
		}
		else {
		    reporter.debugMsg("  WebCopy binary ",null,null,null, DEBUG_KEY);
		    byte[] buf = new byte[4096];
		    int len;
		    while ((len = remoteInputStream.read(buf)) != -1) {
			out.write (buf, 0, len);
		    }
		}
	    }
	    catch (IOException e) {} // allow remote to abrupty close at end of send
	    catch (Exception e) {
			reporter.reportException(e);
	    }
	    finally {

                if (remoteInputStream != null)
		    try { remoteInputStream.close(); } catch (IOException e) {}
                if (out != null) {
		    addCopiedLink(thisLinkObject);
		    try { out.close(); } catch (IOException e) {}
                }
	    }
	}
	//	    System.out.println(" end copyLinks");
    }

	private void addPropertiesToURL(URLConnection urlc, Hashtable requestProperties){
		Enumeration e = requestProperties.keys();
		while (e.hasMoreElements()) {
			String prop = (String) e.nextElement();
			String value = (String) requestProperties.get(prop);
			urlc.setRequestProperty(prop, value);
		}
	}

}
