/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.monitor;

import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.PrintStream;
import java.io.FileNotFoundException;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Properties;
import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpUtils;

import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.interfaces.ScheduleAttributes;
import org.mitre.pam.monitor.SrvPAMBase;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.monitor.ssi.*;
import org.mitre.pam.util.AuthContext;
import org.mitre.pdif.eventParser.ValueList;

/**
 * A servlet that allows the properties of Agents to be edited by using
 * a template web page for each Agent kind.
 *
 * @author Michael DeCorpo
 * @author Jason Mathews, iPAM 2.1.62 5/4/99
 *			  -Minor mods to allow subclassing,
 *                        -addeded security check,
 *                        -added optimization using StringBuffer where appropriate,
 * @author Jason Mathews, iPAM 2.3.64 6/3/99
 *                        -use ValueList.toString() instead of doing it manually,
 *                        -use ValueList.parseValues to parse multi-values rather than
 *                         StringTokenizer, added debugging.
 * @author Jason Mathews, iPAM 2.4.87 10/14/99
 *                        -do not show schedule info if agent is not schedulable.
 *                        -make refresh URL relative -- getRequestURI doesn't necessarily return the same
 *                         hostname that the page was requested from.
 */

public class SrvAgentEditor extends SrvPAMBase {

  public static final String QUERY_BASE = "/PAM/AgentEdit.query";
  
  private static final String TEMPLATEDIR = "AgentEditServlet";
  private static final String IGNOREPROPSFILE = "ignoreproperties.txt";
  private static final String DEFAULTTEMPLATE = "default.shtml";
  private static final String FILEEXTENSION = ".shtml";
  private static final String DEBUG_KEY = "AgentEditor";

  /**
   * This method handles a "GET" request from a client. It looks for 
   * parameter "agent" which is the id of a iPAM Agent and builds a form for 
   * editing the properties of that agent.  It also looks for parameter "uid",
   * but it is currently ignored.
   */
  protected void doGet(HttpServletRequest req, HttpServletResponse res)
       throws ServletException, IOException
  {

	/*
	* Begin access check +++++++++
	*/
	AuthContext context = getAuthContext(req, res);

	if (!context.verifyAccess(reporter, this, "get agent parameters"))
		return;
	/*
	* End access check ------------
	*/

    ServletOutputStream out = res.getOutputStream();
    res.setContentType("text/html");
    res.setHeader("Pragma","no-cache");  // so the browser won't cache the page

    // get the parameters and make sure "agent" and "uid" exist.  Any others
    // are ignored.
    String agent = req.getParameter("agent");
    if (agent == null) {
      printErrorMessage(out,"Parameter \"agent\" not specified.");
      return;
    }
    
    String uid = req.getParameter("uid");
    if (uid == null) {
      printErrorMessage(out,"Parameter \"uid\" not specified.");
      return;
    }
    
    // Get the Agent from the exec
    Agent a = exec.getActiveAgent(agent);
    if (a == null) {
      printErrorMessage(out,"Agent: <b>" + agent + "</b> doesn't exist.<br>");
      return;
    }

    // get the appropriate template file for the agent kind.  If a template
    // for that kind doesn't exist, get the default template file.
    String fname = a.getKind() + FILEEXTENSION;
    File dir = ExecGlobal.getSupportDir(TEMPLATEDIR);
    File templatefile = new File(dir,fname);
    if (!templatefile.exists()) {
      templatefile = new File(dir,DEFAULTTEMPLATE);
      if (!templatefile.exists()) {
	printErrorMessage(out,"The default template file doesn't exist.");
	return;
      }
    }

    // initialize the Properties that will be sent to the parser
    Properties substitutionVar = new Properties();
    Properties substitutionFile = new Properties();
 
    // build a Vector of property names that shouldn't be seen by the user on
    // the web page.  Then go through the vector and remove them from the 
    // Memento
    Hashtable ignores = getIgnores();
    
    // Go through the memento and put each property and its value into the
    // substitutionVar object that will be sent to the parser.  If the property
    // has multiple values, but quotes around the odd numbered values(1 being the
    // first)
    String value;
    for (Enumeration e = a.getProperties(); e.hasMoreElements(); ) {
      String prop = (String) e.nextElement();
    	if (prop.startsWith("bert_"))
    		continue;
    	if (prop.startsWith("meta_"))
    		continue;
    	if (prop.startsWith("cache."))
    		continue;
    	if (ignores.containsKey(prop))
    		continue;
      ValueList vl = a.getPropertyValues(prop);
      // must escape '>' in form values at the very least to not confuse browsers
      value = replaceString(vl.toString(), ">", "&gt;");
/*
      if (vl.size() == 1)
	value = vl.toString();
      else {
	boolean odd = true;
	StringBuffer buf = new StringBuffer();
	for (Enumeration en = vl.elements(); en.hasMoreElements(); ) {
	  if (odd) {
	    buf.append('\"');
	    buf.append(en.nextElement());
	    buf.append('\"');
	  } else {
	    buf.append(en.nextElement());
	  }
	  buf.append(' ');
	  odd = !odd;
	}//for
	value = buf.toString();
      }
*/
      // reporter.debugMsg("set form value", prop,"->", value,DEBUG_KEY);
      substitutionVar.put(prop, value);
    } //for

    // suppress printing of schedule properties if agent type is not schedulable
    if (substitutionVar.containsKey(ScheduleAttributes.SCHEDULE_VERB) &&
		 ! (a instanceof org.mitre.pam.interfaces.Schedulable)) {
	// reporter.log("** Agent " + a + " is not scheduable");
	substitutionVar.remove(ScheduleAttributes.SCHEDULE_VERB);
    }

    // We need to clone the substitutionVar and give the clone to the parser
    // as well because there are some properties that are specific to the 
    // servlet.  The clone will only contain properties of the agent.
    Properties allProps = (Properties) substitutionVar.clone();

    // There are some agent properties that we don't want to show up as 
    // "additional" properties if the template does not ask for their value 
    // at all.  Because the code to build the schedule table is in 
    // org.mitre.pam.executive.Scheduler and not in the class that keeps track
    // of what properties have been asked for in the template, 
    // org.mitre.utils.ssi.ResultFile, it will not know if these fields were 
    // used and always print these fields as "additional" if we do not remove
    // them from the list.  On the downside, if the template for a Juggler
    // doesn't call the #schedule directive, the user will not be able to 
    // change the schedule at all.
    removeScheduleProperties(allProps);
    
    // There are some servlet specific properties that the templates may call.
    // They must be added to the Properties object.
    String reqUri = req.getRequestURI().toString();
    String supportDirUri = ExecGlobal.getSupportDirURLRoot();
    substitutionVar.put("AGENT",agent);
    substitutionVar.put("UID",uid);
    substitutionVar.put("TIMECODE",(new Date()).toString());
    substitutionVar.put("POST_ACTION",QUERY_BASE);
    substitutionVar.put("URL",reqUri);
    substitutionVar.put("PAGE_DIR", supportDirUri+"/"+TEMPLATEDIR+"/");
    substitutionVar.put("REFRESH",buildRefresh(reqUri));
    substitutionVar.put("SUPPORT_DIR", supportDirUri);
    
    // add the id of the exec
    Agent execagent = exec.getExecAsAgent();
    substitutionVar.put("EXECID", execagent.getStringProperty("id","<unknown>"));
    
    // Initialize the parser and run it.
    try {
      SimpleSSI parser = new SimpleSSI(new FileInputStream(templatefile), substitutionVar, substitutionFile, allProps, new PrintStream(out));
      parser.doSubstitutions();
    }
    catch (ParseException pe) {}
    out.flush();
  }
  
  protected void removeScheduleProperties(Hashtable props)
  {
    for (int i=0; i < ScheduleAttributes.attributeList.length; i++)
	props.remove(ScheduleAttributes.attributeList[i]);
//    props.remove("schedule");
//    props.remove("period");
//    props.remove("dayOfWeek");
//    props.remove("hourOfDay");
//    props.remove("minuteOfHour");
  }

    /** fix schedule parameters
    * remove attributes that don't make sense in wrong context
    * e.g. if schedule != weekly and dayOfWeek present then remove it
    */
    protected void normalizeSchedProps(Hashtable props) {
	Object o = props.get(ScheduleAttributes.SCHEDULE_VERB);
	if (o == null) return;
	String schedtype;
	if (o instanceof String)
		schedtype = (String) o;
	else if (o instanceof String[]) {
		String v[] = (String []) o;
		schedtype = v[0];
	} else return; // schedule is an unknown type

	if (schedtype == null || schedtype.length()==0) return;
	boolean isWeekly, isDaily, isHourly;
	isWeekly=isDaily=isHourly = false;
	if (schedtype.equals(ScheduleAttributes.SCHEDULE_WEEKLY))
	      isWeekly = true;
	else if (schedtype.equals(ScheduleAttributes.SCHEDULE_DAILY))
	      isDaily = true;
	else if (schedtype.equals(ScheduleAttributes.SCHEDULE_HOURLY))
	      isHourly = true;
        if (!isWeekly)
		props.remove(ScheduleAttributes.DAY_OF_WEEK);
	if (!schedtype.equals(ScheduleAttributes.SCHEDULE_PERIODIC))
		props.remove(ScheduleAttributes.PERIOD);
	if (!isWeekly && !isDaily)
		props.remove(ScheduleAttributes.HOUR_OF_DAY);
	if (!isWeekly && !isDaily && !isHourly)
		props.remove(ScheduleAttributes.MINUTE_OF_HOUR);
  }

  /**
   * This method will read in the file that contains a list of properties, one 
   * per line, that the servlet can ignore because they are not meant to be seen 
   * by the user.  The names of these properties will be put into a Vector that 
   * is returned.
   */
  protected Hashtable getIgnores()
  {
    Hashtable v = new Hashtable();
    File dir = ExecGlobal.getSupportDir(TEMPLATEDIR);
    File f = new File(dir, IGNOREPROPSFILE);
    
    // open the file
    BufferedReader br = null;
    try {
      br = new BufferedReader(new FileReader(f));
      // read each line adding it to the vector;
      String line;
      while ((line = br.readLine()) != null) {
	if (line.length() != 0)
	  v.put(line, line);
      }
    }
    catch (FileNotFoundException fnfe) {
      reporter.warning("Support file \"" + f + "\" not found");
      reporter.reportException(fnfe);
    }
    catch (IOException e) {
      reporter.reportException(e);
    }
    finally {
      if (br != null) {
	try {
      br.close();
	} catch (IOException ioe) {}
      }
    }
    
    return v;
  }


  /** 
   * This method handles the posting of the form generated by the doGet method.
   * The name of the agent and it's properties are sent in the request.  A handle
   * on the agent in iPAM is gotten and it's properties are changed accordingly.
   */
  protected void doPost(HttpServletRequest req, HttpServletResponse res)
       throws ServletException, IOException
  {
    ServletOutputStream out = res.getOutputStream();
    res.setContentType("text/html");
		
    // put the parameters sent in into a hashtable
    Hashtable h;
    try {
      h = HttpUtils.parsePostData(req.getContentLength(),req.getInputStream());
    }
    catch (IllegalArgumentException iae) {
      printErrorMessage(out,"Unable to parse query parameters.");
      return;
    }

    // get agent, uid, and timecode and make sure they exist.  Currently nothing
    // is done with the uid and timecode, but in the future it may be used for 
    // authentication
    String agent, uid, timecode;
    String s[];
    s = (String []) h.get("agent");
    agent = s[0];
    s = (String []) h.get("uid");
    uid = s[0];
    s = (String []) h.get("timecode");
    timecode = s[0];
    
    if (agent == null) {
      printErrorMessage(out,"agent not specified.");
      return;
    }
    if (uid == null) {
      printErrorMessage(out,"uid not specified.");
      return;
    }
    if (timecode == null) {
      printErrorMessage(out,"timecode not specified.");
      return;
    }

    // get the agent from the exec
    Agent a = exec.getActiveAgent(agent);
    if (a == null) {
      printErrorMessage(out,"Agent: <b>" + agent + "</b> doesn't exist.");
      return;
    }

    // fix schedule parameters
    // e.g. if schedule != weekly and dayOfWeek present then remove it
    normalizeSchedProps(h);

    // get the rest of the agent's properties and set them
    Enumeration e = h.keys();
    while(e.hasMoreElements()) {
      String prop = (String) e.nextElement();
      if (!(prop.equals("agent") || prop.equals("uid") || prop.equals("timecode"))) {
	String v[] = (String []) h.get(prop);
	    Vector vector = ValueList.parseValues(v[0]);
	    if (vector.size() == 1) {
		// handle single value
		String val = (String) vector.elementAt(0);
		a.setProperty(prop, val);
		reporter.debugMsg(prop,"->", val,null,DEBUG_KEY);
	    } else if (vector.size() == 0) {
		a.setProperty(prop, "");
		reporter.debugMsg(prop,"-> <null>", null,null,DEBUG_KEY);
	    } else {
		// handle multiple values
		ValueList vl = new ValueList(vector);
		a.setPropertyValues(prop, vl);
		reporter.debugMsg(prop,"==>" ,vl,null,DEBUG_KEY);
	    }
      }
    } //while
    
    // print a little message and a link to return to the list of agents.
    out.println("Your changes have successfully been made.<br><br>");
    out.println("<a href=\"" + SrvPAMServlet.QUERY_URL + "\">Back</a>");
    out.flush();
  }

  /**
   * This method prints the error message sent in a nice format.
   */
  protected void printErrorMessage(ServletOutputStream out, String msg) throws IOException
  {
    out.println("<html>");
    out.println("<head><title>Agent Edit Error</title></head>");
    out.println("<body bgcolor=\"#FFFFFF\">");
    out.println("<h2>Servlet Error</h2>");
    out.println( msg );
    out.println("</body></html>");
    out.flush();
  }

    /** Returns a new string resulting from replacing all occurrences of oldStr
     * in this string with newStr. 
     *
     * If the character oldStr does not occur in this string, then the
     * string is returned unchanged.
     */
    public static String replaceString(String s, String oldStr, String newStr) {
	int ind = s.indexOf(oldStr);
	if (ind == -1)
	    return s; // oldStr does not appear in s hence no change is needed

	int lastpos = 0;
	StringBuffer buf = new StringBuffer(s.length());
	do {
	    if (ind - lastpos > 0)
		buf.append(s.substring(lastpos, ind));
	    buf.append(newStr);
	    lastpos = ind + oldStr.length();
	    ind = s.indexOf(oldStr, lastpos);
	} while (ind >= lastpos);

	if (lastpos < s.length()) {
	    buf.append(s.substring(lastpos));
	}

	return buf.toString();
    }


  /**
   * This method takes a url and returns an html link to that url.  It is used
   * to build the "Refresh" link.
   */
  private String buildRefresh(String url)
  {
    if (url == null)
      url = QUERY_BASE;
    else {
      // strip off hostname and port from URL to make a relative URL
      // with respect to the document base.
      int ind = url.indexOf(QUERY_BASE);
      if (ind != -1)
	url = url.substring(ind);
    }
    return "<a href=\"" + url + "\">Refresh</a>";
  }
}

