/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.monitor.watcher;
import org.mitre.pam.interfaces.ConnectedAgent;
import org.mitre.pam.interfaces.PIM;
import org.mitre.pam.interfaces.DIM;
import org.mitre.pam.util.alarm.AlarmReporter;


/**
* A reference that aggregates the display block,
* the ConnectedAgent, and local data for the two
*
* @author Nazario Irizarry, Jr.
*/
final class NodeData {
	private int x;
	private int y;
	int nInputs;
	int nOutputs;
	ConnectedAgent agent;
	WatchBlock block;
	private Cluster parentCluster;
	int clusterOrderX;
	int clusterOrderY;
	private boolean needAnimation;
	private static final int STEP_X = 20;
	private static final int STEP_Y = 15;
	private boolean onScreen;
	static int truncateNameLength = 8;
	
	private static String DEBUG_NAME = "nodeData";
	private AlarmReporter reporter = new AlarmReporter(DEBUG_NAME);
	

	NodeData(ConnectedAgent agent, WatchBlock block) {		
		this.agent = agent;
		this.block = block;
		block.setHasInputs(false);
		block.setHasOutputs(false);
		updateConnectionInfo();
		String id = agent.getId();
		block.setKind(agent.getKind());
		block.setName(id);
		if (id.length() > truncateNameLength) {
			block.setDisplayName(id.substring(0, truncateNameLength) + "-");
//			block.setInfoString(id+" ("+agent.getKind()+")");
		}
		block.setPos(0,0);
	}

	Cluster getParentCluster() {
		return parentCluster;
	}

/**
* Has side effect of zeroing the position and offset in the cluster
*/
	void setParentCluster(Cluster parent) {
		if (parent == parentCluster)
			return;
		if (parentCluster != null)
			parentCluster.nodeSaysBye(this);
		parentCluster = parent;
		clusterOrderX = 0;
		clusterOrderY = 0;
		setOffsetInCluster(0,0);
		reporter.debugMsg("new parent: ",parentCluster,"for: ",this,DEBUG_NAME);
	}
	
	void loseParent() {
		reporter.debugMsg("losing parent: ",parentCluster,"for: ",this,DEBUG_NAME);
		parentCluster = null;
		clusterOrderX = 0;
		clusterOrderY = 0;
		needAnimation = false;
	}
	
	void setOffsetInCluster(int x, int y) {
		this.x = x;
		this.y = y;
		needAnimation = true;
	}

	void setOffsetInClusterX(int x) {
		this.x = x;
		needAnimation = true;
	}

	void setOffsetInClusterY(int y) {
		this.y = y;
		needAnimation = true;
	}
	
	int getOffsetInClusterX() {
		return x;
	}

	int getOffsetInClusterY() {
		return y;
	}
	
	void updateConnectionInfo() {
		nInputs = nOutputs = 0;
		if (!(agent instanceof PIM)) {
			nInputs = agent.getSourceList().size();
			block.setHasInputs(true);
		}
		if (!(agent instanceof DIM)) {
			nOutputs = agent.getSinkList().size();
			block.setHasOutputs(true);
		}
		block.adjustSizeForName(); // gets it to update its shape too
	}

	boolean hasNoConnections() {
		return (nInputs == 0) && (nOutputs == 0);
	}
	
	void transferPos() {
		needAnimation = true;
	}
	
	void slide() {
		block.morph();
		if (needAnimation) {
			int xb = block.getX();
			int yb = block.getY();
			int deltaX = 0;
			int deltaY = 0;
			if (parentCluster != null) {
				deltaX = parentCluster.x + x - xb ;
				deltaY = parentCluster.y + y - yb;
			}
			else {
				deltaX = x - xb ;
				deltaY = y - yb;
			}
			if (deltaX == 0 && deltaY == 0) {
				needAnimation = false;
			}
			if (deltaX > STEP_X) deltaX = STEP_X;
			if (deltaX < -STEP_X) deltaX = -STEP_X;
			if (deltaY > STEP_Y) deltaY = STEP_Y;
			if (deltaY < -STEP_Y) deltaY = -STEP_Y;
			block.setPos(xb+deltaX, yb+deltaY);
		}
	}
	
	public String toString() {
		return agent == null ? "node <null>" : "node "+agent.toString();
	}
	
	void setOnScreen(boolean onScreen) {
		if (this.onScreen == onScreen) {
			return;
		}
		this.onScreen = onScreen;
		if (onScreen) {
			agent.getMemento().addMementoListener(block);
		}
		else {
			agent.getMemento().removeMementoListener(block);
		}
	}
}
