/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.productVisitor;

import java.awt.Color;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.File;
import java.net.URL;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Properties;

import org.mitre.pam.util.Codecs;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.interfaces.ProductVisitor;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.store.FileNotLocalException;

/**
 * Visits the element(s) of a product and generates an HTML representation of
 * the product.
 * <p>
 * @version 1
 * @author Tim Taylor modified by N. Irizarry
 *
 * modifications: URL encode filenames; preserve directory structure
 * @version 2, 8 Sept 1998
 * @author Jason Mathews
 */
public class AsHTMLIndex implements ProductVisitor
{
	protected Properties hashOfContents = new Properties();

	/** The product we will visit and process. */
	protected Product inputProduct;

	protected File workingDir;

	/**
	 * The name of an image to be used as the background of the generated HTML
	 * page.
	 */
	protected String bgImageName;

	/** The background color to use for the page. */
	protected Color bgColor = Color.white;

	/** Location where the images we need will be expected to be installed. */
	protected String imagePath;

	/** Counts the nested depth of lists.  Slightly modifies the output for
	 * String and URL content types.
	 */
	protected int listDepth = 0;

	/** Option to serve files remotely which if true converts files to
	 * fully qualified URLs served via iPAM or its peer web server, otherwise
	 * file links are kept relative.
	 */
	protected boolean serveRemotely;

	protected boolean serveFlatFiles = false;

	protected Reporter reporter;

	/**
	 * This constructor allows the path to images to be specified as a string.
	 * @param imagePath  The path to where necessary images are found, can be null
	 */
	public AsHTMLIndex(String imageP, Reporter reporter)
	{
		this (true,null,imageP,reporter);
	}

	/**
	 * This constructor allows the path to images to be specified as a string.
	 * @param remote      Set if index will be served remotely in which case all
	 *  FileRefs are converted to fullly qualified URLs with respect to PAM
	 *  otherwise files are referenced as relative file refs
	 * @param imagePath   The path to where necessary images are found, can be null
	 * @param workingDir  The path to where the index will be served, which can be null be not available
	 */
	public AsHTMLIndex(boolean remote, File workingDir, String imageP, Reporter reporter)
	{
		serveRemotely = remote;
		this.workingDir = workingDir;
		this.reporter = reporter;
		imagePath = imageP;
		if(imagePath == null)
			imagePath = "";
		else
			if(imagePath.length() != 0 && !imagePath.endsWith("/"))
				imagePath += "/";
	}

	/**
	 * get a pseudo table of contents.  The list is in no particular order
	 */
	public Properties getTableOfContents()
	{
		return hashOfContents;
	}

	public void setServeRemote(boolean remote)
	{
		serveRemotely = remote;
	}

	public void setFlatFile(boolean isFlatFile)
	{
		serveFlatFiles = isFlatFile;
	}

	/**
	 * Sets the background of the created page to be an image that will be tiled
	 * to fill the entire page.
	 * @param imageName  The name of the image to use as the background image.
	 */
	public void setBackground(String imageName)
	{
		bgImageName = imagePath + imageName;
	}

	/**
	 * Sets the background color of the created page.
	 * @param c  The color to use for the background of the page.
	 */
	public void setBackground(Color c)
	{
		bgColor = c;
	}

	/**
	 * Sets the input product that will be processed by this visitor when the
	 * writeResults method is called.
	 *@param inputProduct The product to process.
	 */
	public void visit(Product inputProduct)
	{
		this.inputProduct = inputProduct;
	}

	/**
	 * Processes the product specified in a previous call to visit and writes
	 * the results to the specified stream.
	 * @param out  The PrintStream to write the generated output to.
	 * @see visit
	 */
	public void writeResults(PrintStream out)
	{
		process(out);
	}

	/**
	 * Processes the specified product and sends the generated HTML to the
	 * specified output stream.  The output will be indented by the amount
	 * specified.
	 * @param out      The PrintStream to send the output to.
	 */
	protected void process(PrintStream out)
	{
		if(inputProduct == null)
			return;
		startHtml(out);
		generateHeader(out);
		startBody(out);
		generateBody(out);
		endHtml(out);
	}

	protected void startHtml(PrintStream out)
	{
		out.println("<HTML>");
	}

	/**
	 * Prints an HTML document header to the specified stream.  This should
	 * generate a complete <HEAD></HEAD> block.
	 * @param out  The PrintStream to send the output to.
	 */
	protected void generateHeader(PrintStream out)
	{
		out.println("<HEAD>");
		out.println("<TITLE>" + inputProduct.getDescription() + "</TITLE>");
		out.println("</HEAD>");
	}

	protected void startBody(PrintStream out)
	{
		out.print("<BODY");
		if(bgImageName != null)
			out.print(" background=\"" + bgImageName + "\"");
		else
		{
			out.print(" BGCOLOR=#" + Integer.toHexString(bgColor.getRed()) + Integer.toHexString(bgColor.getGreen()) + Integer.toHexString(bgColor.getBlue()));
		}
		out.println('>');
	}

	protected void generateBody(PrintStream out)
	{
		out.println("<P><TABLE BORDER=0 CELLPADDING=1 WIDTH=\"100%\">");
		out.println("<TR>");
		out.println("<TD VALIGN=top width=\"5%\">");
		out.println("<P><IMG SRC=\"" + imagePath + "spacer.gif\" WIDTH=60 HEIGHT=3 ALIGN=bottom>");
		out.println("</TD><TD VALIGN=top width=\"70%\">");
		out.print("<H2><IMG SRC=\"" + imagePath + "world.gif\" WIDTH=34 HEIGHT=34 ALIGN=middle>");
		//    out.println("Browse Agent Findings</H2>");
		out.print((String)inputProduct.getId());
		out.println("</H2>");
		out.println("<P>&nbsp;");
		out.println("</TD><TD width=25%>");
		out.println("<P>");
		out.println("</TD></TR>");
		out.println("<TR>");
		out.println("<TD width=5%><p></TD>");
		out.println("<TD VALIGN=top WIDTH=95%>");
		out.println("<P><TABLE BORDER=0 CELLSPACING=1 CELLPADDING=1 WIDTH=\"98%\">");
		out.println("<TR><TH width=20%><P ALIGN=right>Channel:</TH>");
		String topic = inputProduct.getTopic();
		if(topic != null && topic.length() > 0)
		{
			out.println("<TD><P ALIGN=left>" + inputProduct.getTopic() + "</TD></TR>");
		}
		else
		{
			out.println("<TD><P>&nbsp;</TD></TR>");
		}
		out.println("</TABLE>");
		out.println("</TD><TD>");
		out.print("<P><IMG SRC=\"" + imagePath + "clock.gif\" WIDTH=13 HEIGHT=17 ALIGN=middle>");
		out.println(inputProduct.getModificationDate().toGMTString() + "</TD></TR>");
		out.println("<TR><TD><P></TD><TD COLSPAN=2><P><FONT COLOR=\"#AF0000\">");
		out.println("<IMG SRC=\"" + imagePath + "blrule.gif\" WIDTH=\"96%\" HEIGHT=6 ALIGN=bottom>");
		out.println("</FONT></TD></TR></TABLE></P>");
		out.println("<P>&nbsp;</P>");
		out.println("<P><TABLE BORDER=0 CELLPADDING=1 WIDTH=600>");
		out.println("<TR><TD WIDTH=60><P>");
		out.println("<IMG SRC=\"" + imagePath + "spacer.gif\" WIDTH=60 HEIGHT=3 ALIGN=bottom></TD>");
		out.println("<TD><TABLE BORDER=0 CELLPADDING=1 WIDTH=540>");
		out.println("<TR><TD><P>");
		parseProduct(new Vector(),inputProduct,out);
		out.println("</TD></TR></TABLE></TD></TR></TABLE></P>");
	}

	protected void parseProduct(Vector vRelDirNames, Product product, PrintStream out)
	{
		if(product == null)
			return;
		Object content = product.getContent();
		/*
		 * enumerated product
		 */
		if(content instanceof Enumeration)
		{
			Product root = ((ContainerProduct)product).getRoot();
			/*
			 * enumerated product with index
			 */
			if(root != null)
			{
				vRelDirNames.addElement(product.getId());
				parseProduct(vRelDirNames,root,out);
				vRelDirNames.removeElementAt(vRelDirNames.size() - 1);
			}
			/*
			 * enumerated product without index
			 */
			else
			{
				Enumeration e = (Enumeration)content;
				if(listDepth == 0)
					out.print("<P>");
				out.print("<FONT SIZE=\"+1\">");
				out.print(product.getId());
				out.print("</FONT>");
				String desc = product.getDescription();
				if(desc.length() != 0)
				{
					out.print("<FONT SIZE=\"-1\"> - ");
					out.print(desc);
					out.print("</FONT>");
				}
				out.println("<br><ul>");
				listDepth++;
				// add container ID's to retain the directory structure
				vRelDirNames.addElement(product.getId());
				while(e.hasMoreElements())
				{
					out.print("<li>");
					parseProduct(vRelDirNames,(Product)e.nextElement(),out);
				}
				vRelDirNames.removeElementAt(vRelDirNames.size() - 1);
				out.println("</ul>");
				listDepth--;
			}
		}
		/*
		 * String contents
		 */
		else
			if(content instanceof String)
			{
				if(listDepth == 0)
					out.println("<P>");
				out.print(product.getId());
				String desc = product.getDescription();
				if(desc.length() != 0)
				{
					out.print(" - ");
					out.print(desc);
				}
				String body = (String)content;
				if(body != null && body.length() != 0)
				{
					out.print(" - ");
					out.print(body);
				}
				out.println("<br>");
			}
			/*
			 * URL contents
			 */
			else
				if(content instanceof URL)
				{
					URL url = (URL)content;
					if(listDepth == 0)
						out.println("<P>");
					out.print("<A HREF=\"" + url.toExternalForm() + "\">" + product.getId() + "</A>");
					String desc = product.getDescription();
					if(desc.length() != 0)
					{
						out.print(" - ");
						out.print(desc);
					}
					out.println();
					hashOfContents.put(product.getId(),url.toExternalForm());
				}
				/*
				 * FileRef contents
				 */
				else
					if(content instanceof FileRef)
					{
						FileRef fr = (FileRef)content;
						if(listDepth == 0)
							out.println("<p>");
						// serveRemotely refers to files that are kept locally but are accessed remotely
						// via builtin web server (http)
						String description = product.getDescription();
						if(description == null || description.length() == 0)
						{
							description = "file - " + fr.getSourceName();
						}
						if(serveRemotely)
						{
							try
							{
								String filename = fr.getReferenceAsFullURLString();
								out.println("<A HREF=\"" + filename + "\">" + product.getId() + "</A>" + " - " + description);
								hashOfContents.put(product.getId(),filename);
							}
							catch(FileNotLocalException e)
							{
								out.println(product.getId() + " - " + description);
								reporter.warning("AsHTMLIndex: Non-local file cannot be served remotely: " + fr.toString());
							}
						}
						else
						{
							StringBuffer path = new StringBuffer();
							if(!serveFlatFiles)
							{
								for(int i = 1;i < vRelDirNames.size();i++)
								{
									path.append(Codecs.urlPathEncode((String)vRelDirNames.elementAt(i)));
									path.append('/');
								}
							//for
							}
							path.append(Codecs.urlPathEncode(fr.getFile().getName()));
							String filename = path.toString();
							out.println("<A HREF=\"" + filename + "\">" + product.getId() + "</A>" + " - " + description);
							hashOfContents.put(product.getId(),filename);
						}
					}
					/*
					 * Byte array contents
					 */
					else
						if(content instanceof byte[])
						{
							if(listDepth == 0)
								out.println("<P>");
							out.print(product.getId());
							out.print(" - ");
							out.print(product.getDescription());
							out.print(" Byte array - ");
							AsPlainText.processByteArray(new PrintWriter(out),(byte[])content);
							out.println("<br>");
						}
						/*
						 * unknown content
						 */
						else
						{
							reporter.assert(false,"Unrecognized content '",content.toString(),"'",null);
						}
	}

	protected void endHtml(PrintStream out)
	{
		out.println("</BODY>");
		out.println("</HTML>");
	}

}

