/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.productVisitor;

import java.io.*;
import java.util.Enumeration;
import java.util.Vector;
import java.util.zip.*;
import java.util.Hashtable;

import org.mitre.pam.executive.Exec;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.interfaces.ProductVisitor;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.product.view.ProductStore;
import org.mitre.pam.productVisitor.AsFlatFileList;
import org.mitre.pam.util.StreamCopy;
import org.mitre.pam.interfaces.ProductFilter;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.product.view.BasicProduct;


/*
 * A visitor that creates a zipfile containing the
 *  the contents of its products as well as a text
 *  description of the product.  This zipfile can
 *  be reconstructed using buildProduct()
 *
 * NOTE: the visit(p) method writes to prodZipFile
 *  as a side effect, but the value of prodZipFile
 *  is not guaranteed to be a valid zip file, because
 *  in the event of an exception, prodZipFile
 *  is set to null.
 *
 * @author Srinivas Inguva
 * @author Jones Leung, added a ProductFilter (RenameDuplicateFilter)
 *  to rename duplicate filenames to prevent "java.util.zip.ZipException: duplicate entry"
 * @author John R. Fantry - added the buildFiles() functions
 */


public class AsZipFile implements ProductVisitor {

  private File prodStoreFile;
  private File prodZipFile;
  private Product product;
  private boolean alreadyMade;
  private Reporter reporter;

  public AsZipFile(File zipFile,File psf,Reporter reporter) {
    prodZipFile = zipFile;
    prodStoreFile = psf;
    this.reporter =  reporter;
  }

  public void visit(Product product)  {
  	this.product = product;
  }

  public void makeZip(Product prod, File prodStoreFile, File prodZipFile) throws IOException  {
    Product deepCopy = prod.cloneDeeper();
    ProductStore ps = new ProductStore();
    ps.setRootProduct(deepCopy);

    PrintWriter prodOut = null;
    prodOut = new PrintWriter(new FileOutputStream(prodStoreFile));

    ps.addToStore(deepCopy);
    ps.store(prodOut);
    prodOut.close();

    FileOutputStream fout = null;
    ZipOutputStream zout = null;

    fout = new FileOutputStream(prodZipFile);
    zout = new ZipOutputStream(fout);
    zout.setComment("CDA Product storage for product: "+deepCopy.getUid());
    zout.setMethod(Deflater.DEFLATED);
    zout.setLevel(Deflater.BEST_SPEED);

    /*
    * write the prodStoreFile to the zip file:
    */
    ZipEntry prodFileEntry = makeZipEntry(prodStoreFile,null);
    zout.putNextEntry(prodFileEntry);
    {
    FileInputStream fin = new FileInputStream(prodStoreFile);
    StreamCopy.copyFully(fin,zout,false);
    fin.close();
    }
    zout.closeEntry();

    Hashtable fileList = new Hashtable();
    String root = null;
    fileList = AsZipFile.buildFiles(deepCopy, root);

    Enumeration enum = fileList.keys();
    while(enum.hasMoreElements()) {
      FileRef fr = (FileRef)enum.nextElement();
      File file = fr.getFile();
      String path = (String)fileList.get(fr);
      ZipEntry ze = makeZipEntry(file, path, null);
      zout.putNextEntry(ze);

      FileInputStream fin = new FileInputStream(file);
      StreamCopy.copyFully(fin,zout,false);
      fin.close();

      zout.closeEntry();
    }

    zout.finish();
    zout.close();

    alreadyMade = true;
  }

  /**
  * Builds a hashtable containing FileRef objects as keys and their new paths
  * as values.  The paths are determined by looking at the product structure.
  * This code has been declared static so that all future agents that have a
  * a need for this behavior can share it.
  *
  * @param prod  the product structure to search
  * @param root  a root to use as a prefix for all FileRef objects
  * @returns  a hashtable containing FileRef objects as keys and their new paths
  *           as values
  */
  public static Hashtable buildFiles(Product prod, String root) {
    Hashtable files = new Hashtable();

    AsZipFile.buildFiles(prod, root, files, new Vector());

    return files;
  }

  /**
   * Recursive function which builds new paths based upon the product structure.
   */
  protected static void buildFiles(Product prod, String root, Hashtable files, Vector currentPath) {
    if(prod == null) {
      return;
    }

    Object obj = prod.getContent();

    if(obj instanceof Enumeration) {
      Enumeration enum = (Enumeration)obj;
      currentPath.addElement(prod.getId());

      while(enum.hasMoreElements()) {
        buildFiles((Product)enum.nextElement(), root, files, currentPath);
      }

      currentPath.removeElement(currentPath.lastElement());
    }
    else if(obj instanceof Vector) {
      Vector vec = (Vector)obj;
      currentPath.addElement(prod.getId());

      for(int i=0; i<vec.size(); i++) {
        buildFiles((Product)vec.elementAt(i), root, files, currentPath);
      }

      currentPath.removeElement(currentPath.lastElement());
    }
    else if(obj instanceof FileRef) {
      File src = ((FileRef)obj).getFile();
      String path = AsZipFile.buildPath(root, currentPath);
      files.put(obj, path);
    }
  }

  /**
   * Creates a valid path out of the root and current pathInfo
   */
  protected static String buildPath(String root, Vector pathInfo) {
    StringBuffer path;

    if(root == null) {
      path = new StringBuffer();
    }
    else if(root.endsWith(File.separator)) {
      path = new StringBuffer(root);
    }
    else {
      path = new StringBuffer(root + File.separator);
    }

    for(int i=1; i<pathInfo.size(); i++) {
      path.append(pathInfo.elementAt(i));
      path.append(File.separator);
    }

    return path.toString();
  }

  private ZipEntry makeZipEntry(File file, String path, String comment) {
    ZipEntry ze = new ZipEntry(path + file.getName());

    ze.setComment(comment);
    ze.setMethod(ZipEntry.DEFLATED);
    ze.setSize(file.length());

    return ze;
  }

  private ZipEntry makeZipEntry(File f, String comment) {
    ZipEntry ze = new ZipEntry(f.getName());
    ze.setSize(f.length());
    ze.setMethod(ZipEntry.DEFLATED);
    if (comment != null)
      ze.setComment(comment);

    return ze;
  }

  public File getZipFile() throws IOException {
  	if (alreadyMade)
    	return prodZipFile;

    makeZip(product, prodStoreFile, prodZipFile);
    return prodZipFile;
  }

  /*
  * @param outDir is the output directory
  * @param zFile is the product zip file e.g. "prodStore.zip"
  * @param prodFile is the name of the txt file containing the product information
  *  e.g. "prodStore.txt"
  * @param reporter the agent that will report any messages that may come up
  */
  public static Product buildProduct(File outDir, File zFile, String prodFile, Reporter reporter)  throws IOException{

    if (!outDir.isDirectory() ||  !zFile.isFile() || prodFile == null) {
    	reporter.assert(false, "Incorrect args", null,null,null);
		return null;
    }

    unCompress(zFile,outDir);
    ProductStore prodStore = new ProductStore();

    Product tmpProduct = reconstructProduct(prodStore, prodFile,outDir, reporter);

    reporter.assert(tmpProduct != null,"null product!",null,null,null);

    return(tmpProduct);
  }//end buildProduct();

  private static void unCompress(File zFile, File myDir) throws IOException {
    if(!myDir.exists()) {
      myDir.mkdirs();
    }

    ZipFile zf = new ZipFile(zFile);
    Enumeration e = zf.entries();

    ZipEntry ze;
    File f;
    FileOutputStream tempfos;
    FileInputStream fin;
    String serialized = "";
    String tempString = null;
    FileRef fr;
    while (e.hasMoreElements()) {
      ze = (ZipEntry) e.nextElement();
      String fileName = ze.getName();

      // test to see if the name contains path info
      if(fileName.indexOf(File.separator) != -1) {
        String path = fileName.substring(0, fileName.lastIndexOf(File.separator));
        Vector v = new Vector();
        v.addElement(path);
        String newPath = buildPath(myDir.getAbsolutePath(), v);
        myDir = new File(newPath);
        if(!myDir.exists()) {
          myDir.mkdirs();
        }
        fileName = fileName.substring(fileName.lastIndexOf(File.separator) + 1);
      }

      f = new File(myDir, fileName);
      tempfos = new FileOutputStream(f);
      StreamCopy.copyFully(zf.getInputStream(ze),tempfos);
    }
    zf.close();
  }

  private static Product reconstructProduct(ProductStore prodStore,String productFile,File myDir, Reporter reporter){
    Product retVal = null;
    File prodFile = new File(myDir,productFile);
    try {
      Reader productStream = new FileReader(prodFile);
      prodStore.load(productStream);
      retVal = prodStore.getRootProduct();

      AsFlatFileList affl = new AsFlatFileList();
      affl.visit(retVal);
      Enumeration fileProdList = affl.getFileProductList();

      //Change FileRefs to point to local files that were just uncompressed:
      Object content;
      while (fileProdList.hasMoreElements()) {
		content = ((Product)fileProdList.nextElement()).getContent();
		reporter.assert(content instanceof FileRef,"Unrecognized content",content.toString(),null,null);
		((FileRef)content).resetParent(myDir);
//		((FileRef)content).setKeepFile(false);
		prodFile.delete();
      }
      return retVal;
    } catch (FileNotFoundException e) {
      e.printStackTrace();
      return null;
    } catch (IOException io) {
      io.printStackTrace();
      return null;
    } catch (Exception e2) {
      e2.printStackTrace();
      return null;
    }
  }//end reconstructProduct();

}