/* Copyright 1997, 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.putter;

import java.io.IOException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;

import org.mitre.pam.util.FTP;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Source;
import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.productVisitor.AsFlatFileList;
import org.mitre.pam.productVisitor.AsHTMLIndex;
import org.mitre.pam.store.FileRef;

/**
* This class creates a DIM for sending files to a remote machine using
* the FTP protocol.
* <p>
* This version extends PutProductToDirectory and supports hierarchical output.
*
* @author   John Fantry, Technical Co-op (D510)
* @author   Jason Mathews, many changes (D510)
*/
public class FTPFileDIM extends PutProductToDirectory {
  public final static String USER = "user";
  public final static String PASSWORD = "password";
  public final static String FTP_URL = "ftpUrl";
  private static final String DEBUG_KEY = "FTPFileDIM";

/**
* FTP server
*/
  protected FTP server = null;

  public FTPFileDIM() {
    setDescription("Saves files only to specified FTP directory");
    setProperty(FTP_URL, "ftp://x.y.z/rootDir");
    setProperty(PASSWORD, "<ftp password here>");
    setProperty(USER, "anonymous");

    removeProperty(LOCAL_TARGET_DIRECTORY);

    metaStringInfo(FTP_URL, "ftp://x.y.z/rootDir","Destination directory in URL form");
    metaStringInfo(PASSWORD, "<ftp password here>","Password for FTP account");
    metaStringInfo(USER, "anonymous","FTP account name");
  }

/**
* Callback to receive a command.
* @param  arg Contains the file list that will be copied.
*/
  public void putProduct(Product prod) {
  	try{
		super.putProduct(prod);
	}
	finally {
		if (server != null) {
			try {
			    server.close();
			} catch (IOException ioe) {
			    reportException(ioe);
			}
		    server = null;
		}//if
	}
  }

  /**
  * This function determines what the remote root directory is and creates it if it
  * does not exist.
  */
    protected File getDestDir() throws IOException {

	if (destDir == null) {

	    String rootDir;
	    String url = getStringProperty(FTP_URL,"");

	    if ( !url.toLowerCase().startsWith("ftp://")) {
			throw new IllegalArgumentException("Missing or invalid FTP URL");
	    }

		String host;
		    int pos = url.indexOf('/', 6);
		    if (pos == -1) {
				host = url.substring(6);
				rootDir = "/";
		    }
		    else {
				host = url.substring(6, pos);
				rootDir = url.substring(pos);  // extract path
				if (!rootDir.endsWith("/"))
				    rootDir = rootDir.concat("/");
		    }

	    destDir = new File( rootDir );

	    String user = getStringProperty(USER,"");
	    String password = getStringProperty(PASSWORD,"");

    if ( user.length()==0 || password.length()==0) {
        throw new IllegalArgumentException("Missing login information for ftp");
    }

	    server = new FTP(host, user, password, reporter);
	    server.connect();

	    reporter.debugMsg("getDestDir() rootdir=" , rootDir ," dir=" , destDir, DEBUG_KEY); 
	    reporter.debugMsg("  destDir.path()=" , destDir.getPath(),null,null, DEBUG_KEY); 

		// lets see if the directory already exists
		String responses[] = server.sendCommand(FTP.CWD, rootDir);

		if (!responses[0].startsWith("2")) {
		    // the directory didn't exist so we'll create it
		    String reponses[] = server.sendCommand(FTP.RMKD, rootDir);
		    responses = server.sendCommand(FTP.CWD, rootDir);
		    if (!responses[0].startsWith("2")) {
			throw new IllegalArgumentException("Unable to create destination directory: " + rootDir);
		    }
		}
	}//if

	return destDir;
    }

/**
* Make destination directory given the vector of relative directory names
*
* @return   the File object associated with this remote directory
*/
  protected File makeDestDir(Vector vRelDirNames) throws IOException {

    StringBuffer path = new StringBuffer(destDir.getPath());

      // append each of the directory names onto the root target spec
      // except the first element which is the 'outermost' container id and
      // not part of the path.

    for (int i=1; i < vRelDirNames.size(); i++) {
	path.append("/" + vRelDirNames.elementAt(i));
    }

    String rootDir = path.toString();

    // lets see if the directory already exists
    String responses[] = server.sendCommand(FTP.CWD, rootDir);
	
    if (!responses[0].startsWith("2")) {
	    // the directory didn't exist so we'll create it
	    String reponses[] = server.sendCommand(FTP.RMKD, rootDir);
	    responses = server.sendCommand(FTP.CWD, rootDir);
	    if (!responses[0].startsWith("2"))
		throw new IOException("Unable to create destination directory: " + rootDir);
    }

  return new File(rootDir);
}

  /**
  * This function attempts to write the given file to the remote file system.
  */
  protected void putProduct(Vector relDirNames, String fileName, File src) {

    StringBuffer path = new StringBuffer(destDir.getPath());

      // append each of the directory names onto the root target spec
      // except the first element which is the 'outermost' container id and
      // not part of the path.

    for (int i=1; i < relDirNames.size(); i++) {
      path.append('/');
      path.append(relDirNames.elementAt(i));
    }

    reporter.debugMsg("putProduct() rootDir=" , destDir.getPath(), "\n file=",fileName, DEBUG_KEY);
	reporter.debugMsg( " src=", src, "\n  relDirNames=",relDirNames , DEBUG_KEY);
	reporter.debugMsg( "\n  path=",path, null,null, DEBUG_KEY); 

    try {
      // lets see if the directory already exists
      String responses[] = server.sendCommand(FTP.CWD, path.toString());
      if (responses[0].startsWith("2")) {
        server.sendCommand(FTP.STOR, src);
      }
      else {
        // the directory didn't exist so we'll create it
        String reponses[] = server.sendCommand(FTP.RMKD, path.toString());
        responses = server.sendCommand(FTP.CWD, path.toString());

        // test to see if the directory was successfully created
        if (responses[0].startsWith("2")) {
          server.sendCommand(FTP.STOR, src);
        }
        else {
          // the directory could not be created, log it
	  throw new IOException("FTP can not create remote directory <" + path + ">");
        }
      }
    } catch (IOException ioe) {
      reportException(ioe);
    }
  }

/**
* Generate an index page for this product and save at the ftp directory root
*/
    public void generateIndexPage(Product prod, String nameOfIndex) throws IOException
    {
	File indexFile = new File(getWorkingDir(), nameOfIndex);

	reporter.debugMsg("generateIndexPage() index file=" , indexFile, null,null, DEBUG_KEY);

	FileOutputStream fos = new FileOutputStream(indexFile);

	AsHTMLIndex renderer = new AsHTMLIndex(false, null, null, reporter);
	try {
		renderer.setBackground("ringbinder.jpg");
		renderer.visit(prod);
		renderer.writeResults(new PrintStream(fos));
	}
	finally {
		fos.close();
	}

	String rootDir = getDestDir().getPath();

	// lets see if the directory already exists
	String responses[] = server.sendCommand(FTP.CWD, rootDir);

	if (responses[0].startsWith("2")) {
	    server.sendCommand(FTP.STOR, indexFile);
	}
	else {
	    // the directory didn't exist so we'll create it
	    String reponses[] = server.sendCommand(FTP.RMKD, rootDir);
	    responses = server.sendCommand(FTP.CWD, rootDir);

	    // test to see if the directory was successfully created
	    if (responses[0].startsWith("2")) {
		server.sendCommand(FTP.STOR, indexFile);
	    } else {
		// the directory could not be created, log it
		throw new IOException("FTP can not create remote index file " + indexFile);
	    }
	}

	// copy files that support the index page
	copyTemplates();

	tocTable = renderer.getTableOfContents();
    }//generateIndexPage()

/**
 * Copy supporting files (e.g. buttons, images) required by the index page
 */
    protected void copyTemplates() throws IOException {

      File sourceDir = new File(getStringProperty(SOURCEPATH,""));
      reporter.debugMsg("copyTemplates() src path=" , sourceDir, null,null, DEBUG_KEY);

      // It is assumed that this method is called generateIndexPage
      // only if the index page is correctly written and the
      // remote destination directory is set at the target root.

    /*
    * Copy any "template" or general support files first
    */
      if (sourceDir.exists() && sourceDir.isDirectory()){
	  String fileList[] = sourceDir.list();
	  for(int i=0; i < fileList.length; i++) {
	      File source = new File(sourceDir, fileList[i]);
              server.sendCommand(FTP.STOR, source);
	  }
	} else {
	  log("Source directory " + sourceDir + " does not exist or is not a directory");
	}
  }
 /*
 * for backward compatibility
 */
	public void setProperty(String attribute, String value) {
		if (attribute.equalsIgnoreCase("targetSpec")) {
			attribute = FTP_URL;
		}
		super.setProperty(attribute, value);
	}
}
