/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.putter;

import java.util.*;
import java.io.*;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.AsynchronousAgent;
import org.mitre.pam.product.view.HashtableProduct;
import org.mitre.pam.interfaces.Source;
import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.interfaces.DIM;
import org.mitre.pam.interfaces.Task;
import org.mitre.pam.interfaces.Tasker;
import org.mitre.pam.interfaces.ExecToMonitor;
import org.mitre.pam.interfaces.ScheduleAttributes;

import org.mitre.pam.product.view.BasicProduct;
import org.mitre.pam.product.view.VectorProduct;
import org.mitre.pam.product.view.HashtableProduct;
import org.mitre.pam.productVisitor.AsHTMLIndex;
import org.mitre.pam.productVisitor.AsLocalProduct;
import org.mitre.pam.productVisitor.AsFlatFileList;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.executive.Exec;
import org.mitre.pam.executive.CompletionSinkN;
import org.mitre.pam.store.FileRef;
import org.mitre.pam.store.FileNotLocalException;
//import org.mitre.pam.util.alarm.*;

/**
* Creates a CDF file and an index page of its product. Periodically, 
* a master CDF subscription page (for all PutCDF DIMs) is also created. <p>
*
* This DIM creates an index page and CDF file for its product so that users of Internet
* Explorer or PointCast can subscribe to that information.  The index page is updated whenever
* new content is received. <p>
*
* The schedule that is specified in the CDF file matches the schedule on the juggler
* which provides input to this DIM.<p>
* 
* If the juggler has no schedule, then a default schedule (update every 1 day) is used.<p>
*
* The CDF file and product index pages are not garbage collected because they are placed
* in a persistent product. <p>
* 
* @author Jones Leung
*/
public class PutCDF extends DIMShell1 implements AsynchronousAgent, Runnable {  
    protected static final String SUPPORT_DIR_SUB_URL = "/CDFDIM";  
    protected static final String DEBUG_KEY = "CDFDIM";
    protected static final String VERBOSE_DEBUG_KEY = "CDFDIMV";
    protected static final String SOURCE_SIG = "sourceListSignature";
    
    protected static final String imagePath = ExecGlobal.getSupportDirURLRoot() + SUPPORT_DIR_SUB_URL + "/";

    protected static final String SUB_PAGE_DIR = "CDFsubscriptions";
    protected static final String SUB_PAGE_NAME = "channels.html";    
        
    // default PointCast update frequency = 24 hours
    protected long pointCastUpdateFrequency = 24;
    
    // default CDF update frequency = 1
    protected long interval = 1;
    
    // default CDF interval unit = DAY
    protected String intervalUnit = "DAY";
    
    // persistent product
    protected HashtableProduct indexProduct = new HashtableProduct ("CDFDIM", "");       
    protected HashtableProduct cdfProduct = new HashtableProduct ("CDFDIM", "");    
    
    protected static final String UPDATE_INTERVAL = "updateIntervalSecs";
    
    protected long updateInterval = 600000;  // update interval in milliseconds ( 10 minutes )
    
    // single thread used to update the master CDF subscription page when necessary
    protected static Thread updateThread = null;
        
/**
* Constructor for CDF DIM.
*/
    public PutCDF () {
        super();         
        setDescription ("Creates index page and CDF file for its product");
        setProperty (UPDATE_INTERVAL, "600");
    }    
	

/**
* Update product archive and create index page and CDF file.
*/
    public void putProduct(Product prod) {  
    
    	Agent arg = taskHelper.getCommittedTask().identifyAgentSource();                                                 	    	    	   	    	    
	    // set the schedule specified in the CDF file to match the juggler's schedule
	    updateSchedule(arg);
        
        String jugglerUid = ((Agent)arg).getMemento().getUid(); 
        	    	        
   	    // create index page for the incoming product
   	    String indexName = jugglerUid + ".html";
   	    FileRef indexPage = getWorkingFileRef (indexName, false);
   	    
   	    try {
            writeProduct (prod, new PrintStream (new FileOutputStream (indexPage.getFile())));
        } catch (Exception e) {
            reportException (e);
        }
	        
   	    // create CDF file which points to the index page
   	    String cdfName = jugglerUid + ".cdf";
        FileRef cdfFile = getWorkingFileRef (cdfName, false);
        
        try {
            writeCDF (arg.getId(), arg.getDescription(), indexPage.getReferenceAsFullURLString(), cdfFile.getReferenceAsFullURLString(), new PrintStream (new FileOutputStream (cdfFile.getFile())));
        } catch (Exception e) {
            reportException (e);
        }
        
        // put FileRefs into the persistent product
        reporter.assert (indexPage.getFile() != null, "Null index page!",null,null,null);
        reporter.assert (cdfFile.getFile() != null, "Null CDF file!",null,null,null);
        
        saveIndex (jugglerUid, indexPage);
        saveCDF (jugglerUid, cdfFile);

        // update the master subscription page if the sourceList has changed 
        // since the last time doSend() was executed
        String oldsig = getProperty (SOURCE_SIG);
        String newsig = computeSignature();
        
        reporter.debugMsg("oldsig: " , oldsig,null,null, VERBOSE_DEBUG_KEY );
        reporter.debugMsg("newsig: " , newsig,null,null, VERBOSE_DEBUG_KEY );
        
        if (oldsig == null) {
            // running for the very first time ...
            reporter.debugMsg("Running " , toString() , " for the first time...",null, DEBUG_KEY);
            setProperty (SOURCE_SIG, newsig);
            updateThread.interrupt();
        } else if (!oldsig.equals (newsig)) {
            // source list changed ...
            reporter.debugMsg("Source list for " , toString() , " has changed...",null, DEBUG_KEY);
            setProperty (SOURCE_SIG, newsig);
            updateThread.interrupt();
        }
    }        
    
/**
* Gets an Enumeration of active PutCDF DIMs.<p> 
*/
    protected Enumeration getCDFDIMs() {
        Enumeration modules = ((ExecToMonitor)exec).getListOfActiveModules (ExecToMonitor.DIM);        
        Vector dims = new Vector();
        
        while (modules.hasMoreElements()) {
            DIM c = (DIM) modules.nextElement();
            
            if (c instanceof PutCDF) {                
                dims.addElement (c);
            }
        }
        
        return dims.elements();
    }
    
/**
* Writes a CDF file to a PrintStream. <p>
* @param    channelId       The channel id 
* @param    desc            The description of the channel
* @param    indexURL        The URL of the index page that the product is written to
* @param    cdfURL          The URL of the CDF file pointing to the index page
* @param    out             The PrintStream to write CDF output to
*/
    protected void writeCDF (String channelId, String desc, String indexURL, String cdfURL, PrintStream out) {
        try {       
            reporter.debugMsg("Writing out CDF ...",null,null,null, DEBUG_KEY);
            
            out.println ("<?XML version=\"1.0\"?>");
            out.println ("<CHANNEL ");
// Pointcast-specific attributes
            out.println ("     Title=\"" + channelId + "\"");
            out.println ("     Abstract=\"" + desc + "\"");
            out.println ("     SELF=\"" + cdfURL + "\"");
            out.println ("     Frequency=\"" + pointCastUpdateFrequency + "\"");
            
// Standard CDF tags and attributes           
            out.println ("     HREF=\"" + cdfURL + "\">");
            out.println ("     <TITLE>" + channelId + "</TITLE>");
            out.println ("     <ABSTRACT>" + desc + "</ABSTRACT>");

            out.println ("<ITEM ");
            
// Pointcast-specific attribute            
            out.println ("     Title=\"" + channelId + "\"");

// Standard CDF tags and attributes            
            out.println ("     HREF=\"" + indexURL + "\"");
            out.println ("     Type=\"HTML\">");
            out.println ("<TITLE>" + channelId + "</TITLE>");
            out.println ("</ITEM>");

            out.println ("<SCHEDULE>");
            out.println ("<INTERVALTIME " + intervalUnit + "=\"" + interval + "\"/>");
            out.println ("</SCHEDULE>");
            
            out.println ("</CHANNEL>");  
            
            out.flush();
            out.close();
        } catch (Exception e) {
            reportException (e);
        }
    }
    
/**
* Writes an index page to a PrintStream. <p>
* @param    p               The Product to be written out, if null a dummy page is written
* @param    out             The PrintStream to write product page output to
*/
    protected void writeProduct (Product p, PrintStream out) {
        try {
            if (p != null) {
                reporter.debugMsg("Writing out product ...",null,null,null, DEBUG_KEY);                       
                                    
                /*
                * Create productVisitors for localizing and rendering the product
                */
                AsLocalProduct localizer = new AsLocalProduct(getWorkingDir(), reporter);
                AsHTMLIndex renderer = new AsHTMLIndex(true, null, ExecGlobal.getSupportDirURLRoot() + SUPPORT_DIR_SUB_URL, reporter);

                localizer.visit(p);
                AsFlatFileList findFiles = new AsFlatFileList();            

                /*
                * Localize the product if it is not local, so that it can be served 
                * from a local working dir
                */
                if (!localizer.isAllLocal()) {
                    p = localizer.makeLocalClone(p);
                }
                
                renderer.visit(p);
                renderer.setBackground("ringbinder.jpg");
                renderer.writeResults(out);            
                
                out.flush();
                out.close();            
            } else {
                writeDummyProduct (out);
            }
        } catch (Exception e) {
            reportException (e);
        }
    }
/**
* Writes an empty product page to the PrintStream. <p>
* @param    out         The PrintStream to write the empty product page to
*/
    protected void writeDummyProduct (PrintStream out) {
        out.println ("<HTML>");
        out.println ("<HEAD>");
        out.println ("<TITLE>No content available in this channel</TITLE>");
        out.println ("</HEAD>");
        out.println ("<BODY>");
        out.println ("<H3>There is no content available in this channel.</H3>");
        out.println ("</BODY>");
        out.println ("</HTML>");
        
        out.flush();
        out.close();
    }
    
/**
* Write out a channel (i.e. product link and CDF subscription link). <p>
* @param    channelId       The id on the channel
* @param    desc            The description of the channel
* @param    indexURL        The URL of the index page 
* @param    cdfURL          The URL of the CDF file 
* @param    out             The PrintStream to write to
*/
    protected void writeChannel (String channelId, String desc, String indexURL, String cdfURL, PrintStream out) {    
        out.print ("<tr><td>");
        out.print ("<font size=+1><a href=\"");
        out.print (indexURL);
        out.print ("\">" + channelId + "</a></font><br>");
        out.print (desc);
        out.print ("</td>");
        
        out.print ("<td>");
        out.print ("[<a href=\"");
        out.print (cdfURL);
        out.print ("\">Subscribe</a>]");               
        out.print ("</td></tr>");
                
        out.flush();
    }

/**
* Creates FileRefs for filenames in the DIM's working directory.
* Do not prepend the DIM's working directory to the filename.
* It will automatically be prepended.
*/
    protected FileRef getWorkingFileRef (String filename, boolean keepFile) {               
        String fn = getWorkingDir() + File.separator + filename;
        
        reporter.debugMsg("FileRef created with filename: " , fn,null,null, VERBOSE_DEBUG_KEY);
        File file = new File (fn);
        
        reporter.assert (file != null, "Attempted to put null file into FileRef!",null,null,null);
        reporter.assert (file.getName() != null, "Attempted to use null as sourceName of FileRef!",null,null,null);
        
        FileRef fr = new FileRef(file, file.lastModified(), file.getName());
        
        fr.setKeepFile(keepFile);
        
        reporter.assert (fr != null, "Null FileRef discovered!",null,null,null);
        
        return fr;    
    }
    
/**
* Calculate the schedule to use for the CDF file
*/
    protected void updateSchedule (Agent arg) {            
        // get the source's id
        String sourceId = arg.getId();
        
        reporter.debugMsg("source id: " , sourceId,null,null, VERBOSE_DEBUG_KEY);
        
        // find the schedule for the source 
        // Possible choices: SCHEDULE_NONE, SCHEDULE_PERIODIC, SCHEDULE_DAILY,
        // SCHEDULE_WEEKLY, SCHEDULE_MONTHLY, SCHEDULE_HOURLY, SCHEDULE_ONCE...
        String schedule = arg.getProperty (ScheduleAttributes.SCHEDULE_VERB);
        
         reporter.debugMsg("schedule: " , schedule,null,null, VERBOSE_DEBUG_KEY);
        
        // find the period for the source 
        String period = arg.getProperty (ScheduleAttributes.PERIOD);
        
         reporter.debugMsg("period: " , period,null,null, VERBOSE_DEBUG_KEY);       

        if (period != null) {
             reporter.assert (Long.parseLong (period) > 0, "Period must be greater than 0!",null,null,null);
        }
        
        if (schedule.equals (ScheduleAttributes.SCHEDULE_NONE)) {
            // use default schedule (daily updates)
            // do nothing
        } else if (schedule.equals (ScheduleAttributes.SCHEDULE_PERIODIC)) {
            // set CDF to same schedule
            intervalUnit = "MIN";
            interval = Long.parseLong (period) / 60;            
            pointCastUpdateFrequency = interval / 60;
        } else if (schedule.equals (ScheduleAttributes.SCHEDULE_DAILY)) {
            // set CDF to same schedule
            intervalUnit = "DAY";
            interval = 1;
            pointCastUpdateFrequency = interval * 24;
        } else if (schedule.equals (ScheduleAttributes.SCHEDULE_WEEKLY)) {
            // set CDF to same schedule
            intervalUnit = "DAY";
            interval = 7;
            pointCastUpdateFrequency = interval * 24;
        } else if (schedule.equals (ScheduleAttributes.SCHEDULE_MONTHLY)) {
            // set CDF to same schedule
            intervalUnit = "DAY";
            interval = 30;
            pointCastUpdateFrequency = interval * 24;
        } else if (schedule.equals (ScheduleAttributes.SCHEDULE_HOURLY)) {
            // set CDF to same schedule
            intervalUnit = "HOUR";
            interval = 1;
            pointCastUpdateFrequency = interval;
        } else if (schedule.equals (ScheduleAttributes.SCHEDULE_ONCE)) {
            // log unsupported schedule
            log ("SCHEDULE_ONCE not supported! -- Using default schedule for CDF file (daily updates).");
        } else {
            // log unsupported schedule
            log ("Schedule: '" + schedule + "' not supported! -- Using default schedule for CDF file (daily updates).");
        }       
        
        /*
        * Make sure that interval is non-zero
        */
        if (interval == 0) {
            log ("Interval was set to zero.  Resetting to 1.");
            interval = 1;
        }
        
        /*
        * Make sure that interval is non-zero
        */
        if (pointCastUpdateFrequency == 0) {
            log ("PointCastUpdateFrequency was set to zero.  Resetting to 1.");
            pointCastUpdateFrequency = 1;
        }
        
        reporter.debugMsg("intervalUnit: " , intervalUnit,null,null, VERBOSE_DEBUG_KEY);
        reporter.debugMsg("interval: " , Long.toString(interval),null,null, VERBOSE_DEBUG_KEY);
        reporter.debugMsg("pointCastUpdateFrequency: " , Long.toString(pointCastUpdateFrequency),null,null, VERBOSE_DEBUG_KEY);
    }    
    
/**
* Get the URL of the product page with the specified key (juggler uid).
* @param    key         The URL associated with the key
*/
    public String getIndexURL (String key) {        
        Product p = null;
        
        try {
            p = indexProduct.getElement (key);
        } catch (Exception e) {
            // ignore
        }
        
        try {
            if (p != null) {
                return ((FileRef)p.getContent()).getReferenceAsFullURLString();     
            }            
        } catch (Exception e) {
            reportException (e);
        }
        return null;
    }
    
/**
* Get the URL of the CDF file with the specified key (juggler uid).
* @param    key         The URL associated with the key
*/
    public String getCDFURL (String key) {
        Product p = null;
        
        try {
            p = cdfProduct.getElement (key);
        } catch (Exception e) {
            // ignore
        }
        
        try {
            if (p != null) {
                return ((FileRef)p.getContent()).getReferenceAsFullURLString();
            } 
        } catch (Exception e) {
            reportException (e);            
        }
        return null;
    }
    
/**
* Write out master subscription page.<p>
*/
    public void writeMasterSubPage () {
        reporter.debugMsg("Update master CDF subscription page...",null,null,null, DEBUG_KEY);
        Hashtable usedJugglers = new Hashtable();
        
        try {            
      		File dir = ExecGlobal.getSupportDir(SUB_PAGE_DIR);
        	File file = new File(dir, SUB_PAGE_NAME);            
            FileOutputStream fos = new FileOutputStream (file);
            PrintStream out = new PrintStream (fos);
        
            reporter.debugMsg("Updating Master CDF subscription page...", null,null,null,VERBOSE_DEBUG_KEY);
            reporter.debugMsg("Writing new subscription page: " , file.getAbsolutePath(),null,null, VERBOSE_DEBUG_KEY);
            
            out.println ("<html><head><title>CDF Subscription Page</title>");
            out.println ("<body bgcolor=\"skyblue\">");
        
            out.println("<P><TABLE BORDER=0 CELLPADDING=1 WIDTH=\"100%\">");
            out.println("<TR>");
            out.println("<TD VALIGN=top width=\"5%\">");
            out.println("<P><IMG SRC=\"" + imagePath + "spacer.gif\" WIDTH=60 HEIGHT=3 ALIGN=bottom>");
            out.println("</TD><TD VALIGN=top width=\"70%\">");
            out.print("<H2><IMG SRC=\"" + imagePath + "world.gif\" WIDTH=34 HEIGHT=34 ALIGN=middle>");
            
            out.print("CDF Subscription Page");
            out.println("</H2>");
            out.println("<P>&nbsp;");
            out.println("</TD><TD width=25%>");
            out.println("<P>");
            out.println("</TD></TR>");
            out.println("<TR>");
            out.println("<TD width=5%><p></TD>");
            out.println("<TD VALIGN=top WIDTH=95%>");
            out.println("<P><TABLE BORDER=0 CELLSPACING=1 CELLPADDING=1 WIDTH=\"98%\">");
            out.println("</TABLE>");
            out.println("</TD><TD>");
            out.print("<P><IMG SRC=\"" + imagePath + "clock.gif\" WIDTH=13 HEIGHT=17 ALIGN=middle>");
            out.println((new Date()).toGMTString() + "</TD></TR>");
            out.println("<TR><TD><P></TD><TD COLSPAN=2><P><FONT COLOR=\"#AF0000\">");
            out.println("<IMG SRC=\"" + imagePath + "blrule.gif\" WIDTH=\"96%\" HEIGHT=6 ALIGN=bottom>");
            out.println("</FONT></TD></TR></TABLE></P>");
            out.println("<P>&nbsp;</P>");
            out.println("<P><TABLE BORDER=0 CELLPADDING=1 WIDTH=600>");
            out.println("<TR><TD WIDTH=60><P>");
            out.println("<IMG SRC=\"" + imagePath + "spacer.gif\" WIDTH=60 HEIGHT=3 ALIGN=bottom></TD>");
            out.println("<TD><TABLE BORDER=0 CELLPADDING=1 WIDTH=540>");            
            out.println("<TR><TD><P>");
            
            Enumeration dims = getCDFDIMs();
            int channelcount = 0;
            
            while (dims.hasMoreElements()) {
                
                PutCDF dim = (PutCDF) dims.nextElement();
                Hashtable sourceList = dim.getSourceList();
                Enumeration sources = sourceList.elements();
                
                while (sources.hasMoreElements()) {
                    Agent source = (Agent) sources.nextElement();
                    
                    String jugglerUid = ((Agent) source).getMemento().getUid();
                    String indexURL = dim.getIndexURL (jugglerUid);
                    String cdfURL = dim.getCDFURL(jugglerUid);
                    
                    reporter.debugMsg("indexURL: " , indexURL, null,null, VERBOSE_DEBUG_KEY);
                    reporter.debugMsg("cdfURL: " , cdfURL, null,null, VERBOSE_DEBUG_KEY);
                                        
                    reporter.assert (jugglerUid != null,  "Juggler UID is null!",null,null,null);
                    
                    if (indexURL == null) {
                        // write a dummy product page if one does not already exist
                        FileRef indexPage = dim.getWorkingFileRef (jugglerUid + ".html", false);
                        if (!indexPage.getFile().exists()) {
                            writeDummyProduct (new PrintStream (new FileOutputStream (indexPage.getFile())));
                        }
                        indexURL = indexPage.getReferenceAsFullURLString();
                        dim.saveIndex (jugglerUid, indexPage);
                    }
                    
                    if (cdfURL == null) {
                        // write CDF file in case one does not already exist
                        FileRef cdfFile = dim.getWorkingFileRef (jugglerUid + ".cdf", false);
                        if (!cdfFile.getFile().exists()) {
                            dim.writeCDF (source.getId(), source.getDescription(), indexURL, cdfURL, new PrintStream (new FileOutputStream (cdfFile.getFile())));
                        }
                        cdfURL = cdfFile.getReferenceAsFullURLString();                            
                        dim.saveCDF (jugglerUid, cdfFile);
                    }
                    
                    if (usedJugglers.get (jugglerUid) == null) {            
                        // make sure each channel only gets written out once
                        writeChannel (source.getId(), source.getDescription(), indexURL, cdfURL, out);
                        usedJugglers.put (jugglerUid, "");
                        channelcount++;
                    }                    
                }
            }

            if (channelcount == 0) {
                out.println ("<h3>There are no channels to subscribe to at this time.</h3>");
            }
            
            out.println("</TD></TR></TABLE></TD></TR></TABLE></P>");
            out.println ("</body></html>");        
            
            out.flush();
            out.close();                
        } catch (Exception e) {
            reportException ( e);       
        }
    }
    
/**
* Identifies the persistent product for an item, this
* is the method used by the exec to find out what to save
* on system shutdown.
* @return   this DIM's persistent product
*/
	public Product getPersistentProduct() {	    
	    if (indexProduct == null || cdfProduct == null) {
	        return null;
	    }
	    
    	VectorProduct vp = new VectorProduct (getId(), getDescription());
	    vp.addElement (indexProduct);
	    vp.addElement (cdfProduct);
  		return vp;
	}
/**
* Identifies the persistent product for an item, this
* is the method used by the exec to restore product during
* system restart
* @param    p   the persistent product to be restored 
*/
	public void setPersistentProduct(Product p){	    
        if (p == null) {
            return;
        } else {            
            indexProduct = (HashtableProduct)((VectorProduct) p).elementAt (0);
            cdfProduct = (HashtableProduct)((VectorProduct) p).elementAt (1);
        }               
	}

/**
* Save CDF in persistent product
* @param    jugglerUid      The key under which the product is stored
* @param    cdfFile         The FileRef to be saved in persistent product
*/
    protected void saveCDF (String jugglerUid, FileRef cdfFile) {        
        BasicProduct cdfProd = new BasicProduct (getId(), "");
        cdfProd.setContent (cdfFile);        
        cdfProduct.addElement (jugglerUid, cdfProd);        
    }
    
/**
* Save product index page in persistent product
* @param    jugglerUid      The key under which the product is stored
* @param    indexPage       The FileRef to be saved in persistent product
*/
    protected void saveIndex (String jugglerUid, FileRef indexPage) {
        BasicProduct indexProd = new BasicProduct (getId(), "");        
        indexProd.setContent (indexPage);
        indexProduct.addElement (jugglerUid, indexProd);
    }	
/**
* Update the CDF subscription page on a specified interval
*/
    public void run () {        
        try {
            Thread.sleep (120000); // delay 2 minutes to let exec finish initializing
        } catch (Exception e) {
            // ignore
        }
        
        // find out how often to update the master CDF subscription page
        try {
            updateInterval = 1000*Long.parseLong (getProperty (UPDATE_INTERVAL));
        } catch (Exception e) {
            setProperty (UPDATE_INTERVAL, "600");
            updateInterval = 600000;
        }        
        
        while (true) {
            writeMasterSubPage();
            try {
                Thread.sleep (updateInterval);
            } catch (Exception e) {
                // ignore
            }
        }
    }
/**
* Called after the exec has loaded everything (including the agent's Memento) when it is ready
* to start real work or when this agent is created its 
* properties have been initialized.  This will only be called once.
*/
	public void startService() {
	    // start the single subscription page update thread
	    if (updateThread == null) {
	        updateThread = new Thread(this);
	        updateThread.start();
	    }
	}

/**
* Called when the exec is ready to shutdown or when this agent
* is going to be deleted.   This will only be called once.
*/
	public void stopService() {
	    // stop the single subscription page update thread
	    if (updateThread != null) {
	        updateThread.stop();
	        updateThread = null;
	    }
	}
	
/**
* Called to request that the agent stop injecting asynchronous events into the system.
* However, the agent does not need to stop listening and responding to
* sockets.  An agent should have a queue so that it can process its "stimulus"
* without stimulating the system during a pause. Do not assume that pauseService
* is called before stopService.  The exec may decide to stop and may not issue a pause.
*/
	public void pauseService() {
	    // do nothing
	}
	
/**
* Called when it is ok to pass stimulus to the exec or other agents.
*/
	public void resumeService() {
	    // do nothing
	}
	
/**
* Computes and returns the unique signature of the sourceList
* @return   the unique signature for the sourceList
*/
    public String computeSignature () {
        Hashtable hash = getSourceList();            
        Enumeration enum = hash.elements();
        StringBuffer sig = new StringBuffer ();
        
        while (enum.hasMoreElements()) {
            Object source = enum.nextElement();
            String sourceUid = ((Agent) source).getMemento().getUid();
            sig.append (sourceUid);
        }
        return sig.toString();
    }
}