/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.putter;

import java.util.Properties;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Source;
import org.mitre.pam.product.view.EnumeratedProduct;
import org.mitre.pam.store.FileRef;

/**
* First deletes all files found in the channel directory
* (pointed to by targetSpec), copies all the files contained in the source into the
* channel directory, and then calls the publish program (pointed to by commandPath)
* thereby publishing the channel.
*
* @author   John Fantry, Technical Co-op
* @author   Nazario Irizarry, many additions
*/
public class PutLocalPublisher extends PutProductToDirectory {
  protected final static String NAME_OF_URL_LIST_FILE = "nameOfURLListFile";
  protected final static String ALWAYS_COPY_SUPPORT = "alwaysCopySupportFiles";
  protected final static String COMMANDLINEOPTIONS = "commandLineOptions";
  protected final static String COMMANDPATH = "commandProgram";
  protected final static String KEEP_FILE1 = "properties.txt";
  protected final static String KEEP_FILES = "keepFiles";
  protected final static long TIMEOUT = 1000;
  protected final static int MAXCOUNT = 60;

  public PutLocalPublisher() {
    setDescription("Saves to directory and executes local publishing tool");
    setProperty(NAME_OF_URL_LIST_FILE, "");
    setProperty(COMMANDLINEOPTIONS, "");
    setProperty(COMMANDPATH, "");
    setProperty(KEEP_FILES, KEEP_FILE1);
    setProperty(ALWAYS_COPY_SUPPORT, "false");
    setProperty(MAKE_INDEX_PAGE, "true"); // inherited attribute from PutProductToDirectory

    metaStringInfo(NAME_OF_URL_LIST_FILE, "", "[optional] name of output file to receive list of urls");
    metaStringInfo(COMMANDLINEOPTIONS, "", "[optional] command line options to the publishing tool");
    metaStringInfo(COMMANDPATH, "", "[optional] full path to publishing tool or command to execute");
    metaStringInfo(KEEP_FILES, KEEP_FILE1,"files that may not be deleted");
    metaBooleanInfo(ALWAYS_COPY_SUPPORT, false,"copy templates required for publishing");
    metaBooleanInfo(MAKE_INDEX_PAGE, true, "make an index page if the product does not already have one"); // inherited attribute from PutProductToDirectory

    setProperty(DELETE_BEFORE_OUTPUT, true);
    metaBooleanInfo(DELETE_BEFORE_OUTPUT, true,"If set then delete output directory before writting into it");
  }

  /**
  * Executes the command defined by COMMANDPATH with any paramaters
  * defined by COMMANDLINEOPTIONS. The DIM will wait a total of
  * MAXCOUNT * TIMEOUT milliseconds for the process to start or it
  * will be destroyed.
  */
  protected void postSend(Product prod) throws IOException {
    Product root;

    /* Find out if product has a root element in which case an index
     * already exists.
     */

    if (prod instanceof EnumeratedProduct)
      root = ((EnumeratedProduct) prod).getRoot();
    else if ( prod.getProperty(Product.PROPERTY_ROOT) != null)
      root = prod;
    else root = null;

    String nameOfTOCFile = getStringProperty(NAME_OF_URL_LIST_FILE,"");
    boolean createTOCFile = nameOfTOCFile.length() != 0;
    boolean templatesCopied = false;

    String nameOfIndex = getStringProperty(NAME_OF_INDEX_FILE, DEFAULT_INDEX_NAME);
    boolean indexNameSupplied = nameOfIndex.length() != 0;
    if (!indexNameSupplied)
    	nameOfIndex = DEFAULT_INDEX_NAME;

    boolean makeIndex = getBooleanProperty(MAKE_INDEX_PAGE, true);

    File workingDirectory = getDestDir();

    File index = new File(workingDirectory, nameOfIndex);
    /*
     * If there is no root to the product, then generate one
     */
    if (root == null && makeIndex)
      {
        generateIndexPage(prod, nameOfIndex);
        // support files are copied when index page is generated
        templatesCopied = true;
      }

    /*
     * If product has a root then rename it if needed to match the
     * specified name.
     */
    if (root != null && indexNameSupplied ) {
      
      String orgName = ((FileRef)root.getContent()).getFile().getName();

      if (!orgName.equals(nameOfIndex)) {
	/* can't do this because the original is not in this directory
	   fr.getFile().renameTo(index);
        */
	(new File(workingDirectory, orgName)).renameTo(index);
	tocTable = new Properties();
	tocTable.put(index.getName(), index.getName());
      }
    }

    /*
     * Copy templates
     */
    if (!templatesCopied && getBooleanProperty(ALWAYS_COPY_SUPPORT, false))
      copyTemplates();

    /*
     * Make table of contents file
     */
    if (createTOCFile) {
      File tocFile = new File(workingDirectory, nameOfTOCFile);
      FileOutputStream fos = new FileOutputStream(tocFile);
      tocTable.save(fos,"[tableOfContents]");
      fos.close();
    }

    String command = getProperty(COMMANDPATH);
    if (command != null && command.length() != 0) {
            int count = 0;
	    String line;
	    String prevLine = new String();
	    String channelDir = workingDirectory.getAbsolutePath();
	    String commandLineOptions = getProperty(COMMANDLINEOPTIONS);
	    Runtime runtime = Runtime.getRuntime();

	    try {
	      String commandToIssue = command + ' ' + commandLineOptions + ' ' + channelDir;
	      reporter.debugMsg("Command to ",getId()," --> ",commandToIssue, LOG_COMMANDS);

	      Process p = runtime.exec(commandToIssue);
	      InputStream in = p.getInputStream();
	      BufferedReader br = new BufferedReader(new InputStreamReader(in));

	      while(!br.ready() && count < MAXCOUNT) {  // wait for the process to start
	        Thread.currentThread().sleep(TIMEOUT);
	        count++;
	      }

	      if(count == MAXCOUNT) {
	        p.destroy();                    // process did not start
	        reporter.assert(false, "Could not execute command",null,null,null);
	      } else {

	        while((line = br.readLine()) != null) {
	          prevLine = line;
	        }

	        if(!prevLine.equalsIgnoreCase("done")) {  // something went wrong during publish
	          reporter.assert(false, prevLine,null,null,null);
	        } else {
	          log("Channel " + channelDir + " published successfully");
	        }

	        p.waitFor();                    // DIM won't finish until the process has completely terminated
	      }
	    } catch(IOException ioe) {
	      reportException(ioe);
	    } catch(InterruptedException ie) {
	      reportException(ie);
	    }
	}
  }

/**
* Deletes the contents of the targetSpec directory if the directory
* already exists. The "properties.txt" file is not deleted because
* it is created by Castanet and is required to publish the channel properly.
*/
  protected void preSend(Product arg) throws IOException {
	if (getBooleanProperty(DELETE_BEFORE_OUTPUT, true))
		deleteFilesInTargetSpec(getDestDir(), null);
  }
}
