/* Copyright 1997, 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.putter;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Properties;
import java.util.StringTokenizer;

import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.interfaces.ContainerProduct;
import org.mitre.pam.interfaces.Reporter;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Source;
import org.mitre.pam.productVisitor.AsFlatFileList;
import org.mitre.pam.productVisitor.AsHTMLIndex;
import org.mitre.pam.store.FileRef;

/**
* Creates a directory structure from the target local directory,
* and then saves the product contents into that directory.
*
* @author John Fantry, Technical Co-op (D510)
*
* @author Michael E. Los D510
*
* @author Jason Mathews D510 Make container hierarchy map into directory hierarchy.
*/
public class PutProductToDirectory extends DIMShell1 {
  protected final static String MAKE_INDEX_PAGE = "makeIndexPage";
  protected final static String NAME_OF_INDEX_FILE = "nameOfIndexFile";
  protected final static String DEFAULT_INDEX_NAME = "index.html";
  protected final static String LOCAL_TARGET_DIRECTORY = "localTargetDirectory";
  protected final static String DELETE_BEFORE_OUTPUT = "deleteBeforeOutput";
/**
* name of directory in iPAM support directory in which support files
* for making the index page are expected to be stored
*/
  protected final static String DEFAULT_TARGET = "";
  protected final static String SOURCEPATH = "sourcePath";
  protected final static String DEFAULTPATH = "putProductToDirectory";
  protected Properties tocTable;
  protected File destDir = null;

  public PutProductToDirectory() {
    setDescription("Saves files to a directory");
    setProperty(MAKE_INDEX_PAGE, "false");
    setProperty(NAME_OF_INDEX_FILE, DEFAULT_INDEX_NAME);
    String supportDir = ExecGlobal.getSupportDir(DEFAULTPATH).getAbsolutePath();
    setProperty(SOURCEPATH, supportDir);
    setProperty(LOCAL_TARGET_DIRECTORY, DEFAULT_TARGET);

    metaBooleanInfo(MAKE_INDEX_PAGE, false, "create summary index page for output if it doesn't already exist");
    metaStringInfo(NAME_OF_INDEX_FILE, DEFAULT_INDEX_NAME, "name to user for index file");
    metaStringInfo(SOURCEPATH, supportDir, "support directory, with templates) (don't change)");
    metaStringInfo(LOCAL_TARGET_DIRECTORY, DEFAULT_TARGET,"full path for local directory to use for output");

    setProperty(DELETE_BEFORE_OUTPUT, false);
    metaBooleanInfo(DELETE_BEFORE_OUTPUT, false,"If set then delete output directory before writting into it");
  }

/**
* for backward compatibility
*/
	public void setProperty(String property, String value) {
		if (property.equals("targetSpec")) {
			property = LOCAL_TARGET_DIRECTORY;
		}
		if (property.equals(LOCAL_TARGET_DIRECTORY)) {
			int len = value.length();
			
			if (len > 1 && value.charAt(len-1) == File.separatorChar) {
				value = value.substring(0,len-1);
			}
			
		}
		super.setProperty(property, value);
	}
	
  /**
  * This method can be used to perform any pre-dosend() instructions.
  */
  final protected void preSend(Source arg) throws IOException {
  }

  /**
  * This method can be used to perform any post-dosend() instructions.
  */
  final protected void postSend(Source arg) throws IOException {
  }

/**
* Deletes the contents of the targetSpec directory if the directory
* already exists. The "properties.txt" file is not deleted because
* it is created by Castanet and is required to publish the channel properly.
*/
  protected void preSend(Product arg) throws IOException {
	if (getBooleanProperty(DELETE_BEFORE_OUTPUT, false))
		deleteFilesInTargetSpec(getDestDir(), null);
  }

  /**
  * This method can be used to perform any post-dosend() instructions.
  */
  protected void postSend(Product productToSend) throws IOException {
     // determine whether a summary page should be generated
        boolean makeIndex = getBooleanProperty(MAKE_INDEX_PAGE, false);
        if (makeIndex) {
	    	String nameOfIndex = getStringProperty(NAME_OF_INDEX_FILE, DEFAULT_INDEX_NAME);
	    // check if index name is not supplied and then use default
	    if (nameOfIndex.length() == 0)
			nameOfIndex = DEFAULT_INDEX_NAME;
	    
	    generateIndexPage(productToSend, nameOfIndex);
        }
  }

/**
* @param  the product to send
*/
  public void putProduct(Product productToSend) {
    tocTable = new Properties();

    // reset destDir so that the correct directory is created
    destDir = null;

    try
    {
        preSend(productToSend);

	if (productToSend != null) {
	    getDestDir();
	    Vector relDirNames = new Vector();
	    writeProductToDir(productToSend, relDirNames);
	}

        postSend(productToSend);
    } catch( Exception ex ){
	reportException(ex);
    }
  }


/**
* Make or determine the destination directory. The directory is specified by
* the targetSpec parameter.  If it does not exist or it is empty then
* the current working directory is returned.
*/
    protected File getDestDir() throws IOException {

	if (destDir == null) {
	    String targetSpec = getStringProperty(LOCAL_TARGET_DIRECTORY,DEFAULT_TARGET);
	    if (targetSpec.length() == 0) {
			destDir = getWorkingDir();
			return destDir;
	    }
/*
	    if (!targetSpec.endsWith(File.separator))
		targetSpec += File.separator;
*/
	    destDir = new File(targetSpec);
	    if (destDir.exists()) {
	      if (!destDir.isDirectory())
		throw new IOException("Destination directory must be a valid directory path: "+destDir.getAbsolutePath());
	    } else {
		boolean created = destDir.mkdirs();
		if ( !created ) {
			throw new IOException("Destination directory creation failed: "+destDir.getAbsolutePath());
		}
	    }
	}

	return destDir;
    }

/**
* Make destination directory given the vector of relative directory names
*
* @return   the File object associated with this directory
*/
  protected File makeDestDir(Vector vRelDirNames) throws IOException {
      StringBuffer targetSpec = new StringBuffer(getStringProperty(LOCAL_TARGET_DIRECTORY,DEFAULT_TARGET));

      // Append each of the directory names onto the root target spec
      // except the first element which is the 'outermost' container id and
      // not part of the path.
	
      for (int i=1; i < vRelDirNames.size(); i++) {
		targetSpec.append(File.separator);
		targetSpec.append(vRelDirNames.elementAt(i));
      }//while

      File newDestDir = new File( targetSpec.toString() );

      // create the target director if it doesn't exist
      if (! newDestDir.exists())
	  newDestDir.mkdirs();

	return newDestDir;
  }

  protected void writeProductToDir(Object vProd, Vector vRelDirNames) {
      if (vProd == null)
	  return;

      if (vProd instanceof Product)
	  	writeProductToDir((Product) vProd, vRelDirNames);
  }

  /**
   * @param vProd   Product to be wrtten to destination
   * @param vRelDirNames  Vector in which each element is a directory name.
   *                      Concaatenating the vector elements results in the
   *                      relative path under the root to which the product
   *                      is to be written.
   */
  protected void writeProductToDir(Product vProd, Vector vRelDirNames)
  {
      if (vProd == null)
	  return;

    try {

      Object content = vProd.getContent();
      /*
       * enumeration contents
       */
      if (content instanceof Enumeration) {
	  Enumeration enum = (Enumeration) content;
	  //System.out.println("PutProdToDir:writeProdToDir: EnumeratedProdID; " + vProd.getId());
	  
	  vRelDirNames.addElement(vProd.getId());
	  if (! enum.hasMoreElements()) {	
		makeDestDir(vRelDirNames); // create empty directory
	  } else {
	  while (enum.hasMoreElements()) {
	      writeProductToDir(enum.nextElement(), vRelDirNames);
	  }//while
	}
	  vRelDirNames.removeElement(vRelDirNames.lastElement());
      }

      /*
       * vector contents
       */
      else if (content instanceof Vector) {
	  Vector cv = (Vector) content;
	  //System.out.println("PutProdToDir:writeProdToDir: VectorProdID; " + vProd.getId());

	  vRelDirNames.addElement(vProd.getId());
	  if (cv.size() == 0) {
		makeDestDir(vRelDirNames); // create empty directory
	  } else {
	  for (int i = 0; i< cv.size(); i++) {
	      writeProductToDir(cv.elementAt(i), vRelDirNames);
	  }//for
	}
	  vRelDirNames.removeElement(vRelDirNames.lastElement());
      }
      /*
       * FileRef contents
       */
      else if (content instanceof FileRef)
	  {
	      File source = ((FileRef) content).getFile();

	      //System.out.println("PutProdToDir:writeProdToDir: ProdID; " + vProd.getId() + " source.getName: " + fname);

	      putProduct(vRelDirNames, source.getName(), source);
	  }
    } catch (IOException ex) {
	reportException(ex);
    }
  }//writeProductToDir

/**
* Write the file product to a local file
*/
  protected void putProduct(Vector vRelDirNames, String sTgtFileName, File vSrc)
  {
      try {
      File newDestDir = makeDestDir(vRelDirNames);

      //System.out.println("PutProdToDir:putProduct: target=" + newDestDir + ",f=" + sTgtFileName + ",src=" + vSrc);

      File dest = new File(newDestDir, sTgtFileName);
      if (dest.exists()) {
	  	dest.delete();
      }
	  	FileRef.copyFile(vSrc, dest);
	  //System.out.println("putProd: copied " + vSrc.getPath() + " to " + dest.getPath());
	  	tocTable.put(dest.getName(), dest.getName());
      } catch (IOException ex) {
	  	reportException(ex);
      }
  }//putProduct()

  /**
  * Accepts a string representing a path and checks to see if it is valid.
  * If the path is not valid a modified valid version is returned.
  * @param  source    The path to be checked
  * @returns          A valid path
  */
  protected String pathVerifier(String source) {
    StringBuffer path = new StringBuffer();
    StringTokenizer st = new StringTokenizer(source, File.separator);
    while (st.hasMoreTokens()) {
      StringBuffer sb = new StringBuffer(st.nextToken());
      int len = sb.length();
      //if(len > 32) len = 32;
      // Replace all invalid filename chars with an underscore (_)
      for(int i=0; i < len; i++) {
        char c = sb.charAt(i);
        int d = (int) c;
	// this should NOT replace the : after the drive letter in Windows
        if((d < 48) || ((d > 57) && (d < 65)) || ((d > 90) && (d < 95)) || (d == 96) || (d > 122)) {
          sb.setCharAt(i, '_');
        }
      }
      path.append(File.separator);
      path.append(sb.toString());
    }
    // strip off the root slash (/) to make the path relative
    return path.toString().substring(1);
  }

/**
* Generate an index page for this product and save this file at the root directory
*/
  public void generateIndexPage(Product prod, String nameOfIndex) throws IOException
    {
        File workingDirectory = getDestDir();
        File indexFile = new File(workingDirectory, nameOfIndex);

	// System.out.println("putDir:generateIndexPage() index file=" + indexFile);

        FileOutputStream fos = new FileOutputStream(indexFile);
		AsHTMLIndex renderer;
		try {
			renderer = new AsHTMLIndex(false, workingDirectory, null, reporter);
	        renderer.setBackground("ringbinder.jpg");
	        renderer.visit(prod);
	        renderer.writeResults(new PrintStream(fos));
		} finally {
	        fos.close();
		}

        // copy files that support the index page
        copyTemplates();
        tocTable = renderer.getTableOfContents();
    }//generateIndexPage()


/**
 * Copy supporting files (e.g. buttons, images) required by the index page
 */
    protected void copyTemplates() throws IOException
  {
      File sourceDir = new File(getProperty(SOURCEPATH));

      /*
       * Copy any "template" or general support files first
       */
      if (sourceDir.exists() && sourceDir.isDirectory()){
	  String fileList[] = sourceDir.list();
	  for(int i=0; i < fileList.length; i++) {
	      File source = new File(sourceDir, fileList[i]);
	      File dest = new File(getDestDir(), fileList[i]);
	      FileRef.copyFile(source, dest);
	  }
      }
  }//copytemplates()


  protected String localizeFileSep(char vFileSep, String vPath)
  {
      String sPath = vPath.replace('/', vFileSep);
      return sPath;
  }

/**
* Deletes the contents of the of the given directory and keeps
* files listed in the vFilesToKeep argument.
*
* @param vDir    Directory whose contents will be deleted
* @param vFilesToKeep    String of names of files that will not be deleted
*/
	protected void deleteFilesInTargetSpec(File vDir, String vFilesToKeep) {
		if (!vDir.exists())
			return;
		String[] contents = vDir.list();
		if (contents != null) {
			for(int i=0; i < contents.length; i++) {
				File oldFile = new File(vDir, contents[i]);
				if (oldFile.isDirectory()) {
					deleteFilesInTargetSpec(oldFile, null);
					oldFile.delete();
				}
				else if(oldFile.isFile() && (
						(vFilesToKeep == null) || (vFilesToKeep.indexOf(oldFile.getName()) < 0))) {
					oldFile.delete();
				}
			}//for
		}
	}//deleteFilesInTargetSpec()

}
