/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.putter;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Random;

import org.mitre.pam.executive.ExecGlobal;

import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Source;
import org.mitre.pam.product.view.EnumeratedProduct;
import org.mitre.pam.store.FileRef;
import org.mitre.pdif.eventParser.ValueList;

/**
* Provides an interface to the Starburst Multicast command line inteface (CLI)
* program (pointed to by commandPath).  The Starburst CLI is called
* once for each file that is to be transferred.
*
* author Michael E. Los
*/
public class PutStarBurstCLI extends PutProductToDirectory {

  //--------------------------------------------
  // PUBLIC DATA MEMBERS
  //--------------------------------------------
  // config variables specific to PutStarburstCLI
  /** Multicast IP address on which Starburst will announce a transfer */
  public static final String ANNOUNCE_IPADDR = "announceIPAddress";
  /** Multicast IP address on which the data transfer will be performed */
  public  static final String TRANSFER_IPADDR = "transferIPAddress";
  /** Broadcast IP address */
  public static final String BROADCAST_IPADDR = "broadcastIPAddress";
  /** Maximum time (in minutes) Starburst will allow for the transfer to be performed */
  public static final String MAX_TRANSFER_TIME = "maxTransferTime";
  /** kbps */
  public static final String TRANSFER_DATA_RATE = "transferDataRate";
  /** period of time to allow for file transfer announcements (minutes) */
  public static final String ANNOUNCE_DURATION = "announceDuration";
  /** period of time between file transfer announcements (secs)*/
  public static final String ANNOUNCE_PERIOD = "announcePeriod";

  /** file transfer mode used by StarBurst Multicast (Closed, Unlimited, Limited)*/
  public static final String GROUP_TYPE = "groupType";
  /** Maximum number of times StarBurst Multicast will send the entire file
    * before repeating discrete sets of packets. */
  public static final String MAX_PASSES = "numberFilePasses";
  public static final String ADDRESS_TYPE = "addressType";

  public static final String TARGET_DESCRIPTION = "targetDescription";
  public static final String NUM_TARGETS = "numberOfTargets";
  /** multi-valued attribute: IPAddr, clientName, Person, Phone1, Phone2 */
  public static final String TARGET = "target";

  public static final String RECV_CHAN_IPADDR = "receiveChanIPAddress";
  public static final String TX_CHAN_IPADDR = "transmitChanIPAddress";
  // Number of seconds that SB Multicast server will wait for clients
  // to respond to a status query.
  public static final String STATUS_DELAY = "statusDelay";
  public static final String RANDOM_BACKOFF = "randomBackoff";
  public static final String CONTINUOUS_ANNOUNCE = "continuousAnnounce";

  public final static String ALWAYS_COPY_SUPPORT = "alwaysCopySupportFiles";
  public final static String COMMANDLINEOPTIONS = "commandLineOptions";
  public final static String COMMANDPATH = "commandProgram";
  public final static String KEEP_FILES = "keepFiles";

  //--------------------------------------------
  // PROTECTED DATA MEMBERS
  //--------------------------------------------
  protected static final String GROUP_FILENAME = "groupFilename";
  protected static final String CONFIG_FILENAME = "configFilename";

  // Default names for group and configuration file names
  protected static final String GROUP_FILENAME_VAL = "mftp.grp";
  protected static final String CONFIG_FILENAME_VAL = "mftp.cfg";

  protected static final String sLIMITED = "Limited";
  protected static final String sUNLIMITED = "Unlimited";
  protected static final String sCLOSED = "Closed";
  /** ?? mentioned nowhere except in default StarBurst mcast.grp file */
  protected static final String sUNACKNOWLEDGED = "Unacknowledged";

  protected static final String sMULTICAST = "Multicast";
  protected static final String sBROADCAST = "Broadcast";

  private Random zRand = new Random();

  //--------------------------------------------
  // CONSTRUCTOR
  //--------------------------------------------
  public PutStarBurstCLI() {
    setProperty(ALWAYS_COPY_SUPPORT, "false");
    setProperty(MAKE_INDEX_PAGE, "false"); // inherited attribute from PutProductToDirectory

    setDescription("Saves to directory & executes StarBurst Multicast Transmission");

    // set defaults for names used for StarBurst Tx config and group files
    setProperty(CONFIG_FILENAME, CONFIG_FILENAME_VAL);
    setProperty(GROUP_FILENAME, GROUP_FILENAME_VAL);
    setProperty(TARGET_DESCRIPTION, "UNSPECIFIED");
    setProperty(NUM_TARGETS, "0");
    setProperty(TRANSFER_DATA_RATE, "512");
    setProperty(ANNOUNCE_DURATION, "1");
    setProperty(ANNOUNCE_PERIOD, "2");

    setProperty(RECV_CHAN_IPADDR, "0.0.0.0");
    setProperty(TX_CHAN_IPADDR, "0.0.0.0");
    setProperty(STATUS_DELAY, "2");
    setProperty(RANDOM_BACKOFF, "false");
    setProperty(CONTINUOUS_ANNOUNCE, "disabled");

    String sTemp = getStringProperty(ANNOUNCE_IPADDR,"");
    if (sTemp.length() == 0 )
      setProperty(ANNOUNCE_IPADDR, "224.0.1.88");//default SB announce IP Address

    sTemp = getStringProperty(TRANSFER_IPADDR,"");
    if (sTemp.length() == 0)
      setProperty(TRANSFER_IPADDR, "224.0.2.88");//default SB transfer IP Address

    sTemp = getStringProperty(MAX_TRANSFER_TIME,"");
    if (sTemp.length() == 0)
      setProperty(MAX_TRANSFER_TIME, "300");

    // set a default value in case Broadcast address type is specified
    setProperty(BROADCAST_IPADDR, "255.255.255.255");

    setProperty(GROUP_TYPE, sLIMITED);
    setProperty(ADDRESS_TYPE, sMULTICAST);
    setProperty(MAX_PASSES, "1");
//    setProperty(OVERWRITE, "true");
//    setProperty(HIER_DIR, "true");
    setProperty(KEEP_FILES, "");
    //removeProperty(NAME_OF_URL_LIST_FILE);
  }


  /**
  * Deletes the contents of the targetSpec directory if the directory
  * already exists and the command defined by COMMANDPATH also exists.
  */
  protected void preSend(Product arg)
  {
    destDir = getWorkingDir();
    File command = new File(getStringProperty(COMMANDPATH,""));

    if (destDir.exists() && command.exists()) {
    String keepFiles = getStringProperty(KEEP_FILES,"");
      deleteFilesInTargetSpec(destDir, keepFiles);
    }
  }//preSend()


  /**
  * Executes the command defined by COMMANDPATH with any paramaters
  * defined by COMMANDLINEOPTIONS.
  */
  protected void postSend(Product prod) throws IOException {
    Product root = ((EnumeratedProduct)prod).getRoot();

    boolean templatesCopied = false;

    String nameOfIndex = getStringProperty(NAME_OF_INDEX_FILE,"");
    boolean indexNameSupplied =  nameOfIndex.length() > 0;
    if (!indexNameSupplied)
    	nameOfIndex = DEFAULT_INDEX_NAME;
    boolean makeIndex = getBooleanProperty(MAKE_INDEX_PAGE,true);
    File workingDirectory = getDestDir();

      File index = new File(workingDirectory, nameOfIndex);
      /*
      * If there is no root to the product, then generate one
      */
      if (root == null && makeIndex)
      {
        generateIndexPage(prod, nameOfIndex);
        // support files are copied when index page is generated
        templatesCopied = true;
      }

      /*
      * If product has a root then rename it if needed to match the
      * specified name.
      */
      if (root != null && indexNameSupplied ) {

        String orgName = ((FileRef)root.getContent()).getFile().getName();
        if (!orgName.equals(nameOfIndex)) {
//        	System.out.println("file rename to "+index.getAbsolutePath());
      /* can't do this because the original is not in this directory
        	fr.getFile().renameTo(index);
        */
        	(new File(workingDirectory, orgName)).renameTo(index);
        	tocTable = new Properties();
        	tocTable.put(index.getName(), index.getName());
        	}
      }

      /* Copy templates */
      if (!templatesCopied && getBooleanProperty(ALWAYS_COPY_SUPPORT,true))
        copyTemplates(workingDirectory);

    File tempDir = ExecGlobal.g_getWorkingDir();
    File configFile = null;
    File groupFile = null;
    try {
      configFile = generateSBConfigFile(tempDir, getStringProperty(CONFIG_FILENAME,""));
      groupFile = generateSBGroupFile(tempDir, getStringProperty(GROUP_FILENAME,""));
    } catch(IOException ioe) {
      reportException(ioe);
    }

    String command = getStringProperty(COMMANDPATH,"");
    if (command != null && !command.equals("")) {
	    int count = 0;
	    String line;
	    String prevLine = new String();

      String space = " ";

      StringBuffer cmdLineBuf = new StringBuffer(30);
      cmdLineBuf.append(configFile.getAbsolutePath());
      cmdLineBuf.append(space);
      cmdLineBuf.append(groupFile.getAbsolutePath());
      cmdLineBuf.append(space);
      String tgtDesc = getStringProperty(TARGET_DESCRIPTION,"");
      // if an embedded space is found in description, enclose desc in quotes
      if (tgtDesc.indexOf(" ") > -1)
        cmdLineBuf.append("\"" + tgtDesc+ "\"");
      else
        cmdLineBuf.append(tgtDesc);

      cmdLineBuf.append(space);
      cmdLineBuf.append(getStringProperty(COMMANDLINEOPTIONS,""));

      // if the user specified a final path for the file on the client
      String sTgtSpec = getTargetSpec();
      if (sTgtSpec.equals(""))
        sTgtSpec = null;
      else if (! (sTgtSpec.endsWith("/") || (sTgtSpec.endsWith("\\"))) )
      {
        if (sTgtSpec.indexOf("/") > -1)
          sTgtSpec += "/";
        if (sTgtSpec.indexOf("\\") > -1)
          sTgtSpec += "\\";
        else  // take a guess
          sTgtSpec += "/";
      }

      String commandLineOptions = cmdLineBuf.toString();
	    Runtime runtime = Runtime.getRuntime();

      // Get a list of files in the targetSpec directory
      String [] filenames = workingDirectory.list();
      for ( int i=0; i < filenames.length; i++ )
      {
//        System.out.println(filenames[i]);
//System.out.println("PutStarBurstCLI:postSend: "+ command + space + commandLineOptions + space + workingDirectory.getAbsolutePath() + File.separator + filenames[i]);
        try {
          String sCmdLine = command + space + commandLineOptions;
          if (sTgtSpec != null)
          {
            sCmdLine += space + "-x " + sTgtSpec + filenames[i];
          }

          sCmdLine += space + workingDirectory.getAbsolutePath() + File.separator + filenames[i];

          Process p = runtime.exec(sCmdLine);

          // Act as sink for output from StarBurst Multicast processing
          // It seems to be necessary to absorb the app output in order
          // to get it to run.
          BufferedReader rdr = new BufferedReader(new InputStreamReader(p.getInputStream()));
          String s = null;
          while ( (s = rdr.readLine()) != null) ;
//          try {
//            p.waitFor();
//          } catch (java.lang.InterruptedException ie)
//          {
//            reportException (ie);
//          }
        } catch(IOException ioe) {
          reportException(ioe);
        }
      }//for
    }
  }//postSend()

/**
   * Creates a StarBurst Multicast Server configuration file using
   * attributes provided to the DIM.
   *
   * @param vDir    directory into which to save the configuration file
   * @param vName   name of the configuration file
   *
   * @return        name of the configuration file
   */
  protected File generateSBConfigFile(File vDir, String vName) throws IOException
  {
    File configFile = new File(vDir, vName);
    BufferedWriter writer =  new BufferedWriter(new OutputStreamWriter(
                                        new FileOutputStream(configFile)));
    try
    {
      int nStatusDelay;
      try {
          nStatusDelay = Integer.parseInt(getStringProperty(STATUS_DELAY,"2"));
          if (nStatusDelay < 1) nStatusDelay = 1;
          else if (nStatusDelay > 60) nStatusDelay = 60;
      } catch (NumberFormatException nfe)
      {
        nStatusDelay = 2;
      }

      writer.write("[MFTPSERVER CONFIGURATION FILE 3.04]");
      writer.newLine();
      writer.newLine();
      writer.write("receive_addr= " + getStringProperty(RECV_CHAN_IPADDR,"")); writer.newLine();
      writer.write("transmit_addr= " + getStringProperty(TX_CHAN_IPADDR,"")); writer.newLine();
      writer.write("server_data_rate= " + getStringProperty(TRANSFER_DATA_RATE,"")); writer.newLine();
      writer.write("max_transfer_time= "+ getStringProperty(MAX_TRANSFER_TIME,"")); writer.newLine();
      writer.write("announce_period= "  + getStringProperty(ANNOUNCE_PERIOD,"")); writer.newLine();
      writer.write("announce_duration= "+ getStringProperty(ANNOUNCE_DURATION,"")); writer.newLine();
      writer.write("broadcast_address= "+ getStringProperty(BROADCAST_IPADDR,"")); writer.newLine();
      // 224.0.1.88 is default announce address in Starburst
      writer.write("announce_address= " + getStringProperty(ANNOUNCE_IPADDR,"")); writer.newLine();
      writer.write("transfer_address= " + getStringProperty(TRANSFER_IPADDR,"")); writer.newLine();
      writer.write("status_delay= " + nStatusDelay); writer.newLine();
      writer.write("backoff= " + getStringProperty(RANDOM_BACKOFF,"")); writer.newLine();
      writer.write("port_number= default"); writer.newLine();
      writer.write("mdu_size= default"); writer.newLine();
      writer.write("late_announce= " + getStringProperty(CONTINUOUS_ANNOUNCE,"")); writer.newLine();
    }
    catch( IOException ioe ){
      reportException(ioe);
    }
    finally
    {
      writer.flush();
      writer.close();
    }
    return configFile;
  }

  /**
   * Generates a StarBurst Multicast Group file.  The DIM's configuration
   * variables are expected to contain the data needed to build the
   * file.
   *
   * @return        name of the Group file
   */
  protected File generateSBGroupFile(File vDir, String vName) throws IOException
  {
    File groupFile = new File(vDir, vName);
    BufferedWriter writer =  new BufferedWriter(new OutputStreamWriter(
                                        new FileOutputStream(groupFile)));
    String sGroupType = sLIMITED;
    String gt = getStringProperty(GROUP_TYPE,"");
    if (gt.equalsIgnoreCase(sCLOSED))
      sGroupType = sCLOSED;
    //else if (gt.equalsIgnoreCase(sLIMITED)) //handled by initial value above
    else if (gt.equalsIgnoreCase(sUNLIMITED))
      sGroupType = sUNLIMITED;
    else if (gt.equalsIgnoreCase(sUNACKNOWLEDGED))
      sGroupType = sUNACKNOWLEDGED;

    String sAddrType = sMULTICAST;
    String at = getStringProperty(ADDRESS_TYPE,"");
    if (at.equalsIgnoreCase(sBROADCAST))
      sAddrType = sBROADCAST;

    try
    {
      writer.write("[MFTPSERVER GROUP FILE 3.04]");
      writer.newLine();
      writer.newLine();
      writer.write("group_type= " + sGroupType + " (Closed, Limited, Unlimited, or Unacknowledged)");
      writer.newLine();
      writer.write("max_passes= " + getStringProperty(MAX_PASSES,"") + " (Number of complete file passes) ");
      writer.newLine();
      writer.write("address_type= " + sAddrType + " (Multicast or Broadcast)"); writer.newLine();
      writer.newLine(); writer.newLine();

      // for each specified target
      int numTargets = Integer.parseInt(getStringProperty(NUM_TARGETS,""));
      for ( int i=0; i<numTargets; i++ )
      {
        ValueList vals = getPropertyValues(TARGET+i);

        if (vals !=null)
        {
          String sIPAddr = "";
          String sClient = TARGET+i;
          String sPerson = "";
          String sPhone1 = "";
          String sPhone2 = "";

          int nTokenNum = 0;
          Enumeration enum = vals.elements();
          while ( enum.hasMoreElements() ) {
            String val = (String) enum.nextElement();

//System.out.println("val: " + val + " i: " + i);
            switch ( nTokenNum )
            {
              case 0 :
                sIPAddr = val;
                break;
              case 1 :
                sClient = val;
                break;
              case 2 :
                sPerson = val;
                break;
              case 3 :
                sPhone1 = val;
                break;
              case 4 :
                sPhone2 = val;
                break;
            }//switch
            nTokenNum++;
          }//while

          writer.write("IPAddr: " + sIPAddr); writer.newLine();
          writer.write("Client: " + sClient); writer.newLine();
          writer.write("Person: " + sPerson); writer.newLine();
          writer.write("Phone1: " + sPhone1); writer.newLine();
          writer.write("Phone2: " + sPhone2); writer.newLine();
        }
      }//for
    }
    catch( IOException ioe ){
      reportException(ioe);
    }
    finally
    {
      writer.flush();
      writer.close();
    }
    return groupFile;
  }

/**
 * Get output directory.
 * <P>Overidden so that the working directory is used rather than the
 * TargetSpec directory when PutProductToDirectory.doSend() is invoked.
 */
  protected File getDestDir()
  {
	return getWorkingDir();
  }

  private void copyTemplates(File vDestDir) throws IOException
  {
    File sourceDir = new File(getStringProperty(SOURCEPATH,""));
    String fileList[] = sourceDir.list();

    /*
    * Copy any "template" or general support files first
    */
		if (sourceDir.exists() && sourceDir.isDirectory()){
	        for(int i=0; i<fileList.length; i++) {
	          File source = new File(sourceDir, fileList[i]);
            File dest = new File(vDestDir, fileList[i]);
	          FileRef.copyFile(source, dest);
	        }
	    }
  }//copytemplates()

  /**
   * Deletes the contents of the of the given directory and keeps
   * files listed in the vFilesToKeep argument.
   *
   * @param vDir    Directory whose contents will be deleted
   * @param vFilesToKeep    String of names of files that will not be deleted
  */
  protected void deleteFilesInTargetSpec(File vDir, String vFilesToKeep)
  {
    String[] contents = vDir.list();
    for(int i=0; i<contents.length; i++)
    {
      File oldFile = new File(vDir, contents[i]);
      if(oldFile.isFile() && (vFilesToKeep.indexOf(oldFile.getName()) < 0))
      {
        oldFile.delete();
      }
    }//for
  }//deleteFilesInTargetSpec()
}