package org.mitre.pam.putter;

import org.mitre.pam.interfaces.Product;

import org.mitre.pam.productVisitor.Serializer;

import COM.TIBCO.rv.RvException;
import COM.TIBCO.rv.RvOpaque;
import COM.TIBCO.rv.RvSender;
import COM.TIBCO.rv.RvSession;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;

import java.util.zip.Deflater;
import java.util.zip.DeflaterOutputStream;

/**
* This class is a PIM that will send TIBCO messages. <p>
*
* <pre>
* Config vars:
* daemonHost - the IP Address or hostname of the machine on which
*              your TIB/Rendezvous daemon (RVD) is running (default
*              localhost or 127.0.0.1)
*
* daemonPort - the UDP port number where this DIM should send its messages
*              by communicating with the TIB/Rendezvous daemon (RVD) listening on that
*              port. (default 7500)
*
* subjectName - the static or dynamic "subject name" under which messages
*               sent by this DIM will appear; analogous to a radio
*               broadcast frequency. You should not attempt to send using
*               a wildcard subject name due to unpredictable behavior.
*               (no default value)
*
* compressionLevel - the level of compression to use (0 - 9)
*                     (default 0)
*
* Sample PDIF:
* (daemonHost jleung-pc.mitre.org)
* (daemonPort 7501)
* (subjectName NEWS.BOSNIA)
* (compressionLevel 1)
* </pre>
*
* To prevent RvMsgDispatch:Invalid RVMSG_MAGIC error message when trying to specify
* host and port number for RvSession just use RVA (TIB/Rendezvous agent) instead of
* RVD (TIB/Rendezvous daemon). <p>
*
* The RVA will find and connect to an existing RVD with the specified host
* and port number or it will start an appropriate RVD and connect to it.  <p>
*
* The RVA cannot start RVDs on remote machines unless there is already an
* RVA running on that remote machine. <p>
*
* By default, RVA listens on port 7600.  There should be an RVA running on the
* same machine as iPAM in order for this PIM to connect to an RVD through the RVA. <p>
*
* @author Jones Leung
* @author John R. Fantry - modified class so this it uses DictionaryBaseDIM
*
* @see org.mitre.pam.interfaces.AsynchronousAgent
* @see org.mitre.pam.getter.GetAsyncTIBCO
* @see org.mitre.pam.util.Serializer
*/
public class PutTIBCO extends DictionaryBaseDIM {
  /** The IP Address or hostname of the machine on which the RVD is running */
  protected final static String DAEMON_HOST = "daemonHost";
  protected final static String DEFAULT_HOST = "localhost";

  /** The UDP port where this DIM expects to communicate with the RVD to get messages */
  protected final static String DAEMON_PORT = "daemonPort";
  protected final static String DEFAULT_PORT = "7600";

  /** The subject this DIM is sending messages on */
  protected final static String SUBJECT_NAME = "subjectName";
  protected final static String DEFAULT_SUBJECT = "";

  // The level of compression to use.  Valid values are 0 - 9, with 0 meaning uncompressed
  protected static final String COMPRESSION_LEVEL = "compressionLevel";
  protected static final String DEFAULT_COMPRESSION_LEVEL = "0";

  public PutTIBCO() {
    super();
    setProperty (DAEMON_HOST, DEFAULT_HOST);
    setProperty (DAEMON_PORT, DEFAULT_PORT);
    setProperty (SUBJECT_NAME, DEFAULT_SUBJECT);
    setProperty(COMPRESSION_LEVEL, DEFAULT_COMPRESSION_LEVEL);
    setDescription ("Send TIBCO messages");

    metaStringInfo(DAEMON_HOST, DEFAULT_HOST, "the host that will receive data");
    metaStringInfo(DAEMON_PORT, DEFAULT_PORT, "the port number that the host will be listenting on");
    metaStringInfo(SUBJECT_NAME, DEFAULT_SUBJECT, "the TIBCO subject");
    metaStringInfo(COMPRESSION_LEVEL, DEFAULT_COMPRESSION_LEVEL, "the level of compression (0 - 9) to use, level 0 is uncompressed");
  }

  /**
  * Use Java serialization and/or compression to turn the object into a byte array.  Pass the
  * byte array to sendByteArray().
  *
  * @param the object to be sent
  */
  protected void sendObject(Object obj) {
    int compressionLevel = Integer.parseInt(getProperty(COMPRESSION_LEVEL));
    if(compressionLevel < 0 || compressionLevel > 9) {
      warning("Illegal compression level: " + compressionLevel + ", setting to 0");
      compressionLevel = 0;
    }

    try {
      ByteArrayOutputStream bos = new ByteArrayOutputStream();
      DeflaterOutputStream dos = new DeflaterOutputStream(bos, new Deflater(compressionLevel));
      ObjectOutputStream out = new ObjectOutputStream(dos);
      Serializer.serialize(obj, out);
      out.flush();
      out.close();
      sendByteArray(bos.toByteArray());
    } catch(IOException ioe) {
      reportException(ioe);
      return;
    }
  }

  /**
  * Send a byte array using TIBCO.
  *
  * @param the byte array to be sent
  */
  protected void sendByteArray(byte[] b) {
    if(b == null) {
      return;
    }

    String daemonHost = getProperty (DAEMON_HOST);
    String daemonPort = getProperty (DAEMON_PORT);
    String subjectName = getProperty (SUBJECT_NAME);

    // create an RvSender for this subject and send message
    try {
      RvSession session = new RvSession(daemonHost, (new Integer (daemonPort)).intValue());
      RvOpaque message = new RvOpaque (b);
      RvSender sender = session.newSender (subjectName);
      sender.send (message);
      session.term();
    } catch (RvException rve) {
      reportException (rve);
      return;
    }
  }
}