/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.putter;

import java.io.*;
import java.util.Enumeration;
import java.net.URL;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.rmi.Remote;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.NotBoundException;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.executive.ExecGlobal;
import org.mitre.pam.getter.remote.ConnectorInterface;
import org.mitre.pam.productVisitor.AsZipFile;

/**
* A DIM that takes its product, stores it into
* a zipfile, and notifies another CDA system (using RMI),
* which then pulls the product zipfile and
* reconstructs the product.
*
* @author Srinivas Inguva
**/

public class RemoteConnectorDIM extends DIMShell1 {

  public static final String PRODUCT_FILE = org.mitre.pam.getter.remote.Connector.PRODUCT_FILE_NAME;
  public static String REMOTE_OBJECT_NAME = "connectorObjectName";

  public RemoteConnectorDIM() {
    setDescription("Connector to a remote iPAM using Java RMI");
//    setProperty(PRODUCT_FILE,"prodStore.txt");
    setProperty(REMOTE_OBJECT_NAME,"rmi://rmi.registry.hostname/proxyForPAM");

    metaStringInfo(REMOTE_OBJECT_NAME,"rmi://rmi.registry.hostname/proxyForPAM","The host/rmi service name used by this connector iPAM (must match whatever is used by the receiver)");
  }

  public void putProduct(Product prod) {

//    String prodFile = getProperty(PRODUCT_FILE);
    String prodFile = PRODUCT_FILE;
    File prodStoreFile = new File(getWorkingDir(),prodFile);
    File ProdZipFile = new File(getWorkingDir(),"prodStore.zip");

    AsZipFile asZip = new AsZipFile(ProdZipFile,prodStoreFile, reporter);
    asZip.visit(prod);
    try {
	    ProdZipFile = asZip.getZipFile();
	    if (ProdZipFile == null) {
	      log("ConnectorDIM: Product zipfile was not created!");
	      return;
	    }
	
	    //notify linked pam system of zip file and we are done.
	    notifyProxy(ProdZipFile.getName());
	}
	catch (IOException e) {
		reportException(e);
	}
  }//end doSend();

  private void notifyProxy(String filename) {
    //The whole address where the proxy will look for files is:
    // http://<localhost>/<location>/<filename>,
    //which will usually look like this: 
    // http://129.83.19.1:9090/PAMWorkingDir/23583872387/prodStore.zip
    String location = ExecGlobal.getProtocolHostAndPort() +
      ExecGlobal.getWorkingDirURLRoot() + "/" +
      getMemento().getUid();
    if (!location.endsWith("/"))
      location += "/";
    location += filename;
    
    String remoteObjName = getStringProperty(REMOTE_OBJECT_NAME,"");

    Remote remoteObject = null;
    try {
      remoteObject = Naming.lookup(remoteObjName);
    } catch (RemoteException e) {
      reporter.reportException(e);
      reporter.log("RemoteConnectorDIM: Unknown RemoteException while looking up remote object!");
      return;
    } catch (NotBoundException e) {
      reportException(e);
      reporter.log("RemoteConnectorDIM:  Not Bound to remote server!");
      return;
    } catch (MalformedURLException e) {
      reporter.reportException(e);
      reporter.log("RemoteConnectorDIM:  REMOTE_OBJECT_NAME is malformed: "+remoteObjName);
      return;
    } catch (Exception e) {
      reporter.reportException(e);
      return;
    }
    try {
      if (remoteObject instanceof ConnectorInterface) {
	((ConnectorInterface)remoteObject).notify(new URL(location));
      } else {
	log("RemoteConnectorDIM: received an unknown object from "+remoteObjName);
      }
    } catch (RemoteException e) {
      reporter.reportException(e);
      reporter.log("RemoteConnectorDIM: Unknown RemoteException while notifying remote object!");
      return;
    } catch (MalformedURLException e) {
      reporter.reportException(e);
      reporter.log("RemoteConnectorDIM:  Invalid location: "+location);
      return;
    }

    return;
  }//end notifyProxy();

/**
* get an instance of a working directory based on our source
*/
  public File getWorkingDir(Agent arg) {
    File dir = new File(getWorkingDir(), arg.getMemento().getUid());
    dir.mkdirs();
    
    return dir; 
  }
/**
* get an instance of a working directory based on our source
*/
  public String getFullWorkingDirURL(Agent arg) {
    return getFullWorkingDirURL() + "/" + arg.getMemento().getUid();
  }


}
