/* Copyright 1999 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.service.aim;

import org.mitre.pam.interfaces.AsynchronousAgent;
import org.mitre.pam.interfaces.Product;

import org.mitre.pam.productVisitor.AsPlainText;

import org.mitre.pdif.eventParser.ValueList;


import java.io.PrintWriter;
import java.io.StringWriter;

import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;

/**
* This input agent uses America Online's Instant Messenger service using the toc
* protocol to send messages to AIM users.
*
* When iPAM starts this agent will
* establish a connection with the AIMTOCServer and perform the login.  This connection
* will remain open the entire time that iPAM is running.  The connection will be closed
* when iPAM shuts down.
*
* When a product is received it is sent to the AsPlainText visitor.  The resultant
* string is considered to be the Instant Message.  The receiverList is then retrieved.
* The product is also checked to see if it has a "receiverList" property.  These two
* lists are then merged into a "mailing list" (without duplicates).  The IM is then
* encoded and its length is checked.  If the message is too long it is truncated and a
* string is appended to the end of the message so that the receiver will know that some
* of the message was deleted.  The IM's are then sent out.
*
* @author John R. Fantry
*/
public class AIMDIM extends org.mitre.pam.putter.DIMShell1 implements AIMAgent, AsynchronousAgent {
	protected static final String AIM_TOC_SERVER = "AIMTOCServer";
	protected static final String DEFAULT_AIM_TOC_SERVER = "toc.oscar.aol.com";

	protected static final String AIM_TOC_SERVER_PORT = "AIMTOCServerPort";
	protected static final int DEFAULT_AIM_TOC_SERVER_PORT = 21;  // FTP port

	// must be unique, multiple AIMAgents can NOT share loginNames
	protected static final String LOGIN_NAME = "loginName";
	protected static final String DEFAULT_LOGIN_NAME = "";

	protected static final String PASSWORD = "password";
	protected static final String DEFAULT_PASSWORD = "";

	protected static final String AUTH_HOST = "authorizationHost";
	protected static  String DEFAULT_AUTH_HOST = "login.oscar.aol.com";

	protected static final String AUTH_HOST_PORT = "authorizationHostPort";
	protected static final int DEFAULT_AUTH_HOST_PORT = (int)(Math.random() * 2000.0 + 1025.0);  // any port will work

	// value list of user names to send messages to
	protected static final String RECEIVER_LIST = "receiverList";
	protected static final String DEFAULT_RECEIVER_LIST = "";

	// this value controls whether or not a call to putProduct() will result in a
	// login attempt if there is no active connection
	protected static final String ATTEMPT_RELOGIN = "attemptRelogin";
	//  protected static final boolean DEFAULT_ATTEMPT_RELOGIN = true;

	AIMDriverProxy driver = new AIMDriverProxy(this, reporter);
	protected static final String[] EXAMPLE_RECEIVERS = {"janedoe","johndoe"};

	public AIMDIM() {
		ValueList vl = new ValueList(EXAMPLE_RECEIVERS);

		metaStringInfo(AIM_TOC_SERVER, DEFAULT_AIM_TOC_SERVER, "The AIM TOC server to connect with.");
		metaIntegerInfo(AIM_TOC_SERVER_PORT, DEFAULT_AIM_TOC_SERVER_PORT, "The port to talk to the AIM TOC server on.");
		metaStringInfo(LOGIN_NAME, DEFAULT_LOGIN_NAME, "The AIM login name (screen name).");
		metaSetIsPassword(PASSWORD, "AOL IM login password.");
		metaStringInfo(AUTH_HOST, DEFAULT_AUTH_HOST, "The server that will perform the authorization on the loginName and password.");
		metaIntegerInfo(AUTH_HOST_PORT, DEFAULT_AUTH_HOST_PORT, "The port to talk to the authorization server on.");
		metaMultiStringInfo(RECEIVER_LIST, vl, "List of user names to send messages to.");
		/*    metaBooleanInfo(ATTEMPT_RELOGIN, DEFAULT_ATTEMPT_RELOGIN, "Controls whether or not a call to putProduct() " +
		    								"will result in a login attempt if there is no active connection.");
		*/
	}

	/**
	* Open connection to the AIM server and sign on.  Because the login procedure is somewhat
	* asynchronous (due to the listener thread) there is a while loop in the function that
	* will not return until we know the state of the login attempt.
	*/
	private void makeConnection() {
		if(driver.isConnected()) {
			return;
		}

		String AIMTOCServer = getStringProperty(AIM_TOC_SERVER, DEFAULT_AIM_TOC_SERVER);
		int AIMTOCServerPort = getIntegerProperty(AIM_TOC_SERVER_PORT, DEFAULT_AIM_TOC_SERVER_PORT);
		String loginName = getStringProperty(LOGIN_NAME, "");
		String password = getStringProperty(PASSWORD, "");
		String authHost = getStringProperty(AUTH_HOST, DEFAULT_AUTH_HOST);
		int authPort = getIntegerProperty(AUTH_HOST_PORT, DEFAULT_AUTH_HOST_PORT);

		if(loginName.length()==0 || password.length()==0) {
			warning("loginName/password not defined");
			setEnabled(false);
			return;
		}

		driver.startService(AIMTOCServer, AIMTOCServerPort, loginName, password, authHost, authPort);
	}

	/**
	* Disconnect from server
	*/
	public void disableService() {
		driver.stopService();
	}

	/**
	* Accept a product and send it to aol im
	*/
	public void putProduct(Product prod) {
		makeConnection();

		if(!driver.isConnected()) {
			warning("Login failed.  Dropping product.");
			return;
		}

		Vector receivers = new Vector();

		ValueList vl = getPropertyValues(RECEIVER_LIST);
		Enumeration recvs = vl.elements();

		while(recvs.hasMoreElements()) {
			String receiver = (String)recvs.nextElement();
			if(!receivers.contains(receiver)) {
				receivers.addElement(receiver);
			}
		}

		vl = prod.getPropertyValues(RECEIVER_LIST);
		recvs = vl.elements();

		while(recvs.hasMoreElements()) {
			String receiver = (String)recvs.nextElement();
			if(!receivers.contains(receiver)) {
				receivers.addElement(receiver);
			}
		}

		AsPlainText visitor = new AsPlainText();
		visitor.visit(prod);

		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		visitor.writeResults(pw);

		String msg = sw.toString();

		for(int i=0; i<receivers.size(); i++) {
			driver.sendIM((String)receivers.elementAt(i), msg);
		}
	}

	/**
	* Ignore AIMDriver callback
	*/
	public void processIM(String IM) {
	}

}
