/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.service.resource;

import java.util.Vector;
import java.util.Enumeration;
import org.mitre.pam.interfaces.Memento;
import org.mitre.pam.interfaces.ConnectedAgent;
import org.mitre.pdif.eventParser.ValueList;
import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.daemon.DaemonBase;
import org.mitre.pam.interfaces.WantsConnectionChanges;
import org.mitre.pam.interfaces.ScheduleAttributes;

public final class BytesPerMinuteRscPool extends DaemonBase implements WantsConnectionChanges,ResourcePool, AccessibleAgent {

	
	private static final String DEBUG = "BytesPerMinuteResourcePool";
	public static final String MANAGER = "bytesPerMinuteManager";

	public static final String MANAGER_SLOW = "eightProductAverage";
	public static final String MANAGER_MEDIUM = "thirtyProductAverage";
	public static final String[] MANAGER_OPTIONS = {MANAGER_SLOW, MANAGER_MEDIUM};
	

	private Vector consumerList = new Vector(20);

	private RateConsumer rateConsumer = new RateConsumer(this, DEBUG);
	private boolean needDelegationRecomputeFlag = true;

	public BytesPerMinuteRscPool() {
		rateConsumer.usePoolWeight = true; //allow tally of consumption to use bidOverBudget parameter
		setDescription("A resource pool whose budget allows you to limit the number of bytes per minute moved by individual throttleBytesPerMinute agents");
		setProperty(RateConsumer.HISTORICAL_LEVEL,"0");
		setProperty(MANAGER,MANAGER_SLOW);
		setProperty(RateConsumer.BUDGET, "100000");
		
		metaEnumerationInfo(MANAGER,MANAGER_SLOW,"the name of the budget algorithm",MANAGER_OPTIONS);
		metaDoubleInfo(RateConsumer.BUDGET,100000,"total budget for byte-per-minute throughput for this pool instance");
		metaDoubleInfo(RateConsumer.HISTORICAL_LEVEL,0,"Moving average of usage");
		super.setProperty(ScheduleAttributes.SCHEDULE_VERB, ScheduleAttributes.SCHEDULE_NONE);
	}


/**
* Capture numerical property changes
*/
	public void setProperty(String property, String value) {
		if (property.equalsIgnoreCase(MANAGER)) {
			setManager(value);
		}
		else if (property.equalsIgnoreCase(RateConsumer.BUDGET)) {
			try {
				setBudget(Double.valueOf(value).doubleValue());
			}
			catch (NumberFormatException e) {
				setBudget(100000.0 );
			}
		}
		else if (property.equalsIgnoreCase(RateConsumer.HISTORICAL_LEVEL)) {
			rateConsumer.fudgeConsumptionHistory(value);
		}
		else if (property.equals(ScheduleAttributes.SCHEDULE_VERB)) {
			// can not set schedule on this one
		}
		else {
			super.setProperty(property, value);
		}
	}
/**
* Not implemented.
*/
	synchronized public void setManager(String name){
		if (name.equals(MANAGER_MEDIUM)) {
			rateConsumer.setHistorySize(30);
			super.setProperty(MANAGER, MANAGER_MEDIUM);
		}
		else {
			rateConsumer.setHistorySize(8);
			super.setProperty(MANAGER, MANAGER_SLOW);
		}
		needDelegationRecompute();
	}
/**
* Not implemented.
*/
	public void setConsumerConstraint(String constraintName, Object constaintValue){
	}
/**
* Allows the "children" or "member" consumers
* to request permission
*
* @return 0 if ok to send, or delay time (milliseconds) if need to wait
*/
	synchronized long requestToSend(AccessibleAgent requestor, long time, double sizeRequested) {
		recomputeDelegation();
		RateConsumer requestorModel = requestor.getModel();
		long requestorTime = requestorModel.updateAndCalculateDelay(requestor, time, sizeRequested);
		long poolTime = rateConsumer.updateAndCalculateDelay(requestor, time, sizeRequested);
		if (requestorModel.usePoolWeight) {
			reporter.debugMsg("requestToSend using requestor model - ",requestor,null,null, DEBUG);
			return requestorTime;
		}
		else {
			reporter.debugMsg("requestToSend using pool model - ",requestor,null,null, DEBUG);
			return poolTime;
		}
		
	}
/**
* Retract the request to send
*/
	synchronized void retractRequestToSend(AccessibleAgent requestor) {
		requestor.getModel().retractUpdate(requestor);
		rateConsumer.retractUpdate(requestor);
	}
/**
* Allows the "children" or "member" consumers
* to request an estimate of how long until the pool is within budget.
*
* @return 0 if within budget, or delay time (milliseconds) as estimated wait
*/
	synchronized long estimateDelayToSend(AccessibleAgent requestor, long time) {
		recomputeDelegation();
		return rateConsumer.calculateDelay(time);
	}
/**
* Add a consumer to our list
*/
	void addConsumer(Agent consumer) {
		if (consumerList.contains(consumer))
			return;
		consumerList.addElement(consumer);
		needDelegationRecompute();
		reporter.debugMsg("added pool member - ",consumer,null,null, DEBUG);
	}
/**
* Add a consumer to our list
*/
	void removeConsumer(Agent consumer) {
		consumerList.removeElement(consumer);
		needDelegationRecompute();
		reporter.debugMsg("removed pool member - ",consumer,null,null, DEBUG);
	}
/**
* Get the current budget.
*
* @param resourceDimension (ignored)
* @return the current budget.
*/
	public Number getBudget(String resourceDimension){
		return new Double(rateConsumer.budget);
	}
/**
* Set the allowed pool limit for total number of bytes-per-minute; 
*
* @param resourceDimension for multidimensional pools this is the name of the dimension
* @param newBudget the value of this dimension
*/
	public void setBudget(String resourceDimension, Number newBudget){
		if (resourceDimension.equals(RateConsumer.BUDGET)) {
			setBudget(newBudget.doubleValue());
		}
	}
/**
* Set the allowed pool limit for total number of bytes-per-minute; 
*
* @param resourceDimension for multidimensional pools this is the name of the dimension
* @param newBudget the value of this dimension
*/
	void setBudget(double budget){
		rateConsumer.setBudget(budget);
		needDelegationRecompute();
//		recomputeDelegation();
		wakeupSleepingPoolMembers();
	}
/**
* 
* @return Number whos actual type will depend on the resource.  
* Java can deal abstractly with numbers.
*
*/
	public Number getConsumption(String resourceDimension){
		return new Double(rateConsumer.consumption);
	}
/**
* Set consumption from Number value
*/
	public void setConsumption(String resourceDimension, Number value){
		rateConsumer.fudgeConsumptionHistory(value.doubleValue());
	}
/**
* Allows access to the agents super.setProperty()
*/
	public void superSetProperty(String property, String value) {
		super.setProperty(property, value);
	}

/**
* Trigger an update of consumption metrics whenever there
* is a query
*/
	public String getProperty(String propertyName){
		if (propertyName.equals(RateConsumer.HISTORICAL_LEVEL)) {
			rateConsumer.timeHack();
		}
		return super.getProperty(propertyName);
	}
/**
* Get a property
*/
	public ValueList getPropertyValues(String propertyName){
		if (propertyName.equals(RateConsumer.HISTORICAL_LEVEL)) {
			rateConsumer.timeHack();
		}
		return super.getPropertyValues(propertyName);
	}
/**
* Get a property list
*/
	public Enumeration getProperties(){
		rateConsumer.timeHack();
		return super.getProperties();
	}
/**
* This is intended for read only reference to the memento. To set
* a property use setProperty()
*/
	public Memento getMemento (){
		rateConsumer.timeHack();
		return super.getMemento();
	}
/**
* Allows access to the agents consumption model.
*/
	public RateConsumer getModel(){
		return rateConsumer;
	}
/**
* Allows consumers in the pool to notify when their budget weight has changed
*/
	void consumerBudgetChange(AccessibleAgent consumer) {
		needDelegationRecomputeFlag = true;
	}
/**
* Set a flag to recompute delegated budets
*/
	void needDelegationRecompute() {
		needDelegationRecomputeFlag = true;
	}
/**
* Redistribute delegated budgets
*/
	synchronized private void recomputeDelegation() {
		if (!needDelegationRecomputeFlag)
			return;
		reporter.debugMsg("delegation recomputed",null,null,null, DEBUG);
		needDelegationRecomputeFlag = false;
		int nTotal = 0;
		double cumWeight = 0.0;
		Enumeration enum = consumerList.elements();
		while (enum.hasMoreElements()) {
			AccessibleAgent agent = (AccessibleAgent) enum.nextElement();
			RateConsumer rc = agent.getModel();
			if (rc.usePoolWeight) {
				cumWeight += rc.poolWeight;
				nTotal++;
			}
		}
		if (nTotal == 0)
			return;
		double unitPortion = rateConsumer.budget/cumWeight;
		int desiredHistorySize = rateConsumer.getHistorySize();
		enum = consumerList.elements();
		while (enum.hasMoreElements()) {
			AccessibleAgent agent = (AccessibleAgent) enum.nextElement();
			RateConsumer rc = agent.getModel();
			if (rc.usePoolWeight) {
				rc.budget = unitPortion * rc.poolWeight;
				reporter.debugMsg("new budget for",agent.getId()," = ",Double.toString(rc.budget), DEBUG);
			}
			if (rc.getHistorySize() != desiredHistorySize)
				rc.setHistorySize(desiredHistorySize);
		}
		
	}
/**
* Wakeup up any pool members currently in delay and get them
* to recompute their sleep
*/
	private void wakeupSleepingPoolMembers(){
		reporter.debugMsg("wake up sleeping pool members",null,null,null, DEBUG);
		Enumeration e = consumerList.elements();
		while (e.hasMoreElements()) {
			AccessibleAgent agent = (AccessibleAgent) e.nextElement();
			agent.retractIfNeeded();
		}
		e = consumerList.elements();
		while (e.hasMoreElements()) {
			AccessibleAgent agent = (AccessibleAgent) e.nextElement();
			agent.interruptAndRecalculateDelay();
		}
	}
/**
* Allow the pool to notify its consumers that it has been discarded - not implemented here
*/
	public void poolWasDiscarded(){}
/**
* When we are discarded we need to notify potential consumers that we are gone
*/
	public void youHaveBeenDiscarded(){
		Enumeration enum = consumerList.elements();
		while (enum.hasMoreElements()) {
			AccessibleAgent agent = (AccessibleAgent) enum.nextElement();
			agent.poolWasDiscarded();
		}
	}
/**
* Interrupt a currently delayed thread and recalculate the delay time - not implemented in this class
*/
	public void interruptAndRecalculateDelay(){}
/**
* -- not implemented in this class
*/
	public void retractIfNeeded(){}
/**
* Is this agent in delay -- not implemented in this class
*/
	public boolean isInDelay(){return false;}

	/**
	* An input to the agent was removed
	*/
	public void aSourceWasRemoved(ConnectedAgent source){}
			
	/**
	* An input to the agent was added
	*/
	public void aSourceWasAdded(ConnectedAgent source){}

	/**
	* An output to the agent was removed
	*/
	public void aSinkWasRemoved(ConnectedAgent sink){}
			
	/**
	* An output to the agent was added
	*/
	public void aSinkWasAdded(ConnectedAgent sink){}
}