/* Copyright 1998 The MITRE Corporation. ALL RIGHTS RESERVED */
package org.mitre.pam.service.resource;

import org.mitre.pam.interfaces.Memento;
import java.util.Enumeration;
import java.util.Date;
import org.mitre.pdif.eventParser.ValueList;

import org.mitre.pam.interfaces.WantsConnectionChanges;
import org.mitre.pam.interfaces.ExecToController;
import org.mitre.pam.interfaces.Agent;
import org.mitre.pam.interfaces.Product;
import org.mitre.pam.interfaces.Tasker;
import org.mitre.pam.interfaces.Task;
import org.mitre.pam.interfaces.ScheduledCommand;
import org.mitre.pam.interfaces.ConnectedAgent;
import org.mitre.pam.interfaces.ExecThread;
import org.mitre.pam.aggregator.ChanSchedBase;
import org.mitre.pam.store.IntegrityException;
import org.mitre.pam.interfaces.ScheduleAttributes;
import org.mitre.pam.executive.ExecGlobal;

public class ThrottleBytesPerMinute extends ChanSchedBase implements WantsConnectionChanges,ResourceConsumer, AccessibleAgent {

	private static final String DEBUG = "ThrottleBytesPerMinute";
	public static final String POOL_ID = "bytesPerMinutePoolId";
	BytesPerMinuteRscPool pool;
	Thread threadInDelay; // a reference to the executing but delayed thread
	Object threadInDelayLock = new Object();
	Task commandInDelay;
	boolean interrupted;

	private RateConsumer rateConsumer = new RateConsumer(this, DEBUG);

	public ThrottleBytesPerMinute() {
		setDescription("Restrictes passing products so as to observe a byte-per-minute limit on a specified byte-per-minute pool");
		setProperty(POOL_ID,"");
		setProperty(RateConsumer.USE_POOL_WEIGHT,"false");
		setProperty(RateConsumer.POOL_WEIGHT,"1");
		setProperty(RateConsumer.HISTORICAL_LEVEL,"0");
		setProperty(EXTREME_EXECUTION_TIME, "3600");

		metaIntegerInfo(EXTREME_EXECUTION_TIME, 3600,"If the agent is busy this much time, in seconds, it will be disabled (but allowed to complete)");
		metaStringInfo(POOL_ID,"","Associates this consumer with the bytePerMinuteResource daemon (pool) of the same id");
		metaBooleanInfo(RateConsumer.USE_POOL_WEIGHT,false,"Set to allocate pool budget proportionally, Unset to use budget on first-come-first-served baseis");
		metaDoubleInfo(RateConsumer.POOL_WEIGHT,1,"Relative weight for allocating pool budget (if set)");
		metaDoubleInfo(RateConsumer.HISTORICAL_LEVEL,0,"Moving average of usage");
		super.setProperty(ScheduleAttributes.SCHEDULE_VERB, ScheduleAttributes.SCHEDULE_NONE);
	}

/**
* Before accepting an activation check with our pool to make sure the pool
* is not overruning. If it is don't activate at this time.
*/
	public void doCommittedTask(){
		setCommandInDelay(null);
		setThreadInDelay(null);
		Task task = taskHelper.getCommittedTask();
		if (pool == null || task.getCommand() != Task.ACTIVATE) {
			super.doCommittedTask();
			return;
		}
		long delay = pool.estimateDelayToSend(this, System.currentTimeMillis());
		if (delay <= 500) {
			super.doCommittedTask();
			return;
		}
		/*
		* The following is a simpler algorithm than could be done.
		* We tried delaying the command by the right amount and 
		* then recomputing delay if the budget was changed.  However
		* it is too hard to track multiple commands that are delayed.
		* If you delay just one command the throttle will be open
		* to new commands since it will be idle while delaying the first
		* command.  The best way around this problem is not to track 
		* delayed commands, but to delay for a fixed period and
		* let the command retest when it gets issues again.
		*/
		/*
		* no more than 1 minute delay on a command.
		*/
		delay /= 1000;
		int delayInt = (int) delay;
		if (delay < 1)
			delay = 1;
		if (delay > 60)
			delayInt = 60;

		if (delayInt == delay) {
			log("delaying activation for "+Integer.toString(delayInt) + " sec's, pool "+pool.getId()+" is over budget");
		}
		else {
			log("delaying activation for "+Integer.toString(delayInt) +" of " + Long.toString(delay) + " sec's, pool "+pool.getId()+" is over budget");
		}
		
		setCommandInDelay(task);
		((ScheduledCommand)task).delayFromNow(delayInt);
		
		setState(ExecGlobal.IDLE); //clears committed task
	}
/**
* Set the constraint on this consumer.
*
* @param constraintName is a tag that has meaning to the manager
* @param constraintValue must be of type Number or String
*/
	public void setConsumerConstraint(String constraintName, Object constaintValue){
		if (constraintName.equals(RateConsumer.POOL_WEIGHT)) {
			if (!(constaintValue instanceof Number))
				return;
			Number value = (Number) constaintValue;
			rateConsumer.setPoolWeight(value.doubleValue());
		} else if (constraintName.equals(RateConsumer.USE_POOL_WEIGHT)) {
			boolean value = constaintValue.toString().equalsIgnoreCase("true");
			rateConsumer.setUsePoolWeight(value);
		}
	}
/**
* Set this consumers pool memebership
*/
	private void setPoolId(String poolName) {
		if(poolName == null) {
			poolName = "";
		}
		if (pool != null && pool.getId().equals(poolName)) {
			return;
		}
		youHaveBeenDiscarded();
		super.setProperty(POOL_ID, poolName);
		if(poolName.length() > 0) {
			createPoolIfNotFound(poolName); // also adds consumer to pool
		}
	}
/**
* Set a flag to recompute delegated budets
*/
	void needDelegationRecompute() {
		if (pool != null)
			pool.needDelegationRecompute();
	}
/**
* Intercept sets to pool id
*/
	public void setProperty(String property, String value) {
		if (property.equals(POOL_ID)) {
			setPoolId(value);
		}
		else if (property.equalsIgnoreCase(RateConsumer.HISTORICAL_LEVEL)) {
			rateConsumer.fudgeConsumptionHistory(value);
		}
		else if (property.equalsIgnoreCase(RateConsumer.USE_POOL_WEIGHT)) {
			rateConsumer.setUsePoolWeight(value.equalsIgnoreCase("true"));
			needDelegationRecompute();
		}
		else if (property.equalsIgnoreCase(RateConsumer.POOL_WEIGHT)) {
			rateConsumer.setPoolWeight(value);
			needDelegationRecompute();
		}
		else 
			super.setProperty(property, value);
	}
/**
* Create an instance of the pool if it is not found
*/
	private void createPoolIfNotFound(String poolName) {
		ExecToController controller = (ExecToController) exec;
		Agent maybePool = controller.getActiveAgent(poolName);
		try {
			if (maybePool == null) {
				reporter.debugMsg("creating parent pool ",poolName,null,null,DEBUG);
				controller.declareNode("bytesPerMinuteResourcePool", poolName, null);
			}
		}
		catch (IntegrityException e) {
			reporter.reportException(e);
		}
		Agent aPool = controller.getActiveAgent(poolName);
		if (! (aPool instanceof BytesPerMinuteRscPool)) {
			warning("The name "+poolName+" is in use by an agent that is not of type BytesPerMinuteResourcePool");
			this.pool = null;
			return;
		}
		this.pool = (BytesPerMinuteRscPool) aPool;
		pool.addConsumer(this);
	}
/**
* This is where we implement the throttle function.  
* We delay the thread if we are not ready to pass the product along!
* You should make sure there are sufficient PIM threads so
* that some threads can be delayed.
*/
	public Product buildProduct(Product p) {
		if (p == null) {
			return null;
		}
		if (pool == null) {
			return p;
		}
		delayAsRequired(p.getSizeInBytesAsDouble());
		return p;
	}
/*
* Remember the thread that is in delay
*/
	private void setThreadInDelay(Thread thread) {
		synchronized (threadInDelayLock) {
			threadInDelay = thread;
			commandInDelay = null;
		}
	}
/*
* Remember that we delayed the last task
*/
	private void setCommandInDelay(Task task) {
		synchronized (threadInDelayLock) {
			commandInDelay = task;
			threadInDelay = null;
		}
	}
/**
* Check with our pool and delay as required to control
* the aggregate rate
*/
	private void delayAsRequired(double productSize) {
		while (true) {
			long time = System.currentTimeMillis();
			long delayTime = pool.requestToSend(this, time, productSize);
			if (delayTime <= 0) {
				return;
			}
			String delayStr = "delaying product by " + Long.toString(delayTime/1000) +
				 " seconds, \n\testimated continuation on "+(new Date(time+delayTime)).toString()  ;
			setProgress(delayStr);
			if (delayTime > 2000) {
				reporter.log(delayStr);
			}
			setThreadInDelay(Thread.currentThread());
			interrupted = false;
			try {
				if (threadInDelay instanceof ExecThread) {
					((ExecThread)threadInDelay).goToSleep(delayTime);
				}
				else {
					Thread.sleep(delayTime);
				}
			}
			catch (InterruptedException e) {
				interrupted = true;
			}
			setThreadInDelay(null);
			if (!interrupted)
				return;
		}
	}
/**
* Interrupt a currently delayed thread and recalculate the delay time
*/
	public void interruptAndRecalculateDelay() {
		synchronized (threadInDelayLock) {
			if (commandInDelay != null) {
				reporter.debugMsg("activateASAP ",null,null,null,DEBUG);
				commandInDelay = null;
				activateASAP();
			}
			else if (threadInDelay != null) {
				reporter.debugMsg("interrupt sleeping thread ",null,null,null,DEBUG);
				threadInDelay.interrupt();
			}
		}
	}
/**
* Interrupt a currently delayed thread and recalculate the delay time
*/
	public void retractIfNeeded() {
		synchronized (threadInDelayLock) {
			if (threadInDelay != null) {
				reporter.debugMsg("retract sleeping thread stuff ",null,null,null,DEBUG);
				pool.retractRequestToSend(this);
				interrupted = true;
			}
		}
	}
/**
* Is this agent in delay
*/
	public boolean isInDelay(){
		boolean inDelay = threadInDelay != null || commandInDelay != null;
		reporter.debugMsg(this,inDelay?"in delay":"not in delay",null,null,DEBUG);
		return inDelay;
	}
/**
* Notification that the instance is no longer being used
*/
	public void youHaveBeenDiscarded(){
		if (pool != null) {
			pool.removeConsumer(this);
		}
		pool = null;
	}
	/**
	* An input to the agent was removed
	*/
	public void aSourceWasRemoved(ConnectedAgent source){}
			
	/**
	* An input to the agent was added
	*/
	public void aSourceWasAdded(ConnectedAgent source){}

	/**
	* An output to the agent was removed
	*/
	public void aSinkWasRemoved(ConnectedAgent sink){}
			
	/**
	* An output to the agent was added
	*/
	public void aSinkWasAdded(ConnectedAgent sink){}

/**
* Allows access to the agents super.setProperty()
*/
	public void superSetProperty(String property, String value) {
		super.setProperty(property, value);
	}

/**
* 
* @return Number whos actual type will depend on the resource.  
* Java can deal abstractly with numbers.
*
*/
	public Number getConsumption(String resourceDimension){
		rateConsumer.timeHack();
		return new Double(rateConsumer.consumption);
	}
/**
* Set consumption from Number value
*/
	public void setConsumption(String resourceDimension, Number value){
		rateConsumer.fudgeConsumptionHistory(value.doubleValue());
	}
/**
* Trigger an update of consumption metrics whenever there
* is a query
*/
	public String getProperty(String propertyName){
		if (propertyName.equals(RateConsumer.HISTORICAL_LEVEL)) {
			rateConsumer.timeHack();
		}
		return super.getProperty(propertyName);
	}
/**
* Get a property
*/
	public ValueList getPropertyValues(String propertyName){
		if (propertyName.equals(RateConsumer.HISTORICAL_LEVEL)) {
			rateConsumer.timeHack();
		}
		return super.getPropertyValues(propertyName);
	}
/**
* Get a property list
*/
	public Enumeration getProperties(){
		rateConsumer.timeHack();
		return super.getProperties();
	}
/**
* This is intended for read only reference to the memento. To set
* a property use setProperty()
*/
	public Memento getMemento (){
		rateConsumer.timeHack();
		return super.getMemento();
	}
/**
* Allows access to the agents consumption model.
*/
	public RateConsumer getModel(){
		return rateConsumer;
	}
/**
* Used by the pool to notify us that it has been discarded
*/
	public void poolWasDiscarded(){
		pool.removeConsumer(this);
		pool = null;
	}

}